#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides the forms corresponding to the models.
"""

###########################################################################
# IMPORTS

from django import forms
from django.forms import ModelForm
from django.contrib.auth.forms import UserCreationForm
from django.contrib.auth.tokens import default_token_generator
from django.utils.text import capfirst
from django.utils.translation import ugettext_lazy as _
from django.template import Context, loader
from django.utils.http import int_to_base36

from registration.models import User

###########################################################################
# CLASS

class UserForm(UserCreationForm):

    username = forms.RegexField(label=_("Username"), max_length=30, regex=r'^\w+$',
        error_message = _("This value must contain only letters, numbers and underscores."))

    email = forms.EmailField(label=capfirst('email(optional)'),
                             max_length=75,
                             required=False)

    class Meta:
        model = User
        fields = ('username', 'age', 'gender', 'email')

###########################################################################
# CLASS

class UserModifForm(ModelForm):

    username = forms.RegexField(label=_("Username"), max_length=30, regex=r'^\w+$',
        error_message = _("This value must contain only letters, numbers and underscores."))

    email = forms.EmailField(label=capfirst('email(optional)'),
                             max_length=75,
                             required=False)

    class Meta:
        model = User
        fields = ('username', 'age', 'gender', 'email', 'is_staff')

###########################################################################
# CLASS

class UserModif2Form(ModelForm):

    email = forms.EmailField(label=capfirst('email(optional)'),
                             max_length=75,
                             required=False)

    class Meta:
        model = User
        fields = ('age', 'gender', 'email')

###########################################################################
# CLASS

class PasswordResetForm(forms.Form):

    email = forms.EmailField(label=_("E-mail"), max_length=75)

    def clean_email(self):
        """
        Validates that a user exists with the given e-mail address.
        """

        email = self.cleaned_data["email"]
        self.users_cache = User.objects.filter(email__iexact=email)
        if len(self.users_cache) == 0:
            raise forms.ValidationError(_("That e-mail address doesn't have an associated user account. Are you sure you've registered?"))

        return email

    def save(self, domain_override=None,
              email_template_name='registration/password_reset_email.html',
              use_https=False, token_generator=default_token_generator):
        """
        Generates a one-use only link for resetting password and sends to the user
        """
        from django.core.mail import send_mail

        for user in self.users_cache:

            if not domain_override:
                current_site = Site.objects.get_current()
                site_name = current_site.name
                domain = current_site.domain
            else:
                site_name = domain = domain_override

            t = loader.get_template(email_template_name)
            c = {
                'email': user.email,
                'domain': domain,
                'site_name': site_name,
                'uid': int_to_base36(user.id),
                'user': user,
                'token': token_generator.make_token(user),
                'protocol': use_https and 'https' or 'http',
            }

            send_mail(_("Password reset on %s") % site_name,
                t.render(Context(c)), None, [user.email])
