#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
"""

###########################################################################
# IMPORTS

from django.conf import settings
from django.core.mail import send_mail
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.shortcuts import render_to_response
from django.template import loader, RequestContext #, Context
from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.decorators import login_required
from django.contrib.auth.forms import AuthenticationForm, PasswordChangeForm, PasswordResetForm #, SetPasswordForm
from django.contrib.auth.tokens import default_token_generator
from django.utils.http import int_to_base36
from django.views.decorators.cache import never_cache

from random import Random

from registration.forms import UserForm, UserModif2Form
from registration.models import User

###########################################################################
# GLOBALS

random = Random()
random.jumpahead(8)

###########################################################################

def index(request):
    return login_user(request)

@never_cache
def login_user(request):
    """
    Handles the login form.
    """

    form = None
    message = ''

    redirect_to = request.REQUEST.get('next', '')
    user = request.user
    if user.is_authenticated():
        #message = 'You are logged in.'
        if not redirect_to or '//' in redirect_to or ' ' in redirect_to:
            redirect_to = settings.LOGIN_REDIRECT_URL

        return HttpResponseRedirect(redirect_to)

    if request.POST.get('loginuser', None):

        if request.session.test_cookie_worked():
            request.session.delete_test_cookie()

            if request.POST.get('loginuser', None): # If the form has been submitted...
                form = AuthenticationForm(data=request.POST) # Bound the form to the POST data

                if form.is_valid(): # If all validation rules pass...
                    user = form.get_user()
                    login(request, user)
                    #message = 'You are logged in.'
                    if not redirect_to or '//' in redirect_to or ' ' in redirect_to:
                        redirect_to = settings.LOGIN_REDIRECT_URL

                    return HttpResponseRedirect(redirect_to)

                else:
                    message = 'Please correct the fields.'

        else:
            message = 'Please enable cookies and try again.'

    if form is None:
        form = AuthenticationForm()
        #message = 'Please log in.'

    # generate the test cookie
    request.session.set_test_cookie()

    return render_to_response('registration/login_user.html',
                               {'formLogin': form,
                                'message': message,
                                'next': redirect_to,
                               },
                               context_instance=RequestContext(request))

@never_cache
def create_user(request):
    """
    Handles the registration form.
    """

    form = None
    message = ''

    user = request.user

    if request.POST.get('createuser', None):

        if request.session.test_cookie_worked():
            request.session.delete_test_cookie()

            if request.POST.get('createuser', None): # If the form has been submitted...
                form = UserForm(request.POST) # Bound the form to the POST data

                if form.is_valid(): # If all validation rules pass...
                    # save the new user in the database
                    form.save()

                    # authenticate, then login
                    username = form.cleaned_data["username"]
                    password = form.cleaned_data["password1"]
                    user = authenticate(username=username, password=password)
                    login(request, user)
                    #message = 'New account is created.'
                    redirect_to = settings.LOGIN_REDIRECT_URL
                    return HttpResponseRedirect(redirect_to)

                else:
                    message = 'Please correct the fields.'

        else:
            message = 'Please enable cookies and try again.'

    if form is None:
        form = UserForm() # An unbound form
        #message = 'Please log in.'

    # generate the test cookie
    request.session.set_test_cookie()

    return render_to_response('registration/create_user.html',
                               {'formNewUser': form,
                                'message': message,
                               },
                               context_instance=RequestContext(request))

###########################################################################

@never_cache
def recover_pwd(request):
    """
    Handles the password_recover form.
    """

    form = None
    message = ''

    if request.method == 'POST':

        if request.POST.get('input_recover_pwd', None): # If the form has been submitted...
            form = PasswordResetForm(request.POST) # Bound the form to the POST data

            if form.is_valid(): # If all validation rules pass...
                # save the new user in the database
                #form.save()
                #form.save(domain_override=settings.SERVER_DOMAIN, email_template_name='registration/email_pwd.html')
                #form = None
                message = 'Password was sent to your email.'
                site_name = domain = settings.SERVER_DOMAIN

                for user in form.users_cache:

                    #Generates a random password with the given length and given allowed_chars
                    # Note that default value of allowed_chars does not have "I" or letters
                    # that look like it -- just to avoid confusion.
                    length = 6
                    allowed_chars = 'abcdefghjkmnpqrstuvwxyz23456789'
                    raw_password = ''.join([random.choice(allowed_chars) for i in xrange(0, length, 1)])

                    user.set_password(raw_password)
                    user.save()

                    t = loader.get_template('registration/email_pwd.html')
                    c = {'email': user.email,
                         'domain': domain,
                         'site_name': site_name,
                         'uid': int_to_base36(user.id),
                         'user': user,
                         'token': default_token_generator.make_token(user),
                         'protocol': 'http',
                        }

                    subject = 'Password reset on %s' % site_name
                    #email_body = 'Not yet implemented. Please feel free to create a new account.'
                    email_body = 'username: %s, new password: %s' % (user.username, raw_password) #t.render(Context(c))
                    sender = settings.SERVER_EMAIL
                    send_mail(subject, email_body, sender, [user.email])

                return render_to_response('registration/recover_pwd.html',
                                           {'formRecoverPwd': form,
                                           'message': message,
                                           },
                                           context_instance=RequestContext(request))

            else:
                message = 'Please correct the fields.'

    if form is None:
        form = PasswordResetForm() # An unbound form

    return render_to_response('registration/recover_pwd.html',
                               {'formRecoverPwd': form,
                                'message': message,
                               },
                               context_instance=RequestContext(request))

###########################################################################

@login_required
@never_cache
def change_pwd(request):
    """
    Handles the password_change form.
    """

    form = None
    message = ''

    user = request.user
    if not user.is_authenticated():
        #message = 'You have to login.'
        return HttpResponseRedirect(reverse('registration:login_user'))

    if request.method == 'POST':

        if request.POST.get('input_change_pwd', None): # If the form has been submitted...
            form = PasswordChangeForm(user, request.POST) # Bound the form to the POST data
            #form = SetPasswordForm(user, request.POST) # Bound the form to the POST data

            if form.is_valid(): # If all validation rules pass...
                # save the new user in the database
                form.save()
                form = None
                message = 'Password is changed.'

                return render_to_response('registration/change_pwd.html',
                                           {'formChangePwd': form,
                                            'message': message,
                                           },
                                           context_instance=RequestContext(request))

            else:
                message = 'Please correct the fields.'

    if form is None:
        form = PasswordChangeForm(user) # An unbound form
        #form = SetPasswordForm(user) # An unbound form

    return render_to_response('registration/change_pwd.html',
                               {'formChangePwd': form,
                                'message': message,
                               },
                               context_instance=RequestContext(request))

@login_required
@never_cache
def modify_user(request):
    """
    Handles the modify_user form.
    """

    form = None
    message = ''

    user = request.user
    if not user.is_authenticated():
        #message = 'You have to login.'
        return HttpResponseRedirect(reverse('registration:login_user'))

    # 'request' returns the Django standard USER instance
    # get the customized/subclass USER instance
    user = User.objects.get(pk=user.id)

    if request.method == 'POST':

        if request.POST.get('input_modify_usr', None): # If the form has been submitted...
            form = UserModif2Form(request.POST, instance=user) # Bound the form to the POST data

            if form.is_valid(): # If all validation rules pass...
                # save the new user in the database
                form.save()
                form = None
                message = 'Profile is changed.'

                return render_to_response('registration/modify_user.html',
                                           {'formModifyUsr': form,
                                            'message': message,
                                           },
                                           context_instance=RequestContext(request))

            else:
                message = 'Please correct the fields.'

    if form is None:
        form = UserModif2Form(instance=user) # An unbound form

    return render_to_response('registration/modify_user.html',
                               {'formModifyUsr': form,
                                'message': message,
                               },
                               context_instance=RequestContext(request))

###########################################################################

@login_required
@never_cache
def logout_user(request):

    # clear the session data
    for key in request.session.keys():
        del request.session[key]

    logout(request)

    return render_to_response('registration/logout_user.html',
                               {},
                               context_instance=RequestContext(request))
