#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
"""

###########################################################################
# IMPORTS

from datetime import datetime
from numpy import int32, ndarray, array, append
import sys

from django.conf import settings
from django.db import models

from retrieval_libs.assist.AssistUsr import AssistUsr
from retrieval_libs.display.AnswerMain import AnswerMain

###########################################################################
# FUNCTIONS

def _itrim_data(data, index_from, index_to, index_to2):

    if type(data) is ndarray:
        trimmed = append(data[index_from:index_to], data[0:index_to2], axis=0)
    elif type(data) is list:
        trimmed = data[index_from:index_to] + data[0:index_to2]
    elif type(data) is tuple:
        trimmed = tuple(data[index_from:index_to] + data[0:index_to2])
    else:
        sys.stderr.write('\nUNKNOWN TYPE %s\n' % type(data))
        sys.exit(1)

    return trimmed

def _trim_data(data, index_from, size):

        if data is None:
            return 0, None

        if isinstance(data, dict):
            dataln = len(data['imdbid'])
        elif isinstance(data, (tuple, list, ndarray)):
            dataln = len(data)
        else:
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % type(data))
            sys.exit(1)

        # normalize
        index_from = index_from % dataln

        # wraparound indexes
        index_to = index_from + size
        index_to2 = 0
        if index_to > dataln:
            index_to2 = index_to - dataln
            index_to = dataln

        if isinstance(data, dict):
            trimmed_data = {}
            for key, item in data.iteritems():
                trimmed = _itrim_data(item, index_from, index_to, index_to2)
                trimmed_data[key] = trimmed
        else:
            trimmed_data = _itrim_data(data, index_from, index_to, index_to2)

        return index_from, trimmed_data

def _trim_rankdata(rankdata, index_from, size):

    rankprob = rankdata['improb']

    # sorted arguments, in decreasing order
    sortedarg = rankprob.argsort()
    sortedarg = sortedarg[::-1]

    index_from, sortedarg = _trim_data(sortedarg, index_from, size)

    dispdata = {}
    for key, item in rankdata.iteritems():
        if isinstance(item, ndarray):
            dispdata[key] = item[sortedarg]
        elif isinstance(item, list):
            dispdata[key] = [item[i] for i in sortedarg]
        else:
            sys.stderr.write('\nUNKNOWN ITEM %s\n' % type(item))
            sys.exit(1)

    return index_from, dispdata

###########################################################################
# CLASS

class Answer(object):
    """
    Provide the display information. It gathers the information for
    the current 'window' of results and also some global information.

    HINT: Only a fraction of the results is displayed at a time.
    """

    #######################################################################

    def _update1(self, dispdata):
        """
        Prepare the displayed results for the [FDBK] list.
        """

        self.index_from, dispdata = _trim_data(dispdata, self.index_from, self.SIZE)

        if self.assistusr is not None:
            assistinfo = self.assistusr.fdbk_order(dispdata)
        else:
            assistinfo = None

        self.answers = self.dbanswer.display_info(dispdata, assistinfo)
        self.nb_answers = len(dispdata['imdbid'])

    def _update2(self, rankdata):
        """
        Prepare the displayed results for the [RANK] list.
        HINT: a fraction of the ranked images
        """

        if self.SIZE2 == 0:
            self.answers2 = None
            self.nb_answers2 = 0
            return

        self.index_from2, dispdata = _trim_rankdata(rankdata, self.index_from2, self.SIZE2)

        if self.assistusr is not None:
            assistinfo = self.assistusr.fdbk_order(dispdata)
        else:
            assistinfo = None

        self.answers2 = self.dbanswer.display_info(dispdata, assistinfo)
        self.nb_answers2 = len(rankdata['imdbid'])

    def _update3(self):
        """
        Prepare the displayed results for the [ASSISTANT] list.
        """

        self.answers3 = None

        if self.task:
            target = self.task.shown_target()
            if target:
                dispdata = {}
                dispdbid = array([target,], dtype=int32, order='c')
                dispdata['imdbid'] = dispdbid

                self.answers3 = self.dbanswer.display_info(dispdata, None)

    def _update4(self, tracedata, tt=0):
        """
        Prepare the displayed trace plots.
        HINT: a fraction of the trace plots
        """

        # keep only tracedata with valid 'url'
        # HINT: check if 'url' is missing
        tracedisp = None
        if tracedata:
            tracedisp = []
            for item in tracedata:
                item = list(item)
                if item[3]:
                    item[3] = item[3][tt]
                    tracedisp.append(item)
            if len(tracedisp) == 0:
                tracedisp = None

        if not tracedisp:
            self.nb_answers4 = None
            self.answers4 = None
            self.index_from4 = 0
            self.nb_answers5 = None
            self.answers5 = None
            self.index_from5 = 0
            return

        ###
        self.nb_answers4 = len(tracedisp)
        self.index_from4, reduceddata = _trim_data(tracedisp, self.index_from4, self.SIZE4)
        if self.nb_answers4 > 0:
            self.answers4 = []
            for item in reduceddata:
                answer = {}
                answer['action'] = item[0]
                answer['index'] = item[1]
                answer['trlen'] = item[2]
                answer['url'] = item[3]
                self.answers4.append(answer)

        ###
        self.nb_answers5 = len(tracedisp)
        self.index_from5, reduceddata = _trim_data(tracedisp, self.index_from5, self.SIZE5)
        # HINT: do NOT display if there is only one trace image, it is already displayed by answer4
        if self.nb_answers5 > 1:
            self.answers5 = []
            for item in reduceddata:
                answer = {}
                answer['action'] = item[0]
                answer['index'] = item[1]
                answer['trlen'] = item[2]
                answer['url'] = item[3]
                self.answers5.append(answer)

    #######################################################################

    def init(self, user, task, config, data):
        """
        Initialize the object.
        """

        self.user = user
        self.task = task

        #the index of the first result
        self.index_from = data.get('index_from', 0)
        self.index_from2 = data.get('index_from2', 0)
        self.index_from4 = data.get('index_from4', -1)
        self.index_from5 = data.get('index_from5', 0)

        # number of results displayed on page
        self.SIZE = data.get('ANSWERS_PER_PAGE', 0)
        self.SIZE2 = data.get('ANSWERS_PER_PAGE2', 0)
        self.SIZE4 = data.get('ANSWERS_PER_PAGE4', 0)
        self.SIZE5 = data.get('ANSWERS_PER_PAGE5', 0)

        self.dbanswer = AnswerMain()
        self.dbanswer.init()

        self.answers = None
        self.nb_answers = 0
        self.answers2 = None
        self.nb_answers2 = 0
        self.answers3 = None
        self.nb_answers4 = 0
        self.answers4 = None
        self.nb_answers5 = 0
        self.answers5 = None

        self.assistusr = None
        if self.task:
            if self.task.show_info:
                self.assistusr = AssistUsr()
                self.assistusr.init(self.task, config)

    def data(self):

        data = {}

        data['index_from'] = self.index_from
        data['index_from2'] = self.index_from2
        data['index_from4'] = self.index_from4
        data['index_from5'] = self.index_from5

        return data

    def update(self, rankdata, dispdata, tracedata):
        """
        Identify the results for display (a fraction of the results),
        and prepare the display information.
        """

        if not isinstance(rankdata, (type(None), dict)):
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % type(rankdata))
            sys.exit(1)

        if not isinstance(dispdata, (type(None), dict)):
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % type(dispdata))
            sys.exit(1)

        if dispdata is not None:
            self._update1(dispdata)

        if self.user.is_authenticated():
            if self.user.is_staff and rankdata is not None:
                self._update2(rankdata)

            self._update4(tracedata)

        ###
        self._update3()

    def update2(self,  tracedataH, tt=0):
        """
        Identify the results for display (a fraction of the results),
        and prepare the display information.
        """

        # keep only tracedata with valid 'url'
        # HINT: check if 'url' is missing
        tracedisp = None
        if tracedataH:
            tracedisp = []
            for tracedata in tracedataH:
                for item in tracedata:
                    item = list (item)
                    if item[3]:
                        item[3] = item[3][tt]
                        tracedisp.append(item)
            if len(tracedisp) == 0:
                tracedisp = None

        ###
        self.nb_answers5 = len(tracedisp)
        #self.index_from5, reduceddata = _trim_data(tracedisp, self.index_from5, self.SIZE5)
        self.index_from5 = 0
        reduceddata = tracedisp
        # HINT: do NOT display if there is only one trace image, it is already displayed by answer4
        if self.nb_answers5 > 1:
            self.answers5 = []
            for item in reduceddata:
                answer = {}
                answer['action'] = item[0]
                answer['index'] = item[1]
                answer['trlen'] = item[2]
                answer['url'] = item[3]
                self.answers5.append(answer)

    def displayinfo(self):

        info = {}

        if self.nb_answers > 0:
            info['answers'] = self.answers
            info['nb_answers'] = self.nb_answers
            info['index_from'] = self.index_from

        if self.nb_answers2 > 0:
            info['answers2'] = self.answers2
            info['nb_answers2'] = self.nb_answers2
            info['index_from2'] = self.index_from2

        if self.answers3 is not None:
            info['answers3'] = self.answers3

        if self.nb_answers4 > 0:
            info['answers4'] = self.answers4
            info['nb_answers4'] = self.nb_answers4
            info['index_from4'] = self.index_from4

        if self.nb_answers5 > 0:
            info['answers5'] = self.answers5
            info['nb_answers5'] = self.nb_answers5
            info['index_from5'] = self.index_from5

        return info
