#!/usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Performs the retrieval algorithm. Responsible for the overall workflow
of the retrieval process and for passing data between iterations.
"""

###########################################################################
# IMPORTS

import logging
import sys
from time import time as now

from retrieval_libs.search.SearchMain import SearchMain

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

###########################################################################
# CLASS

class Search(object):
    """
    Performs the retrieval algorithm. Responsible for the overall workflow
    of the retrieval process and for passing data between iterations.
    """

    #######################################################################
    # main functionality

    def init(self, config, session, data):
        """
        Initialize the object.
        """

        self.session = session

        self.dbsearch = SearchMain()
        self.dbsearch.init(
                        config.TRACE,
                        config.SYS_THARGS,
                        config.SYS_WEIGHTS,
                        config.SYS_CONSISTENCY,
                        config.SYS_VERSION)

        # initialize class variables
        self.rankdata = data.get('rankdata', None)
        self.rankdata_0 = data.get('rankdata_0', None)
        self.dispdata = data.get('dispdata', None)
        self.tracedata = data.get('tracedata', None)

    def data(self):
        """
        Return the data/status.
        HINT: stored in the session cache.
        """

        data = {}

        if self.rankdata is not None:
            data['rankdata'] = self.rankdata

        if self.rankdata_0 is not None:
            data['rankdata_0'] = self.rankdata_0

        if self.dispdata is not None:
            data['dispdata'] = self.dispdata

        if self.tracedata is not None:
            data['tracedata'] = self.tracedata

        return data

    def displayinfo(self):
        """
        Prepare the display information.
        """

        info = self.session.displayinfo()

        return info

    #######################################################################

    def _search(self, rankdata, rankdata_0, input_data):
        """
        Perform the retrieval routine (ie. the pipeline of algorithms).
        """

        action = input_data.get('action', None)
        data = input_data.get('data', None)

        if action not in ('new_query', 'new_rnd', 'fdbk_docs', 'fdbk_im'):
            sys.stderr.write('\nUNKNOWN ACTION %s\n' % action)
            sys.exit(1)

        # refresh the CLASS variables
        self.rankdata = None
        self.rankdata_0 = None
        self.dispdata = None
        self.tracedata = None

        if action == 'new_query':
            if data is None or data == '':
                sys.stderr.write('\nERROR %s\n' % action)
                sys.exit(1)

            self.dbsearch._perform_txt_search(data, None)

        elif action == 'fdbk_docs':
            if data is None:
                sys.stderr.write('\nERROR %s\n' % action)
                sys.exit(1)

            #self.dbsearch._perform_txt_search(None, data)
            raise NotImplementedError

        # image relevance feedback
        elif action == 'fdbk_im':
            dispfdbk = input_data.get('dispfdbk', None)
            self.dbsearch._performimgsearch(rankdata, rankdata_0, dispfdbk)
            self.rankdata = self.dbsearch.queryimg.collapse_rankdata(self.rankdata)
            self.tracedata = self.dbsearch.queryimg.tracedata

        elif action == 'new_rnd':
            if data is not None:
                sys.stderr.write('\nERROR %s\n' % action)
                sys.exit(1)

            dispfdbk = input_data.get('dispfdbk', None)
            self.dbsearch._performimgrandom(rankdata, rankdata_0, dispfdbk)
            self.rankdata = self.dbsearch.queryimg.collapse_rankdata(self.rankdata)
            self.tracedata = self.dbsearch.queryimg.tracedata

        else:
            sys.stderr.write('\nUNKNOWN ACTION %s\n' % action)
            sys.exit(1)

    #######################################################################
    #  main functionality

    def search(self, input_data, timing):
        """
        Performs the search in the database, and update the CLASS data.

        action   - the action to be performed on the input data
        query    - the question submitted for searching
        docs     - the documents submitted as relevant; type: (doc_dbid,...)
        images   - the images submitted as relevant; type: (im_dbid,...)
        dispdata - the image set on which the relevant feedback was given
        timing   - the timing of submitted feedback and system performance
        """

        start_time = now()
        logger.debug('    ITERATION START')

        #iteration = self.session.iterations
        actionH = self.session.actionH
        thargsH = self.session.thargsH
        weightsH = self.session.weightsH
        consistencyH = self.session.consistencyH
        dispfdbkH = self.session.dispfdbkH
        traceH = self.session.traceH
        timingH = self.session.timingH

        # update the thargs
        thargs = self.dbsearch.update_thargs(self.rankdata, input_data,
                                        dispfdbkH, thargsH, weightsH)

        # update the weights
        weights = self.dbsearch.update_weights(self.rankdata, input_data,
                                        dispfdbkH, thargsH, weightsH)

        # update the consistency
        consistency = self.dbsearch.update_consistency(self.rankdata, input_data,
                                        dispfdbkH, thargsH, weightsH, consistencyH)

        # update the action history
        actionH.append(input_data)

        # update the timing history
        timingH.append(timing)

        # update the disp/fdbk history
        dispfdbk = input_data.get('dispfdbk', None)
        dispfdbkH.append(dispfdbk)

        # update the thargs history
        thargsH.append(thargs)

        # update the weights history
        weightsH.append(weights)

        # update the consistency history
        consistencyH.append(consistency)

        self.dbsearch.queryimg.set_buffers()
        self.dbsearch.set_data(dispfdbkH, thargsH, weightsH, consistencyH)

        # perform the retrieval routine (ie. the pipeline of algorithms)
        self.dbsearch.search(
                        self.rankdata,
                        self.rankdata_0,
                        input_data)

        # update class variables
        self.rankdata = self.dbsearch.rankdata
        self.rankdata_0 = self.dbsearch.rankdata_0
        self.dispdata = self.dbsearch.dispdata
        self.tracedata = self.dbsearch.tracedata

        self.dbsearch.trim_data()
        self.dbsearch.queryimg.trim_buffers()

        # update the trace history
        traceH.append(self.tracedata)

        # update the iteration counter
        self.session.iterations += 1

        stop_time = now()
        logger.debug('    ITERATION DONE in %f seconds' %
                     (stop_time - start_time))
