#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides the statistics of the system performance.
"""

###########################################################################
# IMPORTS

from datetime import datetime
import logging
from numpy import int32, float32, ndarray, array, zeros, fromstring
import operator
import sys

from retrieval_libs.common.dbformats import str2dispfdbk
from retrieval_libs.dblayer.DbAccess import DbAccess
from retrieval_libs.display.StatisticsData import StatisticsData

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

# implemented as a singleton
dbaccess = DbAccess()

#######################################################################
# FUNCTIONS

def _str2datetime(s):

    to_return = None
    if s is not None:
        try:
            parts = s.split('.')
            dt = datetime.strptime(parts[0], "%Y-%m-%d %H:%M:%S")
            to_return = dt.replace(microsecond = int(parts[1]))
        except:
            pass

    return to_return

def _deltatime(timingH, it):
    """
    Prepare the display information for timing.
    """

    deltatime = 0.0
    if len(timingH) > max(1, it):
        t2 = timingH[it-1][2]
        t2 = _str2datetime(t2)
        if t2 is not None:
            t0 = timingH[it][0]
            t0 = _str2datetime(t0)
            if t0 is not None:
                deltatime = t0 - t2
                deltatime = deltatime.seconds + deltatime.microseconds / 1000000.0

    #print 'deltatime', deltatime
    return deltatime

#######################################################################
# FUNCTIONS

def _filter_data0(dbsessions, iii):
    """
    Keep the 'iii' or the last session per (user, task, config).

    HINT: iii = -1 will return the last session per (user, task, config).
    """

    if dbsessions is None:
        return None

    dtimes = {}
    for session in dbsessions:

        uid, tid, cid, it, ev, dtime = session

        udata = dtimes.get(uid, {})
        utdata = udata.get(tid, [])
        utdata.append(session)
        udata[tid] = utdata
        dtimes[uid] = udata

    sessions = []
    for uid, udata in dtimes.iteritems():
        for utdata in udata.itervalues():
            # order, in increasing order
            utdata = sorted(utdata, key=operator.itemgetter(5), reverse=False)

            if len(utdata) > iii: #and len(utdata) > 2:
                sessions.append(utdata[iii])
            else:
                sessions.append(utdata[-1])

    return sessions

def _filter_data1(dbsessions):
    """
    Keep half of sessions for each user.

    HINT: TODO versions for first half, second half, or random half
    """

    if dbsessions is None:
        return None

    dtimes = {}
    for session in dbsessions:

        uid, tid, cid, it, ev, dtime = session

        udata = dtimes.get(uid, [])
        udata.append(session)
        dtimes[uid] = udata

    sessions = []
    for udata in dtimes.itervalues():
        # order, in increasing order
        udata = sorted(udata, key=operator.itemgetter(5), reverse=False)

        ll = round(float(len(udata))/2) * 2
        # keep the first half of sessions
        for l in xrange(0,ll/2,1):
        # keep the second half of sessions
        #for l in xrange((ll/2 - 1), len(udata), 1):
            sessions.append(udata[l])

    return sessions

def _rearrange_data(dbsessions):

    sessions = _filter_data0(dbsessions, 0)
    sessions1 = []
    sessions10 = _filter_data0(dbsessions, 1)
    for s in sessions10:
        s = list(s)
        s[2] += 5
        sessions1.append(s)

    sessions += sessions1

    sessions2 = []
    sessions20 = _filter_data0(dbsessions, -1)
    for s in sessions20:
        s = list(s)
        s[2] += 10
        sessions2.append(s)

    sessions += sessions2

    configs = {1: 0, 6: 1, 11: 2}
    #configs = {3: 0, 8: 1, 13: 2}
    #configs = {5: 0, 10: 1, 15: 2}

    return sessions, configs

###########################################################################
# CLASS

class Statistics(object):
    """
    Provides the statistics of the system performance.
    """

    #######################################################################
    # main functionality

    def init(self, users, tasks, configs, figids):
        """
        Initialize the object.
        """

        self.dbfig = dbaccess.dbfig()
        self.dbstats = StatisticsData()

        self.users = {}
        for i, uid in enumerate(users):
            self.users[uid] = i

        self.tasks = {}
        for i, tid in enumerate(tasks):
            self.tasks[tid] = i

        self.configs = {}
        for i, cid in enumerate(configs):
            self.configs[cid] = i

        self.figids = figids

        self._config_props = None

        self._evaluation_data = None
        self._trace_data = None
        self._timing_data = None
        self._thargs_data = None
        self._consistency_data = None
        self._displayed_images = None

        self.filter_data = None
        self.norm_utc = None
        self.histo_utc = None

    def update_images(self):
        """
        Statistics for staff users.
        """

        config_pairs = self._get_config_pairs()
        config_props = self._get_config_props()
        evaluation_data = self._get_evaluation_data()

        self.dbstats.init(self.users, self.tasks, self.configs, config_pairs, config_props, evaluation_data)

        fileurls = []
        for figid in self.figids:

            if figid < 10:
                prefix = '0%d' % figid
            else:
                prefix = '%d' % figid

            filename, fileurl = self.dbfig.filename('statistics', prefix, '.png')

            fileurls.append(fileurl)

            if figid == 0:
                self.dbstats.update(filename)
            elif figid == 1:
                timing_data = self._get_timing_data()
                self.dbstats.update13(timing_data, filename)
            elif figid == 2:
                consistency_data = self._get_consistency_data()
                self.dbstats.update16(consistency_data, filename)

        return fileurls

    #######################################################################

    def update_image_user(self, uid, username):
        """
        Statistics for simple users.
        """

        config_pairs = self._get_config_pairs()
        config_props = self._get_config_props()
        evaluation_data = self._get_evaluation_data()

        self.dbstats.init(self.users, self.tasks, self.configs, config_pairs, config_props, evaluation_data)

        fileurls = []

        filename, fileurl = self.dbfig.filename('statistics', 'user_avrg', '.png')
        valid = self.dbstats.update_eval_fdbk(uid, username, filename)

        if valid:
            fileurls.append(fileurl)

        return fileurls

    #######################################################################
    # get functionality

    def _get_config_pairs(self):

        config_pairs = []
        c = 0
        for i, ci in enumerate(self.configs):
            for j, cj in enumerate(self.configs):
                if j > i:
                    config_pairs.append((c, ci, cj))
                    c += 1

        #config_pairs = [(0, 0, 1), (1, 0, 2), (2, 2, 1)] # for textual/visual
        config_pairs = [(0, 1, 2), (1, 4, 2), (2, 4, 1)] # for mass-zoom

        return config_pairs

    def _get_config_props(self):
        """
        """

        if self._config_props:
            return self._config_props

        TABLE_CONFIG = dbaccess.dbtable_config()
        cstr = ', '.join(map(str, self.configs))
        answers = dbaccess.custom_query(
                "SELECT id, name "
                "FROM %s "
                "WHERE id IN (0, %s)"
                % (TABLE_CONFIG, cstr))

        config_props = {}
        config_props['LABELSLONG'] = {}
        config_props['LABELS'] = {}
        config_props['COLORS'] = {}
        config_props['WIDTHS'] = {}
        config_props['MARKERS'] = {}
        for cid, props in answers:
            props = props.split(';')
            if len(props) != 5:
                props = [props[0], '', 'k', '5', '']
                logger.warning('Configs display properties are INCOMPLETE')
                #logger.error('Configs display properties are INCOMPLETE')
                #sys.exit(1)

            #props[0] = props[0].replace(', RI', '')
            #props[0] = props[0].replace('original, mass-zoom', 'mass-zoom')
            #props[0] = '%s, ' % cid + props[0]

            config_props['LABELSLONG'][cid] = props[0]
            config_props['LABELS'][cid] = props[1]
            config_props['COLORS'][cid] = props[2]
            config_props['WIDTHS'][cid] = props[3]
            config_props['MARKERS'][cid] = props[4]

        self._config_props = config_props
        return self._config_props

    def _get_evaluation_data(self):
        """
        """

        if self._evaluation_data:
            return self._evaluation_data

        # initialize data
        shape = (len(self.users), len(self.tasks), len(self.configs), 22)
        data = zeros(shape, dtype=float32, order='c')

        TABLE_SEARCH = dbaccess.dbtable_search()
        ustr = ', '.join(map(str, self.users))
        tstr = ', '.join(map(str, self.tasks))
        cstr = ', '.join(map(str, self.configs))
        sessions = dbaccess.custom_query(
                "SELECT user_id, task_id, config_id, "
                #"iterations, evaluation, datetime, weights "
                "iterations, evaluation, datetime, trace "
                "FROM %s "
                "WHERE user_id IN (0, %s) AND task_id IN (0, %s) AND config_id IN (0, %s)"
                % (TABLE_SEARCH, ustr, tstr, cstr))

        #sessions = _filter_data0(sessions, 4)
        #sessions = _filter_data1(sessions)
        #sessions, configs = _rearrange_data(sessions)

        #rtime = datetime.today()
        #rtime = datetime(2012, 2, 29, 0, 0, 0, 0)

        if sessions:
            for uid, tid, cid, it, ev, dtime, traceH in sessions:

                #if dtime < rtime:
                #    continue

                u = self.users.get(uid, None)
                t = self.tasks.get(tid, None)
                c = self.configs.get(cid, None)
                if u is not None and t is not None and c is not None:

                    #if ev != 'Q':
                    #    continue

                    if it > 20 or ev == 'N':
                        data[u,t,c,21] += 1.0
                        #weightsH = eval(weightsH)
                        #for wi in xrange(0, it + 1, 1):
                        #    data[u,t,c,wi] += 1.0
                        #    wdata[u,t,c,wi] += weightsH[wi][0]

                    else:
                        data[u,t,c,it] += 1.0
                        #weightsH = eval(weightsH)
                        #for wi in xrange(0, it + 1, 1):
                        #    data[u,t,c,wi] += 1.0
                        #    wdata[u,t,c,wi] += weightsH[wi][0]

        #newdata = data.copy()
        #newdata[:,:,0,:]= data[:,:,2,:]
        #newdata[:,:,1,:]= data[:,:,0,:]
        #newdata[:,:,2,:]= data[:,:,1,:]

        self._evaluation_data = data #data, wdata
        return self._evaluation_data

    def _get_timing_data(self):
        """
        """

        if self._timing_data:
            return self._timing_data

        # initialize data
        shape = (len(self.users), len(self.tasks), len(self.configs), 22)
        data = zeros(shape, dtype=float32, order='c')
        norm = zeros(shape, dtype=int32, order='c')

        TABLE_SEARCH = dbaccess.dbtable_search()
        ustr = ', '.join(map(str, self.users))
        tstr = ', '.join(map(str, self.tasks))
        cstr = ', '.join(map(str, self.configs))
        sessions = dbaccess.custom_query(
                "SELECT user_id, task_id, config_id, "
                "timing, datetime "
                "FROM %s "
                "WHERE user_id IN (0, %s) AND task_id IN (0, %s) AND config_id IN (0, %s)"
                % (TABLE_SEARCH, ustr, tstr, cstr))

        #rtime = datetime.today()
        #rtime = datetime(2011, 2, 25, 1, 0, 0, 0)

        if sessions:
            for uid, tid, cid, timingH, dtime in sessions:

                #if dtime < rtime:
                #    continue

                u = self.users.get(uid, None)
                t = self.tasks.get(tid, None)
                c = self.configs.get(cid, None)
                if u is not None and t is not None and c is not None:

                    timingH = eval(timingH)
                    for i in xrange(1, len(timingH), 1):
                        deltatime = _deltatime(timingH, i)
                        data[u,t,c,(i-1)] += deltatime
                        norm[u,t,c,(i-1)] += 1

        # normalize data
        for u in xrange(0, shape[0], 1):
            for t in xrange(0, shape[1], 1):
                for c in xrange(0, shape[2], 1):
                    for i in xrange(0, shape[3], 1):
                        if norm[u,t,c,i] > 0:
                            data[u,t,c,i] /= norm[u,t,c,i]
                        else:
                            # no data
                            data[u,t,c,i] = -0.1

        self._timing_data = data
        return self._timing_data

    def _get_consistency_data(self):
        """
        """

        if self._consistency_data:
            return self._consistency_data

        # initialize data
        shape = (len(self.users), len(self.tasks), len(self.configs), 22)
        data = zeros(shape, dtype=float32, order='c')
        norm = zeros(shape, dtype=int32, order='c')

        TABLE_SEARCH = dbaccess.dbtable_search()
        ustr = ', '.join(map(str, self.users))
        tstr = ', '.join(map(str, self.tasks))
        cstr = ', '.join(map(str, self.configs))
        sessions = dbaccess.custom_query(
                "SELECT user_id, task_id, config_id, "
                "iterations, consistency, datetime "
                "FROM %s "
                "WHERE user_id IN (0, %s) AND task_id IN (0, %s) AND config_id IN (0, %s)"
                % (TABLE_SEARCH, ustr, tstr, cstr))

        #rtime = datetime.today()
        #rtime = datetime(2011, 2, 24, 1, 0, 0, 0)

        data2 = {}
        if sessions:
            for uid, tid, cid, it, consistencyH, dtime in sessions:

                #if dtime < rtime:
                #    continue

                if consistencyH == '':
                    continue

                itmin = 0
                if it < itmin or it > 20:
                    continue

                u = self.users.get(uid, None)
                t = self.tasks.get(tid, None)
                c = self.configs.get(cid, None)
                if u is not None and t is not None and c is not None:

                    consistencyH = map(float32, consistencyH.split('; '))
                    #for i in xrange(0, it + 1, 1):
                    #    data[u,t,c,i] += consistencyH[i]
                    #    norm[u,t,c,i] += 1

                    zoom = 1.0
                    for i in xrange(0, it + 1, 1):
                    #for i in xrange(max(0, it - itmin), it + 1, 1):
                        ishifted = i# - it + 20
                        #data[u,t,c,ishifted] += consistencyH[i]
                        #if i in (0, 1):
                        #    data[u,t,c,ishifted] += 0.5
                        #else:
                        #    data[u,t,c,ishifted] += (consistencyH[i] - 0.5) / 1.5
                        zoom /= consistencyH[i]
                        izoom = min(1.0, zoom)

                        data[u,t,c,ishifted] += izoom
                        norm[u,t,c,ishifted] += 1

                        data2_c = data2.get(c, {})
                        data2_ci = data2_c.get(ishifted, [])
                        data2_ci.append(izoom)
                        data2_c[ishifted] = data2_ci
                        data2[c] = data2_c

        # normalize data
        #shape = data.shape
        #for u in xrange(0, shape[0], 1):
        #    for t in xrange(0, shape[1], 1):
        #        for c in xrange(0, shape[2], 1):
        #            for i in xrange(0, shape[3], 1):
        #                if norm[u,t,c,i] > 0:
        #                    data[u,t,c,i] /= norm[u,t,c,i]
        #                else:
        #                    # no data
        #                    data[u,t,c,i] = -1.0

        #self._consistency_data = norm, data
        self._consistency_data = data, data2
        return self._consistency_data