#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Assigns task-config peers during the testing
"""

###########################################################################
# IMPORTS

import logging
from random import Random
import sys

from django.conf import settings

from retrieval_libs.assist.AutoTarget import AutoTarget
from retrieval_libs.dblayer.DbAccess import DbAccess
from retrieval.models import get_task_model

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

# implemented as a singleton
dbaccess = DbAccess()

random = Random()
random.jumpahead(8)

# the list of databases
ALL_DB = settings.ALL_DB

FIGURES = settings.FIGURES

###########################################################################
# local functionality per db

def _nb_sessions_per_db(db, uid):

    # the list of tasks for testing
    TASKS = ALL_DB[db]['TASKS']
    # the list of configs for testing
    CONFIGS = ALL_DB[db]['CONFIGS']
    # the list of task-configs saved in the database
    TABLE_SEARCH = ALL_DB[db]['TABLE_SEARCH']
    answers = dbaccess.custom_query(
                "SELECT task_id, config_id "
                "FROM %s "
                "WHERE user_id=%s"
                % (TABLE_SEARCH, uid))

    # count the task-config peers done already
    # HINT: filter only the ones included in the current test
    nb_sessions = 0
    for task, config in answers:
        if task in TASKS and config in CONFIGS:
            nb_sessions += 1

    return nb_sessions

def assign_taskconfig_per_db(db, uid):
    """
    Return a (random) taskconfig peer which has NOT been done yet by the user.
    Eventually, the user will be assigned with ALL possible taskconfig peers.
    """

    if uid is None:
        sys.stderr.write('\nUNDEFINED USER\n')
        sys.exit(1)

    # the list of tasks for testing
    TASKS = ALL_DB[db]['TASKS']
    # the list of configs for testing
    CONFIGS = ALL_DB[db]['CONFIGS']
    # the list of task-configs saved in the database
    TABLE_SEARCH = ALL_DB[db]['TABLE_SEARCH']
    answers = dbaccess.custom_query(
        "SELECT task_id, config_id FROM %s WHERE user_id=%s" %
        (TABLE_SEARCH, uid))

    # the number of sessions of one user test
    BLOCK_SIZE = len(TASKS) * len(CONFIGS)

    # the task-config peers done already
    # HINT: filter only the ones included in the current test
    filtered_answers = []
    for answer in answers:
        if answer[0] in TASKS and answer[1] in CONFIGS:
            filtered_answers.append(answer)

    answers = filtered_answers
    nb_sessions = len(answers)

    # all possible task-config peers
    taskconfigs = []
    for task in TASKS:
        for config in CONFIGS:
            taskconfigs.append((task, config))

    # the full list of task-config for testing
    taskconfigs = taskconfigs * (nb_sessions/BLOCK_SIZE + 1)
    random.shuffle(taskconfigs)

    # remove task-config that have been done already
    if answers:
        for tc in answers:
            try:
                taskconfigs.remove(tc)
            except ValueError:
                pass

    # randomize one more time
    random.jumpahead(random.randint(1, 8))
    return random.choice(taskconfigs)

def _reinitialize_tasks_per_db(db):
    """
    Reinitialize the data for the tasks included in the automatic user test.
    """

    Task = get_task_model(db)

    # the list of tasks for testing
    TASKS = ALL_DB[db]['TASKS']

    for taskid in TASKS:

        task = Task.objects.filter(pk=taskid)
        if task:
            task = task[0]
            task.init()
        else:
            sys.stderr.write('\nUNKNOWN TASK %s\n' % taskid)
            sys.exit(1)

        ###
        autotarget = AutoTarget()
        autotarget.init(task)
        target = autotarget.generate_new_target(ALL_DB, FIGURES, db)

        task.reinitialize(target)
        task.save_now()

###########################################################################
# main functionality

def nb_sessions_per_test():
    """
    Returns the number of sessions of one user test.
    """

    # the total number of sessions of one user test
    BLOCK_SIZE = 0

    for db in ALL_DB:
        # the list of tasks for testing
        TASKS = ALL_DB[db]['TASKS']
        # the list of configs for testing
        CONFIGS = ALL_DB[db]['CONFIGS']
        # add the number of sessions of one user test per database
        BLOCK_SIZE += len(TASKS) * len(CONFIGS)

    return BLOCK_SIZE

def nb_sessions_done(u):
    """
    Returns the number of sessions performed by the given user.
    """

    uid = u.id
    if uid is None:
        return None

    nb_sessions = 0
    for db in ALL_DB:
        nb_sessions += _nb_sessions_per_db(db, uid)

    return nb_sessions

def assign_db(uid):
    """"
    Return a (random) taskconfig peer which has NOT been done yet by the user.
    Eventually, the user will be assigned with ALL possible taskconfig peers.
    """

    # the number of sessions performed by the user per database
    nb_sessions_per_db = []
    for db in ALL_DB:
        nb = _nb_sessions_per_db(db, uid)
        nb_sessions_per_db.append(nb)

    # the total number of sessions performed by the user
    nb_sessions = sum(nb_sessions_per_db)

    # the databases of one user test
    BLOCK_SIZE = 0
    one_test_db_names = []
    for db in ALL_DB:
        # the list of tasks for testing
        TASKS = ALL_DB[db]['TASKS']
        # the list of configs for testing
        CONFIGS = ALL_DB[db]['CONFIGS']
        # the number of sessions of one user test per database
        BLOCK_SIZE_PER_DB = len(TASKS) * len(CONFIGS)

        one_test_db_names.extend([db] * BLOCK_SIZE_PER_DB)
        BLOCK_SIZE += BLOCK_SIZE_PER_DB

    # the databases that have to be done by the user
    all_test_db_names = one_test_db_names * (nb_sessions/BLOCK_SIZE + 1)
    random.shuffle(all_test_db_names)

    # remove the databases that have been done already
    for i, db_name in enumerate(ALL_DB):
        for ii in xrange(0, nb_sessions_per_db[i], 1):
            try:
                all_test_db_names.remove(db_name)
            except ValueError:
                pass

    #print 'all_test_db_names', len(all_test_db_names), all_test_db_names
    # randomize one more time
    random.jumpahead(random.randint(1, 8))
    return random.choice(all_test_db_names)

###########################################################################

def reinitialize_tasks():
    """
    Reinitialize the data for the tasks included in the automatic user test.
    HINT: only used by the autoapp.
    """

    for db in ALL_DB:
        _reinitialize_tasks_per_db(db)
