#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
"""

###########################################################################
# IMPORTS

import datetime, re
import logging

from django import http
from django.conf import settings
from django.contrib.auth.decorators import login_required, user_passes_test
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.views.decorators.cache import never_cache

from retrieval.manage.forms import get_task_form, get_config_form
from retrieval.models import get_session_model, \
                             get_task_model, \
                             get_config_model
from registration.forms import UserForm, UserModifForm
from registration.models import User

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

# number of items displayed on a page
ANSWERS_PER_PAGE = 10

# global variables
#active_db

###########################################################################
# cache functionality

# get the session data
def _get_state(request):

    global active_db

    active_db = request.session.get('active_db', None)

###########################################################################
# user functionality

def _create_user(request, user):

    form = None

    # if the form has been submitted...
    if request.POST.get('createuser', None):
        # bound the form to the POST data
        if user is None:
            form = UserForm(request.POST)
        else:
            form = UserModifForm(request.POST, instance=user)

        # if all validation rules pass...
        if form.is_valid():
            form.save()
            form = None
        else:
            message = 'Correct the form fields'
            if user is None or user.id is None:
                formaction = reverse('retrieval:manage:users')
            else:
                formaction = reverse('retrieval:manage:user', args=[str(user.id)])

    if form is None:
        if user is None or user.id is None:
            # an unbound form
            form = UserForm()
            message = 'Create a new user'
            formaction = reverse('retrieval:manage:users')
        else:
            # bound the form to the model data
            form = UserModifForm(instance=user)
            message = 'Modify the selected user'
            formaction = reverse('retrieval:manage:user', args=[str(user.id)])

    return form, formaction, message

def _change_permissions(request, user):

    if request.POST.get('add_expert', None):
        if user:
            # 'request' returns the Django standard USER instance
            # get the customized/subclass USER instance
            user = User.objects.get(pk=user.id)
            user.add_to_group_expert()

    if request.POST.get('remove_expert', None):
        if user:
            # 'request' returns the Django standard USER instance
            # get the customized/subclass USER instance
            user = User.objects.get(pk=user.id)
            user.remove_from_group_expert()

    permnames = []
    if user:
        permnames = user.get_all_permissions()

    return user, permnames

def _delete_user(request, user):

    if request.POST.get('deleteuser', None):
        if user:
            user.delete()
            del user
            user = None

    return user

@login_required
@user_passes_test(lambda u: u.is_staff)
@never_cache
def users(request, user_id=None):

    _get_state(request)

    user = None
    sessions = None
    if user_id:
        user = User.objects.filter(pk=user_id)
        # returns a list, with one element or empty
        if user:
            user = user[0]
            Session = get_session_model(active_db)
            sessions = Session.objects.filter(user=user) \
                                      .order_by('id')[:10]
        else:
            user = None

    user = _delete_user(request, user)
    user, permnames = _change_permissions(request, user)
    form, formaction, message = _create_user(request, user)

    return render_to_response(
        'retrieval/manage/users.html',
        {'users': User.objects.all().order_by('id'),
         'current_user': user,
         'sessions': sessions,
         'form': form,
         'formaction': formaction,
         'message': message,
         'permnames': permnames,
        },
        context_instance=RequestContext(request))

###########################################################################
# task functionality

def _create_task(request, task):

    form = None

    TaskForm = get_task_form(active_db)

    # if the form has been submitted...
    if request.POST.get('createtask', None):
        # bound the form to the POST data
        if task is None:
            form = TaskForm(request.POST)
        else:
            form = TaskForm(request.POST, instance=task)

        # if all validation rules pass...
        if form.is_valid():
            form.save()
            form = None
        else:
            message = 'Correct the form fields'
            if task is None or task.id is None:
                formaction = reverse('retrieval:manage:tasks')
            else:
                formaction = reverse('retrieval:manage:task', args=[str(task.id)])

    if form is None:
        if task is None or task.id is None:
            # an unbound form
            form = TaskForm()
            message = 'Create a new task'
            formaction = reverse('retrieval:manage:tasks')
        else:
            # bound the form to the model data
            form = TaskForm(instance=task)
            message = 'Modify the selected task'
            formaction = reverse('retrieval:manage:task', args=[str(task.id)])

    return form, formaction, message

def _delete_task(request, task):

    if request.POST.get('deletetask', None):
        if task:
            task.delete()
            del task
            task = None

    return task

@login_required
@user_passes_test(lambda u: u.is_staff)
@never_cache
def tasks(request, task_id=None):

    _get_state(request)

    Task = get_task_model(active_db)

    task = None
    target_info = None
    sessions = None
    if task_id:

        task = Task.objects.filter(pk=task_id)
        # returns a list, with one element or empty
        if task:
            task = task[0]
            Session = get_session_model(active_db)
            sessions = Session.objects.filter(task=task) \
                                      .order_by('id') \
                                      .order_by('user')[:10]

            target_info = task.displayinfo2()

        else:
            task = None
            target_info = None

    task = _delete_task(request, task)
    form, formaction, message = _create_task(request, task)

    return render_to_response(
        'retrieval/manage/tasks.html',
        {'tasks': Task.objects.all().order_by('id'),
         'current_task': task,
         'current_target': target_info,
         'sessions': sessions,
         'form': form,
         'formaction': formaction,
         'message': message,
        },
        context_instance=RequestContext(request))

###########################################################################
# config functionality

def _create_config(request, config):

    form = None

    ConfigForm = get_config_form(active_db)

    # if the form has been submitted...
    if request.POST.get('createconfig', None):
        # bound the form to the POST data
        if config is None:
            form = ConfigForm(request.POST)
        else:
            form = ConfigForm(request.POST, instance=config)

        # if all validation rules pass...
        if form.is_valid():
            form.save()
            form = None
        else:
            message = 'Correct the form fields'
            if config is None or config.id is None:
                formaction = reverse('retrieval:manage:configs')
            else:
                formaction = reverse('retrieval:manage:config', args=[str(config.id)])

    if form is None:
        if config is None or config.id is None:
            # an unbound form
            form = ConfigForm()
            message = 'Create a new config'
            formaction = reverse('retrieval:manage:configs')
        else:
            # bound the form to the model data
            form = ConfigForm(instance=config)
            message = 'Modify the selected config'
            formaction = reverse('retrieval:manage:config', args=[str(config.id)])

    return form, formaction, message

def _delete_config(request, config):

    if request.POST.get('deleteconfig', None):
        if config:
            config.delete()
            del config
            config = None

    return config

@login_required
@user_passes_test(lambda u: u.is_staff)
@never_cache
def configs(request, config_id=None):

    _get_state(request)

    Config = get_config_model(active_db)

    config = None
    sessions = None
    if config_id:
        config = Config.objects.filter(pk=config_id)
        # returns a list, with one element or empty
        if config:
            config = config[0]
            Session = get_session_model(active_db)
            sessions = Session.objects.filter(config=config) \
                                      .order_by('id') \
                                      .order_by('user')[:10]
        else:
            config = None

    config = _delete_config(request, config)
    form, formaction, message = _create_config(request, config)

    return render_to_response(
        'retrieval/manage/configs.html',
        {'configs': Config.objects.all().order_by('id'),
         'current_config': config,
         'sessions': sessions,
         'form': form,
         'formaction': formaction,
         'message': message,
        },
        context_instance=RequestContext(request))

###########################################################################
# session functionality

@login_required
@user_passes_test(lambda u: u.is_staff)
@never_cache
def sessions(request, index_from=0):

    _get_state(request)

    Session = get_session_model(active_db)

    sessionsln = Session.objects.count()

    context_data = {}
    if sessionsln > 0:
        # normalize indexes
        index_from = int(index_from) % sessionsln
        index_prev = (index_from - ANSWERS_PER_PAGE) % sessionsln
        index_next = (index_from + ANSWERS_PER_PAGE) % sessionsln
        indexes = range(index_from, index_from + ANSWERS_PER_PAGE, 1)

        # HINT: align indexes to ids
        id_from = index_from + 1
        ids = [i % sessionsln + 1 for i in indexes]

        sessions = Session.objects.filter(pk__in=ids) \
                                  .order_by('id')

        context_data['qsessions'] = {
            'sessions': sessions,
            'size': sessionsln,
            'id_from': id_from,
            'index_prev': index_prev,
            'index_next': index_next}

    return render_to_response(
        'retrieval/manage/sessions.html',
        context_data,
        context_instance=RequestContext(request))

###########################################################################

@login_required
@user_passes_test(lambda u: u.is_staff)
@never_cache
def index(request):

    _get_state(request)

    return render_to_response(
        'retrieval/manage/index.html',
        {},
        context_instance=RequestContext(request))