#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides the system configuration.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, float32, array, fromstring
import sys

from django.conf import settings
from django.db import models
from django.db.models.loading import cache

from retrieval_libs.assist.AutoTarget import AutoTarget
from retrieval_libs.common.dbformats import thargs2str, str2thargs, \
                                            weights2str, str2weights, \
                                            consistency2str, str2consistency

from retrieval.helpers.shortcuts import create_model

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

###########################################################################
# FUNCTION

def get_config_model(active_db):
    """
    Creates the corresponding Config model.
    """

    model = cache.get_model(active_db, 'config')
    if model is None:

        # the list of databases
        ALL_DB = settings.ALL_DB

        if active_db not in ALL_DB:
            sys.stderr.write('\nUNKNOWN DB %s\n' % active_db)
            sys.exit(1)

        fields = {}

        options = {
            'db_table': ALL_DB[active_db]['TABLE_CONFIG'],
            'managed': False,
        }

        model = create_model('Config',
                         super_=(_Config,),
                         module=_Config.__module__,
                         app_label=active_db,
                         fields=fields,
                         options=options)

    return model

###########################################################################
# CLASS

class _Config(models.Model):
    """
    Abstract base-class for the Config models.

    :var name: the label and the display properties
    :var TRACE: the trace size
    :var SYS_THARGS: the image similarity thresholds
    :var SYS_WEIGHTS: the image feature's weights
    :var SYS_CONSISTENCY: the consistency type
    :var SYS_VERSION: the algorithm version
    :var datetime: the time of the last saving in the database
    """

    name = models.CharField('NAME',
                            max_length=100,
                            blank=False,
                            default='',
                            db_column='name')

    TRACE = models.CharField('TRACE',
                             max_length=100,
                             default=None,
                             blank=True)

    SYS_THARGS = models.CharField('SYS_THARGS',
                                 max_length=100,
                                 default='',
                                 blank=False)

    SYS_WEIGHTS = models.CharField('SYS_WEIGHTS',
                                   max_length=100,
                                   default='',
                                   blank=False)

    SYS_CONSISTENCY = models.CharField('SYS_CONSISTENCY',
                                   max_length=100,
                                   default='',
                                   blank=False)

    SYS_VERSION = models.IntegerField('SYS_VERSION',
                                      default=None,
                                      db_column='SYS_VERSION')

    datetime = models.DateTimeField('datetime',
                                    auto_now=True,
                                    db_column='datetime',
                                    editable=False)

    class Meta:
        managed = False
        abstract = True

    #######################################################################
    # local functionality

    # HINT: not needed, changes are only done via the model form.
    #def _format_db(self):
        # put data in 'database' format

    def _format_rt(self):
        # put data in 'run-time' format

        # parse TRACE
        if self.TRACE is not None:
            info = self.TRACE.split('; ')
            if len(info) == 2:
                info[0] = int32(info[0])
                self.TRACE = tuple(info)
            else:
                self.TRACE = None
                #sys.stderr.write('\nPARSING ERROR <TRACE>\n')
                #sys.exit(1)

        # parse SYS_THARGS
        self.SYS_THARGS = str2thargs(self.SYS_THARGS)
        if len(self.SYS_THARGS) == 2:
            self.SYS_THARGS = tuple(self.SYS_THARGS)
        else:
            sys.stderr.write('\nPARSING ERROR <SYS_THARGS>\n')
            sys.exit(1)

        # parse SYS_WEIGHTS
        self.SYS_WEIGHTS = str2weights(self.SYS_WEIGHTS)
        if len(self.SYS_WEIGHTS) == 2:
            self.SYS_WEIGHTS = tuple(self.SYS_WEIGHTS)
        else:
            sys.stderr.write('\nPARSING ERROR <SYS_WEIGHTS>\n')
            sys.exit(1)

        # parse SYS_CONSISTENCY
        self.SYS_CONSISTENCY = str2consistency(self.SYS_CONSISTENCY)
        if len(self.SYS_CONSISTENCY) == 2:
            self.SYS_CONSISTENCY = tuple(self.SYS_CONSISTENCY)
        else:
            sys.stderr.write('\nPARSING ERROR <SYS_CONSISTENCY>\n')
            sys.exit(1)
        
    #######################################################################
    # main functionality

    def init(self):
        """
        Initialize the object.
        """

        self._format_rt()

    def displayinfo(self):
        """
        Return the display information.
        """

        info = {}

        if self.TRACE is None:
            info['TRACE'] = '----'
        else:
            info['TRACE'] = '; '.join(map(str, self.TRACE))

        info['SYS_THARGS'] = thargs2str(self.SYS_THARGS)
        info['SYS_WEIGHTS'] = weights2str(self.SYS_WEIGHTS)
        info['SYS_CONSISTENCY'] = consistency2str(self.SYS_CONSISTENCY)
        info['SYS_VERSION'] = self.SYS_VERSION

        return info
