#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Manages the searching session data.
::

    The session data is accumulated iteration by iteration. In between the
    iterations, the session data is stored in the Django-cache-per-user.
    At the end of the searching session, the session data is usually deleted,
    but in case of testing it stored in the database.
"""

###########################################################################
# IMPORTS

import copy
from datetime import datetime
import logging
from numpy import int32, float32, ndarray, array
import sys

from django.conf import settings
from django.contrib.auth.models import User
from django.db import models
from django.db.models.loading import cache

from retrieval_libs.common.dbformats import thargs2str, str2thargs, \
                                            weights2str, str2weights, \
                                            consistency2str, str2consistency, \
                                            dispfdbk2str, str2dispfdbk
from retrieval.helpers.shortcuts import create_model
from retrieval.models.Task import get_task_model
from retrieval.models.Config import get_config_model

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

EVALUATION_CHOICES = (
    ('Y', 'Successful by chance'),
    ('Q', 'Successful and consistent'),
    ('N', 'Not successful'), # 'Failed'
)

###########################################################################
# FUNCTION

def get_session_model(active_db):
    """
    Provides access to the corresponding Session model.
    """

    model = cache.get_model(active_db, 'session')
    if model is None:

        # the list of databases
        ALL_DB = settings.ALL_DB

        if active_db not in ALL_DB:
            sys.stderr.write('\nUNKNOWN DB %s\n' % active_db)
            sys.exit(1)

        Task = get_task_model(active_db)
        Config = get_config_model(active_db)

        fields = {
            'task': models.ForeignKey(Task),
            'config': models.ForeignKey(Config),
        }

        options = {
            'db_table': ALL_DB[active_db]['TABLE_SEARCH'],
            'managed': False,
        }

        model = create_model('Session',
                         super_=(_Session,),
                         app_label = active_db,
                         module=_Session.__module__,
                         fields=fields,
                         options=options)

    return model

###########################################################################
# CLASS

class _Session(models.Model):
    """
    Abstract base-class for the Session models.

    :var user: the user that performed the session
    :var task: the task given to the user
    :var config: the system configuration
    :var iterations: the number of iterations (redundant)
    :var evaluation: the evaluation given by the user
    :var actionH: the actions performed on the input data
    :var thargsH: the history of the image similarity thresholds
    :var weightsH: the history of the image feature's weights
    :var consistencyH: the history of the consistency scores
    :var dispfdbkH: the displayed image sets and the corresponding relevance feedback that was given
    :var traceH: the evolution of the size of the trace
    :var timingH: the timing of system response and user feedback
    :var datetime: the time of the saving in the database
    """

    user = models.ForeignKey(User)
    #task = models.ForeignKey(Task)
    #config = models.ForeignKey(Config)

    iterations = models.IntegerField('iterations',
                                     default=None,
                                     db_column='iterations')

    evaluation = models.CharField('evaluation',
                                  max_length=1,
                                  choices=EVALUATION_CHOICES,
                                  db_column='evaluation')

    actionH = models.CharField('action',
                               max_length=1000,
                               default='',
                               db_column='action')

    thargsH = models.CharField('thargs',
                                 max_length=10000,
                                 default='',
                                 db_column='thargs')

    weightsH = models.CharField('weights',
                                max_length=1000,
                                default='',
                                db_column='weights')

    consistencyH = models.CharField('consistency',
                                max_length=1000,
                                default='',
                                db_column='consistency')

    dispfdbkH = models.CharField('dispfdbk',
                                 max_length=10000,
                                 default='',
                                 db_column='dispfdbk')

    traceH = models.CharField('trace',
                              max_length=100000000,
                              default='',
                              db_column='trace')

    timingH = models.CharField('timing',
                               max_length=1000,
                               default='',
                               db_column='timing')

    datetime = models.DateTimeField('datetime',
                                    auto_now=False,
                                    db_column='datetime',
                                    editable=False)

    class Meta:
        managed = False
        abstract = True

    #######################################################################
    # local functionality

    def _str2datetime(self, strtime):
        """

        """

        to_return = None
        if strtime is not None:
            try:
                parts = strtime.split('.')
                dt = datetime.strptime(parts[0], "%Y-%m-%d %H:%M:%S")
                to_return = dt.replace(microsecond = int(parts[1]))
            except:
                # fail silently
                pass

        return to_return

    def _strtiming(self, it=-1):
        """
        Prepare the display information for timing.
        """

        strtiming_sys = None
        if len(self.timingH) > max(1, it):
            t2 = self.timingH[it-1].get(2, None)
            t2 = self._str2datetime(t2)
            if t2 is not None:
                t0 = self.timingH[it].get(0, None)
                t0 = self._str2datetime(t0)
                if t0 is not None:
                    strtiming_sys = str(t0 - t2)

        strtiming_usr = None
        if len(self.timingH) > max(0, it):
            t2 = self.timingH[it].get(2, None)
            t2 = self._str2datetime(t2)
            if t2 is not None:
                t0 = self.timingH[it].get(0, None)
                t1 = self.timingH[it].get(1, None)
                t0 = self._str2datetime(t0)
                t1 = self._str2datetime(t1)
                if t1 is not None:
                    strtiming_usr = str(t2 - t1)
                elif t0 is not None:
                    strtiming_sr = str(t2 - t0)

        return strtiming_sys, strtiming_usr

    #######################################################################
    # local functionality

    def _format_db(self):
        """
        Put data in database format.
        """

        actionH = []
        for action in self.actionH:
            action = [action.get(key, None) for key in ('action', 'data')]
            actionH.append(tuple(action))
        strActionH = str(actionH)
        self.actionH = '(' + strActionH[1:-1] + ',)'

        self.thargsH = thargs2str(self.thargsH)
        self.weightsH = weights2str(self.weightsH)
        self.consistencyH = consistency2str(self.consistencyH)
        self.dispfdbkH = dispfdbk2str(self.dispfdbkH)

        # keep only the statistics for each iteration
        # (initial, final and maximum length, COUNT_EXPAND, COUNT_COLLAPSE)
        noneitemsln = 0
        for item in self.traceH:
            if item is None:
                noneitemsln += 1
        if noneitemsln == len(self.traceH):
        #if self.traceH[0] is None:
            self.traceH = ''
            #self.traceH = unicode(self.traceH)
        else:
            strTraceH = []
            #print 'self.traceH', self.traceH
            for tracedata in self.traceH:

                if tracedata is None:
                    strTraceH.append(str(tracedata))
                    continue

                initial_ln = 0
                final_ln = 0
                max_ln = 0
                COUNT_EXPAND = 0
                COUNT_COLLAPSE = 0
                for action, count, traceln, filename in tracedata:
                    if 'initial' in action:
                        initial_ln = traceln
                    elif 'seed' in action:
                        max_ln = max (max_ln, traceln)
                    elif 'expand' in action:
                        COUNT_EXPAND = max (COUNT_EXPAND, count)
                    elif 'collapse' in action:
                        COUNT_COLLAPSE = max (COUNT_COLLAPSE, count)
                    elif 'final' in action:
                        final_ln = traceln
                    else:
                        sys.stderr.write('\nUNKNOWN ACTION %s\n' % action)
                        sys.exit(1)

                stritem = '(%d, %d, %d, %d, %d)' % (initial_ln,
                          final_ln, max_ln, COUNT_EXPAND, COUNT_COLLAPSE)
                strTraceH.append(stritem)

            # HINT: in case of a tuple with only one element add ','
            self.traceH = '(' + ', '.join(strTraceH) + ',)'
            #self.traceH = unicode(self.traceH)

        timingH = []
        #for i, timing in enumerate(self.timingH):
        #    st, ut = self._strtiming(i)
        #    timing['sys'] = st
        #    timing['usr'] = ut
        for timing in self.timingH:
            timing = [timing.get(key, None) for key in (0, 1, 2)]#('sys', 'usr')]
            timingH.append(tuple(timing))
        strtimingH = str(timingH)
        self.timingH = '(' + strtimingH[1:-1] + ',)'

    def _format_rt(self):
        # put data in 'run-time' format

        if self.iterations is None:
            self.iterations = -1

        if self.actionH == '':
            self.actionH = []
        else:
            self.actionH = eval(self.actionH)
            actionH = []
            for action in self.actionH:
                item = {}
                item['action'] = action[0]
                item['data'] = action[1]
                actionH.append(item)
            self.actionH = actionH

        if self.thargsH == '':
            self.thargsH = []
        else:
            self.thargsH = str2thargs(self.thargsH)

        self.weightsH = str2weights(self.weightsH)
        if self.weightsH is None:
            self.weightsH = []

        self.consistencyH = str2consistency(self.consistencyH)
        if self.consistencyH is None:
            self.consistencyH = []

        self.dispfdbkH = str2dispfdbk(self.dispfdbkH)
        if self.dispfdbkH is None:
            self.dispfdbkH = []

        if self.traceH == '':
            self.traceH = []
        else:
            self.traceH = eval(self.traceH)

        ###
        if self.timingH == '':
            self.timingH = []
        else:
            self.timingH = eval(self.timingH)
            timingH = []
            for timing in self.timingH:
                item = {}
                item[0] = timing[0]
                item[1] = timing[1]
                item[2] = timing[2]

                timingH.append(item)
            self.timingH = timingH

    #######################################################################
    # main functionality

    def init(self, data):
        """
        Initialize the object.
        """

        if data:
            # restore data from the cache

            #self.user = data.get('user', None)
            #self.task = data.get('task', None)
            #self.config = data.get('config', None)

            self.iterations = data.get('iterations', -1)
            self.evaluation = data.get('evaluation', None)
            self.actionH = data.get('actionH', [])
            self.thargsH = data.get('thargsH', [])
            self.weightsH = data.get('weightsH', [])
            self.consistencyH = data.get('consistencyH', [])
            self.dispfdbkH = data.get('dispfdbkH', [])
            self.traceH = data.get('traceH', [])
            self.timingH = data.get('timingH', [])
            #self.datetime = data.get('datetime', None)

            if 'id' in data:
                self.id = data['id']

        else:
            # reformat data from the database
            self._format_rt()

    def data(self):
        """
        Return the data/status.
        HINT: This data is stored in the session cache.
        """

        data = {}

        #data['user'] = self.user
        #data['task'] = self.task
        #data['config'] = self.config

        data['iterations'] = self.iterations
        data['evaluation'] = self.evaluation
        data['actionH'] = self.actionH
        data['thargsH'] = self.thargsH
        data['weightsH'] = self.weightsH
        data['consistencyH'] = self.consistencyH
        data['dispfdbkH'] = self.dispfdbkH
        data['traceH'] = self.traceH
        data['timingH'] = self.timingH

        if self.id is not None:
            data['id'] = self.id

        return data

    def update(self, input_data, thargs, weights, tracedata, timing):
        """
        Add the iteration data.
        """

        dispfdbk = input_data.get('dispfdbk', None)

        self.actionH.append(input_data)
        self.thargsH.append(thargs)
        self.weightsH.append(weights)
        #self.consistencyH.append(consistency)
        self.dispfdbkH.append(dispfdbk)
        self.traceH.append(tracedata)
        self.timingH.append(timing)

    def displayinfo(self, it=-1, watching=False):
        """
        Prepare the display information for the given iteration.
        If iteration is not given, it takes the last iteration.
        For watching, it shifts the information about the system.
        """

        if self.iterations == -1:
            sys.stderr.write('\nSESSION IS EMPTY\n')
            sys.exit(1)

        info = {}

        # normalize the index
        it = it % len(self.actionH)

        actiondata = self.actionH[it]
        info['action'] = actiondata['action']
        info['data'] = actiondata.get('data', None)

        dispfdbk = self.dispfdbkH[it]
        info['dispdata'] = dispfdbk[0]
        info['relevant_images'] = dispfdbk[1]

        timing_sys, timing_usr = self._strtiming(it)
        if timing_sys is None:
            info['timing_sys'] = '----'
        else:
            info['timing_sys'] = timing_sys

        if timing_usr is None:
            info['timing_usr'] = '----'
        else:
            info['timing_usr'] = timing_usr

        if not watching: # if searching
            info['iteration'] = it

            if len(self.traceH) > it and \
               self.traceH[it] is not None:
                tracedata = self.traceH[it][0]
                info['tracedata'] = '%s: %s, trace size %s' % (tracedata[0], tracedata[1], tracedata[2])

            strthargs = thargs2str(self.thargsH[it:(it+1)])
            strweights = weights2str(self.weightsH[it:(it+1)])
            strconsistency = consistency2str(self.consistencyH[it:(it+1)])

        else: # if watching
            # HINT: at iteration 0 supress the information
            # otherwise, shift the information about the system
            if it > 0:
                info['iteration'] = '%s out of %s' % (it - 1, self.iterations)

                if len(self.traceH) > (it - 1):
                    tracedata = self.traceH[it - 1]
                    info['tracedata'] = str(tracedata)

                strthargs = None
                strweights = None
                strconsistency = None

                if len(self.thargsH) > (it - 1):
                    strthargs = thargs2str(self.thargsH[(it-1):it])

                if len(self.weightsH) > (it - 1):
                    strweights = weights2str(self.weightsH[(it-1):it])

                if len(self.consistencyH) > (it - 1):
                    strconsistency = consistency2str(self.consistencyH[it:(it+1)])

            else:
                strthargs = None
                strweights = None
                strconsistency = None

        # adjustments
        if strthargs is None:
            info['thargs'] = '----'
        else:
            info['thargs'] = strthargs

        if strweights is None:
            info['weights'] = '----'
        else:
            info['weights'] = strweights

        if strconsistency is None:
            info['consistency'] = '----'
        else:
            info['consistency'] = strconsistency

        if self.id is not None:
            info['id'] = self.id

        return info

    def save_now(self, user, task, config, lastinput, lasttiming):
        """
        Save the object.
        """

        # pre-conditions
        if user is None or \
           not user.is_authenticated() or \
           task is None or \
           config is None:

            sys.stderr.write('\nMISSING DATA\n')
            sys.exit(1)

        # set the foreign keys
        self.user = user
        self.task = task
        self.config = config

        self.update(lastinput, None, None, None, lasttiming)

        if lastinput['action'] == 'evaluation':
            self.evaluation = lastinput.get('data', None)

        datetime = self.timingH[0].get(2, None)
        if datetime is not None:
            self.datetime = datetime

        self._format_db()

        logger.info('    SESSION SAVED WITH EVALUATION %s AFTER %s ITERATIONS' %
                    (self.evaluation, self.iterations))

        super(_Session, self).save()
