#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides the statistics of the system performance.
::

    The model retrieves the session data from the database and generates various plots.
    There is user interface to filter the data and to select the statistics of interest.
"""

###########################################################################
# IMPORTS

from numpy import int32, array, fromstring
import sys

from django.conf import settings
from django.db import models
from django.db.models.loading import cache

from retrieval_libs.dblayer.DbAccess import DbAccess
from retrieval.helpers.shortcuts import create_model
from retrieval.helpers.Statistics import Statistics

###########################################################################
# GLOBALS

# implemented as a singleton
dbaccess = DbAccess()

###########################################################################
# FUNCTION

def get_statistics_model(active_db):
    """
    Provides access to the corresponding Statistics model.
    """

    model = cache.get_model(active_db, 'statistics')
    if model is None:

        # the list of databases
        ALL_DB = settings.ALL_DB

        if active_db not in ALL_DB:
            sys.stderr.write('\nUNKNOWN DB %s\n' % active_db)
            sys.exit(1)

        fields = {}

        options = {
            'db_table': '',
            'managed': False,
        }

        model = create_model('Statistics',
                             super_=(_Statistics,),
                             module=_Statistics.__module__,
                             app_label=active_db,
                             fields=fields,
                             options=options)

    return model

###########################################################################
# CLASS

class _Statistics(models.Model):
    """
    Abstract base-class for the Statistics models.
    
    :var users: the list of users
    :var tasks: the list of tasks
    :var configs: the list of system configurations
    """

    users = models.CharField('users',
                              max_length=10000,
                              default='',
                              blank=True)

    tasks = models.CharField('tasks',
                              max_length=10000,
                              default='',
                              blank=True)

    configs = models.CharField('configs',
                                max_length=10000,
                                default='',
                                blank=True)

    class Meta:
        managed = False
        abstract = True

    #######################################################################
    # local functionality

    def _eval(self, string):

        if string:
            v = fromstring(string, dtype=int32, sep=', ')
            # check if valid
            if 0 in v or len(v.shape) != 1:
                v = None
        else:
            v = None

        return v

    def _reformat_rt(self):

        # users
        users = self._eval(self.users)
        if users is None:
            TABLE_USER = dbaccess.dbtable_user()
            users = dbaccess.custom_query(
                    #"SELECT id FROM %s"
                    "SELECT user_ptr_id FROM %s"
                    % TABLE_USER)
            users = [item[0] for item in users]
            users = array(users, dtype=int32, order='c')

        users.sort()
        self.u = users
        self.users = ', '.join(map(str, users))

        # tasks
        tasks = self._eval(self.tasks)
        if tasks is None:
            TABLE_TASK = dbaccess.dbtable_task()
            tasks = dbaccess.custom_query(
                    "SELECT id FROM %s"
                    % TABLE_TASK)
            tasks = [item[0] for item in tasks]
            tasks = array(tasks, dtype=int32, order='c')

        tasks.sort()
        self.t = tasks
        self.tasks = ', '.join(map(str, tasks))

        # configs
        configs = self._eval(self.configs)
        if configs is None:
            TABLE_CONFIG = dbaccess.dbtable_config()
            configs = dbaccess.custom_query(
                    "SELECT id FROM %s"
                    % TABLE_CONFIG)
            configs = [item[0] for item in configs]
            configs = array(configs, dtype=int32, order='c')

        #configs.sort()
        self.c = configs
        self.configs = ', '.join(map(str, configs))

    #######################################################################
    # main functionality

    def init(self):

        self.users = ''
        self.tasks = ''
        self.configs = ''
        self.figids = [0, 1]

        self.statistics = Statistics()

        self.u = None
        self.t = None
        self.c = None

    def set_data(self, data=None):

        if data is not None:
            self.users = data.get('users', '')
            self.tasks = data.get('tasks', '')
            self.configs = data.get('configs', '')
            self.figids = data.get('figids', [0, 1])

        self._reformat_rt()

    def get_data(self):

        data = {}
        data['users'] = self.users
        data['tasks'] = self.tasks
        data['configs'] = self.configs
        data['figids'] = self.figids

        return data

    #######################################################################

    def update_images(self):

        self.statistics.init(self.u, self.t, self.c, self.figids)

        fileurls = self.statistics.update_images()

        # book the 'active' status for the corresponding radio buttons
        filestat = {}
        for figid in self.figids:
            key = 'im_%d' % figid
            filestat[key] = 'checked'

        return fileurls, filestat

    def update_image_user(self, uid, username):

        self.statistics.init(self.u, self.t, self.c, self.figids)

        fileurls = self.statistics.update_image_user(uid, username)

        return fileurls