#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides the target that is given as reference to the users
during the testing and training.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, float32, array, fromstring
import sys

from django.conf import settings
from django.db import models
from django.db.models.loading import cache

from retrieval_libs.common.dbformats import weights2str, str2weights, \
                                            thargs2str, str2thargs, \
                                            target2str, str2target
from retrieval_libs.display.AnswerMain import AnswerMain
from retrieval.helpers.shortcuts import create_model

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

###########################################################################
# FUNCTION

def get_task_model(active_db):
    """
    Provides access to the corresponding Task model.
    """

    model = cache.get_model(active_db, 'task')
    if model is None:

        # the list of databases
        ALL_DB = settings.ALL_DB

        if active_db not in ALL_DB:
            sys.stderr.write('\nUNKNOWN DB %s\n' % active_db)
            sys.exit(1)

        fields = {
            'TABLE': ALL_DB[active_db]['TABLE'],
        }

        options = {
            'db_table': ALL_DB[active_db]['TABLE_TASK'],
            'managed': False,
        }

        # remove the model from the django cache
        #print 'cache.app_models', cache.app_models
        #try:
        #    del cache.app_models['active_db']['task']
        #except KeyError:
        #    pass

        model = create_model('Task',
                         super_=(_Task,),
                         app_label=active_db,
                         module=_Task.__module__,
                         fields=fields,
                         options=options)

    #print model._meta.db_table, model._meta.app_label, model.__module__

    return model

###########################################################################
# CLASS

class _Task(models.Model):
    """
    Abstract base-class for the Task models.

    :var query: the text query
    :var target: the target image set
    :var USR_THARGS: the image similarity thresholds
    :var USR_WEIGHTS: the image feature's weights
    :var USR_VERSION: the algorithm version
    :var datetime: the time of the last saving in the database
    """

    query = models.CharField('query',
                             max_length=100,
                             blank=True)

    target = models.TextField('target',
                              #max_length=1000000,
                              default='',
                              blank=True,
                              db_column='target')

    USR_THARGS = models.CharField('USR_THARGS',
                                 max_length=100,
                                 default='',
                                 blank=True)

    USR_WEIGHTS = models.CharField('USR_WEIGHTS',
                                   max_length=100,
                                   default='',
                                   blank=True)

    USR_VERSION = models.CharField('USR_VERSION',
                                   max_length=100,
                                   default='',
                                   blank=True)

    datetime = models.DateTimeField('datetime',
                                    auto_now=True,
                                    db_column='datetime',
                                    editable=False)

    class Meta:
        managed = False
        abstract = True

    #######################################################################
    # local functionality

    def _format_db(self):
        """
        Put data in database format.
        """

        # HINT: NOT necessary to reformat
        #self.query
        #self.datetime
        #self.USR_VERSION

        self.target = target2str(self.target)
        self.USR_THARGS = thargs2str(self.USR_THARGS)
        self.USR_WEIGHTS = weights2str(self.USR_WEIGHTS)

    def _format_rt(self):
        """
        Put data in run-time format.
        """

        # parse target
        self.target = str2target(self.target)

        # parse USR_THARGS
        self.USR_THARGS = str2thargs(self.USR_THARGS)
        if len(self.USR_THARGS) == 2:
            self.USR_THARGS = tuple(self.USR_THARGS)
        else:
            sys.stderr.write('\nPARSING ERROR <USR_THARGS>\n')
            sys.exit(1)

        # parse USR_WEIGHTS
        self.USR_WEIGHTS = str2weights(self.USR_WEIGHTS)
        if len(self.USR_WEIGHTS) == 2:
            self.USR_WEIGHTS = tuple(self.USR_WEIGHTS)
        else:
            sys.stderr.write('\nPARSING ERROR <USR_WEIGHTS>\n')
            sys.exit(1)

    #######################################################################
    # main functionality

    # HINT: this will reformat data also for the model form
    #def __init__(self, *args, **kwargs):
    #    super(_Task, self).__init__(*args, **kwargs)
    #    self._format_rt()

    def init(self, data=None):
        """
        Initialize the object.
        HINT: restored from the session cache.
        """

        if data is None:
            self.show_query = True
            self.show_target = False
            self.show_info = False
        else:
            self.show_query = data.get('show_query', True)
            self.show_target = data.get('show_target', False)
            self.show_info = data.get('show_info', False)

            if 'id' in data:
                self.id = data['id']

        self._format_rt()

    def data(self):
        """
        Return the data/status.
        HINT: stored in the session cache.
        """

        data = {}

        data['show_query'] = self.show_query
        data['show_target'] = self.show_target
        data['show_info'] = self.show_info

        if self.id is not None:
            data['id'] = self.id

        return data

    def reinitialize(self, target):
        """
        Reinitialize the task data.
        HINT: useful for the autoapp.
        """

        self.target = target

    def save_now(self):
        """
        Save the model in the database table.

        HINT: it is used only by the automatic reinitializations.
              changes via the model form are saved directly.
        """

        self._format_db()

        super(_Task, self).save()

    #######################################################################

    def shown_target(self):
        """
        Return the target for display.
        """

        target = None
        if self.show_target:
            target = self.target[0]

        return target

    def displayinfo(self):
        """
        Prepare the display information.
        """

        info = {}

        if self.show_query:
            info['query'] = self.query

        if self.show_target:
            info['target'] = self.target[0]

        return info

    def displayinfo2(self):
        """
        Prepare the display information.
        """

        info = {}

        if self.target:
            target = str2target(self.target)
            dispdata = {'imdbid': target}
            dbanswer = AnswerMain()
            dbanswer.init()
            info = dbanswer.display_info(dispdata, None)

        return info