#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
"""

###########################################################################
# IMPORTS

import logging
from datetime import datetime
from numpy import int32
import operator
from random import Random
from time import time as now
import sys

from django.conf import settings
from django.core.urlresolvers import reverse
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.shortcuts import render_to_response
from django.shortcuts import get_object_or_404
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django.template import RequestContext
from django.views.decorators.cache import never_cache

from retrieval_libs.dblayer.DataBuffer import DataBuffer

from retrieval.helpers.Search import Search
from retrieval.helpers.Answer import Answer
from retrieval.helpers.UserTestController import nb_sessions_done, \
                                                 nb_sessions_per_test, \
                                                 assign_taskconfig_per_db
from retrieval.models import get_session_model, \
                             get_task_model, \
                             get_config_model

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

# implemented as a singleton
databuffer = DataBuffer()

random = Random()
random.jumpahead(8)

# the list of databases
ALL_DB = settings.ALL_DB

# number of items displayed on page
ANSWERS_PER_PAGE = 8
ANSWERS_PER_PAGE2 = 9
ANSWERS_PER_PAGE4 = 1
ANSWERS_PER_PAGE5 = 6

# the maximum number of iterations for the sessions during testing
ITER_MAX = settings.ITER_MAX
#warning_iter_max = False

#input_data = {'action': '', 'data': '', 'dispfdbk': '',}

# cached variables
#active_db
#active_mode
#buffer_search
#buffer_session
#buffer_config
#buffer_task
#buffer_dblayer
#buffer_answer
#buffer_timing = {0: responseSendingTime, 1: responseCallbackTime, 2: requestArrivalTime}

###########################################################################
# initialize buffers

def _init_searching():
    # only initializations

    buffer_c = buffer_config.get(active_mode, None)
    if buffer_c is None:
        buffer_c = {}
        buffer_config[active_mode] = buffer_c

    if 'id' not in buffer_c:
        buffer_c['id'] = ALL_DB[active_db]['DEFAULT_CONFIG']

    # buffer_task is empty
    # HINT: search has no task

def _init_testing(u):
    # only initializations

    buffer_c = buffer_config.get(active_mode, None)
    if buffer_c is None:
        buffer_c = {}
        buffer_config[active_mode] = buffer_c

    buffer_t = buffer_task.get(active_mode, None)
    if buffer_t is None:
        buffer_t = {}
        buffer_task[active_mode] = buffer_t

    # assign a taskconfig peer
    tid, cid = assign_taskconfig_per_db(active_db, u.id)
    buffer_t['id'] = tid
    buffer_c['id'] = cid

    buffer_t['show_query'] = True
    buffer_t['show_target'] = True
    buffer_t['show_info'] = False

def _init_training():
    # only initializations

    buffer_c = buffer_config.get(active_mode, None)
    if buffer_c is None:
        buffer_c = {}
        buffer_config[active_mode] = buffer_c

    buffer_t = buffer_task.get(active_mode, None)
    if buffer_t is None:
        buffer_t = {}
        buffer_task[active_mode] = buffer_t

    if 'id' not in buffer_c:
        buffer_c['id'] = ALL_DB[active_db]['DEFAULT_CONFIG']

    if 'id' not in buffer_t:

        #tid, cid = assign_taskconfig_per_db(active_db, u.id)

        # the list of tasks for training
        tasks = [1,]

        # randomize
        tid = random.choice(tasks)

        buffer_t['id'] = tid
        buffer_t['show_query'] = True
        buffer_t['show_target'] = True
        buffer_t['show_info'] = True

###########################################################################
# cache functionality

def _get_state(request, action, data):
    """
    Get the input data and update the necessary data from the cache.
    """

    global input_data

    global active_db
    global active_mode
    global buffer_search
    global buffer_session
    global buffer_config
    global buffer_task
    global buffer_answer
    global buffer_dblayer
    global buffer_timing

    cache = request.session

    active_db = cache.get('active_db', None)
    if active_db is None:
        sys.stderr.write('\nUNKNOWN DB %s\n' % active_db)
        sys.exit(1)

    # get the configuration
    active_mode = cache.get('active_mode', None)
    buffer_config = cache.get('config', {})
    buffer_task = cache.get('task', {})

    input_data = {}

    # for convenience
    buffer_search = cache.get('search', {})
    dispdata = buffer_search.get('dispdata', None)
    input_data['dispfdbk'] = (dispdata, None)

    # get action and data from forms
    if request.POST.get('input_evalY', None):
        input_data['action'] = 'evaluation'
        input_data['data'] = 'Y'

    elif request.POST.get('input_evalQ', None):
        input_data['action'] = 'evaluation'
        input_data['data'] = 'Q'

    # HINT: consider the actions AFTER the request POST data
    elif action is 'search':
        input_data['action'] = 'new_rnd'
        active_mode = 'search'
        _init_searching()

    elif action == 'test':
        input_data['action'] = 'new_rnd'
        active_mode = 'test'
        u = request.user
        _init_testing(u)

    elif action == 'train':
        input_data['action'] = 'new_rnd'
        active_mode = 'train'
        _init_training()

    #elif request.POST.get('fdbk_im', None):
    elif action == 'fdbk_im':
        # CLEAN if input/buffer data are inconsistent
        flag = True
        if isinstance(dispdata, dict):
            dbid = dispdata.get('imdbid', None)
            if dbid is not None:
                dataln = len(data)
                if dataln == 1:
                    idbid = data[0]
                    if idbid not in dbid and idbid > 0:
                        flag = False
                else:
                    for idbid in data:
                        if idbid not in dbid:
                            flag = False
                            break
            else:
                flag = False
        else:
            flag = False

        if flag:
            input_data['action'] = action
            input_data['data'] = data[0]
            input_data['dispfdbk'] = (dispdata, data[0])
        else:
            logger.warning('INCONSISTENT DATA %s, %s' %
                           (str(dispdata), str(data)))

            input_data['action'] = 'new_rnd'

    else:
        sys.stderr.write('\nUNKNOWN ACTION %s\n' % action)
        sys.exit(1)

    #logger.info('    ACTIVE MODE %s' % active_mode)
    if active_mode is None:
        sys.stderr.write('\nUNKNOWN MODE %s\n' % active_mode)
        sys.exit(1)

    # get the session buffers
    if input_data['action'] in ('new_query', 'new_rnd'):
        # reset for NEW session
        input_data['dispfdbk'] = (None, None)

        buffer_search = {}
        buffer_session = {}
        buffer_dblayer = {}
        buffer_timing = {}

        # clean up the databuffer for the next session
        databuffer.del_data()
        # pass the session buffer to the databuffer
        databuffer.set_data(buffer_dblayer)

    else:
        buffer_search = cache.get('search', {})
        buffer_session = cache.get('session', {})
        buffer_dblayer = cache.get('dblayer', {})
        buffer_timing = cache.get('timing', {})

        # clean up the databuffer for the next session
        databuffer.del_data()
        # pass the session buffer to the databuffer
        databuffer.set_data(buffer_dblayer)

    # refresh the buffer_answer
    #buffer_answer = cache.get('answer', {})
    buffer_answer = {}

def _set_state(request):
    """
    Store the data in the cache.
    """

    start_time = now()

    cache = request.session

    cache['active_mode'] = active_mode
    cache['search'] = buffer_search
    cache['session'] = buffer_session
    cache['config'] = buffer_config
    cache['task'] = buffer_task
    cache['answer'] = buffer_answer
    cache['dblayer'] = buffer_dblayer
    cache['timing'] = buffer_timing

    stop_time = now()
    logger.debug('DONE in %f seconds' % (stop_time - start_time))

def _del_state():
    """
    Delete the global variables.
    """

    global input_data
    global warning_iter_max

    global active_db
    global active_mode
    global buffer_search
    global buffer_session
    global buffer_config
    global buffer_task
    global buffer_answer
    global buffer_dblayer
    global buffer_timing

    del input_data
    del warning_iter_max

    del active_db
    del active_mode
    del buffer_search
    del buffer_session
    del buffer_config
    del buffer_task
    del buffer_answer
    del buffer_dblayer
    del buffer_timing

###########################################################################
# local functionality

def _prepare_config(active_db):

    Config = get_config_model(active_db)

    buffer_c = buffer_config.get(active_mode, None)
    if buffer_c is None:
        #c = None
        sys.stderr.write('\nUNKNOWN CONFIG\n')
        sys.exit(1)
    else:
        configid = buffer_c.get('id', 0)
        c = Config.objects.filter(pk=configid)
        if c:
            c = c[0]
            c.init()
        else:
            # relevant for the action 'config_new'
            c = Config.objects.get(pk=1)
            c.init()
            # update buffers
            buffer_c['id'] = c.id

    return c

def _prepare_task(active_db):

    Task = get_task_model(active_db)

    buffer_t = buffer_task.get(active_mode, None)
    if buffer_t is None:
        # in case of searching
        t = None
    else:
        # in case of testing/training
        taskid = buffer_t.get('id', 0)
        t = Task.objects.filter(pk=taskid)
        if t:
            t = t[0]
            t.init(buffer_t)
        else:
            # relevant for the action 'task_new'
            t = Task.objects.get(pk=1)
            t.init(buffer_t)
            # update buffers
            buffer_t['id'] = t.id

    return t

def _prepare_session(active_db, user, task, config):

    Session = get_session_model(active_db)

    sid = buffer_session.get('id', 0)
    s = Session.objects.filter(pk=sid)
    if s:
        s = s[0]
        s.init(buffer_session)
    else:
        # create a new Session
        s = Session()
        s.init(buffer_session)

    return s

def _prepare_search(s, u, t, c):

    global input_data

    global buffer_search
    global buffer_session
    global buffer_dblayer
    global buffer_timing
    global warning_iter_max

    warning_iter_max = False

    # if the number of iterations reaches the maximum during testing,
    # a negative evaluation is inserted automatically
    # HINT: the input action CAN be an evaluation already
    if active_mode == 'test' and \
       s.iterations >= ITER_MAX and \
       input_data['action'] != 'evaluation':

        #input_data = {} # the other fields should remain unchanged
        input_data['action'] = 'evaluation'
        input_data['data'] = 'N'

        # warning when the input action is changed automatically
        warning_iter_max = True

    if active_mode == 'test' and input_data['action'] == 'evaluation':
        # set the last iteration data and save the session
        s.save_now(u, t, c, input_data, buffer_timing)

        # remove the session
        s = None
        search = None

        # clean up the buffers for the next session
        buffer_search = {}
        buffer_session = {}
        buffer_dblayer = {}
        buffer_timing = {}

        # clean up the databuffer for the next session
        databuffer.del_data()
        # pass the session buffer to the databuffer
        databuffer.set_data(buffer_dblayer)

    else:
        search = Search()
        search.init(c, s, buffer_search)

    return search, s, t, c

def _prepare_answer(user, task, config):
    # initialize Answer object

    buffer_answer['ANSWERS_PER_PAGE'] = ANSWERS_PER_PAGE
    buffer_answer['ANSWERS_PER_PAGE2'] = ANSWERS_PER_PAGE2
    buffer_answer['ANSWERS_PER_PAGE4'] = ANSWERS_PER_PAGE4
    buffer_answer['ANSWERS_PER_PAGE5'] = ANSWERS_PER_PAGE5

    a = Answer()
    a.init(user, task, config, buffer_answer)

    return a

def _perform(search, s, u, t, c, a):

    global buffer_search
    global buffer_session
    global buffer_dblayer
    global buffer_answer

    # progress info
    if t is not None:
        logger.info('    ACTIVE DB %s, USER %s, TASK %d, CONFIG %d' %
                    (ALL_DB[active_db]['TABLE'], u.username, t.id, c.id))
    else:
        logger.info('    ACTIVE DB %s, USER %s, CONFIG %d' %
                    (ALL_DB[active_db]['TABLE'], u.username, c.id))

    # HINT: this triggers a fresh start after saving
    search.search(input_data, buffer_timing)

    rankdata = search.rankdata
    dispdata = search.dispdata
    tracedata = search.tracedata

    a.update(rankdata, dispdata, tracedata)

    # update buffers
    buffer_search = search.data()
    buffer_session = s.data()
    buffer_answer = a.data()
    buffer_dblayer = databuffer.get_data()

# render the results
@never_cache
def _render(request, search, u, t, c, a):

    context_data = {}

    if search is not None:
        qs = search.displayinfo()
        context_data['qs'] = qs
        query = qs.get('query', '')
    else:
        query = ''

    context_data['qc'] = c.displayinfo()

    if active_mode == 'search':
        context_data['BASE'] = 'retrieval/search/base.html'

    elif active_mode == 'test':
        context_data['BASE'] = 'retrieval/search/base_test.html'
        context_data['iter_max'] = ITER_MAX
        context_data['qt'] = t.displayinfo()

        nb_sessions = nb_sessions_done(u)
        context_data['nb_sessions'] = nb_sessions

        # HINT: one could deny the access to the testing if one test is complete
        # here, the user is made aware of his progress after each complete test
        block_size = nb_sessions_per_test()
        if input_data['action'] == 'evaluation' and \
           nb_sessions % block_size == 0:

            return render_to_response(
                'retrieval/help/info_test_done.html',
                context_data,
                context_instance=RequestContext(request))

        # warning the user
        if warning_iter_max:

            return render_to_response(
                'retrieval/help/info_iter_max.html',
                context_data,
                context_instance=RequestContext(request))

        if active_mode == 'test' and search is None:
            redirect_to = reverse('retrieval:main:test')

            return HttpResponseRedirect(redirect_to)

    elif active_mode == 'train':
        context_data['BASE'] = 'retrieval/search/base_train.html'
        context_data['qt'] = t.displayinfo()

    else:
        sys.stderr.write('\nUNKNOWN MODE %s\n' % active_mode)
        sys.exit(1)

    #if a is not None:
    context_data['qa'] = a.displayinfo()

    context_data['active_mode'] = active_mode

    # pass the data to the template and render it
    return render_to_response(
        'retrieval/search/index.html',
        context_data,
        context_instance=RequestContext(request))

###########################################################################
# main functionality

# Pipeline of the operations in the view functions:
#
#  - receive the user request
#  - parse the user request
#  - get the necessary data from the cache
#  - perform the required operations
#  - prepare the template context
#  - store the modified data in the cache
#  - send the system response

# get the user input, update the webpage and, then, store the session data
def _index(request, action, data=None):

    global buffer_timing

    # get the requestArrivalTime before any other operation
    requestArrivalTime = str(datetime.now())

    logger.info('    SYSTEM START')
    start_time = now()

    # request.session.set_expiry(300)
    _get_state(request, action, data)

    buffer_timing[2] = requestArrivalTime

    # initialize the objects
    user = request.user
    config = _prepare_config(active_db)
    task = _prepare_task(active_db)
    session = _prepare_session(active_db, user, task, config)
    search, session, task, config = _prepare_search(session, user, task, config)

    if search is not None:
        answer = _prepare_answer(user, task, config)
        _perform(search, session, user, task, config, answer)
    else:
        answer = None

    # reset buffer_timing for next iteration
    buffer_timing = {}
    buffer_timing[0] = str(datetime.now())

    _set_state(request)
    render = _render(request, search, user, task, config, answer)
    _del_state()

    stop_time = now()
    logger.info('    SYSTEM DONE in %f seconds' % (stop_time - start_time))

    return render

###########################################################################
# default callback

def index(request):
    return _index(request, action='search')

#def search(request):
#    return _index(request, action='search')

@login_required
def test(request):
    return _index(request, action='test')

@login_required
def train(request):
    return _index(request, action='train')

###########################################################################
# search functions

# perform the image search algorithm
# dbid is 0 when 'NO ONE ALIKE' fdbk is given
def fdbk_im(request, dbid=0):

    images = (int32(dbid),)
    #logger.info('fdbk_im: IMAGES %s' % str(images))

    return _index(request, action='fdbk_im', data=images)

###########################################################################

def _get_state2(request, action, data):
    # get the request, and update the data

    global input_data

    global active_db
    global active_mode
    global buffer_search
    global buffer_session
    global buffer_config
    global buffer_task
    global buffer_answer

    cache = request.session

    active_db = cache.get('active_db', None)
    if active_db is None:
        sys.stderr.write('\nUNKNOWN DB %s\n' % active_db)
        sys.exit(1)

    active_mode = cache.get('active_mode', None)
    if active_mode is None:
        sys.stderr.write('\nUNKNOWN MODE %s\n' % active_mode)
        sys.exit(1)

    input_data = {}
    input_data['action'] = action

    buffer_search = cache.get('search', {})
    buffer_session = cache.get('session', {})
    buffer_config = cache.get('config', {})
    buffer_task = cache.get('task', {})
    buffer_answer = cache.get('answer', {})

    if action == 'task_new':
        buffer_t = buffer_task.get(active_mode, {})
        buffer_t['id'] = buffer_t.get('id', 0) + 1

    elif action == 'task_query':
        buffer_t = buffer_task.get(active_mode, None)
        if buffer_t is not None:
            show_query = buffer_t.get('show_query', True)
            buffer_t['show_query'] = not show_query

    elif action == 'task_target':
        buffer_t = buffer_task.get(active_mode, None)
        if buffer_t is not None:
            show_target = buffer_t.get('show_target', False)
            buffer_t['show_target'] = not show_target

    elif action == 'task_info':
        buffer_t = buffer_task.get(active_mode, None)
        if buffer_t is not None:
            show_info = buffer_t.get('show_info', False)
            buffer_t['show_info'] = not show_info

    elif action == 'config_new':
        buffer_c = buffer_config.get(active_mode, {})
        buffer_c['id'] = buffer_c.get('id', 0) + 1
        # Now the session proceeds with the newly set configuration
        # HINT: here one may like to reset the session buffers
        #buffer_search = {}
        #buffer_session = {}

    elif action == 'next':
        buffer_answer['index_from'] = data + ANSWERS_PER_PAGE
    elif action == 'prev':
        buffer_answer['index_from'] = data - ANSWERS_PER_PAGE
    elif action == 'next2':
        buffer_answer['index_from2'] = data + ANSWERS_PER_PAGE2
    elif action == 'prev2':
        buffer_answer['index_from2'] = data - ANSWERS_PER_PAGE2
    elif action == 'next4':
        buffer_answer['index_from4'] = data + ANSWERS_PER_PAGE4
    elif action == 'prev4':
        buffer_answer['index_from4'] = data - ANSWERS_PER_PAGE4
    elif action == 'next5':
        buffer_answer['index_from5'] = data + ANSWERS_PER_PAGE5
    elif action == 'prev5':
        buffer_answer['index_from5'] = data - ANSWERS_PER_PAGE5
    else:
        sys.stderr.write('\nUNKNOWN ACTION %s\n' % action)
        sys.exit(1)

def _set_state2(request):
    # store the session data

    cache = request.session

    if buffer_search:
        cache['search'] = buffer_search
    elif 'search' in cache:
        del cache['search']

    if buffer_session:
        cache['session'] = buffer_session
    elif 'session' in cache:
        del cache['session']

    cache['config'] = buffer_config
    cache['task'] = buffer_task
    cache['answer'] = buffer_answer

def _del_state2():
    #

    global input_data

    global active_db
    global active_mode
    global buffer_search
    global buffer_session
    global buffer_config
    global buffer_task
    global buffer_answer

    del input_data

    del active_db
    del active_mode
    del buffer_search
    del buffer_session
    del buffer_config
    del buffer_task
    del buffer_answer

###########################################################################

def _perform2(search, s, u, t, c, a):

    global buffer_answer

    rankdata = search.rankdata
    dispdata = search.dispdata
    tracedata = search.tracedata

    a.update(rankdata, dispdata, tracedata)

    # update buffers
    buffer_answer = a.data()

###########################################################################

def _index2(request, action, data=None):
    # get the user input, update the webpage and, then, store the session data

    logger.info('    SYSTEM START')
    start_time = now()

    _get_state2(request, action, data)

    # initialize the objects
    user = request.user
    config = _prepare_config(active_db)
    task = _prepare_task(active_db)
    session = _prepare_session(active_db, user, task, config)
    search, session, task, config = _prepare_search(session, user, task, config)
    answer = _prepare_answer(user, task, config)

    _perform2(search, session, user, task, config, answer)
    _set_state2(request)
    render = _render(request, search, user, task, config, answer)
    _del_state2()

    stop_time = now()
    logger.info('    SYSTEM DONE in %f seconds' % (stop_time - start_time))

    return render

###########################################################################
# navigation through the items (each page displays a fraction of items)

# dispdata
def next(request, index):
    index = int(index)
    return _index2(request, action='next', data=index)

def prev(request, index):
    index = int(index)
    return _index2(request, action='prev', data=index)

# rankdata
def next2(request, index):
    index = int(index)
    return _index2(request, action='next2', data=index)

def prev2(request, index):
    index = int(index)
    return _index2(request, action='prev2', data=index)

# trace plots
def next4(request, index):
    index = int(index)
    return _index2(request, action='next4', data=index)

def prev4(request, index):
    index = int(index)
    return _index2(request, action='prev4', data=index)

def next5(request, index):
    index = int(index)
    return _index2(request, action='next5', data=index)

def prev5(request, index):
    index = int(index)
    return _index2(request, action='prev5', data=index)

###########################################################################
# config functions

def config_new(request):
    # request a new configuration
    return _index2(request, action='config_new')

###########################################################################
# task functions

def task_new(request):
    # request a new task
    return _index2(request, action='task_new')

def task_query(request):
    # toggle the display of the task query
    return _index2(request, action='task_query')

def task_target(request):
    # toggle the display of the task target
    return _index2(request, action='task_target')

def task_info(request):
    # toggle the display of the task info
    return _index2(request, action='task_info')

###########################################################################
# automatic callback when webpage is loaded at the client

def update(request):
    """
    This method is executed after any action on the webpage.
    Therefore just save the timing of the first callback.
    """

    #global buffer_timing

    buffer_timing = request.session.get('timing', {})

    if 1 not in buffer_timing:
        buffer_timing[1] = str(datetime.now())

    request.session['timing'] = buffer_timing

    return HttpResponse('', mimetype='text/plain')
