#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
"""

###########################################################################
# IMPORTS

import logging
from datetime import datetime
from numpy import int32
import sys
from time import time as now

from django.conf import settings
from django.core.urlresolvers import reverse
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.shortcuts import render_to_response
from django.shortcuts import get_object_or_404
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django.template import RequestContext
from django.views.decorators.cache import never_cache

from django.contrib.auth.models import User, AnonymousUser

from retrieval.helpers.Answer import Answer
from retrieval.helpers.Watch import Watch
from retrieval.models import get_session_model, \
                                      get_task_model, \
                                      get_config_model

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

# the list of databases
ALL_DB = settings.ALL_DB

# number of results displayed on page
ANSWERS_PER_PAGE = 8

# cached variables
#active_db
#active_mode
#buffer_watch
#buffer_session
#buffer_config
#buffer_task

###########################################################################

def _init_watching():
    # only initializations

    buffer_c = buffer_config.get(active_mode, None)
    if buffer_c is None:
        buffer_c = {}
        buffer_config[active_mode] = buffer_c

    buffer_t = buffer_task.get(active_mode, None)
    if buffer_t is None:
        buffer_t = {}
        buffer_task[active_mode] = buffer_t

###########################################################################

def _get_state(request, action, data):
    # get the request, and update the data

    global active_db
    global active_mode
    global buffer_watch
    global buffer_session
    global buffer_config
    global buffer_task

    cache = request.session

    active_db = cache.get('active_db', None)
    if active_db is None:
        sys.stderr.write('\nUNKNOWN DB %s\n' % active_db)
        sys.exit(1)

    # get the configuration
    active_mode = cache.get('active_mode', None)
    buffer_config = cache.get('config', {})
    buffer_task = cache.get('task', {})

    if request.POST.get('prev_iteration', None):
        buffer_watch = cache.get('watch', {})
        buffer_session = cache.get('session', {})
        iteration = buffer_watch.get('iteration', 1)
        buffer_watch['iteration'] = iteration - 1

    elif request.POST.get('next_iteration', None):
        buffer_watch = cache.get('watch', {})
        buffer_session = cache.get('session', {})
        iteration = buffer_watch.get('iteration', 1)
        buffer_watch['iteration'] = iteration + 1

    # HINT: consider the actions AFTER the request data
    elif action == 'new_session':
        active_mode = 'watch'
        _init_watching()
        # reset for NEW session
        buffer_watch = {}
        buffer_session = {}

    # in case of a webpage refresh
    elif action is None:
        active_mode = 'watch'
        _init_watching()
        # reset for NEW session
        buffer_watch = {}
        buffer_session = {}

    else:
        sys.stderr.write('\nUNKNOWN ACTION %s\n' % action)
        sys.exit(1)

def _set_state(request):

    cache = request.session

    if active_mode:
        cache['active_mode'] = active_mode

    if buffer_watch:
        cache['watch'] = buffer_watch

    if buffer_session:
        cache['session'] = buffer_session

    if buffer_config:
        cache['config'] = buffer_config

    if buffer_task:
        cache['task'] = buffer_task

def _del_state():

    global active_db
    global active_mode
    global buffer_watch
    global buffer_session
    global buffer_config
    global buffer_task

    del active_db
    del active_mode
    del buffer_watch
    del buffer_session
    del buffer_config
    del buffer_task

###########################################################################

def _prepare_user(s=None):

    if s is None:
        u = AnonymousUser()
    else:
        u = s.user

    return u

def _prepare_config(s=None):

    buffer_c = buffer_config.get(active_mode, None)
    if buffer_c is None:
        sys.stderr.write('\nUNKNOWN CONFIG\n')
        sys.exit(1)

    elif s is None:
        Config = get_config_model(active_db)
        configid = buffer_c.get('id', 0)
        c = Config.objects.get(pk=configid)
        c.init()

    else:
        c = s.config
        c.init()
        # update buffers
        buffer_c['id'] = c.id

    return c

def _prepare_task(s=None):

    buffer_t = buffer_task.get(active_mode, None)
    if buffer_t is None:
        #t = None
        sys.stderr.write('\nUNKNOWN TASK\n')
        sys.exit(1)

    elif s is None:
        Task = get_task_model(active_db)
        taskid = buffer_t.get('id', 0)
        t = Task.objects.get(pk=taskid)
        t.init(buffer_t)

    else:
        t = s.task
        t.init(buffer_t)
        # update buffers
        buffer_t['id'] = t.id

    return t

def _prepare_session(sid=None):

    if sid is None:
        sid = buffer_session.get('id', None)

    if sid is None:
        sys.stderr.write('\nUNDEFINED SESSION %s\n' % sid)
        exit(1)

    Session = get_session_model(active_db)
    s = Session.objects.filter(pk=sid)
    if s:
        s = s[0]
        s.init(buffer_session)
    else:
        sys.stderr.write('\nUNKNOWN SESSION %s\n' % sid)
        exit(1)

    return s

def _prepare_watch(s):

    watch = Watch()
    watch.init(s, buffer_watch)

    return watch

def _prepare_answer(user, task, config):

    buffer_answer = {}
    buffer_answer['ANSWERS_PER_PAGE'] = ANSWERS_PER_PAGE

    a = Answer()
    a.init(user, task, config, buffer_answer)

    return a

###########################################################################

def _perform(watch, s, u, t, c, a, uuu):

    global buffer_session

    # progress info
    if t is not None:
        logger.info('    ACTIVE DB %s, USER %s, TASK %d, CONFIG %d' %
                    (ALL_DB[active_db]['TABLE'], u.username, t.id, c.id))
    else:
        logger.info('    ACTIVE DB %s, USER %s, CONFIG %d' %
                    (ALL_DB[active_db]['TABLE'], u.username, c.id))

    watch.update()

    rankdata = watch.rankdata
    dispdata = watch.dispdata
    tracedata = watch.tracedata

    a.update(rankdata, dispdata, tracedata)

    # update buffers
    buffer_session = s.data()

@never_cache
def _render(request, watch, u, t, c, a):

    context_data = {}

    context_data['qs'] = watch.displayinfo()
    context_data['qt'] = t.displayinfo()
    context_data['qc'] = c.displayinfo()
    context_data['qa'] = a.displayinfo()

    # pass the data to the template and render it
    return render_to_response(
        'retrieval/watch/index.html',
        context_data,
        context_instance=RequestContext(request))

###########################################################################

def _index(request, action, data):

    logger.info('    SYSTEM START')
    start_time = now()

    _get_state(request, action, data)

    # initialize the objects
    session = _prepare_session(data)
    user = _prepare_user(session)
    config = _prepare_config(session)
    task = _prepare_task(session)
    answer = _prepare_answer(user, task, config)
    watch = _prepare_watch(session)

    _perform(watch, session, user, task, config, answer, request.user)
    _set_state(request)
    render = _render(request, watch, user, task, config, answer)
    _del_state()

    stop_time = now()
    logger.info('    SYSTEM DONE in %f seconds\n\n' % (stop_time - start_time))

    return render

###########################################################################
# default callback

def index(request):
    dbid = int32(1)
    return _index(request, action='new_session', data=dbid)

@login_required
def session(request, dbid):
    dbid = int32(dbid)
    return _index(request, action='new_session', data=dbid)

def example(request):
    dbid = int32(1)
    return _index(request, action='new_session', data=dbid)

###########################################################################

def _get_state2(request, action):
    # get the request, and update the data

    global active_db
    global active_mode
    global buffer_watch
    global buffer_session
    global buffer_config
    global buffer_task

    cache = request.session

    active_db = cache.get('active_db', None)
    if active_db is None:
        sys.stderr.write('\nUNKNOWN DB %s\n' % active_db)
        sys.exit(1)

    active_mode = cache.get('active_mode', None)
    if active_mode is None:
        sys.stderr.write('\nUNKNOWN MODE %s\n' % active_mode)
        sys.exit(1)

    buffer_watch = cache.get('watch', {})
    buffer_session = cache.get('session', {})
    buffer_config = cache.get('config', {})
    buffer_task = cache.get('task', {})

    if action == 'task_target':
        buffer_t = buffer_task.get(active_mode, None)
        if buffer_t is not None:
            tmp = buffer_t.get('show_target', False)
            buffer_t['show_target'] = not tmp

    elif action == 'task_info':
        buffer_t = buffer_task.get(active_mode, None)
        if buffer_t is not None:
            tmp = buffer_t.get('show_info', False)
            buffer_t['show_info'] = not tmp

    else:
        sys.stderr.write('\nUNKNOWN ACTION %s\n' % action)
        sys.exit(1)

def _set_state2(request):

    cache = request.session

    cache['task'] = buffer_task

def _perform2(watch, a):

    watch.update()

    rankdata = watch.rankdata
    dispdata = watch.dispdata
    tracedata = watch.tracedata

    a.update(rankdata, dispdata, tracedata)

    # update buffers
    # HINT: buffers are NOT modified

###########################################################################
# get the user input, update the webpage and, then, store the data

def _index2(request, action):

    logger.info('    SYSTEM START')
    start_time = now()

    _get_state2(request, action)

    # initialize the objects
    session = _prepare_session()
    user = _prepare_user(session)
    config = _prepare_config(session)
    task = _prepare_task(session)
    answer = _prepare_answer(user, task, config)
    watch = _prepare_watch(session)

    _perform2(watch, answer)
    _set_state2(request)
    render = _render(request, watch, user, task, config, answer)
    _del_state()

    stop_time = now()
    logger.info('    SYSTEM DONE in %f seconds\n\n' % (stop_time - start_time))

    return render

###########################################################################
# task functions

def task_target(request):
    # toggle the display of the task target
    return _index2(request, action='task_target')

def task_info(request):
    # toggle the display of the task info
    return _index2(request, action='task_info')

###########################################################################
# irreversibile functions

@login_required
def delete(request, dbid):

    _get_state2(request, None)

    dbid = int32(dbid)

    #initialize the objects
    session = _prepare_session(dbid)
    session.delete()

    #_set_state2(request)
    _del_state()

    return HttpResponseRedirect(reverse('retrieval:manage:sessions'))
