#!/usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Encapsulates the functionality for the Django-view and the Html-page that
corresponds to the user intelligence on how to use the application:

- what are the actions required in a certain state
- what are the buttons that trigger the actions
"""

###########################################################################
# EXPORT

__all__ = ['first_iteration', 'fdbk_iteration']

###########################################################################
# IMPORTS

import logging
import sys
from time import time as now

from django.core.urlresolvers import reverse
from django.template import RequestContext

from retrieval_libs.assist.AutoUsr import AutoUsr

from retrieval.models import get_task_model, get_config_model

###########################################################################
# GLOBALS

logger = logging.getLogger('assistant')

# global variables
#active_db
#active_mode
#buffer_search
#buffer_session
#buffer_config
#buffer_task

###########################################################################
# cache functionality

def _get_state(response):
    """
    Get data from the cache.
    """

    global active_db
    global active_mode
    global buffer_search
    global buffer_session
    global buffer_config
    global buffer_task

    cache = response.client.session
    #print 'request', response.request

    active_db = cache.get('active_db', None)
    active_mode = cache.get('active_mode', None)
    buffer_search = cache.get('search', {})
    buffer_session = cache.get('session', {})
    buffer_config = cache.get('config', {})
    buffer_task = cache.get('task', {})

    if active_db is None:
        sys.stderr.write('\UNKNOWN DB\n')
        sys.exit(1)

    if active_mode != 'test':
        sys.stderr.write('\nMISMATCH MODE %s\n' % active_mode)
        sys.exit(1)

# HINT: automatic user do not modify the cache
#def _set_state(response):
#    """
#    Store data in the cache.
#    """

#######################################################################
#  local functionality

def _prepare_task(active_db):
    """
    Retrieve the current task from the session cache.
    """

    buffer_t = buffer_task.get(active_mode, {})
    taskid = buffer_t.get('id', None)
    if taskid is None:
        sys.stderr.write('\nUNDEFINED TASK\n')
        sys.exit(1)

    Task = get_task_model(active_db)

    task = Task.objects.filter(pk=taskid)
    if task:
        task = task[0]
        task.init(buffer_t)
    else:
        sys.stderr.write('\nUNKNOWN TASK\n')
        sys.exit(1)

    return task

def _prepare_config(active_db):
    """
    Retrieve the current task from the session cache.
    """

    buffer_c = buffer_config.get(active_mode, {})
    configid = buffer_c.get('id', None)
    if configid is None:
        sys.stderr.write('\nUNDEFINED CONFIG\n')
        sys.exit(1)

    Config = get_config_model(active_db)

    config = Config.objects.filter(pk=configid)
    if config:
        config = config[0]
        config.init()
    else:
        sys.stderr.write('\nUNKNOWN CONFIG\n')
        sys.exit(1)

    return config

###########################################################################
# main functionality

def first_iteration():
    """
    Return the AUTO user refresh action/request.
    HINT: It can be an initial query action as well.
    """

    logger.info('    BROWSER START')
    start_time = now()

    action = reverse('retrieval:main:test')
    data = {} #{'query_txt': ''}

    stop_time = now()
    logger.info('    ACTION %s' % action)
    logger.info('    BROWSER DONE in %f seconds' % (stop_time - start_time))

    return action, data

def fdbk_iteration(response):
    """
    Return the AUTO user relevance feedback action/request.
    """

    logger.info('    BROWSER START')
    start_time = now()

    _get_state(response)

    task = _prepare_task(active_db)
    config = _prepare_config(active_db)

    thargs = buffer_session['thargsH'][-1]
    weights = buffer_session['weightsH'][-1]

    dispdata = buffer_search.get('dispdata', None)
    if dispdata is None or 'imdbid' not in dispdata:
        sys.stderr.write('\nMISSING DISPDATA\n')
        sys.exit(1)

    # pass the view data to the user
    autousr = AutoUsr()
    autousr.init(task, config)
    autousr.set_thargs(thargs)
    autousr.set_weights(weights)

    # get the user answer
    imdbid = autousr.fdbk_im(None, dispdata)

    # translate the user answer into the view action
    if isinstance(imdbid, tuple):
        action = reverse('retrieval:search:fdbk_im', kwargs={'dbid': imdbid[0]})
        data = {}
    elif imdbid is None:
        action = reverse('retrieval:search:test')
        data = {u'input_evalY': u'One image is relevant!',}
    else:
        sys.stderr.write('\nUNKNOWN ACTION\n')
        sys.exit(1)

    stop_time = now()
    if logger.getEffectiveLevel() < logging.DEBUG:
        strinfo = '\n\n    BROWSER RETURNS FDBKIM %s ON DISPDATA:\n' % imdbid
        for key, value in dispdata.iteritems():
            strinfo += '    %s: %s\n' % (key, value)

        logger.debug(strinfo)

    logger.info('    ACTION %s' % action)
    logger.info('    BROWSER DONE in %f seconds' % (stop_time - start_time))

    return action, data
