#!/usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Manages the workflow of the automatic user actions and the system
responses during the automatic evaluation tests. The main user actions
that are involved are:

- start a new searching session
- provide the relevance feedback
"""

###########################################################################
# IMPORTS

import logging
import os
from time import time as now
import cProfile
import pstats

# HINT: Set up the Django-powered application.
from django.core.management import setup_environ
import heat.settings.staging as settings
setup_environ(settings)

# HINT: The test environment provides a dummy web client that can interact
#       with the Django-powered application programmatically. It simulates
#       the GET/POST requests on a URL and the corresponding HTTP responses.
from django.test.utils import setup_test_environment
setup_test_environment()

from django.test import Client

from retrieval.helpers.UserTestController import reinitialize_tasks

from heat.autoapp import browser

###########################################################################
# GLOBALS

logger = logging.getLogger('workflow')

USERNAME = settings.AUTOUSR_NAME
PASSWORD = settings.AUTOUSR_PWD
TESTS_MAX = settings.TESTS_MAX
REINIT_LOOP = settings.REINIT_LOOP
PROFILE = settings.PROFILE
PROFILE_FILE = settings.PROFILE_FILE

###########################################################################
# CLASS

class Workflow(object):
    """
    Manages the workflow of the automatic user actions and the system
    responses during the automatic evaluation tests. The main user actions
    that are involved are:

    - start a new searching session
    - provide the relevance feedback
    """

    #######################################################################
    #  local functionality

    def _first_iteration(self):
        """
        Perform the first iteration (i.e. pass the AUTO user initial query
        to the retrieval system).
        """

        logger.info('    WORKFLOW BROWSER START')
        start_time = now()

        # prepare the request
        action, data = browser.first_iteration()

        stop_time = now()
        logger.info(
                '    WORKFLOW BROWSER DONE in %f seconds\n\n'
                % (stop_time - start_time))

        logger.info('    WORKFLOW CLIENT START')
        start_time = now()

        # process the request
        # pass the request to the CLIENT, and get the response
        # HINT: self.client is referenced by response.client
        response = self.client.post(action, data, follow=True)
        #print 'response.context', response.context

        # progress info
        username = response.context['user']
        nb_sessions = response.context['nb_sessions']
        logger.info('    USERNAME: %s, SESSIONS: %d' %
                    (username, nb_sessions))

        stop_time = now()
        logger.info(
                '    WORKFLOW CLIENT DONE in %f seconds\n\n'
                % (stop_time - start_time))

        return response

    def _fdbk_iteration(self, response):
        """
        Perform one iteration (i.e. pass the AUTO user relevance feedback
        to the retrieval system)
        """

        logger.info('    WORKFLOW BROWSER START')
        start_time = now()

        # prepare request
        action, data = browser.fdbk_iteration(response)

        stop_time = now()
        logger.info(
                '    WORKFLOW BROWSER DONE in %f seconds\n\n'
                % (stop_time - start_time))

        logger.info('    WORKFLOW CLIENT START')
        start_time = now()

        # process the request
        # pass the request to the CLIENT, and get the response
        # HINT: self.client is referenced by response.client
        response = self.client.post(action, data, follow=True)

        # progress info
        username = response.context['user']
        nb_sessions = response.context['nb_sessions']
        logger.info(
                '    CONTEXT: USER %s, SESSIONS %d'
                % (username, nb_sessions))

        stop_time = now()
        logger.info(
                '    WORKFLOW CLIENT DONE in %f seconds\n\n'
                % (stop_time - start_time))

        return response

    def _test(self):
        """
        Perform the automatic user actions and the system responses.
        """

        # refresh action
        response = self._first_iteration()

        test_done = False
        while not test_done:

            warning_iter_max = False
            for template in response.template:
                # skip the warning and continue the test
                if template.name == 'retrieval/help/info_iter_max.html':
                    warning_iter_max = True

            if warning_iter_max:
                # refresh action
                response = self._first_iteration()
            else:
                # relevance feedback action
                response = self._fdbk_iteration(response)

            # EXIT criterion
            for template in response.template:
                if template.name == 'retrieval/help/info_test_done.html':
                    test_done = True

                #if PROFILE:
                #    warning_iter_max = True

    #######################################################################
    #  main functionality

    def __init__(self):

        self.client = Client()

    def perform(self):
        """
        Perform user tests with the automatic user.
        """

        self.client.login(username=USERNAME, password=PASSWORD)

        for i in xrange(0, TESTS_MAX, 1):

            if i % REINIT_LOOP == 0:# and os.getenv('SGE_TASK_ID'):

                logger.info('    WORKFLOW REINIT START')
                start_time = now()

                # randomize the tasks
                reinitialize_tasks()

                stop_time = now()
                logger.info(
                        '    WORKFLOW REINIT DONE in %f seconds\n\n'
                        % (stop_time - start_time))

            # perform one user test
            self._test()

            logger.info('    AUTOUSER DID PERFORM %d TESTS\n\n' % (i + 1))

        self.client.logout()

###########################################################################

if __name__ == '__main__':
    # launch the automatic application

    workflow = Workflow()

    if PROFILE:
        cProfile.run('workflow.perform()', PROFILE_FILE)

        profile = pstats.Stats(PROFILE_FILE)
        profile.sort_stats('time') #'cumulative', 'calls', 'time'
        #profile.print_stats(0.2, 'retrieval_app')
        profile.print_stats(0.9, 'thdist')

    else:
        workflow.perform()