#! /usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Uploads the ImageNet_1M collection.

- ImageNet collection is provided freely at <http://www.image-net.org/>
- It is assumed that the ImageNet files have beed downloaded
"""

###########################################################################
# IMPORTS

from datetime import datetime
import logging
from numpy import int32, zeros
import os
from random import Random
from scipy.io import loadmat
import sys
import tarfile

from django.core.management import setup_environ
import heat.settings.maintenance as settings
setup_environ(settings)

from retrieval_libs.dblayer.DbAccess import DbAccess
from retrieval_libs.dblayer.FileAccess import create_dir
from retrieval_libs.dbutils.dbimg_check import update_valid_url
from retrieval_libs.dbutils.dbmodels_init import init_config_table, \
                                                 init_task_table, \
                                                 init_session_table
from retrieval_libs.dbutils.dbtree_create import create_subtree
from retrieval_libs.dbutils.dbtree_update import update_children, \
                                                 update_subtree_size, \
                                                 check_children_data

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

# implemented as a singleton
dbaccess = DbAccess()

ALL_DB = settings.ALL_DB
FIGURES = settings.FIGURES
TABLE_USER = settings.TABLE_USER

# HINT: keep it commented out, to avoid any execution by mistake
#ACTIVE_DB = 'ImageNet_1M'

random = Random()
random.jumpahead(8)

# First, one should download the ImageNet files. In Linux, use wget command.
# WNIDS: http://www.image-net.org/api/text/imagenet.sbow.obtain_synset_list
# SBOWS: http://www.image-net.org/api/download/imagenet.sbow.synset?wnid=[%s]
DIR_DATA = os.path.join(settings.DIR_DATA, 'ImageNet')

SYNSETS = os.path.join(DIR_DATA, 'www.image-net.org/api/text/imagenet.sbow.obtain_synset_wordlist')
DIR_SBOW = os.path.join(DIR_DATA, 'www.image-net.org/downloads/features/sbow')
DIR_URLS = os.path.join(DIR_DATA, 'urls')
URLS = [os.path.join(DIR_DATA, 'imagenet_spring10_urls.tgz'),]

# temporary file created for convenience
WNIDS = os.path.join(DIR_DATA, 'wnids.txt')

###########################################################################
# local functionality

def _extract_wnids():
    """
    Parse the SYNSETS file and get the wnids of the collection at hand.

    EXTRA: Save separately the wnid's list in the WNIDS file.
    """

    wnids = []

    in_filep = open(SYNSETS, 'r')
    out_filep = open(WNIDS, 'w')

    #for line in out_filep:
    #    wnid = line.split('\n')[0]
    #    wnids.append(wnid)

    for i, line in enumerate(in_filep):

        # skip the empty lines, if any
        if 'wnid' not in line:
            continue

        part2 = line.split('wnid=')[1]
        wnid = part2.split('">')[0]

        wnids.append(wnid)

        out_filep.write('%s\n' % wnid)
        logger.debug('    INDEX %s WNID %s' % (i, wnid))

    in_filep.close()
    out_filep.close()

    return wnids

def _read_sbow(wnid):
    """
    Read the ImageNet *.sbow.mat file corresponding to the given wnid.
    """

    filename = DIR_SBOW + '/%s.sbow.mat' % wnid

    if not os.access(filename, os.F_OK):
        #logger.debug('    MISSING SBOW FILE: %s' % filename)
        return None

    # read Matlab data file
    X = loadmat(filename, struct_as_record=False, squeeze_me=True)

    # access data
    data = X['image_sbow']
    #dataln = len(data)
    #item = data[1]
    #itemid = item.ID
    #sbow = item.sbow.word
    #logger.debug(
    #        'SBOW ID %s DTYPE %s SHAPE %s DTYPE[1] %s'
    #        % (itemid, type(sbow), sbow.shape, type(sbow[1]))

    return data

def _sbow_to_hist(sbow):
    """
    Translate the given membership-like vector into a histogram-like vector.
    """

    sbowln = sbow.shape[0]

    # HINT: the SIFT words dictionary is of dimension 1000
    feat = zeros((1000,), dtype=int32, order='c')

    for i in xrange(0, sbowln, 1):
        feat[sbow[i]] += 1

    return feat

def _extract_urls(wnid):
    """
    Return the urls corresponding to the given wnid.
    """

    urls = {}

    filename = DIR_URLS + '/%s.urls.txt' % wnid
    urlp = open(filename, 'r')
    for line in urlp:
        content = line.split(' ')
        im_wnid = content[0]
        info = content[1]
        im_url = info.split('\n')[0]

        urls[im_wnid] = im_url

    urlp.close()
    urlp = None

    return urls

###########################################################################

def generate_url_files():
    """
    Generate the url files corresponding to sbow files.
    """

    # get the wnids of the collection at hand
    wnids = _extract_wnids()

    for filename in URLS:

        # open a compressed file for reading
        tar_filep = tarfile.open(filename, 'r|gz')

        wnid_curr = None
        url_filep = None

        # loop the compressed files/members in the tar file
        member = tar_filep.next()
        while member is not None:

            memberp = tar_filep.extractfile(member)

            # loop the lines in the compressed files/members
            for line in memberp:

                content = line.split('\t')
                info = content[0]
                wnid, imid = info.split('_')
                info = content[1]
                im_url = info.split('\n')[0]

                # detect when passing to a new wnid
                # HINT: close the current url_file and initialize a new one
                if wnid_curr != wnid:
                    wnid_curr = wnid

                    if url_filep is not None:
                        url_filep.close()
                        url_filep = None

                    if wnid_curr in wnids:
                        filename = DIR_URLS + '/%s.urls.txt' % wnid_curr
                        create_dir(filename)
                        url_filep = open(filename, 'a')
                    else:
                        url_filep = None

                # skip images that are not in the given wnid's list
                if url_filep is None:
                    continue

                url_filep.write('%s_%s %s\n' % (wnid, imid, im_url))

            wnid_curr = None
            if url_filep is not None:
                 url_filep.close()
                 url_filep = None

            memberp.close()
            member = tar_filep.next()

###########################################################################
# load functionality

def initialize_database(active_db):
    """
    Initialize the database for the ImageNet_1M collection.
    """

    TABLE_CONFIG = ALL_DB[active_db]['TABLE_CONFIG']
    TABLE_TASK = ALL_DB[active_db]['TABLE_TASK']
    TABLE_SEARCH = ALL_DB[active_db]['TABLE_SEARCH']

    # create TABLE_CONFIG, TABLE_TASK and TABLE_SEARCH
    init_config_table(TABLE_CONFIG)
    init_task_table(TABLE_TASK)
    init_session_table(TABLE_SEARCH, TABLE_USER, TABLE_TASK, TABLE_CONFIG)

    # initialize TABLE_CONFIG
    creation_time = datetime.now()

    default_configs = (
            ('HEAT-500;;k;3;', '500; FIXED', '1.0; FIXED', '2'),
            ('HEAT-1000;;y;3;', '1000; FIXED', '1.0; FIXED', '2'),
            ('HEAT-1500;;g;3;', '1500; FIXED', '1.0; FIXED', '2'),
            ('HEAT-1000, mass-zoom;;r;3;', '1000; FIXED', '1.0; ADAPTIVE', '2'),
            ('random;;k;1;', '', '1.0; FIXED', '-1'),
)

    for display_props, trace, consistency, version in default_configs:
        odbid = dbaccess.custom_query(
                "INSERT INTO %s "
                "(name, TRACE, SYS_THARGS, SYS_WEIGHTS, SYS_CONSISTENCY, SYS_VERSION, datetime) "
                "VALUES "
                "('%s', '%s', '%s', '%s', '%s', '%s', '%s')"
                % (TABLE_CONFIG,
                    display_props,
                    trace,
                   '75.0 0.06 75.0 0.29, 1.0 1.0 1.0 1.0; FIXED',
                   '1.0, 0.0; FIXED',
                   consistency,
                   version,
                   creation_time))

    # initialize TABLE_TASK
    creation_time = datetime.now()

    default_tasks = (
            ('dog/dogs in forefront', '1'),
            ('electronic device/devices', '1'),
            ('boat/boats/ferryboats', '1'),
            ('fruits & vegetables', '1'),
            ('item/items of furniture', '1'),
            ('shops & shopping centers', '1'))

    for query, target in default_tasks:
        odbid = dbaccess.custom_query(
                "INSERT INTO %s "
                "(query, target, USR_THARGS, USR_WEIGHTS, USR_VERSION, datetime) "
                "VALUES "
                "('%s', '%s', '%s', '%s', '%s', '%s')"
                % (TABLE_TASK,
                   query,
                   target,
                   '75.0 0.06 75.0 0.29, 1.0 1.0 1.0 1.0; FIXED',
                   '1.0, 0.0; FIXED',
                   'IDEAL',
                   creation_time))

    # initialize dbimg
    dbimg = dbaccess.dbimg()
    dbimg.clear_table()
    dbimg.init_table()

    # initialize dbtree
    dbtree = dbaccess.dbtree()
    dbtree.clear_table()
    dbtree.init_table()

def upload_database(dbimg):
    """
    Upload the ImageNet_1M collection.

    HINT: Parse and upload the sbow files into the dbimg table.
    """

    # in case of a crash, one can continue the uploading from a restart point
    flag_skip_0 = True
    flag_skip_1 = True
    WNID_RESTART = 'n02134084'
    IMID_RESTART = 'n02134084_5858'

    # get the wnids of the collection at hand
    wnids = _extract_wnids()

    for i, wnid in enumerate(wnids):

        # restart point
        if flag_skip_0:
            if wnid == WNID_RESTART:
                flag_skip_0 = False
                # do NOT 'continue' here
            else:
                logger.debug('    SKIPPED INDEX %s WNID %s' % (i, wnid))
                continue

        # get urls
        urldata = _extract_urls(wnid)

        if not urldata:
            logger.debug('    MISSING URL FILE %s' % wnid)
            continue

        # get sbows
        sbowdata = _read_sbow(wnid)
        if sbowdata is None:
            logger.debug('    MISSING SBOW FILE %s' % wnid)
            continue

        # load dbimg
        for ii in xrange(0, len(sbowdata), 1):

            im_sbowdata = sbowdata[ii]

            im_wnid = im_sbowdata.ID

            # restart point
            if flag_skip_1:
                if im_wnid == IMID_RESTART:
                    flag_skip_1 = False
                    continue
                else:
                    continue

            # check url compatibility
            im_url = urldata.get(im_wnid, None)
            if im_url is None:
                logger.debug('    MISSING URL %s' % im_url)
                continue
            elif '%' in im_url or '\'' in im_url or '"' in im_url:
                logger.debug('    INCOMPATIBLE URL %s' % im_url)
                continue
            else:
                try:
                    im_url = im_url.encode('utf-8')
                except UnicodeDecodeError:
                    logger.debug('    INCOMPATIBLE URL %s' % im_url)
                    continue

                if len(im_url) > 255:
                    logger.debug('    INCOMPATIBLE URL %s' % im_url)
                    continue

            # image feature vectors
            im_sbow = im_sbowdata.sbow.word
            im_feat = _sbow_to_hist(im_sbow)
            im_feat0 = ' '.join([str(item) for item in im_feat])

            im_feat1 = '0 0'

            # file with the pre-computed image similarity distances
            im_distfile = im_wnid.replace('_', '/')

            # insert in the database
            dbimg.insert_img(im_url, 0, im_distfile, im_feat0, im_feat1)
            logger.debug(
                    '    INSERTED INDEX %s/%s WNID %s'
                    % (i, ii, im_wnid))

        del urldata
        del sbowdata

###########################################################################
# the main if we test/execute this module

if __name__ == '__main__':

    dbaccess.init(ALL_DB, FIGURES, ACTIVE_DB)

    #generate_url_files()

    # creates the tables in the database
    #initialize_database(ACTIVE_DB)

    dbimg = dbaccess.dbimg()
    dbtree = dbaccess.dbtree()

    # upload the data into the tables
    #upload_database(dbimg)

    # HINT: time consuming operation
    #update_valid_url(dbimg)

    # create the hierarchical organization
    # HINT: the root node has dbid = 1
    dbtree.insert_root()
    create_subtree(dbimg, dbtree, [1,])

    # optimize the tables
    # HINT: do not optimize the tables right at the creation time since there
    #       is no need to maintain the table_indexes during the uploading
    dbimg.optimize_table()
    dbtree.optimize_table()

    # complete the database with redundant data
    update_subtree_size(dbtree, 1)
    update_children(dbtree, [1,])
    #check_children_data(dbtree)
