#! /usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Uploads the ImageNet_60K collection.

- ImageNet_60K is generated by sub-sampling ImageNet_1M
- It is assumed that ImageNet_1M was already set up
"""

###########################################################################
# IMPORTS

from datetime import datetime
import logging
from random import Random

from django.core.management import setup_environ
import heat.settings.maintenance as settings
setup_environ(settings)

from retrieval_libs.dblayer.DbAccess import DbAccess
from retrieval_libs.dbutils.dbimg_distdata import initialize_distdata, \
                                                  compute_distdata
from retrieval_libs.dbutils.dbmodels_init import init_config_table, \
                                                 init_task_table, \
                                                 init_session_table
from retrieval_libs.dbutils.dbtree_create import create_subtree
from retrieval_libs.dbutils.dbtree_update import update_children, \
                                                 update_subtree_size, \
                                                 check_children_data

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

# implemented as a singleton
dbaccess = DbAccess()

ALL_DB = settings.ALL_DB
FIGURES = settings.FIGURES
TABLE_USER = settings.TABLE_USER

# HINT: keep it commented out, to avoid any execution by mistake
#ACTIVE_DB = 'ImageNet_60K'

# the large collection to copy from and the percentage to copy
LARGE_TABLE_IMG = 'ImageNet_1M_img'
TABLE_SIZE = 60000

random = Random()
random.jumpahead(8)

###########################################################################
# local functionality

def initialize_database(active_db):
    """
    Initialize the database for the ImageNet_33K collection.
    """

    TABLE_CONFIG = ALL_DB[active_db]['TABLE_CONFIG']
    TABLE_TASK = ALL_DB[active_db]['TABLE_TASK']
    TABLE_SEARCH = ALL_DB[active_db]['TABLE_SEARCH']

    # create TABLE_CONFIG, TABLE_TASK and TABLE_SEARCH
    init_config_table(TABLE_CONFIG)
    init_task_table(TABLE_TASK)
    init_session_table(TABLE_SEARCH, TABLE_USER, TABLE_TASK, TABLE_CONFIG)

    # initialize TABLE_CONFIG
    creation_time = datetime.now()

    default_configs = (
            ('original;;k;3;', '', '1.0; FIXED', '0'),
            ('original, mass-zoom;;r;3;', '', '1.0; ADAPTIVE', '0'),
            ('random;;k;1;', '', '1.0; FIXED', '-1'),
    )

    for display_props, trace, consistency, version in default_configs:
        odbid = dbaccess.custom_query(
                "INSERT INTO %s "
                "(name, TRACE, SYS_THARGS, SYS_WEIGHTS, SYS_CONSISTENCY, SYS_VERSION, datetime) "
                "VALUES "
                "('%s', '%s', '%s', '%s', '%s', '%s', '%s')"
                % (TABLE_CONFIG,
                    display_props,
                    trace,
                    '75.0 0.06 75.0 0.29, 1.0 1.0 1.0 1.0; FIXED',
                    '1.0, 0.0; FIXED',
                    consistency,
                    version,
                    creation_time))

    # initialize TABLE_TASK
    creation_time = datetime.now()

    default_tasks = (
            ('dog/dogs in forefront', ''),
            ('electronic device/devices', ''),
            ('boat/boats/ferryboats', ''),
            ('fruits & vegetables', ''),
            ('item/items of furniture', ''),
            ('shops & shopping centers', ''))

    for query, target in default_tasks:
        odbid = dbaccess.custom_query(
                "INSERT INTO %s "
                "(query, target, USR_THARGS, USR_WEIGHTS, USR_VERSION, datetime) "
                "VALUES "
                "('%s', '%s', '%s', '%s', '%s', '%s')"
                % (TABLE_TASK,
                    query,
                    target,
                    '75.0 0.06 75.0 0.29, 1.0 1.0 1.0 1.0; FIXED',
                    '1.0, 0.0; FIXED',
                    'IDEAL',
                    creation_time))

    # initialize dbimg
    dbimg = dbaccess.dbimg()
    dbimg.clear_table()
    dbimg.init_table()

    # initialize dbtree
    dbtree = dbaccess.dbtree()
    dbtree.clear_table()
    dbtree.init_table()

def upload_database():
    """
    Generate and load a limited collection.

    HINT: It samples randomly without considering the wnid categories.
    """

    # get all the images with valid urls
    answers = dbaccess.custom_query(
                "SELECT id FROM %s WHERE valid = 1"
                % LARGE_TABLE_IMG)
    dbid = [item[0] for item in answers]
    # get the size of the collection at hand
    dbidln = len(dbid)
    # get the size of the limited collection
    limited_dbidln = min(TABLE_SIZE, dbidln)
    logger.debug('    DBIDLN %s' % limited_dbidln)

    # select randomly the images for the limited collection
    limited_dbid = random.sample(dbid, limited_dbidln)

    # insert into the database
    for idbid in limited_dbid:
        cdbid = dbaccess.custom_query(
                "INSERT INTO %s_img ("
                "valid, pathToOrig, pathToDist, featdata0, featdata1"
                ") SELECT "
                "valid, pathToOrig, pathToDist, featdata0, featdata1 "
                "FROM %s WHERE id = %d"
                % (dbaccess.dbtable(), LARGE_TABLE_IMG, idbid))

        logger.debug('    INSERTED DBID %s' % cdbid)

###########################################################################
# the main if we test/execute this module

if __name__ == '__main__':

    dbaccess.init(ALL_DB, FIGURES, ACTIVE_DB)

    #initialize_database(ACTIVE_DB)

    dbimg = dbaccess.dbimg()
    dbtree = dbaccess.dbtree()

    # upload the data into the tables
    upload_database()

    # create the hierarchical organization
    # HINT: the root node has dbid = 1
    dbtree.insert_root()
    create_subtree(dbimg, dbtree, [1,])

    # optimize the tables
    # HINT: do not optimize the tables right at the creation time since there
    #       is no need to maintain the table_indexes during the uploading
    dbimg.optimize_table()
    dbtree.optimize_table()

    # complete the database with redundant data
    update_subtree_size(dbtree, 1)
    update_children(dbtree, [1,])
    #check_children_data(dbtree)

    # pre-compute the image similarity distances
    initialize_distdata(dbimg)
    compute_distdata(dbimg)
