#! /usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Uploads the ImageSynthetic collection, where each image contains a 2D point
and the image features are the 2D coordinates of that point.

The hierarchical organization of the collection is as follows. All the points
are within the square (0.0, 0.0, 1.0, 1.0), and this is the root square. For
any node, its children are obtained by dividing its square into 4 squares, and
its representative image is the center of its square.

For example, the root node represents the full square (0.0, 0.0, 1.0, 1.0),
and its representative image is the central point (0.5, 0.5). Then, its 4
children nodes would represent the squares:

- (0.0, 0.0, 0.5, 0.5) - lower left
- (0.0, 0.5, 0.5, 1.0) - upper left
- (0.5, 0.5, 1.0, 1.0) - upper right
- (0.5, 0.0, 1.0, 0.5) - lower right
"""

###########################################################################
# IMPORTS

from datetime import datetime
import logging

from django.core.management import setup_environ
import heat.settings.maintenance as settings
setup_environ(settings)

from retrieval_libs.dblayer.DbAccess import DbAccess
from retrieval_libs.dbutils.dbimg_distdata import initialize_distdata, \
                                                  compute_distdata
from retrieval_libs.dbutils.dbmodels_init import  init_config_table, \
                                                  init_task_table, \
                                                  init_session_table
from retrieval_libs.dbutils.dbtree_update import update_children, \
                                                 update_subtree_size, \
                                                 check_children_data

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

# implemented as a singleton
dbaccess = DbAccess()

ALL_DB = settings.ALL_DB
FIGURES = settings.FIGURES
TABLE_USER = settings.TABLE_USER

# HINT: keep it commented out, to avoid any execution by mistake
#ACTIVE_DB = 'ImageSynthetic_22K'

###
XSIZE = 2
YSIZE = 2
XRES = 1.0 / 256.0
YRES = 1.0 / 256.0

###########################################################################
# main functionality

def initialize_database(active_db):
    """
    Initialize the database for the ImageSynthetic collection.
    """

    TABLE_CONFIG = ALL_DB[active_db]['TABLE_CONFIG']
    TABLE_TASK = ALL_DB[active_db]['TABLE_TASK']
    TABLE_SEARCH = ALL_DB[active_db]['TABLE_SEARCH']

    # create TABLE_CONFIG, TABLE_TASK and TABLE_SEARCH
    init_config_table(TABLE_CONFIG)
    init_task_table(TABLE_TASK)
    init_session_table(TABLE_SEARCH, TABLE_USER, TABLE_TASK, TABLE_CONFIG)

    # initialize TABLE_CONFIG
    creation_time = datetime.now()

    default_configs = (
            ('original;;k;3;', '', '1.0; FIXED', '0'),
            ('HEAT-100;;r;3;', '100; FIXED', '1.0; FIXED', '2'),
            ('HEAT-500;;r;3;', '500; FIXED', '1.0; FIXED', '2'),
            ('original, mass-zoom;;k;3;', '', '1.0; ADAPTIVE', '0'),
            ('HEAT-100, mass-zoom;;r;3;', '100; FIXED', '1.0; ADAPTIVE', '2'),
            ('random;;k;1;', '', '1.0; FIXED', '-1'),
)

    for display_props, trace, consistency, version in default_configs:
        odbid = dbaccess.custom_query(
                "INSERT INTO %s "
                "(name, TRACE, SYS_THARGS, SYS_WEIGHTS, SYS_CONSISTENCY, SYS_VERSION, datetime) "
                "VALUES "
                "('%s', '%s', '%s', '%s', '%s', '%s', '%s')"
                % (TABLE_CONFIG,
                    display_props,
                    trace,
                    '0.10 0.06 0.10 0.29, 1.0 1.0 1.0 1.0; FIXED',
                    '1.0, 0.0; FIXED',
                    consistency,
                    version,
                    creation_time))

    # initialize TABLE_TASK
    creation_time = datetime.now()

    default_tasks = (
            ('right upper corner', '21845'),
            ('central point', '12737'))

    for query, target in default_tasks:
        odbid = dbaccess.custom_query(
                "INSERT INTO %s "
                "(query, target, USR_THARGS, USR_WEIGHTS, USR_VERSION, datetime) "
                "VALUES "
                "('%s', '%s', '%s', '%s', '%s', '%s')"
                % (TABLE_TASK,
                    query,
                    target,
                    '0.10 0.06 0.10 0.29, 1.0 1.0 1.0 1.0; FIXED',
                    '1.0, 0.0; FIXED',
                    'IDEAL',
                    creation_time))

    # initialize TABLE_SEARCH
    # HINT: Insert a search session as an example.
    TABLE_SEARCH = ALL_DB[active_db]['TABLE_SEARCH']

    action = "(('new_rnd', None), ('fdbk_im', 20292), ('fdbk_im', 20291), ('fdbk_im', 21630), ('fdbk_im', 21613), ('fdbk_im', 21622), ('fdbk_im', 21632), ('fdbk_im', 21629), ('evaluation', 'Y'),)"

    dispfdbk = "None, None; 6064 19045 20292 8929 1552 3743 15400 14414, 20292; 20291 17269 17942 10042 7198 13171 4829 15, 20291; 20293 3066 9413 7466 14042 12804 9587 21630, 21630; 20294 21613 6233 5067 724 10344 14765 16710, 21613; 21622 20226 1522 11443 8363 13455 16187 4370, 21622; 21632 20277 10941 5496 15842 1311 16435 7875, 21632; 21629 20478 2771 4479 12348 13672 6864 17861, 21629; 21669 19755 528 18230 11615 15105 6644 1858, None"

    trace = "((21845, 0, 0, 0, 0), (21845, 0, 0, 0, 0), (21845, 0, 0, 0, 0), (21845, 0, 0, 0, 0), (21845, 0, 0, 0, 0), (21845, 0, 0, 0, 0), (21845, 0, 0, 0, 0), (21845, 0, 0, 0, 0), None,)"

    timing = "((None, None, '2011-10-30 12:54:01.522901'), ('2011-10-30 12:54:02.829759', None, '2011-10-30 12:54:15.568015'), ('2011-10-30 12:54:17.022851', None, '2011-10-30 12:54:26.886429'), ('2011-10-30 12:54:28.443419', None, '2011-10-30 12:54:37.684007'), ('2011-10-30 12:54:38.750302', None, '2011-10-30 12:54:44.646245'), ('2011-10-30 12:54:45.784500', None, '2011-10-30 12:55:01.050068'), ('2011-10-30 12:55:02.164400', None, '2011-10-30 12:55:15.415544'), ('2011-10-30 12:55:16.481489', None, '2011-10-30 12:55:32.441099'), ('2011-10-30 12:55:33.804870', None, '2011-10-30 12:55:42.914236'),)"

    creation_time = "2011-10-30 12:54:01"
    #creation_time = datetime.now()

    odbid = dbaccess.custom_query(
                "INSERT INTO %s "
                "(id, user_id, task_id, config_id, iterations, evaluation, "
                "action, dispfdbk, trace, timing, datetime) "
                "VALUES "
                "(%d, %d, %d, %d, %d, '%s', '%s', '%s', '%s', '%s', '%s')"
                % (TABLE_SEARCH,
                    1, 1, 1, 1, 7, 'Y',
                    action, dispfdbk, trace, timing, creation_time))

    # initialize dbimg
    dbimg = dbaccess.dbimg()
    dbimg.clear_table()
    dbimg.init_table()

    # initialize dbtree
    dbtree = dbaccess.dbtree()
    dbtree.clear_table()
    dbtree.init_table()

def upload_database(dbid, name, x0, y0, x1, y1):
    """
    Upload the ImageSynthetic collection.

    HINT: Create and insert the given dbid and recursively, its children.
    """

    dbimg = dbaccess.dbimg()
    dbtree = dbaccess.dbtree()

    # the image feature vector
    xcoord = (x1 + x0) / 2.0
    ycoord = (y1 + y0) / 2.0
    imfeat = '%.10f %.10f' % (xcoord, ycoord)

    # the radius of the node
    dx = (x1 - x0) / 2.0
    dy = (y1 - y0) / 2.0
    radius = '(%.10f, %.10f)' % (dx, dy)

    if dx < XRES or dy < YRES:
        # exit the recursive routine
        return None

    # create the seed image
    imfile = name + '.png'
    ddfile = name + '.bin'
    seed_dbid = dbimg.insert_img(imfile, 0, ddfile, imfeat, '0 0')

    # create the corresponding node
    if dbid == 0:
        # insert the root node
        cdbid = dbtree.insert_node(0, '0', seed_dbid, radius, dbid, 0, '', '')
    else:
        # insert the node
        cdbid = dbtree.insert_node(-1, '', seed_dbid, radius, dbid, 0, '', '')

    # call the routines for the children
    children = []
    for i in xrange(0, XSIZE, 1):
        for j in xrange(0, YSIZE, 1):

            cname = '%s/%d%d' % (name, i, j)

            cx0 = x0 + i * dx
            cy0 = y0 + j * dy
            cx1 = cx0 + dx
            cy1 = cy0 + dy

            idbid = upload_database(cdbid, cname, cx0, cy0, cx1, cy1)

            # book the children nodes
            if idbid is not None:
                children.append(idbid)

    # update the node with the created children data
    dbtree.update_children_data(cdbid, children)

    logger.info('    INSERTED DBID %s' % cdbid)

    # return the inserted node
    return cdbid

###########################################################################
# the main if we test/execute this module

if __name__ == '__main__':

    dbaccess.init(ALL_DB, FIGURES, ACTIVE_DB)

    logger.info('    INITIALIZE DB - CREATE TABLES')
    initialize_database(ACTIVE_DB)

    dbimg = dbaccess.dbimg()
    dbtree = dbaccess.dbtree()

    logger.info('    UPLOAD DB - GENERATE SIZE %s' % COLLECTION_SIZE)
    logger.info('    UPDATE DB - CREATE HIERARCHICAL TREE')
    # generate the collection starting from the root node and
    # upload the data into the tables
    # HINT: The hierarchical organization is created in this step too.
    upload_database(0, '00', 0.0, 0.0, 1.0, 1.0)

    logger.info('    OPTIMIZE DB - BTREE INDEXES AND UNIQUE KEYS')
    # HINT: do not optimize the tables right at the creation time since there
    #       is no need to maintain the table_indexes during the uploading
    dbimg.optimize_table()
    dbtree.optimize_table()

    logger.info('    UPDATE DB - FILL-IN REDUNDANT DATA')
    update_subtree_size(dbtree, 1)
    update_children(dbtree, [1,])
    #check_children_data(dbtree)

    logger.info('    UPDATE DB - PRE-COMPUTE SIMILARITY DISTANCES')
    initialize_distdata(dbimg)
    compute_distdata(dbimg)
