#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Generates a target. This is used to reinitialize the tasks for the
automatic user test.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, float32, array
from random import Random
import string
import sys
from time import time as now

from retrieval_libs.dblayer.DbAccess import DbAccess

###########################################################################
# GLOBALS

logger = logging.getLogger('assistant')

# implemented as a singleton
dbaccess = DbAccess()

random = Random()
random.jumpahead(8)

###########################################################################
# CLASS

class AutoTarget(object):
    """
    Generates a target. This is used to reinitialize the tasks for the
    automatic user test.
    """

    #######################################################################
    # target based on the hierarchical tree

    def _leaf_image_rnd(self, table):
        """
        Returns the image of a random leaf node.
        """

        query = "SELECT count(*) FROM %s_img" % table
        answer = dbaccess.custom_query(query)
        size = answer[0][0]

        target = None
        while target is None:

            dbid = random.randint(1, size)

            answer = dbaccess.custom_query(
                    "SELECT id, valid "
                    "FROM %s_img "
                    "WHERE id=%d"
                    % (table, dbid))

            if answer[0][1] == 1:
                try:
                    answer = dbaccess.custom_query(
                            "SELECT id, children_size "
                            "FROM %s_tree "
                            "WHERE seed_dbid=%d"
                            % (table, dbid))

                    # if there is no sub-tree
                    if len(answer) == 0 or answer[0][1] == 0:
                        target = dbid

                except:
                    # if not a representative image in the tree
                    target = dbid

        target = [target,]
        #print 'target', len(target), target

        return target

    def _node_stats(self, table, level, omega_min, omega_max):
        """
        Computes some statistics at the given level in the hierarchical tree.
        """

        query = "SELECT id FROM %s_tree WHERE depth=%d" % (table, level)
        answer = dbaccess.custom_query(query)
        tndbid = [item[0] for item in answer]

        dbtree = dbaccess.dbtree()
        omegalns = []
        radiuses = []
        totalc = len(tndbid)
        limitc = 0

        print 'table', table, 'level', level, 'nodes', totalc

        for dbid in tndbid:
            # return the size of \Omega(N) set of images
            omegaln = dbtree.subtree_size(dbid)
            omegalns.append(omegaln)
            # return the radius of \Omega(N) set of images
            radius = dbtree.radius(dbid)
            radiuses.append(radius)

            if omega_min <= omegaln and omegaln <= omega_max:
                limitc += 1

        omegalns = array(omegalns, dtype=int32, order='c')
        if not isinstance(radiuses[0], float32):
            radiuses = []
        radiuses = array(radiuses, dtype=int32, order='c')

        print 'omegalnarr', omegalns.mean(), omegalns.var(), omegalns.std()
        print 'radiusarr', radiuses.mean(), radiuses.var(), radiuses.std()
        print 'all nodes', totalc, 'nodes within limits', limitc

    def _node_omega(self, table, level, omega_min, omega_max):
        """
        Choose randomly a node N at the given level and within the given
        size limits and return the set of images \Omega(N)
        """

        dbtree = dbaccess.dbtree()

        answers = dbaccess.custom_query(
            "SELECT id, seed_dbid FROM %s_tree WHERE depth=%d" %
            (table, level))

        omegaln = 0
        counter = 0
        while omegaln < omega_min or omegaln > omega_max:
            selectedanswer = random.sample(answers, 1)
            tndbid = [int32(selectedanswer[0][0])]
            imdbid = int32(selectedanswer[0][1])

            omega = dbtree.omega(tndbid)
            omega = [int32(item) for item in omega]
            omegaln = len(omega)

            counter += 1
            if counter % 1000 == 0:
                sys.stderr.write('\nNODE SAMPLING EXCEEDED %s\n' % counter)
                self._node_stats(table, level, omega_min, omega_max)
                #sys.stderr.write('\nNODE SAMPLING WAS UNSUCCESSFUL\n')
                #sys.exit(1)

        omega.remove(imdbid)
        omega.insert(0, imdbid)

        return omega

    #######################################################################
    # main functionality

    def init(self, task):
        """
        Initialize the object.
        """

        self.taskid = task.id

    def generate_new_target(self, ALL_DB, FIGURES, active_db):
        """
        Generate a new target.
        """

        #logger.debug('START')
        #start_time = now()

        dbaccess.init(ALL_DB, FIGURES, active_db)

        db_name = dbaccess.dbtable()

        if db_name.startswith('ImageNet_1M',):
            #target = self._leaf_image_rnd(db_name)
            target = self._node_omega(db_name, level=5, omega_min=200, omega_max=300)
            #self._node_stats(db_name, level=5, omega_min=200, omega_max=300)

        elif db_name.startswith('ImageNet_33K') or \
             db_name.startswith('ImageNet_60K'):
            target = self._leaf_image_rnd(db_name)
            #target = self._node_omega(db_name, level=4, omega_min=1, omega_max=10000)
            #self._node_stats(db_name, level=3, omega_min=10, omega_max=700)

        elif db_name.startswith('ImageSynthetic_22K'):
            #target = self._leaf_image_rnd(db_name)
            target = self._node_omega(db_name, level=4, omega_min=1, omega_max=1000)
            #self._node_stats(db_name, level=3, omega_min=1, omega_max=100)

        else:
            sys.stderr.write('\nUNKNOWN TABLE %s\n' % db_name)
            sys.exit(1)

        target = array(target, dtype=int32, order='c')

        #stop_time = now()
        #logger.debug('DONE in %f seconds' % (stop_time - start_time))

        return target
