#!/usr/bin/python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Simulates the user relevance feedback decisions.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, int64, ndarray, array
import random
import string
import sys
from time import time as now

from retrieval_libs.assist.Behavior import IdeBehavior, RndBehavior
from retrieval_libs.dblayer.DbAccess import DbAccess

###########################################################################
# GLOBALS

logger = logging.getLogger('assistant')

# implemented as a singleton
dbaccess = DbAccess()

###########################################################################
# CLASS

class AutoUsr(object):
    """
    Simulates the user relevance feedback decisions.

    HINT: The distinction between AutoUsr and Behavior is not very strong.
    Behavior subclasses could be AutoUsr subclasses, but maybe we would like
    to simulate an AutoUsr with varying Behavior.
    """

    #######################################################################
    # main functionality

    def init(self, task, config):
        """
        Initialize the object.
        """

        self.id = task.id
        #self.query = task.query
        self.target = task.target
        if not isinstance(self.target, ndarray):
            logger.error('    UNKNOWN TARGET TYPE %s' % type(self.target))
            sys.stderr.write('\nTask %s does not define the target.\n\n' % self.id)
            sys.exit(1)

        dbimg = dbaccess.dbimg_autousr()

        # create the UsrBehavior object
        mode = task.USR_VERSION
        if mode == 'IDEAL':
            self.behavior = IdeBehavior()
        elif mode == 'RANDOM':
            self.behavior = RndBehavior()
        else:
            sys.stderr.write('\nUNKNOWN USR_VERSION %s\n' % mode)
            sys.exit(1)

        self.behavior.init(dbimg, task)

    def set_thargs(self, thargs):
        """
        Fwd the thargs.
        """

        self.behavior.set_thargs(thargs)

    def set_weights(self, weights):
        """
        Fwd the weights.
        """

        self.behavior.set_weights(weights)

    #######################################################################
    # fdbk functionality

    def fdbk_doc(self, rankdata, dispdata):
        """
        Give [doc] relevance feedback on the displayed images.
        """

        raise NotImplementedError

    def fdbk_im(self, rankdata, dispdata):
        """
        Give [img] relevance feedback on the displayed images.
        """

        dispdbid = dispdata['imdbid']

        #if rankdata is not None:
        #    rankdbid = rankdata['imdbid']
        #    dispdbid = dispdbid + rankdbid[0:3]

        # check if one image of target is in dispdbid
        # loop over dispdbid that is presumably smaller than target
        for imdbid in dispdbid:
            if imdbid in self.target:
                logger.info(
                        '    TASK %d, TARGET (ID%d, LN%d), TARGET WAS REACHED'
                        % (self.id, self.target[0], len(self.target)))
                return None

        # determine relevance feedback on dispdbid
        fdbk_dbid, fdbk_dist = self.behavior.fdbk_im(self.target, dispdbid)

        # progress info
        if fdbk_dbid == '':
            logger.info(
                    '    TASK %d, TARGET (ID%d, LN%d), NEW RANDOM SET'
                    % (self.id, self.target[0], len(self.target)))
        elif fdbk_dbid == 0:
            logger.info(
                    '    TASK %d, TARGET (ID%d, LN%d), FDBK NO ONE ALIKE, DDW %s'
                    % (self.id, self.target[0], len(self.target), str(fdbk_dist)))
        elif fdbk_dbid is not None:
            logger.info(
                    '    TASK %d, TARGET (ID%d, LN%d), FDBK %s, DDW %s'
                    % (self.id, self.target[0], len(self.target), str(fdbk_dbid), str(fdbk_dist)))
        else:
            sys.stderr.write('\nUNKNOWN FDBK %s\n' % fdbk_dbid)
            sys.exit(1)

        return fdbk_dbid

    def fdbk_order(self, dispdata):
        """
        Return the sorted indexes for fdbkdbid, ordered on the minimum
        distances to any image in refdbid, in increasing order.
        """

        dispdbid = dispdata.get('imdbid', None)

        fdbk_order = self.behavior.fdbk_order(self.target, dispdbid)

        return fdbk_order
