#!/usr/bin/python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Behavior       -     Gives the relevance feedback (i.e. selects
                     a relevant image from the given set of images).

IdeBehavior    -     Gives the ideal relevance feedback.
RndBehavior    -     Gives a random relevance feedback.
"""

###########################################################################
# IMPORTS

import logging
from numpy import float32, array
from random import Random
import sys

from retrieval_libs.common.thdist cimport fdbkdist

###########################################################################
# GLOBALS

logger = logging.getLogger('assistant')

random = Random()
random.jumpahead(8)

###########################################################################
# CLASS

cdef class Behavior(object):
    """
    Gives the relevance feedback (i.e. selects a relevant image
    from the given set of images).
    """

    #######################################################################
    # local functionality

    cdef tuple _allthargs(self):
        # return the set of all possible thargs

        if self.thargs_type == 'FIXED':
            thargs = self.thargs_init

        elif self.thargs_type == 'SYNCHRON':
            thargs = self.thargs_curr

        else:
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % self.thargs_type)
            sys.exit(1)

        #logger.info('AUTOUSR THARGS %s' % (', '.join(map(str, thargs))))

        return thargs

    cdef cnp.ndarray _allweights(self):
        # return the set of all possible weights

        if self.weights_type == 'FIXED':
            weights = self.weights_init

        elif self.weights_type == 'SYNCHRON':
            weights = self.weights_curr

        else:
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % self.weights_type)
            sys.exit(1)

        #logger.info('AUTOUSR WEIGHTS %s' % (', '.join(map(str, weights))))

        return weights

    #######################################################################
    # main functionality

    def init(self, dbimg, task):
        """
        Initialize the object.

        HINT: The task is known here only because it was misused for
        the automatic user settings.
        """

        self.dbimg = dbimg

        self.thargs_init = task.USR_THARGS[0]
        self.thargs_type = task.USR_THARGS[1]
        self.thargs_curr = None

        self.weights_init = task.USR_WEIGHTS[0]
        self.weights_type = task.USR_WEIGHTS[1]
        self.weights_curr = None

    def set_thargs(self, thargs):

        self.thargs_curr = thargs

    def set_weights(self, weights):

        self.weights_curr = weights

    #######################################################################

    def fdbk_im(self, refdbid, fdbkdbid):
        """
        Return the relevance feedback on fdbkdbid,
        given the target set of images refdbid.
        """

        raise NotImplementedError

    def fdbk_order(self, refdbid, fdbkdbid):
        """
        Return the sorted indexes for fdbkdbid, ordered on the minimum
        distances to any image in refdbid, in increasing order.
        """

        raise NotImplementedError

###########################################################################
# CLASS

cdef class IdeBehavior(Behavior):
    """
    Gives the ideal relevance feedback.
    """

    def fdbk_im(self, refdbid, fdbkdbid):
        """
        Return the relevance feedback on fdbkdbid,
        given the target set of images refdbid.

        The closest image (from fdbkdbid) to refdbid is chosen.
        If necessary, ONE image is chosen randomly.
        """

        dd = self.dbimg.distdata(refdbid[0:1], fdbkdbid, -1, False, False)

        allthargs = self._allthargs()
        allweights = self._allweights()
        dispdist = fdbkdist(dd, allthargs, allweights)

        mindist = dispdist[0]
        candidates = []

        for idbid, idist in zip(fdbkdbid, dispdist):
            if mindist > idist:
                mindist = idist
                # reset candidates
                candidates = []
                candidates.append(idbid)
            elif mindist == idist:
                candidates.append(idbid)
            #else:
            #    pass

        nb_candidates = len(candidates)
        if nb_candidates > 0:
            # choose randomly one candidate
            fdbk_dbid = random.choice(candidates)
        else:
            sys.stderr.write('\nNO CANDIDATE\n')
            sys.exit(1)

        return (fdbk_dbid,), mindist

    def fdbk_order(self, refdbid, fdbkdbid):
        """
        Return the sorted indexes for fdbkdbid, ordered on the minimum
        distances to any image in refdbid, in increasing order.

        The minimum distances are the minimum for any weights.
        """

        dd = self.dbimg.distdata(refdbid[0:1], fdbkdbid, -1, False, False)

        allthargs = self._allthargs()
        allweights = self._allweights()
        dispdist = fdbkdist(dd, allthargs, allweights)

        # sorted arguments, in increasing order
        sortedarg = dispdist.argsort()

        fdbkorder = []
        #fdbkorder = zeros(sortedarg.shape, dtype=int, order='C')
        for i, index in enumerate(sortedarg):
            idbid = fdbkdbid[index]
            fdbkorder.append((idbid, i))
            #fdbkorder[index] = i

        return dict(fdbkorder)
        #return tuple(fdbkorder)

###########################################################################
# CLASS

cdef class RndBehavior(Behavior):
    """
    Gives a random relevance feedback.
    """

    def fdbk_im(self, refdbid, fdbkdbid):
        """
        Return the relevance feedback on fdbkdbid,
        given the target set of images refdbid.

        An image is randomly chosen.
        """

        return random.choice(fdbkdbid)
