#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides methods to manage the buffered data.
"""

###########################################################################
# IMPORTS

import logging
from numpy import argmin
import sys

from retrieval_libs.common.index cimport get_index1, get_index2, get_index00

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

#######################################################################
# local functionality

def _progress_info1(BUFFER, index):

    key = BUFFER[2][index]

    keystr = '('

    if key.shape[0] <= 8:
        keystr += ', '.join([str(item) for item in key])
    else:
        keystr += ', '.join([str(item) for item in key[0:8]]) + ', ...'

    keystr += ')'

    logger.debug('    BUFFER SIZE %d, KEY PRIORITY %d,\nKEY %s' %
                 (len(BUFFER[3]), BUFFER[1][index], keystr))

def _progress_info2(BUFFER, index):

    key = BUFFER[2][index]

    keystr = []

    for ikey in key:

        ikeystr = '('

        if ikey.shape[0] <= 8:
            ikeystr += ', '.join([str(item) for item in ikey])
        else:
            ikeystr += ', '.join([str(item) for item in ikey[0:8]]) + ', ...'

        ikeystr += ')'

        keystr.append(ikeystr)

    keystr = '(' + ', '.join([item for item in keystr]) + ',)'

    logger.debug('    BUFFER SIZE %d, KEY PRIORITY %d,\nKEY %s' %
                 (len(BUFFER[3]), BUFFER[1][index], keystr))

#######################################################################
# FUNCTIONS

cdef buffer_set1(list BUFFER, cnp.ndarray key, tuple data):
    """
    Buffer the key and the corresponding data.
    Form: [MAXLN, [ACCESSCOUNT], [KEYS], [DATA],]
    """

    cdef cnp.int32_t existing_index, MAXLN

    MAXLN = BUFFER[0]
    if MAXLN == 0:
        return

    # save data into the buffer
    existing_index = get_index1(key, BUFFER[2])
    if existing_index >= 0:
        logger.error('\n\nBUFFER %s' % BUFFER)
        sys.stderr.write('\nERROR %s\n' % key)
        sys.exit(1)

    # decrement the access count for all
    # HINT: older data get lower priority
    for i in xrange(0, len(BUFFER[1]), 1):
        BUFFER[1][i] -= 1

    if len(BUFFER[2]) >= MAXLN:
        index = argmin(BUFFER[1])
        BUFFER[1][index] = MAXLN
        # HINT: there is NO need to duplicate for ownership
        BUFFER[2][index] = key
        BUFFER[3][index] = data
    else:
        BUFFER[1].append(MAXLN)
        # HINT: there is NO need to duplicate for ownership
        BUFFER[2].append(key)
        BUFFER[3].append(data)

cdef tuple buffer_get1(list BUFFER, cnp.ndarray key):
    """
    If there is data corresponding to the key, return it;
    Otherwise, return 'None'.
    """

    cdef cnp.int32_t index
    cdef tuple data

    data = None
    index = get_index1(key, BUFFER[2])
    if index >= 0:
        data = BUFFER[3][index]
        # HINT: accessed data get higher priority
        if BUFFER[1][index] < BUFFER[0]:
            BUFFER[1][index] += 1

        if logger.getEffectiveLevel() < logging.DEBUG:
            _progress_info1(BUFFER, index)

    return data

cdef buffer_del1(list BUFFER, cnp.ndarray key):
    """
    If there is data corresponding to the key, delete it.
    """

    cdef cnp.int32_t index

    index = get_index1(key, BUFFER[2])
    if index >= 0:
        del BUFFER[1][index]
        del BUFFER[2][index]
        del BUFFER[3][index]

###########################################################################

cdef buffer_set2(list BUFFER, tuple key, tuple data):
    """
    Buffer the key and the corresponding data.
    Form: [MAXLN, [ACCESSCOUNT], [KEYS], [DATA],]
    """

    cdef cnp.int32_t existing_index, MAXLN

    MAXLN = BUFFER[0]
    if MAXLN == 0:
        return

    # pre-condition
    existing_index = get_index2(key, BUFFER[2])
    if existing_index >= 0:
        logger.error('\n\nBUFFER %s' % BUFFER)
        sys.stderr.write('\nERROR %s\n' % key)
        sys.exit(1)

    # decrement the access count for all
    # HINT: older data get lower priority
    for i in xrange(0, len(BUFFER[1]), 1):
        BUFFER[1][i] -= 1

    if len(BUFFER[2]) >= MAXLN:
        index = argmin(BUFFER[1])
        BUFFER[1][index] = MAXLN
        # HINT: there is NO need to duplicate for ownership
        BUFFER[2][index] = key
        BUFFER[3][index] = data
    else:
        BUFFER[1].append(MAXLN)
        # HINT: there is NO need to duplicate for ownership
        BUFFER[2].append(key)
        BUFFER[3].append(data)

cdef tuple buffer_get2(list BUFFER, tuple key):
    """
    If there is data corresponding to the key, return it;
    Otherwise, return 'None'.
    """

    cdef tuple data
    cdef cnp.int32_t index

    data = None
    index = get_index2(key, BUFFER[2])
    if index >= 0:
        data = BUFFER[3][index]
        # HINT: accessed data get higher priority
        if BUFFER[1][index] < BUFFER[0]:
            BUFFER[1][index] += 1

        if logger.getEffectiveLevel() < logging.DEBUG:
            _progress_info2(BUFFER, index)

    return data

cdef buffer_del2(list BUFFER, tuple key):
    """
    If there is data corresponding to the key, delete it.
    """

    cdef cnp.int32_t index

    index = get_index2(key, BUFFER[2])
    if index >= 0:
        del BUFFER[1][index]
        del BUFFER[2][index]
        del BUFFER[3][index]

#######################################################################

cdef buffer_set00(list BUFFER, cnp.int32_t key, cnp.ndarray data):
    """
    Buffer the key and the corresponding data.
    Form: [MAXLN, [ACCESSCOUNT], [KEYS], [DATA],]
    """

    cdef cnp.int32_t existing_index, MAXLN

    MAXLN = BUFFER[0]
    if MAXLN == 0:
        return

    # save data into the buffer
    existing_index = get_index00(key, BUFFER[2], False)
    if existing_index >= 0:
        logger.error('\n\nBUFFER %s' % BUFFER)
        sys.stderr.write('\nERROR %s\n' % key)
        sys.exit(1)

    # decrement the access count for all
    # HINT: older data get lower priority
    for i in xrange(0, len(BUFFER[1]), 1):
        BUFFER[1][i] -= 1

    if len(BUFFER[2]) >= MAXLN:
        index = argmin(BUFFER[1])
        BUFFER[1][index] = MAXLN
        # HINT: there is NO need to duplicate for ownership
        BUFFER[2][index] = key
        BUFFER[3][index] = data
    else:
        BUFFER[1].append(MAXLN)
        # HINT: there is NO need to duplicate for ownership
        BUFFER[2].append(key)
        BUFFER[3].append(data)

cdef cnp.ndarray buffer_get00(list BUFFER, cnp.int32_t key):
    """
    If there is data corresponding to the key, return it;
    Otherwise, return 'None'.
    """

    cdef cnp.int32_t index
    cdef cnp.ndarray data

    data = None
    index = get_index00(key, BUFFER[2], False)
    if index >= 0:
        data = BUFFER[3][index]
        # HINT: accessed data get higher priority
        if BUFFER[1][index] < BUFFER[0]:
            BUFFER[1][index] += 1

        if logger.getEffectiveLevel() < logging.DEBUG:
            _progress_info1(BUFFER, index)

    return data

cdef buffer_del00(list BUFFER, cnp.int32_t key):
    """
    If there is data corresponding to the key, delete it.
    """

    cdef cnp.int32_t index

    index = get_index00(key, BUFFER[2], False)
    if index >= 0:
        del BUFFER[1][index]
        del BUFFER[2][index]
        del BUFFER[3][index]
