#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides methods to select candidate images or nodes.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, array
import operator
from random import Random
import sys

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

random = Random()
random.jumpahead(8)

cdef cnp.float32_t PACCURACY = 0.000001

cdef cnp.int32_t CANDIDATES2_LN = 0

#######################################################################
# local functionality

cdef cnp.int32_t _candidate(
        dict rankdata,
        cnp.int32_t candidate_clix,
        cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] skipdisp):
    """
    Return the ONE index with the highest probability (i.e. argmax) within the
    cluster clix.
    If clix == -1, the candidates are not yet clustered.
    If necessary, the ONE index is chosen randomly.
    """

    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] rankdbid = rankdata['imdbid']
    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] rankprob = rankdata['improb']
    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] rankclix = rankdata['tnclix']

    cdef cnp.int32_t i, ni = rankprob.shape[0]
    cdef cnp.int32_t idbid, ii, ini = skipdisp.shape[0]
    cdef cnp.float32_t pmaxmin, pmax, pcurr
    cdef cnp.int32_t trix
    cdef list candidates, candidates2
    cdef bint skipflag

    # keeps the indexes of maximum prob within the accuracy range
    candidates = []
    # keeps the indexes of the first CANDIDATES2_LN highest prob
    candidates2 = []
    pmax = 0.0
    pmaxmin = pmax - PACCURACY
    for i from 0 <= i < ni:

        # skip the images from other clusters
        if rankclix[i] != candidate_clix:
            continue

        # skip the images already displayed
        skipflag = False
        idbid = rankdbid[i]
        for ii from 0 <= ii < ini:
            if skipdisp[ii] == idbid:
                skipflag = True
                break
        if skipflag == True:
            continue

        pcurr = rankprob[i]

        # append until it reach its maxlen
        if len(candidates2) < CANDIDATES2_LN:
            candidates2.append((i, pcurr))

        if pcurr > pmaxmin:
            if pcurr > pmax:
                pmax = pcurr
                pmaxmin = pmax - PACCURACY
                # reset candidates
                #candidates = []
                for c in xrange(len(candidates) - 1, -1, -1):
                    if candidates[c][1] < pmaxmin:
                        del candidates[c]

            candidates.append((i, pcurr))

            if len(candidates2) > 0 and len(candidates2) >= CANDIDATES2_LN:
                candidates2 = sorted(candidates2, key=operator.itemgetter(1), reverse=False)
                del candidates2[0]

            candidates2.append((i, pcurr))

    # HINT: candidates and candidates2 could be merged and sampled

    if len(candidates2) > len(candidates):
        # choose randomly one candidate
        trix = random.choice(candidates2)[0]
    elif len(candidates) > 0:
        # choose randomly one candidate
        trix = random.choice(candidates)[0]
    else:
        trix = int32(-1)

    logger.debug('Candidates: index(not DbId!) %d, size %d/%d, pmax %f' %
                 (trix, len(candidates), len(candidates2), pmax))

    return trix

###########################################################################

cdef cnp.ndarray skipped_images(list dispfdbkH):
    """
    Return the set of images that should be skipped.
    """

    # no skipping
    #return array([], dtype=int32, order='c')

    skipdisp = []
    if dispfdbkH is not None:
        for disp, fdbk in dispfdbkH:
            if isinstance(disp, dict):
                skipdisp.extend(disp['imdbid'])
            elif isinstance(disp, (list, tuple)):
                skipdisp.extend(disp)
    skipdisp = array(skipdisp, dtype=int32, order='c')

    return skipdisp

cdef cnp.int32_t candidate_image(cnp.int32_t sdclix, dict rankdata, cnp.ndarray skipdisp):
    """
    Return (the index of) the candidate image.
    """

    cdef cnp.int32_t sdtrix

    if sdclix == 0:
        # no skipping
        #skipdisp = array([], dtype=int32, order='c')
        sdtrix = _candidate(rankdata, -1, skipdisp)
    else:
        # no skipping
        #skipdisp = array([], dtype=int32, order='c')
        sdtrix = _candidate(rankdata, -1, skipdisp)

    return sdtrix

##########################################################################

cdef list candidate_nodes(dict tracedata, cnp.int32_t size):
    """
    Return 'size' indexes with the lowest level
    If necessary, the 'size' indexes are chosen randomly
    """

    cdef list tndbid = tracedata['tndbid']
    cdef list tnlevel = tracedata['tnlevel']
    cdef list tnaccess = tracedata['tnaccess']

    cdef cnp.int32_t i, ni = len(tndbid)
    cdef cnp.int32_t pmin, pcurr
    cdef list trix, candidates

    candidates = []
    pmin = -1
    for i from 0 <= i < ni:

        if tnaccess[i] < 0:
            continue

        pcurr = tnlevel[i]
        if pcurr < pmin or pmin < 0:
            pmin = pcurr
            # reset candidates
            candidates = []
            candidates.append(i)
        elif pcurr == pmin:
            candidates.append(i)
        else:
            pass

    candidatesln = min (size, len(candidates))
    if candidatesln > 0:
        trix = random.sample(candidates, candidatesln)
    else:
        trix = None

    return trix

cdef list candidate_nodes2(cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] data):
    """
    Return the indexes with the lowest probability
    """

    cdef cnp.int32_t size = 1

    cdef cnp.int32_t i, ni = data.shape[0]
    cdef cnp.float32_t pminmin, pmin, pcurr
    #cdef cnp.int_t ix = -1
    cdef list candidates, ix

    candidates = []
    pminmin = data[0]
    pmin = pminmin + PACCURACY
    for i from 0 <= i < ni:

        pcurr = data[i]
        if pcurr < pmin:
            if pcurr < pminmin:
                pminmin = pcurr
                pmin = pminmin + PACCURACY
                # reset candidates
                candidates = []

            candidates.append(i)

    #print 'candidates', candidates
    if len(candidates) > 0:
        # choose randomly one candidate
        #ix = random.choice(candidates)
        ix = candidates
    else:
        logger.error('    VAR: %d, %d, %f, %f, %s' % (i, ni, pmin, pminmin, str(data)))
        sys.stderr.write('\nERROR\n')
        sys.exit(1)

    return ix

###########################################################################

cdef cnp.int32_t candidate_max(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] allprob):
    """
    Return the ONE index with the highest probability (i.e. argmax)
    If necessary, the ONE index is chosen randomly
    """

    cdef cnp.int32_t i, ni = allprob.shape[0]
    cdef cnp.float32_t pmaxmax, pmax, pcurr
    cdef cnp.int_t trix = -1
    cdef list candidates

    candidates = []
    pmaxmax = 0.0
    pmax = pmaxmax - PACCURACY
    for i from 0 <= i < ni:

        pcurr = allprob[i]
        if pcurr > pmax:
            if pcurr > pmaxmax:
                pmaxmax = pcurr
                pmax = pmaxmax - PACCURACY
                # reset candidates
                candidates = []

            candidates.append(i)

    if len(candidates) > 0:
        # ... choose one candidate
        trix = min(candidates)
    else:
        #trix = int32(-1)
        print 'allprob', allprob
        sys.stderr.write('\nNO CANDIDATE\n\n')
        sys.exit(1)

    logger.debug('    CANDIDATES SELECTED %s, SIZE %s, %s' %
                 (trix, len(candidates), candidates))

    return trix
