#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides methods to compose the newdata from data and deltadata
according to the indexes in trix.

.. note:: data and deltadata are NOT changed in place.
"""

###########################################################################
# IMPORTS

from numpy import int32, float32, empty
import sys

###########################################################################
# FUNCTIONS

cdef cnp.ndarray extend_int32_ndim1(
        cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] data,
        list trix,
        cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] deltadata):
    """
    newdata[trix]
    """

    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] newdata

    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_copy_to
    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_copy_from
    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_add

    cdef cnp.int32_t trix_to_copyln, trix_to_addln, newdataln
    cdef cnp.int32_t i, trix_to, trix_from

    # trix
    trix_to_copy_to, trix_to_copy_from, trix_to_add = trix
    trix_to_copyln = len(trix_to_copy_to)
    trix_to_addln = len(trix_to_add)

    # newdata
    newdataln = trix_to_copyln + trix_to_addln
    newdata = empty((newdataln,), dtype=int32, order='c')

    for i from 0 <= i < trix_to_copyln:
        trix_to = trix_to_copy_to[i]
        trix_from = trix_to_copy_from[i]
        newdata[trix_to] = data[trix_from]

    for i from 0 <= i < trix_to_addln:
        trix_to = trix_to_add[i]
        newdata[trix_to] = deltadata[i]

    return newdata

cdef cnp.ndarray extend_int64_ndim1(
        cnp.ndarray[cnp.int_t, ndim=1, mode='c'] data,
        list trix,
        cnp.ndarray[cnp.int_t, ndim=1, mode='c'] deltadata):
    """
    newdata[trix]
    """

    cdef cnp.ndarray[cnp.int_t, ndim=1, mode='c'] newdata

    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_copy_to
    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_copy_from
    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_add

    cdef cnp.int32_t trix_to_copyln, trix_to_addln, newdataln
    cdef cnp.int32_t i, trix_to, trix_from

    # trix
    trix_to_copy_to, trix_to_copy_from, trix_to_add = trix
    trix_to_copyln = len(trix_to_copy_to)
    trix_to_addln = len(trix_to_add)

    # newdata
    newdataln = trix_to_copyln + trix_to_addln
    newdata = empty((newdataln,), dtype=int, order='c')

    for i from 0 <= i < trix_to_copyln:
        trix_to = trix_to_copy_to[i]
        trix_from = trix_to_copy_from[i]
        newdata[trix_to] = data[trix_from]

    for i from 0 <= i < trix_to_addln:
        trix_to = trix_to_add[i]
        newdata[trix_to] = deltadata[i]

    return newdata

cdef cnp.ndarray extend_float32_ndim1(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] data,
        list trix,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] deltadata):
    """
    newdata[trix]
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] newdata

    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_copy_to
    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_copy_from
    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_add

    cdef cnp.int32_t trix_to_copyln, trix_to_addln, newdataln
    cdef cnp.int32_t i, trix_to, trix_from

    # trix
    trix_to_copy_to, trix_to_copy_from, trix_to_add = trix
    trix_to_copyln = len(trix_to_copy_to)
    trix_to_addln = len(trix_to_add)

    # newdata
    newdataln = trix_to_copyln + trix_to_addln
    newdata = empty((newdataln,), dtype=float32, order='c')

    for i from 0 <= i < trix_to_copyln:
        trix_to = trix_to_copy_to[i]
        trix_from = trix_to_copy_from[i]
        newdata[trix_to] = data[trix_from]

    for i from 0 <= i < trix_to_addln:
        trix_to = trix_to_add[i]
        newdata[trix_to] = deltadata[i]

    return newdata

###########################################################################

cdef cnp.ndarray extend_float32_ndim2(
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] data,
        list trix,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] deltadata):
    """
    newdata[trix,:]
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] newdata

    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_copy_to
    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_copy_from
    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_add

    cdef cnp.int32_t trix_to_copyln, trix_to_addln, newdataln
    cdef cnp.int32_t i, trix_to, trix_from

    cdef cnp.int32_t t, nt

    # trix
    trix_to_copy_to, trix_to_copy_from, trix_to_add = trix
    trix_to_copyln = len(trix_to_copy_to)
    trix_to_addln = len(trix_to_add)

    # shape
    nt = data.shape[1]

    # newdata
    newdataln = trix_to_copyln + trix_to_addln
    newdata = empty((newdataln, nt), dtype=float32, order='c')

    for i from 0 <= i < trix_to_copyln:
        trix_to = trix_to_copy_to[i]
        trix_from = trix_to_copy_from[i]
        #newdata[trix_to,:] = data[trix_from,:]
        for t from 0 <= t < nt:
            newdata[trix_to,t] = data[trix_from,t]

    for i from 0 <= i < trix_to_addln:
        trix_to = trix_to_add[i]
        #newdata[trix_to,:] = deltadata[i,:]
        for t from 0 <= t < nt:
            newdata[trix_to,t] = deltadata[i,t]

    return newdata
