#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides methods to get the index of an item within a list.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32
import sys

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

cdef cnp.float32_t ACCURACY = 0.000001

#######################################################################
# FUNCTIONS

cdef cnp.int32_t get_index00(
        cnp.int32_t xdbid,
        list ddbid,
        bint FLAG_MATCH):
    """
    Return the index of xdbid within ddbid.
    If not found and FLAG_MATCH is True, raise system error.
    If not found and FLAG_MATCH is False, return -1.

    xdbid int32
    ddbid list
    """

    cdef cnp.int32_t i, ddbid_ln

    ddbid_ln = len(ddbid)
    for i from 0 <= i < ddbid_ln:
        if ddbid[i] == xdbid:
            return i

    # no match
    if FLAG_MATCH == True:
        sys.stderr.write('\nNO MATCH\n')
        sys.exit(1)

    # no match
    return int32(-1)

cdef cnp.int32_t get_index0(
        cnp.int32_t xdbid,
        cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] ddbid,
        bint FLAG_MATCH):
    """
    Return the index of xdbid within ddbid.
    If not found and FLAG_MATCH is True, raise system error.
    If not found and FLAG_MATCH is False, return -1.

    xdbid int32
    ddbid ndarray(dtype=int32, dim=1)
    """

    cdef cnp.int32_t i, ddbid_ln

    ddbid_ln = ddbid.shape[0]
    for i from 0 <= i < ddbid_ln:
        if ddbid[i] == xdbid:
            return i

    # no match
    if FLAG_MATCH == True:
        sys.stderr.write('\nNO MATCH\n')
        sys.exit(1)

    # no match
    return int32(-1)

cdef cnp.int32_t get_index1(
        cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] xdbid,
        list ddbid):
    """
    Return the index of xdbid within ddbid.
    If not found, return -1.

    xdbid ndarray(dtype=int32, dim=1)
    ddbid list of ndarray(dtype=int32, dim=1)
    """

    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] idbid

    cdef cnp.int32_t xdbid_ln, ddbid_ln
    cdef cnp.int32_t i, curr
    cdef bint equality

    xdbid_ln = xdbid.shape[0]
    ddbid_ln = len(ddbid)

    for i from 0 <= i < ddbid_ln:

        idbid = ddbid[i]

        # HINT: match if they reference the same data
        if &idbid.data == &xdbid.data:
            # easy match
            #logger.debug('    THE SAME DATA POINTER %d' % (xdbid_ln))
            return i

        # HINT: no match if they are of different length
        if idbid.shape[0] != xdbid_ln:
            # no match
            continue

        # HINT: loop until they are different at one index
        equality = True
        for curr from 0 <= curr < xdbid_ln:
            if idbid[curr] != xdbid[curr]:
                equality = False
                break

        # match if the equality test was passed successfully
        if equality == True:
            return i

    # no match
    return int32(-1)

cdef cnp.int32_t get_index2(tuple xdbid, list ddbid):
    """
    Return the index of xdbid within ddbid.
    If not found, return -1.

    xdbid tuple of ndarray(dtype=int32, dim=1)
    ddbid list of tuples of ndarray(dtype=int32, dim=1)
    """

    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] xdbid_curr
    cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] idbid_curr

    cdef cnp.int32_t i
    cdef cnp.int32_t curr, xdbid_ln
    cdef cnp.int32_t curr2, xdbid_curr_ln
    cdef bint equality, equality2

    #xdbid_type = type(xdbid)
    xdbid_ln = len(xdbid)
    #if xdbid_ln == 0:
    #    # no match
    #    return int32(-1)

    for i, idbid in enumerate(ddbid):

        # HINT: match if they reference the same object
        #if &idbid == &xdbid:
        #    return i

        # HINT: no match if they are of different type
        #if type(idbid) != xdbid_type:
        #    # no match
        #    continue

        # HINT: no match if they are of different length
        #if len(idbid) != xdbid_ln:
        #    # no match
        #    continue

        # HINT: loop the tuple items
        equality = True
        for curr from 0 <= curr < xdbid_ln:
            idbid_curr = idbid[curr]
            xdbid_curr = xdbid[curr]
            xdbid_curr_ln = xdbid_curr.shape[0]

            # HINT: match if they reference the same data
            if &idbid_curr.data == &xdbid_curr.data:
                # easy match
                #logger.debug('    THE SAME DATA POINTER %d' % (xdbid_curr_ln))
                continue

            # HINT: no match if they are of different type
            #if type(idbid_curr) != type(xdbid_curr):
            #    # no match
            #    equality = False
            #    continue

            # HINT: no match if they are of different length
            if idbid_curr.shape[0] != xdbid_curr_ln:
                # no match
                equality = False
                break

            # HINT: loop until they are different at one index
            equality2 = True
            for curr2 from 0 <= curr2 < xdbid_curr_ln:
                if abs(idbid_curr[curr2] - xdbid_curr[curr2]) > ACCURACY:
                    equality2 = False
                    break

            # no match if the equality test has failed
            if equality2 == False:
                equality = False
                break

        # match if the equality test was passed successfully
        if equality == True:
            return i

    # no match
    return int32(-1)
