#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides methods to compute the posteriori probabilities.

.. note:: prob = (prob * pp) / (prob * pp + (1.0 - prob) * pm)

    where:

    - prob are the probabilities of relevance
    - pp and pm are the P_{+} and P_{-} probabilities
"""

###########################################################################
# IMPORTS

import logging
from numpy import float32, array
import sys

from libc.math cimport abs as c_abs

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

cdef cnp.float32_t ACCURACY = 0.0 #pow(10, -30)

###########################################################################

cdef update_prob0(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] prob,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pp,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pm):
    """
    Update IN PLACE the posteriori probabilities.
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] probM

    # HINT: prob = (prob * pp) / (prob * pp + (1.0 - prob) * pm)
    probM = 1.0 - prob
    probM *= pm
    prob *= pp
    probM += prob
    prob /= probM

cdef update_prob(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] prob,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pp,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pm):
    """
    Update IN PLACE the posteriori probabilities.
    """

    cdef cnp.int32_t i, ni = prob.shape[0]
    cdef cnp.float32_t probi, probip
    #cdef cnp.int32_t count = 0

    # HINT: prob = (prob * pp) / (prob * pp + (1.0 - prob) * pm)
    for i from 0 <= i < ni:
        if c_abs(pp[i] - pm[i]) > ACCURACY:
            probi = prob[i]
            probip = probi * pp[i]
            prob[i] = probip / (probip + (1.0 - probi) * pm[i])
            #count += 1

    #logger.info('    COMPUTED %s OUT OF %s' % (count, ni))

cdef update_prob_full(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] prob,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pp,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pm):
    """
    Update IN PLACE the posteriori probabilities.
    """

    cdef cnp.int32_t i, ni = prob.shape[0]
    cdef cnp.float32_t probi, probip

    # HINT: prob = (prob * pp) / (prob * pp + (1.0 - prob) * pm)
    for i from 0 <= i < ni:
        probi = prob[i]
        probip = probi * pp[i]
        prob[i] = probip / (probip + (1.0 - probi) * pm[i])

###########################################################################

cdef dwdate_prob(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] prob,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pp,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pm):
    """
    Dwdate IN PLACE the posteriori probabilities.
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] probM

    # HINT: prob = (prob * pm) / (prob * pm + (1.0 - prob) * pp)
    probM = 1.0 - prob
    probM *= pp
    prob *= pm
    probM += prob
    prob /= probM

cdef dwdate_prob0(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] prob,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pp,
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pm):
    """
    Dwdate IN PLACE the posteriori probabilities.
    """

    cdef cnp.int32_t i, ni = prob.shape[0]
    cdef cnp.float32_t probi, probim
    #cdef cnp.int32_t count = 0

    # HINT: prob = (prob * pm) / (prob * pm + (1.0 - prob) * pp)
    for i from 0 <= i < ni:
        if c_abs(pp[i] - pm[i]) > ACCURACY:
            probi = prob[i]
            probim = probi * pm[i]
            prob[i] = probim / (probim + (1.0 - probi) * pp[i])
            #count += 1

    #logger.debug('    COMPUTED %s OUT OF %s' % (count, ni))

###########################################################################

cdef cnp.ndarray update_prob2(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] prob,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] pp,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] pm):
    """
    Update the posteriori probabilities.
    HINT: ipp and ipm are changed in place.
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] probM

    # HINT: prob = (prob * pp) / (prob * pp + (1.0 - prob) * pm)
    probM = 1.0 - prob
    pm *= probM
    pp *= prob
    pm += pp
    pp /= pm

    return pp

cdef cnp.ndarray update_prob20(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] prob,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] pp,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] pm):
    """
    Update the posteriori probabilities.
    HINT: ipp and ipm are changed in place.
    """

    cdef cnp.int32_t w, nw = prob.shape[0]
    cdef cnp.int32_t i, ni = prob.shape[1]
    cdef cnp.float32_t probi, probip
    #cdef cnp.int32_t count = 0

    # HINT: prob = (prob * pp) / (prob * pp + (1.0 - prob) * pm)
    for w from 0 <= w < nw:
        for i from 0 <= i < ni:
            if c_abs(pp[w,i] - pm[w,i]) > ACCURACY:
                probi = prob[w,i]
                probip = probi * pp[w,i]
                pp[w,i] = probip / (probip + (1.0 - probi) * pm[w,i])
                #count += 1
            else:
                pp[w,i] = prob[w,i]

    #logger.debug('    COMPUTED %s OUT OF %s' % (count, ni))

    return pp

###########################################################################

cdef cnp.ndarray dwdate_prob2(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] prob,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] pp,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] pm):
    """
    Dwdate the posteriori probabilities.
    HINT: ipp and ipm are changed in place.
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] probM

    # HINT: prob = (prob * pm) / (prob * pm + (1.0 - prob) * pp)
    probM = 1.0 - prob
    pp *= probM
    pm *= prob
    pp += pm
    pm /= pp

    #oprob = squeeze(pm)

    return pm[0,:]

cdef cnp.ndarray dwdate_prob20(
        cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] prob,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] pp,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] pm):
    """
    Dwdate the posteriori probabilities.
    HINT: ipp and ipm are changed in place.
    """

    cdef cnp.int32_t w, nw = prob.shape[0]
    cdef cnp.int32_t i, ni = prob.shape[1]
    cdef cnp.float32_t probi, probim
    #cdef cnp.int32_t count = 0

    # HINT: prob = (prob * pm) / (prob * pm + (1.0 - prob) * pp)
    for w from 0 <= w < nw:
        for i from 0 <= i < ni:
            if c_abs(pp[w,i] - pm[w,i]) > ACCURACY:
                probi = prob[w,i]
                probim = probi * pm[w,i]
                pm[w,i] = probim / (probim + (1.0 - probi) * pp[w,i])
                #count += 1
            else:
                pm[w,i] = prob[w,i]

    #logger.debug('    COMPUTED %s OUT OF %s' % (count, ni))

    #oprob = squeeze(pm)

    return pm[0,:]
