#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

# decorator for excluding functions from profiling @cython.profile(False)

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides methods to compute the image similarity distances based on the
image feature vectors and the positive and the negative thresholds.
"""

###########################################################################
# IMPORTS

import logging
from numpy import float32, ones, ones_like, empty
from time import time as now
import sys

from libc.math cimport abs as c_abs

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

cdef cnp.float32_t ACCURACY = pow(10, -30)

###########################################################################
# FUNCTIONS

cdef tuple _coef_p(cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] THARGS):
    """
    Get the positive thresholds.
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] tpX, tpY, coefA, coefB

    tpX = THARGS[0]
    tpY = THARGS[1]

    #coefA = - (1.0 - tpY) / tpX
    coefA = ones_like(tpX)
    coefA -= tpY
    coefA /= - tpX

    # coefB = 1.0
    coefB = ones_like(tpX)

    return coefA, coefB, tpX, tpY

cdef tuple _coef_m(cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] THARGS):
    """
    Get the negative thresholds.
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] tmX, tmY, coefA, coefB

    tmX = THARGS[2]
    tmY = THARGS[3]

    #coefA = (1.0 - tmY) / tmX
    coefA = ones_like(tmX)
    coefA -= tmY
    coefA /= tmX

    # coefB = tmY
    coefB = ones_like(tmX)
    coefB[:] = tmY

    # thY = 1.0
    thY = ones_like(tmX)

    return coefA, coefB, tmX, thY

cpdef tuple get_thcoefs(cnp.ndarray THARGS):
    """
    Get the positive and negative thresholds.
    """

    THPCOEFS = _coef_p(THARGS)
    THMCOEFS = _coef_m(THARGS)

    return THPCOEFS, THMCOEFS

###########################################################################
# local functionality

cdef cnp.ndarray _iddc( tuple distdata, tuple thargs):
    """
    Return the calibrated distances.

    Form: array(k:, h:, t:) where:
      - k corresponding to the index in kdbid
      - h corresponding to the index in hdbid
      - t corresponding to the index in feature type (0/1/-1)
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] coefA, coefB, thX, thY

    cdef cnp.ndarray[cnp.float32_t, ndim=3, mode='c'] ddc
    cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] dist_k

    cdef cnp.int32_t k, nk, h, nh, t, nt
    cdef cnp.float32_t coefA_t, coefB_t, thX_t, thY_t

    nk = len(distdata)
    nh = distdata[0].shape[0]
    nt = distdata[0].shape[1]

    # initialize an empty array
    ddc = empty((nk, nh, nt), dtype=float32, order='c')

    for k from 0 <= k < nk:
        dist_k = distdata[k]

        coefA, coefB, thX, thY = thargs[1][1]

        for t from 0 <= t < nt:
            coefA_t = coefA[t]
            #coefB_t = coefB[t]

            if coefA_t > 0.0:
                for h from 0 <= h < nh:
                    #ddc[k,h,t] = coefA_t * dist_k[h,t]# + coefB_t
                    ddc[k,h,t] = dist_k[h,t]
            else:
                for h from 0 <= h < nh:
                    #ddc[k,h,t] = 0.0#coefB_t
                    ddc[k,h,t] = dist_k[h,t]

    return ddc

cdef cnp.ndarray _iddw(
        cnp.ndarray[cnp.float32_t, ndim=3, mode='c'] ddc,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] allweights):
    """
    Return the weighted distances for all given weights.

    Form: array(w:, k:, h:) where:
      - w corresponding to the index in weights
      - k corresponding to the index in kdbid
      - h corresponding to the index in hdbid
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=3, mode='c'] ddw
    cdef cnp.int32_t t, nt, k, nk, h, nh, w, nw
    cdef cnp.float32_t weight
    cdef cnp.npy_intp *shape
    cdef bint initialized

    shape = ddc.shape
    nk = shape[0]
    nh = shape[1]
    nt = shape[2]
    nw = allweights.shape[0]

    ddw = empty((nw, nk, nh), dtype=float32, order='c')

    # LOOP for all ~possible weights
    for w from 0 <= w < nw:

        initialized = False
        for t from 0 <= t < nt:
            weight = allweights[w,t]

            # apply the weight and sum up
            if c_abs(weight - 0.0) < ACCURACY:
                continue
                #if initialized == False:
                #    initialized = True
                #    for k from 0 <= k < nk:
                #        for h from 0 <= h < nh:
                #            ddw[w,k,h] = 0.0
                #else:
                #    pass

            elif c_abs(1.0 - weight) < ACCURACY:
                for k from 0 <= k < nk:
                    for h from 0 <= h < nh:
                        ddw[w,k,h] = ddc[k,h,t]
                break
                #if initialized == False:
                #    initialized = True
                #    for k from 0 <= k < nk:
                #        for h from 0 <= h < nh:
                #            ddw[w,k,h] = ddc[k,h,t]
                #else:
                #    for k from 0 <= k < nk:
                #        for h from 0 <= h < nh:
                #            ddw[w,k,h] += ddc[k,h,t]

            else:
                if initialized == False:
                    initialized = True
                    for k from 0 <= k < nk:
                        for h from 0 <= h < nh:
                            ddw[w,k,h] = ddc[k,h,t] * weight
                else:
                    for k from 0 <= k < nk:
                        for h from 0 <= h < nh:
                            ddw[w,k,h] += ddc[k,h,t] * weight

    return ddw

cdef tuple _iddpddm( tuple distdata, tuple thargs):
    """
    Return the pozitive and negative similarity distances.

    Form: array(k:, h:, t:) where:
      - k corresponding to the index in kdbid
      - h corresponding to the index in hdbid
      - t corresponding to the index in feature type (0/1/-1)
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] coefpA, coefpB, thpX, thpY
    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] coefmA, coefmB, thmX, thmY

    cdef cnp.ndarray[cnp.float32_t, ndim=3, mode='c'] ddp, ddm
    cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] dist_k

    cdef cnp.int32_t k, nk, h, nh, t, nt
    cdef cnp.float32_t coefpA_t, coefpB_t, thpX_t, thpY_t
    cdef cnp.float32_t coefmA_t, coefmB_t, thmX_t, thmY_t
    cdef cnp.float32_t dist_kht
    #cdef cnp.int32_t counter

    nk = len(distdata)
    nh = distdata[0].shape[0]
    nt = distdata[0].shape[1]

    # initialize an empty array
    ddp = empty((nk, nh, nt), dtype=float32, order='c')
    ddm = empty((nk, nh, nt), dtype=float32, order='c')

    #counter = 0
    for k from 0 <= k < nk:
        dist_k = distdata[k]

        coefpA, coefpB, thpX, thpY = thargs[1][0]
        coefmA, coefmB, thmX, thmY = thargs[1][1]

        for t from 0 <= t < nt:
            coefpA_t = coefpA[t]
            coefpB_t = coefpB[t]
            thpX_t = thpX[t]
            thpY_t = thpY[t]
            coefmA_t = coefmA[t]
            coefmB_t = coefmB[t]
            thmX_t = thmX[t]
            thmY_t = thmY[t]

            for h from 0 <= h < nh:
                dist_kht = dist_k[h,t]

                # compute it only if it is below the threshold thX
                # otherwise set the value of the threshold thY
                if dist_kht < thpX_t:
                    ddp[k,h,t] = coefpA_t * dist_kht + coefpB_t
                    #counter += 1
                else:
                    ddp[k,h,t] = thpY_t

                # compute it only if it is below the threshold thX
                # otherwise set the value of the threshold thY
                if dist_kht < thmX_t:
                    ddm[k,h,t] = coefmA_t * dist_kht + coefmB_t
                    #counter += 1
                else:
                    ddm[k,h,t] = thmY_t

    #if nh > 8:
    #    logger.debug('    DDP SHAPE %s, THRESHOLD %s, COUNTER %s' %
    #                 ((nk, nh, nt), thX, counter))

    return ddp, ddm

###########################################################################
# queryimg functionality

cdef cnp.ndarray ddw(
        tuple distdata,
        tuple thargs,
        cnp.ndarray weights):
    """
    Return the calibrated-weighted distances.
    """

    cdef cnp.ndarray ddc, ddw

    # get the calibrated distances
    ddc = _iddc(distdata, thargs)
    # apply the weights and sum up
    ddw = _iddw(ddc, weights)

    return ddw

cdef tuple ddpddm(
        tuple distdata,
        tuple thargs,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] weights):
    """
    Return the pozitive and negative similarity probabilities.
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=3, mode='c'] ddp, ddm

    # HINT: ddp, ddm can be owned and changed in place
    ddp, ddm = _iddpddm(distdata, thargs)

    ddp *= weights
    ddpw = ddp.sum(2)

    ddm *= weights
    ddmw = ddm.sum(2)

    return ddpw, ddmw

cdef tuple pppm(
        tuple distdata,
        cnp.int32_t kindex,
        tuple thargs,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] weights):
    """
    Return the pozitive and negative similarity probabilities.
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=3, mode='c'] ddp, ddm
    cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pp, pm

    cdef cnp.int32_t nh

    # HINT: ddp, ddm can be owned and changed in place
    ddp, ddm = _iddpddm(distdata, thargs)

    ddp *= weights
    ddpw = ddp.sum(2)

    ddm *= weights
    ddmw = ddm.sum(2)

    if kindex >= 0:
        pp = ddpw[kindex,:]
        pp /= ddpw.sum(0)
        pm = ddmw[kindex,:]
        pm /= ddmw.sum(0)
    else:
        sys.stderr.write('\nUNKNOWN INDEX\n')
        sys.exit(1)

    return pp, pm

cdef tuple pppm2(
        tuple distdata,
        cnp.int32_t kindex,
        tuple thargs,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] allweights):
    """
    Return the pozitive and negative similarity probabilities.
    """

    cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] weights
    cdef cnp.ndarray[cnp.float32_t, ndim=3, mode='c'] ddp, ddm
    cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] allpp, allpm

    cdef cnp.int32_t i, ni, nk, nh, nt

    ni = allweights.shape[0]
    nk = len(distdata[0])
    nh = distdata[0].shape[0]
    nt = distdata[0].shape[1]

    allpp = empty((ni, nh), dtype=float32, order='c')
    allpm = empty((ni, nh), dtype=float32, order='c')

    # WARNING: DO NOT CHANGE IN PLACE ddp, ddm
    ddp, ddm = _iddpddm(distdata, thargs)

    # LOOP for all ~possible weights
    for i from 0 <= i < ni:

        # set the current weights
        weights = allweights[i:(i+1),:]

        pp = allpp[i,:]
        pm = allpm[i,:]

        ddpw = ddp * weights
        ddpw = ddpw.sum(2)

        ddmw = ddm * weights
        ddmw = ddmw.sum(2)

        if kindex >= 0:
            pp[:] = ddpw[kindex,:]
            pp /= ddpw.sum(0)
            pm[:] = ddmw[kindex,:]
            pm /= ddmw.sum(0)
        else:
            sys.stderr.write('\nUNKNOWN INDEX\n')
            sys.exit(1)

    return allpp, allpm

cdef tuple pppm3(
        tuple distdata,
        cnp.int32_t kindex,
        tuple allthargs,
        cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] weights):
    """
    Return the pozitive and negative similarity probabilities.
    """

    cdef tuple thargs
    cdef cnp.ndarray[cnp.float32_t, ndim=3, mode='c'] ddp, ddm
    cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] allpp, allpm

    cdef cnp.int32_t i, ni, nh

    ni = len(allthargs)
    nh = distdata[0].shape[0]

    allpp = empty((ni, nh), dtype=float32, order='c')
    allpm = empty((ni, nh), dtype=float32, order='c')

    # LOOP for all ~possible thresholds
    for i from 0 <= i < ni:

        # set the current thresholds
        thargs = allthargs[i]

        # HINT: ddp, ddm can be owned and changed in place
        ddp, ddm = _iddpddm(distdata, thargs)

        pp = allpp[i,:]
        pm = allpm[i,:]

        ddp *= weights
        ddpw = ddp.sum(2)

        ddm *= weights
        ddmw = ddm.sum(2)

        if kindex >= 0:
            pp[:] = ddpw[kindex,:]
            pp /= ddpw.sum(0)
            pm[:] = ddmw[kindex,:]
            pm /= ddmw.sum(0)
        else:
            sys.stderr.write('\nUNKNOWN INDEX\n')
            sys.exit(1)

    return allpp, allpm

###########################################################################
# autousr functionality

cdef cnp.ndarray _ddw_min(
        tuple distdata,
        tuple thargs,
        cnp.ndarray allweights):
    """
    Return the minimum individual distances for all given weights.
    """

    dd = ddw(distdata, thargs, allweights)
    # the minimum distances from each of fdbkdbid, for each weights
    dd = dd.min(1)
    # the minimum distances, for any weights
    dd = dd.min(0)

    return dd

cdef cnp.ndarray _ddw_cum(
        tuple distdata,
        tuple thargs,
        cnp.ndarray allweights):
    """
    Return the minimum cumulated distances for all given weights.
    """

    dd = ddw(distdata, thargs, allweights)
    # the sum of distances from each of fdbkdbid, for each weights
    dd = dd.sum(1)
    # the minimum of distances, for any weights
    dd = dd.min(0)

    return dd

cdef cnp.ndarray fdbkdist(
        tuple distdata,
        tuple thargs,
        cnp.ndarray allweights):
    """
    Return the minimum distances for all given weights.
    """

    #fdbkdist = _ddw_min(distdata, thargs, allweights)
    fdbkdist = _ddw_cum(distdata, thargs, allweights)

    return fdbkdist
