#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides the DbTable instances.
"""

###########################################################################
# IMPORTS

import logging
import sys

from retrieval_libs.dblayer.DbCnx import DbCnx
from retrieval_libs.dblayer.DbTableImg cimport DbTableImg, DbTableImgSynthetic
from retrieval_libs.dblayer.DbTableTree cimport DbTableTree
from retrieval_libs.dblayer.FileManager import FileManager

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

###########################################################################
# FUNCTIONS

def _collections(ALL_DB):
    """
    Returns the available collections.
    FORM: [{db_name:'', db_tbl:''},...]
    """

    collections = []
    for db_name in ALL_DB:
        db_table = ALL_DB[db_name]['TABLE']
        info = {'db_name':db_name, 'db_tbl':db_table}
        collections.append(info)

    return collections

###########################################################################
# CLASS

class DbAccess(object):
    """
    Provides the DbTable instances.

    .. note:: Implemented as a very simple singleton.
    """

    #######################################################################

    __single = None # the one, true Singleton

    def __new__(classtype, *args, **kwargs):
        # Check to see if a __single exists already for this class
        # Compare class types instead of just looking for None so
        # that subclasses will create their own __single objects
        if classtype != type(classtype.__single):
          classtype.__single = object.__new__(classtype, *args, **kwargs)
          #logger.info('    DbAccess: NEW INSTANCE')

        return classtype.__single

    #######################################################################
    # main functionality

    def __init__(self):
        """
        Initialize the object.
        """

        # initialize the connection object
        # HINT: __init__ is called for each "reference" to the singleton
        try:
            self.dbcnx
        except AttributeError:
            self.dbcnx = DbCnx()

    def custom_query(self, query):
        """
        One single function to execute and commit an atomic query.
        """

        return self.dbcnx.custom_query(query)

    def init(self, ALL_DB, FIGURES, active_db):
        """
        Initialize the object.
        """

        if active_db is None:
            sys.stderr.write('\nACTIVE_DB IS NONE\n')
            sys.exit(1)

        # HINT: do not use __init__ since it is called for each "instance"
        try:
            self.active_db
        except AttributeError:
            self.active_db = None

        # if already initialized, skip it
        if self.active_db == active_db:
            return

        self.active_db = active_db
        self.collections = _collections(ALL_DB)

        db_data = ALL_DB[active_db]
        self.table = db_data['TABLE']
        self.DIR_DIST = db_data.get('DIR_DIST', None)
        self.DIR_IMG = db_data.get('DIR_IMG', None)

        self.TABLE_USER = ALL_DB[active_db]['TABLE_USER']
        self.TABLE_SEARCH = ALL_DB[active_db]['TABLE_SEARCH']
        self.TABLE_TASK = ALL_DB[active_db]['TABLE_TASK']
        self.TABLE_CONFIG = ALL_DB[active_db]['TABLE_CONFIG']

        self.FIGURES = FIGURES

        logger.info('    DbAccess: INIT: %s, %s' % (self.active_db, self.table))

        # tables
        self._dbimg = None
        self._dbimg_autousr = None
        self._dbtree = None
        # check the tables
        if logger.getEffectiveLevel() < logging.DEBUG:
            self.dbdoc()
            self.dbtxt()
            self.dbimg()
            self.dbtree()

        # files
        self._dbfig = None

    def dbtable(self):
        return self.table

    def dbtable_user(self):
        return self.TABLE_USER

    def dbtable_search(self):
        return self.TABLE_SEARCH

    def dbtable_task(self):
        return self.TABLE_TASK

    def dbtable_config(self):
        return self.TABLE_CONFIG

    #######################################################################
    # tables functionality

    def dbimg(self):

        if not self._dbimg:
            if self.table.startswith('ImageSynthetic'):
                self._dbimg = DbTableImgSynthetic()
                self._dbimg.init(self.table, self.dbcnx, self.DIR_DIST, self.DIR_IMG)
            else:
                self._dbimg = DbTableImg()
                self._dbimg.init(self.table, self.dbcnx, self.DIR_DIST)

        return self._dbimg

    def dbimg_autousr(self):

        if not self._dbimg_autousr:
            if self.table.startswith('ImageSynthetic'):
                self._dbimg_autousr = DbTableImgSynthetic()
                self._dbimg_autousr.init(self.table, self.dbcnx, self.DIR_DIST, self.DIR_IMG, DDBUFFER_ID='DD_AUTOUSR')
            else:
                self._dbimg_autousr = DbTableImg()
                self._dbimg_autousr.init(self.table, self.dbcnx, self.DIR_DIST, DDBUFFER_ID='DD_AUTOUSR')

        return self._dbimg_autousr

    def dbtree(self):

        if not self._dbtree:
            self._dbtree = DbTableTree()
            self._dbtree.init(self.table, self.dbcnx)

        return self._dbtree

    #######################################################################
    # files functionality

    def dbfig(self):

        if not self._dbfig:
            self._dbfig = FileManager(self.FIGURES)

        return self._dbfig

    #######################################################################

    def displayinfo(self):
        """
        Prepare the display information.
        """

        info = {}

        info['collections'] = self.collections
        info['active_db'] = self.active_db
        info['collection_tbl'] = self.table

        dbimg = self.dbimg()
        info['collection_size'] = dbimg.size()

        return {'qdb': info}
