#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Manages the generic interaction with the tables.
"""

###########################################################################
# IMPORTS

import datetime
import logging
from numpy import int32, ndarray, array, argsort
import sys

from retrieval_libs.dblayer.DbCnx import DbCnx

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

###########################################################################
# CLASS

cdef class DbTable(object):
    """
    Manages the generic interaction with the tables.
    """

    #######################################################################
    # local functionality

    def _check_table(self, table):
        """
        Return True if the table exists and is valid.
        """

         # first, ask the buffer
        valid = self.VALID.get(table, None)

        if valid is None:
            # otherwise, ask the database
            # check if table exists
            self.cnx.execute("CHECK TABLE %s FAST QUICK" % table)
            result = self.cnx.fetchall()
            if 'Error' not in result[0]:
                valid = True
                if logger.getEffectiveLevel() < logging.DEBUG:
                    logger.debug('    VALID TABLE %s' % table)
            else:
                valid = False
                if logger.getEffectiveLevel() < logging.DEBUG:
                    logger.warning('    MISSING TABLE %s' % table)

            # then, add it to the buffer
            self.VALID[table] = valid

        #else:
            #logger.debug('    VALID TABLE %s' % table)

        #elif not valid:
            #logger.error('    MISSING TABLE %s' % table)

        return valid

    #######################################################################
    # main functionality

    def __init__(self):

        super(DbTable, self).__init__()

        #self.suffix = None
        #self.table = None
        self.SIZE = {}
        self.VALID = {}

    def init(self, table, cnx):
        """
        Initialize the object.
        """

        self.cnx = cnx

        self.table = table + '_' + self.suffix

    def is_valid(self):
        """
        Return True if the table exists and is valid.
        """

        valid = self._check_table(self.table)

        return valid

    #######################################################################
    # configure the database

    def init_table(self):
        """
        Initialize the table.
        """

        raise NotImplementedError

    def clear_table(self):
        """
        Clear the table.
        """

        table = self.table

        valid = self._check_table(table)
        if valid == True:
            # delete the table
            #self.cnx.execute("DROP TABLE %s" % table)

            # just rename the table, for safety
            t = datetime.datetime.now()
            table_bkp = table + '_' + t.strftime('%H')#%M%S')

            self.cnx.execute(
                    "ALTER TABLE %s RENAME TO %s"
                    % (table, table_bkp))

            # duplicate the table
            #self.cnx.execute("CREATE TABLE %s LIKE %s" % (table_bkp, table))
            #self.cnx.execute(
            #        "INSERT INTO %s SELECT * FROM %s"
            #        % (table_bkp, table))

            self.cnx.commit()

    #######################################################################
    # generic exchange data with the database

    def size(self):
        """
        Return the number of rows.
        """

        table = self.table

        # first, ask the buffer
        size = self.SIZE.get(table, None)
        if size is not None:
            return size

        # otherwise, ask the database
        self.cnx.execute("SELECT COUNT(*) FROM %s" % table)
        size = self.cnx.fetchone()[0]

        # then, add it to the buffer
        self.SIZE[table] = size

        return size

    def all_dbid(self):
        """
        Get all the ids. Format: tuple, sorted
        """

        values = self.get_data('id', '', None)

        values = map(lambda x: int32(x[0]), values)

        return tuple(values)

    cpdef cnp.ndarray all_ids(self):
        """
        Get all the ids. Format: ndarray, sorted
        """

        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] ids

        cdef tuple values, val
        cdef list tmp

        values = self.get_data('id', '', None)

        tmp = [val[0] for val in values]
        ids = array(tmp, dtype=int32, order='c')

        return ids

    #######################################################################
    # generic exchange data with the database

    cpdef get_data(self, str columns, str key, value, bint order=True):
        """
        Get the given columns for the rows that match the given [key, value].
        If value is None, return the given columns for all the rows.
        """

        #logger.debug(
        #        '\n\n    MYSQL QUERY\n    columns %s\n    key %s\n'
        #        '    value %s\n    order %s\n'
        #        % (str(columns), str(key), str(value), order))

        table = self.table

        valid = self._check_table(table)
        if valid == False:
            return None

        # prepare the query, send it to the MySQL engine and fetch the data
        # HINT: if necessary, reformat data
        if value is None:
            self.cnx.execute(
                    "SELECT %s FROM %s ORDER BY id ASC"
                    % (columns, table))

            # get the result as a tuple
            data = self.cnx.fetchall()

            if len(data) == 0 or len(data[0]) == 0:
                logger.error(
                        '\n\n    MYSQL QUERY\n    columns %s\n    key %s\n'
                        '    value %s\n    order %s\n'
                        % (str(columns), str(key), str(value), order))
                logger.error(
                        '\n\n    MYSQL QUERY DATA\n    data %s\n'
                        % str(data))
                sys.stderr.write('\nMYSQL QUERY DATA IS EMPTY\n')
                sys.exit(1)

        elif isinstance(value, (tuple, list, ndarray)):
            valueln = len(value)

            if valueln > 1:
                # HINT: this works correctly ONLY FOR keys that can be ordered
                if key not in ('id', 'seed_dbid'):
                    sys.stderr.write('\nKEY %s CANNOT BE ORDERED\n' % key)
                    sys.exit(1)

                strvalue = '(' + ', '.join(map(str, value)) + ')'

                # get the raw data (ORDERED BY id ASC)
                # HINT: the 'unordered' query would NOT preserve the original
                # order anyway, so we impose a (known) order by id.
                # if necessary, the original order is recovered few lines below
                self.cnx.execute(
                        "SELECT %s FROM %s WHERE %s IN %s ORDER BY %s ASC"
                        % (columns, table, key, strvalue, key))

                # get the result as a tuple
                data = self.cnx.fetchall()

                if len(data) != len(value):
                    logger.error(
                            '\n\n    MYSQL QUERY\n    columns %s\n    key %s\n'
                            '    value %s\n    order %s\n'
                            % (str(columns), str(key), str(value), order))
                    logger.error(
                            '\n\n    MYSQL QUERY DATA\n    data %s\n'
                            % str(data))
                    sys.stderr.write(
                            '\nMYSQL QUERY DATA MISMATCH %s != %s\n'
                            % (len(value), len(data)))
                    sys.exit(1)

                if order == True:
                    # extra-sorting operation for PRESERVING the order
                    myorder = argsort(argsort(value))
                    data = tuple([data[i] for i in myorder])

            elif valueln == 1:
                strvalue = str(value[0])
                self.cnx.execute(
                        "SELECT %s FROM %s WHERE %s = '%s'"
                        % (columns, table, key, strvalue))

                # get the result as a tuple
                data = self.cnx.fetchall()

                if len(data) != 1:
                    logger.error(
                            '\n\n    MYSQL QUERY\n    columns %s\n    key %s\n'
                            '    value %s\n    order %s\n'
                            % (str(columns), str(key), str(value), order))
                    logger.error(
                            '\n\n    MYSQL QUERY DATA\n    data %s\n'
                            % str(data))
                    sys.stderr.write(
                            '\nMYSQL QUERY DATA MISMATCH %s != %s\n'
                            % (len(value), len(data)))
                    sys.exit(1)

            else:
                logger.error(
                        '\n\n    MYSQL QUERY\n    columns %s\n    key %s\n'
                        '    value %s\n    order %s\n' %
                        (str(columns), str(key), str(value), order))
                sys.stderr.write('\nVALUE IS EMPTY\n')
                sys.exit(1)

        else:
            strvalue = str(value)
            self.cnx.execute(
                    "SELECT %s FROM %s WHERE %s = '%s'"
                    % (columns, table, key, strvalue))

            # get the result as a tuple
            data = self.cnx.fetchall()

            # reformat data
            if len(data) == 1 and len(data[0]) == 2:
                # if only one value to return => compute it directly
                # WARNING: this works correctly ONLY FOR two columns
                if columns.count(',') != 1:
                    sys.stderr.write('\nTOO MANY COLUMNS "%s"\n' % columns)
                    sys.exit(1)

                data = data[0][1]

            elif len(data) == 0 or len(data[0]) == 0:
                data = None

        return data

    #######################################################################

    def custom_query(self, query):
        """
        Execute and commit an atomic query.
        """

        return self.cnx.custom_query(query % self.table)