#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True
#cython: nonecheck=False
#cython: cdivision=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Manages the table of the image data, e.g. filenames, features and distances.

Among other stuff, it encapsulates the functionality for providing the
image similarity distances. It knows to choose between computing the
distances on-the-fly or reading the files with the pre-computed distances.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, float32, ndarray, empty, array, zeros, \
                  append, argmin, argsort, copy, fromstring, reshape, squeeze
import operator
import os
import sys
from time import time as now

from libc.math cimport abs as c_abs
from libc.math cimport sqrt as c_sqrt
from libc.math cimport log as c_log

from retrieval_libs.common.buffer cimport buffer_set1, buffer_get1, buffer_set2, buffer_get2
from retrieval_libs.common.extend cimport extend_float32_ndim2
from retrieval_libs.common.index cimport get_index0
from retrieval_libs.dblayer.DataBuffer import DataBuffer
from retrieval_libs.dblayer.DbCnx import DbCnx
from retrieval_libs.dblayer.DbTable cimport DbTable
from retrieval_libs.dblayer.FileAccess import create_dir, read_bindata
from retrieval_libs.display.SyntheticImage import generate_image2D

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

# implemented as a singleton
cdef object databuffer = DataBuffer()

###########################################################################
# CLASS

cdef class DbTableImg(DbTable):
    """
    Manages the table of the image data, e.g. filenames, features and distances.

    Among other stuff, it encapsulates the functionality for providing the
    image similarity distances. It knows to choose between computing the
    distances on-the-fly or reading the files with the pre-computed distances.
    """

    #######################################################################
    # buffer management

    cdef _init_buffers(self):

        self.FFBUFFER = [2, [], [], []]
        self.DDBUFFER = [2, [], [], []]

    cdef set_buffers(self):
        """
        Restore the data.
        """

        cdef list BUFFER

        BUFFER = databuffer.get_buffer(self.FFBUFFER_ID)

        if BUFFER is not None:
            self.FFBUFFER = BUFFER
        else:
            # increase the buffer size
            self.FFBUFFER[0] = 512
            # pass reference
            databuffer.set_buffer(self.FFBUFFER_ID, self.FFBUFFER)

        BUFFER = databuffer.get_buffer(self.DDBUFFER_ID)

        if BUFFER is not None:
            self.DDBUFFER = BUFFER
        else:
            # increase the buffer size
            self.DDBUFFER[0] = 512
            # pass reference
            databuffer.set_buffer(self.DDBUFFER_ID, self.DDBUFFER)

    cdef trim_buffers(self):
        """
        Store the data.

        HINT: Keep only the data that is needed in the next iteration. This
        data could be recomputed but it is much more efficient to re-use it.
        """

        cdef list BUFFER

        # HINT: keep only one item, namely the features of the current trace
        BUFFER = [self.FFBUFFER[0], [], [], []]
        sortedarg = argsort(self.FFBUFFER[1])
        for i in sortedarg[::-1]:
            if len(self.FFBUFFER[2][i]) > 8:
                BUFFER[1].append(self.FFBUFFER[1][i])
                BUFFER[2].append(self.FFBUFFER[2][i])
                BUFFER[3].append(self.FFBUFFER[3][i])

                if len(BUFFER[1]) >= 1:
                    break

        self.FFBUFFER = BUFFER
        # pass reference
        databuffer.set_buffer(self.FFBUFFER_ID, self.FFBUFFER)

        logger.debug('    EXPORTED FFBUFFER SIZE: %d' % len(self.FFBUFFER[2]))

        # HINT: keep only eight items, namely the distances between each seed
        # (i.e. image that will be displayed) and the current trace
        BUFFER = [self.DDBUFFER[0], [], [], []]
        sortedarg = argsort(self.DDBUFFER[1])
        for i in sortedarg[::-1]:
            if len(self.DDBUFFER[2][i][1]) > 8:
                BUFFER[1].append(self.DDBUFFER[1][i])
                BUFFER[2].append(self.DDBUFFER[2][i])
                BUFFER[3].append(self.DDBUFFER[3][i])

                if len(BUFFER[1]) >= 8:
                    break

        self.DDBUFFER = BUFFER
        # pass reference
        databuffer.set_buffer(self.DDBUFFER_ID, self.DDBUFFER)

        logger.debug('    EXPORTED DDBUFFER SIZE: %d' % len(self.DDBUFFER[2]))

    ############################################################
    # main functionality

    def __init__(self):
        """
        Initialize the object.
        """

        super(DbTableImg, self).__init__()

        self.suffix = 'img'
        self.DIR_DIST = None
        self.FFBUFFER_ID = 'FF'
        self.DDBUFFER_ID = 'DD'

        self._init_buffers()

    def init(self, table, cnx, DIR_DIST, DDBUFFER_ID=''):
        """
        Initialize the object.
        """

        super(DbTableImg, self).init(table, cnx)

        self.DIR_DIST = DIR_DIST

        if DDBUFFER_ID != '':
            self.DDBUFFER_ID = DDBUFFER_ID

        self.all_hdbid = None
        if self.is_valid() == True:
            try:
                self.has_featdata = self._has_featdata(1)
            except:
                self.has_featdata = None
        else:
            self.has_featdata = None

    #######################################################################
    # configure the database

    def init_table(self):
        """
        Initialize the table.
        """

        self.cnx.execute(
                "CREATE TABLE %s ("
                "id INTEGER NOT NULL auto_increment, "
                "valid TINYINT(1) NOT NULL default '1', "
                "pathToOrig VARCHAR(255) NOT NULL, "
                "pathToDist VARCHAR(255) NOT NULL, "
                "featdata0 BLOB NOT NULL, "
                "featdata1 BLOB NOT NULL, "
                "parentnode_dbid INTEGER NOT NULL default '0', "
                "PRIMARY KEY (id)"#, "
                #"UNIQUE KEY pathToDist (pathToDist)"
                ")"
                % self.table)

        self.cnx.commit()

    def optimize_table(self):
        """
        Optimize the table.
        """

        self.cnx.execute(
                "CREATE UNIQUE INDEX pathToDist_idx ON %s (pathToDist)"
                % self.table)

        self.cnx.commit()

    #######################################################################
    # set functionality

    def insert_img(self, pathorig, pathdist, feat0, feat1):
        """
        Add the image data.
        """

        self.cnx.execute(
                "INSERT INTO %s ("
                "pathToOrig, pathToDist, "
                "featdata0, featdata1"
                ") VALUES ("
                "'%s', '%s', '%s', '%s'"
                ")"
                % (self.table, pathorig, pathdist, feat0, feat1))

        self.cnx.commit()

        # retrieve the dbid
        self.cnx.execute("SELECT LAST_INSERT_ID()")
        dbid = self.cnx.fetchone()[0]

        return dbid

    def update_dist(self, dbid, pathtodist):
        """
        Update the image distances filename.
        """

        self.cnx.execute(
                "UPDATE %s "
                "SET pathToDist = '%s' "
                "WHERE id = %d"
                % (self.table, pathtodist, dbid))

        self.cnx.commit()

    def update_feat(self, dbid, feat0, feat1):
        """
        Update the image feature vectors.
        """

        self.cnx.execute(
                "UPDATE %s "
                "SET featdata0 = '%s', featdata1 = '%s' "
                "WHERE id = %d"
                % (self.table, feat0, feat1, dbid))

        self.cnx.commit()

    def update_featdata(self, dbid, feat0, feat1, line0, line1):
        """
        Update the image feature vectors.
        """

        self.cnx.execute(
                "UPDATE %s "
                "SET featdata0 = '%s', featdata1 = '%s' "
                "WHERE id = %d"
                % (self.table, line0, line1, dbid))

        self.cnx.commit()

        # progress info
        if logger.getEffectiveLevel() <= logging.DEBUG:
            # retrieve the node dbid
            self.cnx.execute("SELECT LAST_INSERT_ID()")
            node_dbid = self.cnx.fetchone()[0]

            self.cnx.execute(
                    "SELECT featdata0 FROM %s WHERE id = %d"
                    % (self.table, node_dbid))

            data = self.cnx.fetchall()

            logger.debug(
                    'DBID %d DATA %s TYPE %s'
                    % (node_dbid, data, type(data)))

    def update_valid_url(self, dbid, valid):
        """
        Update the image valid flag.
        """

        self.cnx.execute(
                "UPDATE %s "
                "SET valid = '%s' "
                "WHERE id = %d"
                % (self.table, valid, dbid))

        self.cnx.commit()

    #######################################################################
    # image information

    def im_valid(self, dbid):
        """
        Get the valid flag of the image with the given id.
        """

        return self.get_data('id, valid', 'id', dbid, True)

    def info(self, dbid):
        """
        Get info for the image with the given id.
        """

        if not isinstance(dbid, (tuple, list, ndarray)):
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % type(dbid))
            sys.exit(1)

        return self.get_data('id, pathToOrig', 'id', dbid, True)

    def im_path(self, dbid):
        """
        Get the path of the image with the given id.
        """

        return self.get_data('id, pathToOrig', 'id', dbid, True)

    def im_path2(self, dbid):
        """
        Get the path of the image with the given id.
        """

        return self.get_data('id, valid, pathToOrig', 'id', dbid, True)

    def dbid(self, path):
        """
        Get the dbid of the image with the given path.
        """

        return self.get_data('pathToOrig, id', 'pathToOrig', path, True)

    def all_valid_ids(self):
        """
        Get all the ids. Format: ndarray, sorted
        """

        values = self.get_data('id', 'valid', 1, True)

        ids = [item[0] for item in values]
        ids = array(ids, dtype=int32, order='c')

        return ids

    def parentnode_dbid(self, dbid):
        """
        Get the dbid of the parent node of the given images.
        """

        values = self.get_data('parentnode_dbid', 'id', dbid, True)

        ids = [item[0] for item in values]
        ids = array(ids, dtype=int32, order='c')

        return ids

    #######################################################################
    # image features

    cdef tuple _ifeatdata(self, cnp.ndarray dbid, cnp.int32_t tt):
        """
        Get the feature vector list of the image with the given id,
        for the corresponding feature type.
        """

        cdef cnp.ndarray[cnp.int_t, ndim=1, mode='c'] myorder
        cdef cnp.ndarray tdata
        cdef tuple data
        cdef list featdata
        cdef str field
        cdef cnp.int32_t t, nt, l0, ln

        if tt == -1:
            field = 'featdata0, featdata1'
            nt = 2
        elif tt == 0:
            field = 'featdata0'
            nt = 1
        elif tt == 1:
            field = 'featdata1'
            nt = 1
        else:
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % tt)
            sys.exit(1)

        # get the raw data (ORDERED BY id ASC)
        data = self.get_data(field, 'id', dbid, False)

        # RECOVER the order
        # HINT: mixed with the string operations
        myorder = argsort(argsort(dbid))

        featdata = []
        for t in xrange(0, nt, 1):
            # WARNING: fromstring returns ndarray even for empty strings
            tdatastr = ' '.join([data[i][t] for i in myorder])
            tdata = fromstring(tdatastr, dtype=float32, sep=' ')
            l0 = data[0][t].count(' ') + 1
            ln = tdata.shape[0]
            tdata = reshape(tdata, (ln/l0, l0), order='c')

            featdata.append(tdata)

        return tuple(featdata)

    cpdef tuple featdata(self, cnp.ndarray dbid, cnp.int32_t tt, bint buffered):
        """
        Get the pre-computed feature vectors of the image with the given id,
        for the corresponding feature type.
        """

        cdef tuple data

        #logger.debug('START')
        #start_time = now()

        # first, ask the buffer
        data = buffer_get1(self.FFBUFFER, dbid)
        if data is not None:
            #stop_time = now()
            #logger.debug('    DONE [BUFFER] in %f seconds'
            #             % (stop_time - start_time))
            return data

        # otherwise, retrieve the data
        data = self._ifeatdata(dbid, tt)

        # then, add to the buffer
        if buffered == True and tt == -1:
            # WARNING: buffer key CANNOT distinguish different 't'
            buffer_set1(self.FFBUFFER, dbid, data)

        #stop_time = now()
        #logger.debug('    DONE in %f seconds' % (stop_time - start_time))

        return data

    cdef tuple distdata_feat(self, tuple kfeat, tuple hfeat):

        cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] dd_k
        cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] kfeat_t, hfeat_t

        cdef cnp.int32_t k, nk, h, nh, t, nt, i, ni
        cdef cnp.float32_t tmp, tmpsum
        cdef list dd

        nk = kfeat[0].shape[0]
        nh = hfeat[0].shape[0]
        nt = len(kfeat)

        # initialize empty arrays
        dd = []
        for k from 0 <= k < nk:
            dd_k = empty((nh, nt), dtype=float32, order='c')
            dd.append(dd_k)

        for t from 0 <= t < nt:
            kfeat_t = kfeat[t]
            hfeat_t = hfeat[t]
            # length of the vectors of type t
            ni = kfeat_t.shape[1]

            for k from 0 <= k < nk:
                dd_k = dd[k]

                for h from 0 <= h < nh:

                    # compute the Euclidean distance between the vectors
                    tmpsum = 0.0
                    for i from 0 <= i < ni:
                        tmp = kfeat_t[k,i] - hfeat_t[h,i]
                        tmpsum += tmp * tmp

                    # HINT: tmpsum can be infinitesimal negative due to float32 accuracy
                    if tmpsum > 0.0:
                        dd_k[h,t] = c_sqrt(tmpsum)
                    else:
                        dd_k[h,t] = 0.0

        return tuple(dd)

    #######################################################################
    # pre-computed image distances

    def distfile(self, dbid):
        """
        Get the image distance filenames (i.e. containing the pre-computed
        distances).
        """

        if self.DIR_DIST is None:
            sys.stderr.write('\nUNDEFINED DIR_DIST\n')
            sys.exit(1)

        pathToDist = self.get_data('id, pathToDist', 'id', dbid, True)

        relpath = []

        for i, rpath in pathToDist:
            # if necessary, add the default extension
            if '.' not in rpath:
                rpath += '.bin'
            # get the absolute path
            rpath = os.path.join(self.DIR_DIST, rpath)

            relpath.append((i, rpath))

        return relpath

    def _has_featdata(self, dbid):
        """
        Return True if the image dbid has feature's data.
        Otherwise, return False.
        """

        to_return = False
        data = self.get_data('featdata0', 'id', dbid, False)
        if data[0][0] != '':
            to_return = True

        return to_return

    def hasdist(self, dbid):
        """
        Return True if the image dbid has a distance file.
        Otherwise, return False.
        """

        hasdist = False
        filename = self.distfile([dbid,])
        if os.access(filename, os.F_OK):
            hasdist = True

        return hasdist

    cdef tuple _distdata_dist(self,
            cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] kdbid,
            cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] hdbid,
            cnp.int32_t tt):
        """
        Get the pre-computed distances of the image with the given id,
        for the corresponding feature type.

        Form: tuple(array(hdbid:,tt:),...)
        """

        cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] odd
        cdef cnp.int32_t idbid, dbid, all_hdbidln, trix
        cdef list htrix, dd

        # identify the corresponding trix, if necessary
        htrix = []
        if hdbid.shape[0] != self.size():
            # initialize the dbid, if necessary
            if self.all_hdbid is None:
                self.all_hdbid = self.all_ids()

            # identify the trix corresponding to the hdbid
            all_hdbidln = len(self.all_hdbid)
            for dbid in hdbid:

                # identify the corresponding trix
                trix = dbid - 1 # for efficiency, this is the case for mm_0
                if all_hdbidln <= trix or self.all_hdbid[trix] != dbid:
                    # if the guess was not correct
                    trix = get_index0(dbid, self.all_hdbid, False)
                if trix == -1:
                    sys.stderr.write('\nUNKNOWN DBID %s\n' % dbid)
                    sys.exit(1)

                htrix.append(trix)

        # get the distdata
        dd = []
        filenames = self.distfile(kdbid)
        for idbid, filename in filenames:

            idd = read_bindata(filename)

            # trim distdata, if necessary
            if htrix:
                idd = idd[htrix,:]

            if idd.shape[1] == 2:
                if tt in (0, 1):
                    idd = squeeze(idd[:,tt])
                else:
                    pass
            elif idd.shape[1] == 1:
                if tt == -1:
                    odd = zeros((idd.shape[0], 2), dtype=float32, order='c')
                    odd[:,0] = idd[:,0]
                    idd = odd
                elif tt == 0:
                    pass
                elif tt == 1:
                    sys.stderr.write('\nUNAVAILABLE DIST t=%s\n' % tt)
                    sys.exit(1)
            else:
                sys.stderr.write('\nUNKNOWN DISTDATA SHAPE\n')
                sys.exit(1)

            dd.append(idd)

        return tuple(dd)

    cpdef tuple distdata(self,
            cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] kdbid,
            cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] hdbid,
            cnp.int32_t tt, bint buffered, bint special):
        """
        Get the pre-computed distances of the image with the given id,
        for the corresponding feature type.

        Form: tuple(array(hdbid:,tt:),...)
        """

        cdef list datalist
        cdef tuple key, data, ikey, idata
        cdef cnp.int32_t i, j, kdbidln, hdbidln

        logger.debug('    START')
        start_time = now()

        if tt not in (-1, 0, 1):
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % tt)
            sys.exit(1)

        hdbidln = hdbid.shape[0]

        # first, ask the buffer
        key = (kdbid, hdbid)
        data = buffer_get2(self.DDBUFFER, key)
        if data is not None:
            stop_time = now()
            logger.debug(
                    '    DISTDATA DONE [BUFFER] in %f seconds %s %s'
                    % (stop_time - start_time, kdbid, hdbidln))
            return data

        # second, ask the buffer for items/parts, and concatenate them
        # HINT: check if it is worth to go into this special case
        if hdbidln > 500 and special == True and tt == -1:
            kdbidln = kdbid.shape[0]
            if kdbidln > 1 and kdbidln < 9:
                #logger.debug('DD SPECIAL kdbid %s hdbidln %s' % (kdbid, hdbidln))

                datalist = None
                for i in xrange((kdbidln - 1), 0, -1):
                    ikey = (kdbid[:i], hdbid)
                    idata = buffer_get2(self.DDBUFFER, ikey)
                    if idata is not None:
                        # HINT: do NOT deepcopy, it will NOT be modified in place
                        datalist = list(idata)
                        logger.debug('DD1 %s %s %s' % (i, kdbid, hdbidln))
                        break

                    #logger.debug('DD1 NOT %s %s %s' % (i, kdbid, len(hdbid)))

                if datalist is not None:
                    for j in xrange(i, kdbidln, 1):
                        ikey = (kdbid[j:(j+1)], hdbid)
                        idata = buffer_get2(self.DDBUFFER, ikey)
                        if idata is None:
                            # if one is not buffered, give up and get them all at once
                            idata = self.distdata(kdbid[j:], hdbid, tt, False, False)
                            datalist.extend(idata)
                            logger.debug('DD2 %s %s %s' % (j, kdbid, hdbidln))
                            break
                        else:
                            # if buffered, just append it
                            datalist.extend(idata)
                            logger.debug('DD2 %s %s %s' % (j, kdbid, hdbidln))

                    # data was successfully concatenated
                    data = tuple(datalist)

        # otherwise, compute the distances
        if data is None:
            if hdbidln == self.size():
                # read the files with pre-computed distances
                data = self._distdata_dist(kdbid, hdbid, tt)

            elif self.has_featdata == True:
                # compute the distances on-the-fly
                kfeat = self.featdata(kdbid, tt, True)
                hfeat = self.featdata(hdbid, tt, True)
                data = self.distdata_feat(kfeat, hfeat)

                #data0 = self._distdata_dist(kdbid, hdbid, tt)
                #print 'kdbid', kdbid, hdbid
                #print 'data', [sum(abs(data[i] - data0[i])) for i in xrange(0, len(data), 1)]

            elif self.has_featdata == False:
                # read the files with pre-computed distances
                data = self._distdata_dist(kdbid, hdbid, tt)
                #print 'prec', kdbid, hdbid

            else:
                pass

        # then, add to the buffer
        if buffered == True and tt == -1:
            # WARNING: buffer key CANNOT distinguish different 't'
            buffer_set2(self.DDBUFFER, key, data)

        stop_time = now()
        logger.debug(
                '    DISTDATA DONE in %f seconds %s %s'
                % (stop_time - start_time, kdbid, hdbidln))

        return data

    def compute_distdata(self, kdbid, hdbid):
        """
        Compute the distances.
        """

        # compute the distances on-the-fly
        kfeat = self.featdata(kdbid, -1, True)
        hfeat = self.featdata(hdbid, -1, True)

        data = self.distdata_feat(kfeat, hfeat)

        return data

    #######################################################################

    cdef expcol_featdata(self,
            cnp.ndarray newdbid,
            cnp.ndarray dbid, list trix, cnp.ndarray deltadbid):
        """
        Prepare efficiently the FF databuffer for the new trace.
        """

        cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] item, deltaitem, newitem

        cdef list newfeatdatalist
        cdef tuple featdata, deltafeatdata, newfeatdata

        #logger.debug('    START')
        start_time = now()

        # first, ask the buffer
        newfeatdata = buffer_get1(self.FFBUFFER, newdbid)
        if newfeatdata is not None:
            stop_time = now()
            logger.debug(
                    '    DONE [COINCIDENCE] in %f seconds'
                    % (stop_time - start_time))
            return

        # retrieve the current data
        featdata = buffer_get1(self.FFBUFFER, dbid)
        if featdata is None:
            logger.warning('\n    MISSING RANK FEATDATA\n')
            featdata = self.featdata(dbid, -1, False)
            #sys.stderr.write('\nMISSING RANK FEATDATA\n')
            #sys.exit(1)
            #return

        # retrieve the 'delta' data
        if len(deltadbid) > 0:
            deltafeatdata = self.featdata(deltadbid, -1, True)
        else:
            farray = array([[]], dtype=float32, order='c')
            deltafeatdata = (farray, farray)

        # NEWFEATDATA: extend each feature data
        newfeatdatalist = []
        for i, item in enumerate(featdata):
            deltaitem = deltafeatdata[i]
            newitem = extend_float32_ndim2(item, trix, deltaitem)
            newfeatdatalist.append(newitem)
        # reformat
        newfeatdata = tuple(newfeatdatalist)

        # then, add to the buffer
        buffer_set1(self.FFBUFFER, newdbid, newfeatdata)

        stop_time = now()
        logger.debug(
                '    expcol_featdata DONE in %f seconds'
                % (stop_time - start_time))

    cdef expcol_dd(self,
            cnp.ndarray seeddbid, cnp.ndarray newdbid,
            cnp.ndarray dbid, list trix, cnp.ndarray deltadbid):
        """
        Prepare efficiently the DD databuffer for the new trace.
        """

        cdef cnp.ndarray kdbid, hdbid
        cdef tuple dd, deltadd, newdd
        cdef tuple key

        cdef list newddlist

        #logger.info('    START')
        start_time = now()

        ###
        kdbid = seeddbid

        # first, ask the buffer (for pure coincidence)
        key = (kdbid, newdbid)
        data = buffer_get2(self.DDBUFFER, key)
        if data is not None:
            #stop_time = now()
            #logger.debug(
            #        '    DONE [BUFFER] in %f seconds'
            #        % (stop_time - start_time))
            return

        # otherwise, compose the new dd by merging the old/new parts
        hdbid = dbid
        dd = self.distdata(kdbid, hdbid, -1, False, True)
        if len(deltadbid) > 0:
            deltadd = self.distdata(kdbid, deltadbid, -1, False, False)
        else:
            farray = array([[]], dtype=float32, order='c')
            deltadd = (farray, farray)

        # NEWDD: extend each distance item
        newddlist = []
        for i, item in enumerate(dd):
            deltaitem = deltadd[i]
            newitem = extend_float32_ndim2(item, trix, deltaitem)
            newddlist.append(newitem)
        # reformat
        newdd = tuple(newddlist)

        # then, add to the buffer
        buffer_set2(self.DDBUFFER, key, newdd)
        #print 'EXCOL', kdbid, newdd.shape[0], newdd.shape[1], newdd.shape[2]

        stop_time = now()
        logger.debug(
                '    expcol_dd DONE in %f seconds'
                % (stop_time - start_time))

###########################################################################
# CLASS

cdef class DbTableImgSynthetic(DbTableImg):
    """
    DbTableImgSynthetic is a sub-class of DbTableImg.

    .. note:: It generates the synthetic images on-the-fly.
    """

    ############################################################
    # main functionality

    def init(self, table, cnx, DIR_DIST, DIR_IMG, DDBUFFER_ID=''):
        """
        Initialize the object.
        """

        super(DbTableImgSynthetic, self).init(table, cnx, DIR_DIST, DDBUFFER_ID=DDBUFFER_ID)

        self.DIR_IMG = DIR_IMG

    #######################################################################
    # image information

    def info(self, dbid):
        """
        Get all info for the image with the given dbid.

        .. note:: If the image file does not exists, it is generated on the fly.
        """

        if not isinstance(dbid, (list, tuple, ndarray)):
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % type(dbid))
            sys.exit(1)

        data = self.get_data('id, pathToOrig', 'id', dbid, True)

        # generate the corresponding images, if necessary
        # HINT: images might be created already
        for index, item in enumerate(data):
            filename = self.DIR_IMG + '/' + item[1]

            # check if the file exists already
            if not os.access(filename, os.F_OK):

                logger.debug('    SYNTHETIC IMAGE %s' % filename)

                imfeat = self.featdata(dbid, -1, False)
                coord2D = ([imfeat[0][index,0]], [imfeat[0][index,1]])
                create_dir(filename)
                generate_image2D(coord2D, filename)

                # check if the file exists now
                if not os.access(filename, os.F_OK):
                    logger.warning('    FAILED TO CREATE %s' % filename)

        return data