#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Manages the table of the hierarchical tree-like organization of the images.
"""

###########################################################################
# IMPORTS

import datetime
import logging
from numpy import int32, float32, ndarray, array, zeros, ones, \
                  append, argsort, fromstring
import sys

from retrieval_libs.dblayer.DbCnx import DbCnx
from retrieval_libs.dblayer.DbTable cimport DbTable

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

###########################################################################
# CLASS

cdef class DbTableTree(DbTable):
    """
    Manages the table of the hierarchical tree-like organization of the images.
    """

    ############################################################
    # main functionality

    def __init__(self):
        """
        Initialize the object.
        """

        DbTable.__init__(self)

        self.suffix = 'tree'

    #######################################################################
    # configure the table

    def init_table(self):
        """
        Initialize the table.
        """

        self.cnx.execute(
                "CREATE TABLE %s("
                "id INTEGER NOT NULL auto_increment, "
                "depth INTEGER NOT NULL default '-1', "
                "name VARCHAR(100) NOT NULL, "
                "statistics TINYTEXT, "
                "seed_dbid INTEGER NOT NULL, "
                "parent_dbid INTEGER NOT NULL, "
                "subtree_size INTEGER NOT NULL, "
                "children_size INTEGER NOT NULL, "
                "children_table VARCHAR(20) NOT NULL, "
                "children_dbid LONGBLOB NOT NULL, "
                "PRIMARY KEY (id)"#, "
                #"UNIQUE KEY seed_dbid (seed_dbid)"
                ")"
                % self.table)

        self.cnx.commit()

    def optimize_table(self):
        """
        Optimize the table.
        """

        self.cnx.execute(
                "CREATE UNIQUE INDEX seed_dbid_idx ON %s (seed_dbid)"
                % self.table)

        self.cnx.execute(
                "CREATE UNIQUE INDEX name_idx ON %s (name)"
                % self.table)

        self.cnx.commit()

    #######################################################################
    # set functionality

    def insert_node(self,
            depth, name, seed_dbid, statistics, parent_dbid,
            children_size ,children_table, children_dbid):
        """
        """

        self.cnx.execute(
                "INSERT INTO %s ("
                "depth, name, seed_dbid, subtree_size, statistics, parent_dbid, "
                "children_size, children_table, children_dbid"
                ") VALUES ("
                "%d, '%s', %d, 0, '%s', %d, %d, '%s', '%s'"
                ")"
                % (self.table, depth, name, seed_dbid, statistics,
                parent_dbid, children_size, children_table, children_dbid))

        self.cnx.commit()

        # retrieve the node dbid
        self.cnx.execute("SELECT LAST_INSERT_ID()")
        dbid = self.cnx.fetchone()[0]

        return dbid

    def insert_root(self):
        """
        Add the ROOT node.
        """

        dbid = 1
        level = 0
        name = '0'
        seed_dbid = 0
        subtree_size = 0
        radius = '0.0'
        parent_dbid = 0
        children_size = 0
        children_table = ''
        children_dbid = ''

        # save the node data
        self.cnx.execute(
                "INSERT INTO %s ("
                "id, depth, name, seed_dbid, subtree_size, statistics, "
                "parent_dbid, children_size, children_table, children_dbid"
                ") VALUES ("
                "%d, %d, '%s', %d, %d, '%s', %d, %d, '%s', '%s'"
                ")"
                % (self.table,
                    dbid, level, name, seed_dbid, subtree_size, radius,
                    parent_dbid, children_size, children_table, children_dbid))

        self.cnx.commit()

    def insert_child(self, name, parent_name, children_table, seed_dbid, radius, children_dbid):
        """
        Add new node to the table.
        """

        parent_dbid = self.get_data('name, id', 'name', parent_name, True)
        parent_level = self.get_data('name, depth', 'name', parent_name, True)
        level = parent_level + 1

        if children_dbid is None:
            subtree_size = 0
            children_size = 0
            children_dbid = ''
            radius = '0.0'
        else:
            subtree_size = 0
            children_size = len(children_dbid)
            children_dbid = ' '.join(map(str, children_dbid))
            radius = str(radius)

        # save the node data
        self.cnx.execute(
                "INSERT INTO %s ("
                "depth, name, seed_dbid, subtree_size, statistics, parent_dbid, "
                "children_size, children_table, children_dbid"
                ") VALUES ("
                "%d, '%s', %d, %d, '%s', %d, %d, '%s', '%s'"
                ")"
                % (self.table,
                level, name, seed_dbid, subtree_size, radius, parent_dbid,
                children_size, children_table, children_dbid))

        self.cnx.commit()

        # retrieve the node dbid
        self.cnx.execute("SELECT LAST_INSERT_ID()")
        dbid = self.cnx.fetchone()[0]

        return dbid

    def insert_leaf(self, dbid, cdbid):
        """
        Insert a leaf image cdbid into the children of dbid.
        """

        # check the children table
        tbl = self.get_data('id, children_table', 'id', dbid)
        if tbl is not None and tbl != '' and tbl != 'img':
            sys.stderr.write('\nNode %s refers to another table %s\n' % (dbid, tbl))
            sys.exit(1)

        # get the current children
        children = self.get_data('id, children_dbid', 'id', dbid)
        if children is not None and len(children) > 0:
            children = fromstring(children, dtype=int32, sep=' ')
            children = list(children)
        else:
            children = []

        # add the new given child
        if cdbid not in children:
            children.append(cdbid)

        children.sort()

        # update the table
        children_size = len(children)
        children = ' '.join(map(str, children))

        self.cnx.execute(
                "UPDATE %s "
                "SET children_table = '%s', "
                "children_size = '%s', "
                "children_dbid = '%s' "
                "WHERE id = '%s'"
                % (self.table, 'img', children_size, children, dbid))

        self.cnx.commit()

        logger.info('    DBID %s' % dbid)

    def update_parent(self, dbid, already_updated=None):
        """
        Update the parent of the given dbid.
        """

        if already_updated and dbid in already_updated:
            return

        children_dbid = self.get_data('id', 'parent_dbid', dbid, True)
        if children_dbid is None:
            logger.debug('    DBID %s is a leaf node' % dbid)

            self.cnx.execute(
                    "UPDATE %s "
                    "SET children_table = '%s', "
                    "children_size = %d, "
                    "children_dbid= '%s' "
                    "WHERE id = %d"
                    % (self.table, 'img', 0, '', dbid))

            return

        children_dbid = [int32(c[0]) for c in children_dbid]

        children_dbid.sort()

        children_table = self.suffix
        children_size = len(children_dbid)
        children_dbidstr = ' '.join(map(str, children_dbid))

        self.cnx.execute(
                "UPDATE %s "
                "SET children_table = '%s', "
                "children_size = %d, "
                "children_dbid = '%s' "
                "WHERE id = %d"
                % (self.table,
                children_table, children_size, children_dbidstr, dbid))

        self.cnx.commit()

        logger.debug('    DBID %s is a non-leaf node' % dbid)

    def update_children(self, dbid):
        """
        Update the children of the given dbid.
        """

        pname = self.get_data('id, name', 'id', dbid, True)
        plevel = self.get_data('id, depth', 'id', dbid, True)
        ptbl = self.get_data('id, children_table', 'id', dbid, True)
        if ptbl != self.suffix:
            logger.debug('    Node %s refers to another table %s' % (pname, ptbl))
            return

        childlist = self.get_data('id, children_dbid', 'id', dbid, True)
        if len(childlist) > 0:
            childlist = fromstring(childlist, dtype=int32, sep=' ')
        else:
            # degenerated array
            childlist = array([], dtype=int32, order='c')

        for i, cdbid in enumerate(childlist):

            clevel = plevel + 1
            cname = pname + ('%d' % i)
            parent_dbid = dbid

            self.cnx.execute(
                    "UPDATE %s "
                    "SET depth = %d, name = '%s', parent_dbid = %d "
                    "WHERE id=%d"
                    % (self.table, clevel, cname, dbid, cdbid))

        self.cnx.commit()

    def update_children_data(self, dbid, children):
        """
        Update the children of the given dbid.
        """

        children_size = len(children)

        if children_size == 0:
            children_dbid = ''
            children_table = 'img'
        else:
            children.sort()
            children_dbid = ' '.join(map(str, children))
            children_table = self.suffix

        self.cnx.execute(
                "UPDATE %s "
                "SET children_table = '%s', "
                "children_size = '%s', "
                "children_dbid = '%s' "
                "WHERE id = '%s'"
                % (self.table,
                children_table, children_size, children_dbid, dbid))

        self.cnx.commit()

    def update_subtree_size(self, dbid, subtree_size):

        self.cnx.execute(
                "UPDATE %s "
                "SET subtree_size = %d "
                "WHERE id = %d"
                % (self.table, subtree_size, dbid))

        self.cnx.commit()

    def delete_node(self, dbid):
        """
        Delete node and its sub-tree.
        """

        pdbid = self.get_data('id, parent_dbid', 'id', dbid, True)

        # delete the node data
        self.cnx.execute("DELETE FROM %s WHERE id = %d" % (self.table, dbid))

        # update the parent
        children_dbid = self.get_data('id, children_dbid', 'id', pdbid, True)
        ctbl = self.get_data('id, children_table', 'id', pdbid, True)

        if len(children_dbid) > 0:
            children_dbid = fromstring(children_dbid, dtype=int32, sep=' ')
        else:
            # degenerated array
            children_dbid = array([], dtype=int32, order='c')

        children_dbid = children_dbid.tolist()
        children_dbid.remove(dbid)
        children_size = len(children_dbid)
        children_dbid = ' '.join(map(str, children_dbid))

        if children_size == 0:
            # delete recursively
            self.delete_node(pdbid)

        else:
            self.cnx.execute(
                    "UPDATE %s "
                    "SET children_table = '%s', "
                    "children_size = %d, "
                    "children_dbid = '%s' "
                    "WHERE id = '%s'"
                    % (self.table,
                    ctbl, children_size, children_dbid, pdbid))

            self.cnx.commit()

        logger.info('    DELETED %s' % dbid)

    #######################################################################
    # get functionality

    def size(self):
        """
        Return the size of the table.
        """

        return self._size(self.table)

    def dbid(self, name):
        """
        Get the dbid of the node with the given name.
        """

        return self.get_data('name, id', 'name', name, True)

    def name(self, dbid):
        """
        Get the name of the node with the given id.
        """

        return self.get_data('id, name', 'id', dbid, True)

    def nb_leaves(self):
        """
        Get the number of leave nodes.
        """

        data = self.get_data ('COUNT(*)', 'children_table', ('img',), True)

        return data[0][0]

    def children_table(self, dbid):
        """
        Get the children table of the node with the given id.
        """

        return self.get_data('id, children_table', 'id', dbid, True)

    cpdef children_size(self, dbid):
        """
        Get the children size of the node with the given id.
        """

        return self.get_data('id, children_size', 'id', dbid, True)

    def subtree_size(self, dbid):
        """
        Get the size of subtree of the node with the given id.
        """

        return self.get_data('id, subtree_size', 'id', dbid, True)

    def radius(self, dbid):
        """
        Get the radius of the node with the given id.
        """

        data = self.get_data('id, statistics', 'id', dbid, True)
        try:
            radius = float32(data)
        except ValueError:
            radius = data

        return radius

    def level(self, dbid):
        """
        Get the level of the node with the given id.
        """

        return self.get_data('id, depth', 'id', dbid, True)

    def level_of_seed_dbid(self, dbid):
        """
        Get the level of the node with the given seed_dbid.
        """

        return self.get_data('seed_dbid, depth', 'seed_dbid', dbid, True)
        
    def seed_dbid(self, dbid):
        """
        Get the seed_dbid of the node with the given id.
        """

        return self.get_data('id, seed_dbid', 'id', dbid, True)

    def tndbid(self, dbid):
        """
        Get the node of the seed with the given id.
        """

        data = self.get_data('seed_dbid, id', 'seed_dbid', dbid, False)
        tndbid = [int32(item[1]) for item in data]

        data = self.tracedata(tndbid)
        return data

    cpdef chsize_level(self, dbid):
        """
        Get the children size of the node with the given id.
        """

        return self.get_data('id, children_size, depth', 'id', dbid, True)

    #######################################################################

    cpdef children_data(self, dbid):
        """
        Get the children-related data of the node with the given id.
        """

        if not isinstance(dbid, (list, tuple, ndarray)):
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % type(dbid))
            sys.exit(1)

        return self.get_data('id, depth, children_table, children_size, children_dbid', 'id', dbid, True)

    #######################################################################
    # trace functionality

    cdef dict tracedata(self, cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] dbid):
        """
        Get the trace-related data of the node with the given id.
        """

        cdef cnp.ndarray[cnp.int_t, ndim=2, mode='c'] data
        cdef cnp.ndarray[cnp.int_t, ndim=1, mode='c'] idata
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] tndbid, imdbid, tnsize, pndbid, chsize, tnlevel
        cdef cnp.ndarray[cnp.int_t, ndim=1, mode='c'] myorder
        cdef cnp.int32_t myorderln, i, imy
        cdef tuple dbdata
        cdef dict tracedata

        dbdata = self.get_data('id, seed_dbid, subtree_size, parent_dbid, children_size, depth', 'id', dbid, False)
        try:
            data = array(dbdata, dtype=int, order='c')
        except:
            logger.error('\n\n DBDATA\n %s\n', str(dbdata))
            sys.stderr.write('\nCONVERSION TO NUMPY ARRAY FAILED\n')
            sys.exit(1)

        # HINT: extra-sorting operation for PRESERVING the order
        myorder = argsort(argsort(dbid))
        myorderln = myorder.shape[0]
        if myorderln > 0:
            shape = (myorderln,)
            imdbid = zeros(shape, dtype=int32, order='c')
            tndbid = zeros(shape, dtype=int32, order='c')
            tnsize = zeros(shape, dtype=int32, order='c')
            pndbid = zeros(shape, dtype=int32, order='c')
            chsize = zeros(shape, dtype=int32, order='c')
            tnlevel = zeros(shape, dtype=int32, order='c')

            for i from 0 <= i < myorderln:
                imy = myorder[i]
                idata = data[imy]

                tndbid[i] = idata[0]
                imdbid[i] = idata[1]
                tnsize[i] = idata[2]
                pndbid[i] = idata[3]
                chsize[i] = idata[4]
                tnlevel[i] = idata[5]

            tracedata = {}
            tracedata['tndbid'] = tndbid
            tracedata['imdbid'] = imdbid
            tracedata['tnsize'] = tnsize
            tracedata['pndbid'] = pndbid
            tracedata['chsize'] = chsize
            tracedata['tnlevel'] = tnlevel

        else:
            tracedata = None

        return tracedata

    cdef dict expand_tracedata(self, cnp.ndarray dbid):

        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] item, deltaitem
        cdef list leafimdbid, tmp, leafpndbid
        cdef dict noleafdata, leafdata
        cdef dict data
        cdef tuple childdata
        cdef cnp.int32_t leafdataln

        childdata = self.children_data(dbid)

        noleafdbid = []
        leafimdbid = []
        leafpndbid = []
        for idbid, ilevel, ictbl, ichsize, icdbid in childdata:

            # HINT: check if ichsize larger than 0 because
            # fromstring misbehaves for empty string

            if ictbl == 'tree':
                icdbid = fromstring(icdbid, dtype=int32, sep=' ')
                noleafdbid.extend(icdbid)

            elif ictbl == 'img' and ichsize == 0:
                noleafdbid.append(idbid)

            elif ictbl == 'img' and ichsize > 0:
                icdbid = fromstring(icdbid, dtype=int32, sep=' ')
                leafimdbid.extend(icdbid)
                tmp = [idbid] * len(icdbid)
                leafpndbid.extend(tmp)

            else:
                sys.stderr.write('\nUNKNOWN TABLE %s\n' % ictbl)
                sys.exit(1)

        leafdata = None
        leafdataln = len(leafimdbid)
        if leafdataln > 0:
            leafdata = {}
            shape = (leafdataln,)
            leafdata['imdbid'] = array(leafimdbid, dtype=int32, order='c')
            # set INVALID tndbid, chsize, tnlevel
            leafdata['tndbid'] = zeros(shape, dtype=int32, order='c')
            leafdata['tnsize'] = ones(shape, dtype=int32, order='c')
            leafdata['pndbid'] = array(leafpndbid, dtype=int32, order='c')
            leafdata['chsize'] = zeros(shape, dtype=int32, order='c')
            leafdata['tnlevel'] = zeros(shape, dtype=int32, order='c')
            leafdata['tnlevel'][:] = -1

        noleafdata = None
        if len(noleafdbid) > 0:
            noleafdbid = array(noleafdbid, dtype=int32, order='c')
            noleafdata = self.tracedata(noleafdbid)

        data = None
        if leafdata is None:
            data = noleafdata
        elif noleafdata is None:
            data = leafdata
        else:
            data = {}
            for key in leafdata.iterkeys():
                item = leafdata[key]
                deltaitem = noleafdata[key]
                data[key] = append(item, deltaitem, axis=0)

        return data

    cpdef list omega(self, dbid):
        """
        Return the list of images represented by the given tree nodes.
        """

        if not isinstance(dbid, (list, tuple, ndarray)):
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % type(dbid))
            sys.exit(1)

        omega = []

        # get the children-related data of the node with the given id
        data = self.get_data('id, seed_dbid, children_table, children_size, children_dbid', 'id', dbid, True)

        for idbid, iseed, ictbl, ichsize, icdbid in data:

            if ictbl == 'tree' and ichsize > 0:
                omega.append(iseed)
                icdbid = fromstring(icdbid, dtype=int32, sep=' ')
                # recursive call
                iomega = self.omega(icdbid)
                omega.extend(iomega)

            elif ictbl == 'img' and ichsize == 0:
                omega.append(iseed)

            elif ictbl == 'img' and ichsize > 0:
                omega.append(iseed)
                icdbid = fromstring(icdbid, dtype=int32, sep=' ')
                omega.extend(icdbid)

            else:
                sys.stderr.write(
                        '\nERROR DBID %s CH_TBL %s CH_SIZE %s CH_LIST %s\n' %
                        (idbid, ictbl, ichsize, icdbid))
                sys.exit(1)

        return omega

    def namedata(self, dbid):
        """
        """

        if not isinstance(dbid, (list, tuple, ndarray)):
            sys.stderr.write('\nUNKNOWN TYPE %s\n' % type(dbid))
            sys.exit(1)

        namedata = {}
        namedata['imdbid'] = []
        namedata['tnname'] = []
        namedata['pnname'] = []
        namedata['tndbid'] = []
        namedata['pndbid'] = []
        namedata['tnlevel'] = []
        namedata['chsize'] = []

        # get the children-related data of the node with the given id
        data = self.get_data('id, name, depth, parent_dbid, seed_dbid, children_table, children_size, children_dbid', 'id', dbid, True)

        noleafdbid = []

        for idbid, iname, idepth, ipndbid, iseed, ictbl, ichsize, icdbid in data:

            ipnname = self.get_data('id, name', 'id', ipndbid, True)

            # skip the root node
            if idbid > 1:
                namedata['imdbid'].append(iseed)
                namedata['tnname'].append(iname)
                namedata['pnname'].append(ipnname)
                namedata['tndbid'].append(idbid)
                namedata['pndbid'].append(ipndbid)
                namedata['tnlevel'].append(idepth)
                namedata['chsize'].append(ichsize)

            if ictbl == 'tree' and ichsize > 0:
                icdbid = fromstring(icdbid, dtype=int32, sep=' ')
                noleafdbid.extend(icdbid)

            elif ictbl == 'img' and ichsize == 0:
                pass

            elif ictbl == 'img' and ichsize > 0:
                icdbid = fromstring(icdbid, dtype=int32, sep=' ')

                namedata['imdbid'].extend(icdbid)
                namedata['tnname'].extend([''] * len(icdbid))
                namedata['pnname'].extend([iname] * len(icdbid))
                namedata['tndbid'].extend([0] * len(icdbid))
                namedata['pndbid'].extend([idbid] * len(icdbid))
                namedata['tnlevel'].extend([idepth + 1] * len(icdbid))
                namedata['chsize'].extend([0] * len(icdbid))

            else:
                sys.stderr.write(
                        '\nERROR DBID %s CH_TBL %s CH_SIZE %s CH_LIST %s\n'
                        % (idbid, ictbl, ichsize, icdbid))
                sys.exit(1)

        if noleafdbid:
            namedata2 = self.namedata(noleafdbid)
            for key in namedata:
                namedata[key].extend(namedata2[key])

        return namedata