#!/usr/bin/env python2.5

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides methods to access the data files for R/W/A operations.
"""

###########################################################################
# IMPORTS

import cPickle
import fcntl
import logging
import os
import sys
import time

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

###########################################################################
# FUNCTIONS

def f_lock(filename):
    """
    Lock the file and return its unique id that should be used later for
    unlocking the file.

    The locking mechanism ASSUMES that the file exists already.
    """

    if not os.access(filename, os.F_OK):
        sys.stderr.write('\nMISSING %s\n' % filename)
        sys.exit(1)

    # locking routine
    TIMEOUT = 3600 # seconds
    for failure_counter in xrange(0, TIMEOUT + 1, 1):
        try:
            fid = open(filename, 'r')
            fcntl.flock(fid, fcntl.LOCK_EX | fcntl.LOCK_NB)
            logger.debug('    ... locked successfully: %s' % filename)

        except IOError:
            # file is locked by someone else; try again after 1 second
            logger.info('    ... waiting for unlocking: %s' % filename)
            time.sleep(1)

        else:
            # file locked successfully; exit (properly) the locking routine
            break

    # check if the locking routine finished properly or it failed
    if failure_counter >= TIMEOUT:
        sys.stderr.write('\nTIMEOUT %s\n' % filename)
        sys.exit(1)

    return fid

###########################################################################

def create_dir(filename):
    """
    Create the directory tree implied by the given filename.
    """

    logger.debug('    FILE %s' % filename)

    dirname = os.path.dirname(filename)

    if os.path.isdir(dirname):
        return

    try:
        logger.debug('    CREATE DIR %s' % filename)
        os.makedirs(dirname)

    except OSError:
        if os.path.isdir(dirname):
            # created in the meanwhile by other thread
            logger.info('    CREATED IN THE MEANWHILE %s' % filename)
        else:
            logger.error('    ERROR %s' % filename)

#    if not os.access(dirname, W_OK):
#        os.chmod(dirname, 0755)

def create_file(filename):
    """
    Create an empty file with the given filename.
    """

    logger.debug('    FILE %s' % filename)

    create_dir(filename)

    if os.access(filename, os.F_OK):
        return

    try:
        f = open(filename, 'w')
        try:
            f.write('')
        except (KeyboardInterrupt, SystemExit):
            f.close()
            f = open(filename, 'w')
            f.write('')
            logger.info('    ... KeyboardInterrupt: File closed properly')
            raise
        finally:
            f.close()

    except IOError:
        logger.error('    ERROR %s' % filename)

    finally:
        pass

###########################################################################

def num(filename):
    """
    Return the number of lines in the given file.
    """

    f = open(filename, 'r')

    #nb_lines = 0
    #for line in f:
    #    nb_lines += 1

    buf_size = 1024 * 1024
    read_f = f.read # loop optimization
    buf = read_f(buf_size)
    while buf:
        nb_lines += buf.count('\n')
        buf = read_f(buf_size)

    f.close()

    return nb_lines

###########################################################################

def read_txtdata(filename, lock=False):
    """
    Return a list containing all the lines of the given file,
    with the newline removed.

    If the file EXISTS, its content is ASSUMED to be 100% correct.
    """

    if not os.access(filename, os.F_OK):
        logger.error('    MISSING %s' % filename)
        # if the file do NOT exists, it is considered as ERROR
        #sys.stderr.write('\nMISSING %s\n' % filename)
        #sys.exit(1)
        return None

    logger.debug('    EXISTING %s' % filename)

    if lock:
        fid_lock = f_lock(filename)

    try:
        f = open(filename, 'r')
        try:
            content = f.readlines()
            # remove the newlines
            for i in xrange(0, len(content), 1):
                content[i] = content[i].replace('\n', '')

        except (KeyboardInterrupt, SystemExit):
            content = None
            logger.info('    ... KeyboardInterrupt: File closed properly')
            raise
        finally:
            f.close()

    except IOError:
        sys.stderr.write('\nERROR %s\n' % filename)
        sys.exit(1)

    except EOFError:
        sys.stderr.write('\nEMPTY %s\n' % filename)
        sys.exit(1)

    finally:
        if lock:
            fid_lock.close()

    return content

def save_txtdata(filename, data, lock=False):
    """
    Save the data in the given file, in text format.
    """

    logger.info('    FILE %s' % filename)

    if lock:
        fid_lock = f_lock(filename)
    else:
        create_dir(filename)

    try:
        f = open(filename, 'w')
        try:
            f.write('%s' % data)
        except (KeyboardInterrupt, SystemExit):
            f.close()
            f = open(filename, 'w')
            f.write('%s' % data)
            logger.info('    ... KeyboardInterrupt: File closed properly')
            raise
        finally:
            f.close()

    except IOError:
        sys.stderr.write('\nERROR %s\n' % filename)
        sys.exit(1)

    finally:
        if lock:
            fid_lock.close()

###########################################################################

def read_bindata(filename, lock=False):
    """
    Return the data in the given file in binary format.

    If the file EXISTS, its content is ASSUMED to be 100% correct.
    """

    # if the file do NOT exists, it is considered as ERROR
    if not os.access(filename, os.F_OK):
        sys.stderr.write('\nMISSING %s\n' % filename)
        sys.exit(1)

    logger.debug('    READING %s' % filename)

    if lock:
        fid_lock = f_lock(filename)

    try:
        f = open(filename, 'rb')
        try:
            content = cPickle.load(f)
        except (KeyboardInterrupt, SystemExit):
            content = None
            logger.info('    ... KeyboardInterrupt: File closed properly')
            raise
        finally:
            f.close()

    except IOError:
        sys.stderr.write('\nERROR %s\n' % filename)
        sys.exit(1)

    except EOFError:
        sys.stderr.write('\nEMPTY %s\n' % filename)
        sys.exit(1)

    finally:
        if lock:
            fid_lock.close()

    return content

def save_bindata(filename, data, lock=False):
    """
    Save the data in the given file in binary format.
    """

    logger.debug('    FILE %s' % filename)

    create_dir(filename)
    #create_file(filename)

    if lock:
        fid_lock = f_lock(filename)

    try:
        f = open(filename, 'wb')
        try:
            cPickle.dump(data, f, 2)
        except (KeyboardInterrupt, SystemExit):
            f.close()
            f = open(filename, 'wb')
            cPickle.dump(data, f, 2)
            logger.info('    ... KeyboardInterrupt: File closed properly')
            raise
        finally:
            f.close()

    except IOError:
        sys.stderr.write('\nERROR %s\n' % filename)
        sys.exit(1)

    finally:
        if lock:
            fid_lock.close()
