#!/usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Generates unique file names.
"""

###########################################################################
# IMPORTS

from datetime import datetime, timedelta
from os import listdir, path, stat, access, makedirs, remove, rmdir
from os import W_OK, F_OK
from random import Random

###########################################################################
# GLOBALS

random = Random()
random.jumpahead(8)

###########################################################################
# CLASS

class FileManager(object):
    """
    Generates unique file names.
    """

    #######################################################################
    # local functionality

    def _clean(self, root_dir, max_age):
        """
        Delete all expired directories and files.
        """

        if access(root_dir, W_OK):

            basetime = datetime.now() - timedelta(seconds=max_age)

            for dirname in listdir(root_dir):
                dirname = path.join(root_dir, dirname)
                # get the time of the last modification
                st_mtime = stat(dirname).st_mtime
                lasttime = datetime.fromtimestamp(st_mtime)

                if basetime > lasttime:
                    try:
                        for filename in listdir(dirname):
                            filename = path.join(dirname, filename)
                            remove(filename)
                        rmdir(dirname)
                    except:
                        pass

    def _delete_similar_files(self, root_dir, prefix, max_age):
        """
        Delete all expired similar files.
        """

        if access(root_dir, W_OK):

            basetime = datetime.now() - timedelta(seconds=max_age)

            dirname = root_dir
            #for dirname in listdir(root_dir):
            #    dirname = path.join(root_dir, dirname)
            # get the time of the last modification
            st_mtime = stat(dirname).st_mtime
            lasttime = datetime.fromtimestamp(st_mtime)

            if basetime > lasttime:
                try:
                    for filename in listdir(dirname):
                        if prefix in filename:
                            filename = path.join(dirname, filename)
                            remove(filename)
                    rmdir(dirname)
                except:
                    pass

    #######################################################################
    # main functionality

    def __init__(self, settings):
        """
        Initialize the object.
        """

        self.settings = {
            'root_dir': None,
            'username': 'anonymous',
            'root_url': None,
            'auto_cleanup': False,
            'max_age': 3600, # seconds
            'alphabet': "abdeghkmnqrt",
        }

        # get options from the settings
        try:
            for k in settings.keys():
                self.settings[k] = settings[k]
        except:
            pass

    def filename(self, relpath, prefix, suffix):
        """
        Generate a unique filename.
        Form: DIR_ROOT/relpath/prefixNAME.suffix
        """

        root_dir = path.join(self.settings['root_dir'], self.settings['username'])
        root_url = path.join(self.settings['root_url'], self.settings['username'])

        # if necessary, create directory
        abspath = path.join(root_dir, relpath)
        if not path.isdir(abspath):
            makedirs(abspath)

        if self.settings['auto_cleanup']:
            max_age = self.settings['max_age']
            #self._clean(abspath, max_age)
            self._delete_similar_files(abspath, prefix, max_age)

        alphabet = self.settings['alphabet']
        t = datetime.now()
        basename = t.strftime('%H%M%S') + '_' + prefix + '_'
        while True:
            filename = basename + ''.join(random.sample(alphabet, 3)) + suffix
            # HINT: check again if there is already a file with this name
            # ... the randomness space should be sufficient for our usage
            filepath = path.join(abspath, filename)
            if not access(filepath, F_OK):
                break

        fileurl = path.join(root_url, relpath, filename)

        return filepath, fileurl

    def filename2(self, relpath, filename):
        """
        Return the full path filename.
        Form: DIR_ROOT/relpath/filename
        """

        root_dir = path.join(self.settings['root_dir'], self.settings['username'])
        root_url = path.join(self.settings['root_url'], self.settings['username'])

        # if necessary, create directory
        abspath = path.join(root_dir, relpath)
        if not path.isdir(abspath):
            makedirs(abspath)

        filepath = path.join(abspath, filename)
        fileurl = path.join(root_url, relpath, filename)

        return filepath, fileurl
