#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Pre-compute the image similarity distances.
"""

###########################################################################
# IMPORTS

cimport numpy as cnp

import logging
from numpy import float32, int32, zeros, empty_like, array, empty, reshape, squeeze, append
import os
import random
import sys
import time
from time import time as now

from retrieval_libs.dblayer.FileAccess import f_lock, create_file, read_bindata, save_bindata
from retrieval_libs.common.buffer cimport buffer_set00, buffer_get00

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

HBLOCKLN = 60000#33000#22000#
KBLOCKLN = 6000#3300#2200#
KSUBBLOCKLN = 6000#3300#2200#

cdef cnp.float32_t DISTMAX = 1000000.0
cdef cnp.int32_t HBUFFER_LN = 5000
cdef cnp.int32_t CHNGMAX = 2000

cdef cnp.float32_t PACCURACY = 0.001

###########################################################################
# local functionality

def _i_block(num, BLOCKLN):
    """
    Get the (i_from, i_to indexes of the) block for processing.
    """

    if os.getenv('SGE_TASK_ID') is None:
        # data is processed in one block
        i_from = 1
        i_to = num + 1
    else:
        # data is processed in blocks, one block per job
        job_id = int(os.getenv('SGE_TASK_ID'))
        job_index = job_id - 1

        nb_blocks = num / BLOCKLN + 1

        i_from = (job_index % nb_blocks) * BLOCKLN + 1
        i_to = min(i_from + BLOCKLN, num + 1)

        #stepsize = int(os.getenv('SGE_TASK_STEPSIZE'))
        #if stepsize != 1:
        #    sys.stderr.write('\nSGE_TASK_STEPSIZE %s != 1\n' % stepsize)
        #    sys.exit(1)

    return i_from, i_to

###########################################################################
# main functionality

def initialize_distdata(dbimg):
    """
    Initialize the image distance files.
    """

    num = dbimg.size()

    # if not necessary, skip it
    if num < HBLOCKLN:
        return

    i_from, i_to = _i_block(num, KBLOCKLN)
    logger.info('    DBID [%s, %s]' % (i_from, i_to))

    # determine the number of feature types
    tmp = array([1,], dtype=int32, order='c')
    tmpdata = dbimg.featdata(tmp, -1, False)
    nt = len(tmpdata)

    # initialize with invalid distances
    initial_dist = empty((num, nt), dtype=float32, order='c')
    initial_dist[:] = float32(-1)

    #
    nb_ii = (i_to - i_from) / KSUBBLOCKLN + 1
    for ii in xrange(0, nb_ii, 1):
        ii_from = i_from + ii * KSUBBLOCKLN
        ii_to = min(ii_from + KSUBBLOCKLN, i_to)

        if ii_from >= ii_to:
            continue

        kdbid = range(ii_from, ii_to, 1)
        kdbid = array(kdbid, dtype=int32, order='c')

        for idbid, ipath in dbimg.distfile(kdbid):
            logger.debug('    DBID %s PATH %s' % (idbid, ipath))
            create_file(ipath)
            save_bindata(ipath, initial_dist)

def compute_distdata(dbimg):
    """
    Compute the image similarity distances.
    """

    num = dbimg.size()

    i_from, i_to = _i_block(num, KBLOCKLN)
    j_from, j_to = _i_block(num, HBLOCKLN)

    hdbid = range(j_from, j_to, 1)
    hdbid = array(hdbid, dtype=int32, order='c')

    logger.debug(
            '    DBID [%s, %s] x [%s, %s]'
            % (i_from, i_to, j_from, j_to))

    nb_ii = (i_to - i_from) / KSUBBLOCKLN + 1
    for ii in xrange(0, nb_ii, 1):
        ii_from = i_from + ii * KSUBBLOCKLN
        ii_to = min(ii_from + KSUBBLOCKLN, i_to)

        if ii_from >= ii_to:
            continue

        logger.info(
                '    DBID [%s, %s] x [%s, %s]'
                % (ii_from, ii_to, j_from, j_to))

        kdbid = range(ii_from, ii_to, 1)
        kdbid = array(kdbid, dtype=int32, order='c')

        distdata = dbimg.compute_distdata(kdbid, hdbid)
        distfile = dbimg.distfile(kdbid)
        for i, idbid in enumerate(kdbid):

            #
            xdbid, ipath = distfile[i]
            if xdbid != idbid:
                sys.exit(1)

            if HBLOCKLN >= num:
                idist = distdata[i][:,0:1]
                save_bindata(ipath, idist)
            else:
                idist = read_bindata(ipath)
                idist[j_from:j_to,:] = distdata[i]
                save_bindata(ipath, idist)

            #logger.debug('    DBID %s PATH %s' % (idbid, ipath))

def check_distdata(dbimg):
    """
    Check the image similarity distances.
    """

    num = dbimg.size()

    hdbid = range(1, num + 1, 1)
    hdbid = array(hdbid, dtype=int32, order='c')

    i_from, i_to = _i_block(num, KBLOCKLN)
    logger.info('    DBID [%s, %s]' % (i_from, i_to))

    nb_ii = (i_to - i_from) / KSUBBLOCKLN + 1
    for ii in xrange(0, nb_ii, 1):
        ii_from = i_from + ii * KSUBBLOCKLN
        ii_to = min(ii_from + KSUBBLOCKLN, i_to)
        logger.info('    DBID [%s, %s]' % (ii_from, ii_to))

        if ii_from >= ii_to:
            continue

        kdbid = range(ii_from, ii_to, 1)
        kdbid = array(kdbid, dtype=int32, order='c')

        for idbid, ipath in dbimg.distfile(kdbid):

            try:
                idist = read_bindata(ipath, lock=False)
            except:
                idist = None

            #if idist is None or idist.shape[0] != num:
                #for t in xrange(0, idist.shape[1], 1):
            #    t = 0
            #    for i in xrange(0, idist.shape[0], 1):
            #        if idist[i,t] < 0.0:
                        #logger.error('ERROR %s %s %s %s' % (ipath, idbid, i, t))
                        #sys.exit(1)

            if idist is None or idist.shape[0] != num:
                idbid = array([idbid,], dtype=int32, order='c')
                distdata = dbimg.compute_distdata(idbid, hdbid)
                idist = distdata[0][:,0:1]
                save_bindata(ipath, idist, lock=False)

            if len(idist.shape) == 1:
                print idist.shape
                idist = reshape(idist, (idist.shape[0], 1), order='c')
                save_bindata(ipath, idist, lock=False)

        logger.info('    DBID [%s, %s]' % (ii_from, ii_to))