#! /usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Creates the dbtree based on an approximated top-down k-means algorithm.

HINT: The data is read and processed in blocks. Moreover, the algorithm
      has two steps: 1. perform clustering on the sample to get a good
      starting code; 2. perform clustering on the all data, initialized
      with the 'good' code;

      It assumes that the root node has dbid == 1.
"""

###########################################################################
# IMPORTS

cimport numpy as cnp

import logging
from numpy import int32, float32, array, zeros, zeros_like, argmin, fromstring, mean, where
import os
from random import Random
import sys

from math import sqrt

from retrieval_libs.dblayer.DbTableImg cimport DbTableImg

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

random = Random()
random.jumpahead(8)

REINIT = False

KMEANS_SIZE = 8
BLOCK_SIZE = 50000

RADIUS_MIN = 0.000001

###########################################################################
# local functionality

def _kpoints(data, k):

    n = len(data)
    p = random.sample(range(0, n, 1), k)
    code = data[p,:].copy()

    return code

def _vq(dbimg, data, code_book):

    cdef DbTableImg c_dbimg = dbimg

    n = len(data)
    k = len(code_book)

    label = zeros(n, dtype=int32, order='c')
    min_dist = zeros(n, dtype=float32, order='c')

    for i in xrange(0, n, 1):
        dd = c_dbimg.distdata_feat((data[i:(i+1),:],), (code_book,))
        dist = dd[0][:,0]
        label[i] = argmin(dist)
        min_dist[i] = dist[label[i]]

    return label, min_dist

def _kmeans2(dbimg, data, code_book, niter):

    k = len(code_book)

    for i in xrange(0, niter, 1):
        # Compute the nearest neighbour for each obs
        # using the current code book
        label = _vq(dbimg, data, code_book)[0]
        # Update the code by computing centroids using the new code book
        for j in xrange(0, k, 1):
            mbs = where(label==j)
            if mbs[0].size > 0:
                code_book[j] = mean(data[mbs], axis=0)
            else:
                logger.debug('One of the clusters is empty')

    return code_book, label

def _kmeans(dbimg, all_dbid, code, niter, nc):
    """
    Run k-means with a given initial codebook.

    HINT: This is the 'scipy.cluster.vq.kmeans2' method modified only in that
          the feature data is read and processed in blocks, and it is not
          required to be read at once at the method call.
    """

    all_dbidln = len(all_dbid)

    for i in xrange(0, niter, 1):
        # Compute the nearest neighbour for each obs
        # using the current code book

#        all_label = None
        #all_dist = None
        all_mean = {}
        all_size = {}

        for istart in xrange(0, all_dbidln, BLOCK_SIZE):
            iend = min(istart + BLOCK_SIZE, all_dbidln)
            tmp_dbid = all_dbid[istart:iend]

            featdata = dbimg.featdata(tmp_dbid, tt=0, buffered=False)
            data = featdata[0]

            label, dist = _vq(dbimg, data, code)
            #print 'label', label.shape
#            if all_label is None:
#                all_label = label
                #all_dist = dist
#            else:
#                all_label = append(all_label, label, axis=0)
                #all_dist = append(all_dist, dist, axis=0)

            #print 'all_label', all_label.shape#, all_dist.shape

            # update the code by computing centroids using the new code book
            for j in xrange(0, nc, 1):
                mbs = where(label==j)
                if mbs[0].size > 0:
                    jmean = mean(data[mbs], axis=0)
                    jln = mbs[0].size
                    #print 'jmean', jmean.shape

                    alljmean = all_mean.get(j, [])
                    alljmean.append(jmean)
                    all_mean[j] = alljmean

                    alljln = all_size.get(j, [])
                    alljln.append(jln)
                    all_size[j] = alljln

                else:
                    logger.debug('One of the clusters is empty')

            logger.debug('PROGRESS INFO %s %s' % (istart, iend))

        #for j in xrange(0, nc, 1):
        # HINT: the empty clusters remain unchanged
        for j in all_mean.iterkeys():
            alljmean = all_mean[j]
            alljsize = all_size[j]

            #print 'allj', j, len(alljmean), len(alljsize)

            jme = zeros_like(alljmean[0])
            jsi = 0
            for ix in xrange(0, len(alljmean), 1):
                jme += alljmean[ix] * alljsize[ix]
                jsi += alljsize[ix]

            code[j] = jme / jsi

            #print 'code', j, code[j]

    return code#,all_label #, all_dist

###########################################################################
# local functionality

def _perform_kmeans_approx(dbimg, dbtree, all_dbid):
    """
    Approximated top-down k-means algorithm.

    HINT: The data is read and processed in blocks. Moreover, the algorithm
          has two steps: 1. perform clustering on the sample to get a good
          starting code; 2. perform clustering on the all data, initialized
          with the 'good' code;
    """

    clu_dbid = random.sample(all_dbid, BLOCK_SIZE)
    clu_dbid = array(clu_dbid, dtype=int32, order='c')

    featdata = dbimg.featdata(clu_dbid, tt=0, buffered=False)
    descriptors = featdata[0]

    # perform clustering on the sample to get a good starting code
    code_book = _kpoints(descriptors, KMEANS_SIZE)
    clusters, labels = _kmeans2(dbimg, descriptors, code_book, niter=30)

    clusters = array(clusters, dtype=float32, order='c')
    labels = None

    # perform clustering on the all data, initialized with the 'good' code
    clusters = _kmeans(dbimg, all_dbid, clusters, 2, KMEANS_SIZE)

    # prepare data for storage
    centroids = {}
    radius = {}
    children = {}
    all_dbidln = len(all_dbid)
    for istart in xrange(0, all_dbidln, BLOCK_SIZE):
        iend = min (istart + BLOCK_SIZE, all_dbidln)
        tmp_dbid = all_dbid[istart:iend]

        featdata = dbimg.featdata(tmp_dbid, tt=0, buffered=False)
        descriptors = featdata[0]
        membership, dist = _vq(dbimg, descriptors, clusters)

        ###
        for i, clix in enumerate(membership):

            idist = dist[i]
            idbid = tmp_dbid[i]

            # centroid, i.e. the closest descriptor to the 'true' center
            cldi = centroids.get(clix, None)
            if cldi is None:
                cldi = [idbid, idist]
            else:
                if idist < cldi[1]:
                    cldi = [idbid, idist]
                elif idist == cldi[1]:
                    cldi[0] = random.choice([cldi[0], idbid])

            centroids[clix] = cldi

            # radius
            distmax = radius.get(clix, 0.0)
            if idist > distmax:
                distmax = idist

            radius[clix] = distmax

            # children
            clch = children.get(clix, [])
            clch.append(idbid)

            children[clix] = clch

        logger.debug('PROGRESS INFO %s %s' % (istart, iend))

    # prepare data for storage
    count = 0
    for clix in centroids.iterkeys():

        # remove the seed from the children
        cldi = centroids.get(clix, None)
        if cldi is not None:
            cl_seed = cldi[0]
        cl_dbid = children.get(clix, None)
        if cl_dbid is not None:
            if cldi is not None:
                cl_dbid.remove(cl_seed)

            logger.debug('CLUSTER %s CHSIZE %s' % (clix, len(cl_dbid)))

            count += len(cl_dbid)

            if len(cl_dbid) == 0:
                cl_dbid = None

            children[clix] = cl_dbid

    logger.debug('Check point: GIVEN %s CLUSTERED %s' % (len(all_dbid), count))

    return centroids, radius, children

def _perform_kmeans_exact(dbimg, dbtree, all_dbid):
    """
    Exact k-means algorithm.
    """

    cdef DbTableImg c_dbimg = dbimg

    featdata = dbimg.featdata(all_dbid, tt=0, buffered=False)
    descriptors = featdata[0]

    # perform clustering
    code_book = _kpoints(descriptors, KMEANS_SIZE)
    clusters, labels = _kmeans2(dbimg, descriptors, code_book, niter=20)

    # prepare data for storage
    centroids = {}
    radius = {}
    children = {}
    for i, clix in enumerate(labels):

        dd = c_dbimg.distdata_feat((descriptors[i:(i+1),:],), (clusters[clix:(clix+1),:],))
        idist = dd[0][0,0]

        idbid = all_dbid[i]

        # centroid, i.e. the closest descriptor to the 'true' center
        cldi = centroids.get(clix, None)
        if cldi is None:
            cldi = [idbid, idist]
        else:
            if idist < cldi[1]:
                cldi = [idbid, idist]
            elif idist == cldi[1]:
                cldi[0] = random.choice([cldi[0], idbid])

        centroids[clix] = cldi

        # radius
        distmax = radius.get(clix, 0.0)
        if idist > distmax:
            distmax = idist

        radius[clix] = distmax

        # children
        clch = children.get(clix, [])
        clch.append(idbid)

        children[clix] = clch

    # prepare data for storage
    count = 0
    for clix in centroids.iterkeys():

        # remove the seed from the children
        cl_seeds = centroids.get(clix, None)
        if cl_seeds is not None:
            cl_seed = cl_seeds[0]
        cl_dbid = children.get(clix, None)
        if cl_dbid is not None:
            if cl_seeds is not None:
                cl_dbid.remove(cl_seed)

            logger.debug('CLUSTER %s CHSIZE %s' % (clix, len(cl_dbid)))

            # count the cluster size, including the seed
            count += len(cl_dbid) + 1

            if len(cl_dbid) == 0:
                cl_dbid = None

            children[clix] = cl_dbid

    logger.debug('Check point: GIVEN %s CLUSTERED %s' % (len(all_dbid), count))

    return centroids, radius, children

###########################################################################
# local functionality

def create_subtree(dbimg, dbtree, dbids):
    """
    Perform the approximated top-down k-means algorithm starting from the
    nodes in dbids and save the resulted clusters and nodes in the database.

    HINT: The given nodes (in dbids) are leaf nodes in dbtree.
          Clustering is performed on each of these nodes and the new
          resulted clusters are inserted as leaf nodes. Then, the splitted
          nodes (in dbids) are updated as non-leaf nodes.
          Recursively, the clustering is performed on the new leaf nodes.

          It assumes that the root node has dbid == 1.
    """

    for dbid in dbids:

        name = dbtree.name(dbid)

        ch_tbl = dbtree.children_table(dbid)
        if ch_tbl == 'tree':
            logger.debug('    Node %s is NOT a leaf node' % (name))
            # HINT: Go deeper in the tree if the node is already clustered.
            #       This is used only when the routine is restarted after it
            #       was intrerrupted and the dbtree was left incomplete.
            ch_data = dbtree.children_data([dbid])
            ch_dbid = ch_data[0][4]
            ch_dbid = fromstring(ch_dbid, dtype=int32, sep=' ')
            create_subtree(dbimg, dbtree, ch_dbid)
            # proceed with the next node
            continue

        ch_size = dbtree.children_size(dbid)
        # HINT: Split the root node anyway, since its ch_size is not updated yet.
        if dbid != 1 and ch_size < KMEANS_SIZE:
            logger.debug('    Node %s is a genuine leaf node' % (name))
            # proceed with the next node
            continue

        radius = dbtree.radius(dbid)
        # HINT: stop clustering the nodes with almost identic images.
        if dbid != 1 and radius <= RADIUS_MIN:
            logger.debug('    Node %s is a genuine leaf node' % (name))
            # proceed with the next node
            continue

        if dbid != 1:
            ch_data = dbtree.children_data([dbid])
            ch_dbid = ch_data[0][4]
            ch_dbid = fromstring(ch_dbid, dtype=int32, sep=' ')
        else: # the root node
            ch_dbid = dbimg.all_valid_ids()

        # perform the clustering
        if len(ch_dbid) <= BLOCK_SIZE:
            centroids, radiuses, children = _perform_kmeans_exact(dbimg, dbtree, ch_dbid)
        else:
            centroids, radiuses, children = _perform_kmeans_approx(dbimg, dbtree, ch_dbid)

        # insert in the database
        centroidsln = len(centroids)
        new_dbids = []
        #for i in xrange(0, centroidsln, 1):
        for i in centroids.iterkeys():
            icentroid = centroids[i][0]
            iradius = radiuses[i]
            ichildren = children[i]
            if ichildren is not None:
                ichildren.sort()
                ichildren = tuple(ichildren)
                ichildren_size = len(ichildren)
                ichildren_dbid = ' '.join(map(str, ichildren))
            else:
                ichildren_size = 0
                ichildren_dbid = ''

            iname = name + ('%d' % i)

            # insert the child node
            idbid = dbtree.insert_node(-1, iname, icentroid, iradius, dbid, ichildren_size, 'img', ichildren_dbid)

            # book the children nodes
            new_dbids.append(idbid)

            logger.debug('    Child node %s is inserted' % (iname))

        # update the parent node
        dbtree.update_children_data(dbid, new_dbids)
        logger.debug('    Parent node %s is updated' % (name))

        # call the routine for the newly defined leaf nodes
        create_subtree(dbimg, dbtree, new_dbids)