#! /usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Updates the dbtree with some redundant information like parent-children
inter-relations, names, depths and sizes/cardinalities of the nodes.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, fromstring
import os
import sys

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

###########################################################################

def update_parent(dbtree):
    """
    Update the children sizes, tables and dbids.
    """

    all_dbid = dbtree.all_dbid()

    already_updated = []
    for dbid in all_dbid:
        dbtree.update_parent(dbid, already_updated)
        already_updated.append(dbid)
        logger.debug('    DBID %s' % dbid)

def update_children(dbtree, pdbid):
    """
    Update the depths and names.
    """

    for dbid in pdbid:

        children_table = dbtree.children_table(dbid)
        if children_table == 'img':
            logger.debug('    1: DBID %s has NO children' % dbid)
            # proceed with the next node
            continue

        children_size = dbtree.children_size(dbid)
        if children_size == 0:
            logger.debug('    2: DBID %s has NO children' % dbid)
            # proceed with the next node
            continue

        ###
        dbtree.update_children(dbid)
        logger.debug('    DBID %s' % dbid)

        ###
        children_data = dbtree.children_data([dbid,])
        cdbid = fromstring(children_data[0][4], dtype=int32, sep=' ')
        update_children(dbtree, cdbid)

def update_subtree_size(dbtree, dbid):
    """
    Update the children subtree sizes.
    """

    # HINT: count the seed as well
    subtree_size = 1

    children_table = dbtree.children_table(dbid)
    if children_table == 'img':
        subtree_size += dbtree.children_size(dbid)

    elif children_table == 'tree':
        children_data = dbtree.children_data([dbid,])
        childlist = fromstring(children_data[0][4], dtype=int32, sep=' ')
        for cdbid in childlist:
            subtree_size += update_subtree_size(dbtree, cdbid)

    else:
        sys.stderr.write('\nUNKNOWN TABLE %s\n' % children_table)
        sys.exit(1)

    dbtree.update_subtree_size(dbid, subtree_size)
    logger.debug('    DBID %s, SIZE %d' % (dbid, subtree_size))

    return subtree_size

def check_children_data(dbtree):
    """
    Check the integrity of the children data.
    """

    all_dbid = dbtree.all_dbid()

    for dbid in all_dbid:

        children_data = dbtree.children_data((dbid,))

        for idbid, idepth, ictbl, ichsize, icdbidstr in children_data:

            if ictbl not in ('tree', 'img'):
                sys.stderr.write('\nUNKNOWN TABLE %s\n' % ictbl)
                sys.exit(1)

            if ichsize == 0 and len(icdbidstr) == 0:
                logger.debug('    DBID %d' % idbid)
            else:
                icdbid = fromstring(icdbidstr, dtype=int32, sep=' ')
                if 0 in icdbid or len(icdbid) != ichsize:
                    logger.error(
                        'DBID %d, CHILDREN_TABLE %s, CHILDREN_DBID %s'
                        % (idbid, ictbl, icdbidstr))
