#!/usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides the statistics of the system performance.
"""

###########################################################################
# IMPORTS

import logging
import sys

from numpy import int32, float32, array, zeros, zeros_like, arange, mean, std
import operator
from random import Random

from scipy.special import bdtrc
from scipy.stats import wilcoxon, ttest_ind, ttest_rel
from scipy.misc.common import factorial

from retrieval_libs.display.StatisticsGfx import init, figure_0, figure_12, \
        figure_15, figure_eval_fdbk, figure_displayed_images

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

random = Random()
random.jumpahead(8)

###########################################################################
# FUNCTIONS

def _histo_utc(norm, histo):
    # iterations histogram for the same[user, task, config]

    # NOT normalize
    histo = histo * 1.0
    norm = histo.sum(3)
    return norm, histo

    histo = histo * 1.0
    if norm is None:
        norm = histo.sum(3)
    else:
        norm = norm * 1.0

    histo_shape = histo.shape
    if len(histo_shape) == 4:
        for u in xrange(0, histo_shape[0], 1):
            for t in xrange(0, histo_shape[1], 1):
                for c in xrange(0, histo_shape[2], 1):
                    if norm[u,t,c] > 0:
                        for i in xrange(0, histo_shape[3], 1):
                            histo[u,t,c,i] /= norm[u,t,c]
                        #
                        norm[u,t,c] = 1

    elif len(histo_shape) == 3:
        for u in xrange(0, histo_shape[0], 1):
            for t in xrange(0, histo_shape[1], 1):
                for c in xrange(0, histo_shape[2], 1):
                    if norm[u,t,c] > 0:
                        histo[u,t,c] /= norm[u,t,c]
                        norm[u,t,c] = 1

    else:
        pass

    return norm, histo

def _histo_uc(norm, histo):
    # iterations histogram for the same[user, config]

    histo = histo.sum(1)
    norm = norm.sum(1)

    for u in xrange(0, histo.shape[0], 1):
        for c in xrange(0, histo.shape[1], 1):
            if norm[u,c] > 0:

                if len(histo.shape) > 2:
                    for i in xrange(0, histo.shape[2], 1):
                        histo[u,c,i] /= norm[u,c]
                else:
                    histo[u,c] /= norm[u,c]
                #
                norm[u,c] = 1

    return norm, histo

def _histo_tc(norm, histo):
    # iterations histogram for the same[task, config]

    histo = histo.sum(0)
    norm = norm.sum(0)

    for t in xrange(0, histo.shape[0], 1):
        for c in xrange(0, histo.shape[1], 1):
            if norm[t,c] > 0:

                if len(histo.shape) > 2:
                    for i in xrange(0, histo.shape[2], 1):
                        histo[t,c,i] /= norm[t,c]
                else:
                    histo[t,c] /= norm[t,c]
                #
                norm[t,c] = 1

    return norm, histo

def _histo_c(norm, histo):
    # iterations histogram for the same[config]

    histo = histo.sum(0)
    norm = norm.sum(0)

    for c in xrange(0, histo.shape[0], 1):
        if norm[c] > 0:

            if len(histo.shape) > 1:
                for i in xrange(0, histo.shape[1], 1):
                    histo[c,i] /= norm[c]
            else:
                histo[c] /= norm[c]
            #
            norm[c] = 1

    return norm, histo

def _variance(norm, histo):
    # iterations histogram for the same[config]

    histo = histo.sum(0)
    norm = norm.sum(0)

    for c in xrange(0, histo.shape[0], 1):
        if norm[c] > 0:

            if len(histo.shape) > 1:
                for i in xrange(0, histo.shape[1], 1):
                    histo[c,i] /= norm[c]
            else:
                histo[c] /= norm[c]
            #
            norm[c] = 1

    return norm, histo

###########################################################################
# FUNCTIONS

def _filter_data1(data):
    # keep only ONE sample for the same [user, task, config]
    # data[user,task,config,iter]
    # HINT: TODO versions for min, max, or random

    # make a copy
    data = data * 1.0

    # number of samples for the same [user, task, config]
    norm = data.sum(3)

    # keep only ONE sample for the same [user, task, config]
    # randomize the choice
    for u in xrange(0, data.shape[0], 1):
        for t in xrange(0, data.shape[1], 1):
            for c in xrange(0, data.shape[2], 1):

                if norm[u,t,c] > 1:
                    candidates = []
                    for i in xrange(0,data.shape[3],1):
                        candidates += [i] * data[u,t,c,i]
                        data[u,t,c,i] = 0
                    lucky = random.choice(candidates)
                    #if c == 0:
                    #lucky = min(candidates)
                    #else:
                    #    lucky = max(candidates)
                    data[u,t,c,lucky] = 1
                    #norm[u,t,c] = 1

    return data

def _filter_data2(data):
    # filter out the first/second half of the users (in terms of overall average number of iterations)

    # make a copy
    data = data * 1.0

    # iterations histogram for the same[user, task, config]
    norm, histo = _histo_utc(None, data)

    # average number of iterations
    for i in xrange(0, histo.shape[3], 1):
        histo[:,:,:,i] *= i
    itermeans = histo.sum(3).sum(2).sum(1)
    norm = norm.sum(2).sum(1)

    for u in xrange(0, norm.shape[0], 1):
        if norm[u] > 0:
            itermeans[u] /= norm[u]
            norm[u] = 1

    ulist = []
    for u, itermean in enumerate(itermeans):
        # do not add users with no sessions
        if norm[u] > 0:
            ulist.append([u, itermean])

    # order, in increasing order
    ulist = sorted(ulist, key=operator.itemgetter(1), reverse=False)

    itermeans = itermeans.sum(0)
    norm = norm.sum(0)
    itermeans /= norm
    for l in xrange(0,len(ulist),1):
        if ulist[l][1] <= itermeans:
        #if ulist[l][1] >= itermeans:
            u = ulist[l][0]
#            data[u,:,:,:] = 0

    #print 'ulist', itermeans, ulist

    ll = int(round(float(len(ulist)) / 2))
    for l in xrange(0, ll, 1):
    #for l in xrange(ll, len(ulist), 1):
        u = ulist[l][0]
        data[u,:,:,:] = 0

    return data

def _filter_data(data):

    # make a copy
    data = data * 1.0

    #data = _filter_data1(data)
    #data = _filter_data2(data)

    return data

###########################################################################
# CLASS

class StatisticsData(object):
    """
    Provides the statistics of the system performance.
    """

    #######################################################################
    # local functionality

    def _dif(self, norm, itermeans, delta=0.0):
        # difference between configurations

        shape = (itermeans.shape[0], itermeans.shape[1], len(self.config_pairs))
        dif = zeros(shape, dtype=float32, order='c')
        tot = zeros(shape, dtype=float32, order='c')
        for u in xrange(0, itermeans.shape[0], 1):
            for t in xrange(0, itermeans.shape[1], 1):

                for c, c0, c1 in self.config_pairs:
                    c0 = self.configs[c0]
                    c1 = self.configs[c1]

                    if norm[u,t,c0] > 0 and norm[u,t,c1] > 0:
                        diftmp = itermeans[u,t,c1] - itermeans[u,t,c0]
                        if diftmp >= delta:
                            dif[u,t,c] = diftmp
                            tot[u,t,c] = 1
                        elif diftmp <= -delta:
                            dif[u,t,c] = diftmp
                            tot[u,t,c] = 1

        return tot, dif

    def _gap(self, norm, itermeans, delta=0.0):
        # (binary difference)/gap between configurations

        shape = (itermeans.shape[0], itermeans.shape[1], len(self.config_pairs))
        gap = zeros(shape, dtype=float32, order='c')
        tot = zeros(shape, dtype=float32, order='c')
        for u in xrange(0, itermeans.shape[0], 1):
            for t in xrange(0, itermeans.shape[1], 1):

                for c, c0, c1 in self.config_pairs:
                    c0 = self.configs[c0]
                    c1 = self.configs[c1]

                    if norm[u,t,c0] > 0 and norm[u,t,c1] > 0:
                        diftmp = itermeans[u,t,c1] - itermeans[u,t,c0]
                        if diftmp >= delta:
                            gap[u,t,c] = 1
                            tot[u,t,c] = 1
                        elif diftmp <= -delta:
                            tot[u,t,c] = 1

        return tot, gap

    def _get_filter_data(self):

        if self._filter_data is None:
            self._filter_data = _filter_data(self.data)

        return self._filter_data

    def _get_user_data(self, uid):

        for u, uidu in enumerate(self.users):
            if uidu == uid:
                break

        data = self._get_filter_data()

        userdata = zeros_like(data)
        if u < userdata.shape[0]:
            userdata[u,:,:,:] = data[u,:,:,:] * 1.0

        return userdata

    def _get_histo_utc(self):

        if self.norm_utc is None or self.histo_utc is None:
            data = self._get_filter_data()
            self.norm_utc, self.histo_utc = _histo_utc(None, data)

        return self.norm_utc, self.histo_utc

    def _get_itermeans_utc(self):

        if self.itermeans_utc is None:
            #data = self._get_filter_data()
            norm, histo = self._get_histo_utc()

            histo = histo * 1.0
            for i in xrange(0, histo.shape[3], 1):
                histo[:,:,:,i] *= i

            self.itermeans_utc = histo.sum(3)

        return self.itermeans_utc

    def _get_gap_utc(self):

        if self.norm_gap is None or self.histo_gap is None:
            #data = self._get_filter_data()
            norm, histo = self._get_histo_utc()
            itermeans = self._get_itermeans_utc()

            delta = self.delta
            self.norm_gap, self.histo_gap = self._gap(norm, itermeans, delta)

        return self.norm_gap, self.histo_gap

    def _get_dif_utc(self):

        if self.norm_dif is None or self.histo_dif is None:
            #data = self._get_filter_data()
            norm, histo = self._get_histo_utc()
            itermeans = self._get_itermeans_utc()

            delta = self.delta
            self.norm_dif, self.histo_dif = self._dif(norm, itermeans, delta)

        return self.norm_dif, self.histo_dif

    #######################################################################
    # main functionality

    def init(self, users, tasks, configs, config_pairs, config_props, delta, data):

        self.users = users
        self.tasks = tasks
        self.configs = configs
        self.config_pairs = config_pairs

        init(config_props)

        self.delta = delta
        self.data = data

        #
        self._filter_data = None
        self.norm_utc = None
        self.histo_utc = None
        self.itermeans_utc = None
        self.norm_gap = None
        self.histo_gap = None
        self.norm_dif = None
        self.histo_dif = None

    #######################################################################

    def update(self, filename):
        # cumulative histogram

        data = self._get_filter_data()
        utisum = data.sum(3).sum(1).sum(0)
        norm, histo = self._get_histo_utc()
        utsum = norm.sum(1).sum(0)

        #norm, histo = _histo_uc(norm, histo)
        #norm, histo = _histo_tc(norm, histo)
        newshape = (norm.shape[0]*norm.shape[1], norm.shape[2])
        norm = norm.reshape(newshape)
        newshape = (histo.shape[0]*histo.shape[1], histo.shape[2], histo.shape[3])
        histo = histo.reshape(newshape)
        norm, histo = _histo_c(norm, histo)

        # cumulative histogram
        for i in xrange(1, histo.shape[1], 1):
            histo[:,i] += histo[:,(i-1)]

        yinfo = 'Cumulative percentage of successful sessions'
        tinfo = 'Cumulative histogram for norm(u,t)(i)'
        figure_0(self.configs, yinfo, tinfo, utisum, utsum, histo, filename)

    #######################################################################

    def update_eval_fdbk(self, uid, username, filename):
        # cumulative histogram
        # take only data for the 'adaptive' configuration

        # for the current user
        data1 = self._get_user_data(uid)
        norm1, histo1 = _histo_utc(None, data1)
        utsum1 = norm1.sum(1).sum(0)

        newshape = (norm1.shape[0]*norm1.shape[1], norm1.shape[2])
        norm1 = norm1.reshape(newshape)
        newshape = (histo1.shape[0]*histo1.shape[1], histo1.shape[2], histo1.shape[3])
        histo1 = histo1.reshape(newshape)
        norm1, histo1 = _histo_c(norm1, histo1)

        for i in xrange(1, histo1.shape[1], 1):
            histo1 [0,i] += histo1 [0,(i-1)]

        # check if the user has sufficient data
        #if utsum1.sum() < 1:
        #    return False

        # for all users
        data = self._get_filter_data()
        norm, histo = self._get_histo_utc()
        utsum = norm.sum(1).sum(0)

        newshape = (norm.shape[0]*norm.shape[1], norm.shape[2])
        norm = norm.reshape(newshape)
        newshape = (histo.shape[0]*histo.shape[1], histo.shape[2], histo.shape[3])
        histo = histo.reshape(newshape)
        norm, histo = _histo_c(norm, histo)

        for i in xrange(1, histo.shape[1], 1):
            histo[0,i] += histo[0,(i-1)]

        # put them together
        utsum[1] = utsum1 [0]
        #utsum[2] = 0.0
        histo[1,:] = histo1 [0,:]
        #histo[2,:] = 0.0
        utsum1 [0] = utsum[1]
        utsum1 [1] = utsum[0]
        utsum1 [2:] = 0.0
        histo1 [0,:] = histo[1,:]
        histo1 [1,:] = histo[0,:]
        histo1 [2:,:] = 0.0

        yinfo = 'rate of success'
        tinfo = 'Cumulative histogram'
        figure_eval_fdbk(username, self.configs, yinfo, tinfo, utsum1, histo1, filename)

        return True

    #######################################################################

    def update13(self, data, filename):
        # histogram

        #data = self._get_filter_data()
#        utisum = data.sum(3).sum(1).sum(0)
#        norm, histo = _histo_utc(None, data)
#        utsum = norm.sum(1).sum(0)

        #norm, histo = _histo_uc(norm, histo)
#        norm, histo = _histo_tc(norm, histo)
        #newshape = (norm.shape[0]*norm.shape[1], norm.shape[2])
        #norm = norm.reshape(newshape)
        #newshape = (histo.shape[0]*histo.shape[1], histo.shape[2], histo.shape[3])
        #histo = histo.reshape(newshape)
#        norm, histo = _histo_c(norm, histo)

        # cumulative histogram
        #for i in xrange(1, histo.shape[1], 1):
        #    histo[:,i] += histo[:,(i-1)]

        yinfo = 'Timing of the system response(in seconds)'
        tinfo = ''
        figure_12(self.configs, yinfo, tinfo, data, filename)

    #######################################################################

    def _prepare_data16(self, data1, percentage=1.0):
        # data[user,task,config,iter]

        data0, data = data1

        shape = data0.shape
        odata = zeros((shape[2], shape[3], 2), dtype=float32, order='c')
        odata[:,:,0] = -1.0

        for c, data_c in data.iteritems():
            for i, data_ci in data_c.iteritems():
                odata[c,i,0] = mean(data_ci)
                odata[c,i,1] = std(data_ci)
                print c, i, odata[c,i,0], odata[c,i,1]

        return odata

    def update16(self, data, filename):
        # histogram

        data = self._prepare_data16(data, percentage=1.0)

        #yinfo = 'Cumulative consistency score'
        yinfo = 'Zoom value'
        tinfo = ''
        figure_15(self.configs, yinfo, tinfo, data, filename)

    def update_displayed_images(self, data, filename):
        # histogram

        xinfo = 'Number of display times of an image'
        yinfo = 'PDF of displayed images'
        tinfo = ''
        figure_displayed_images(self.configs, tinfo, xinfo, yinfo, data, filename)
