#!/usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Provides the trace representation for display.
"""

###########################################################################
# IMPORTS

#import logging
import math
from numpy import float32, array, empty, sin, cos, cumsum, histogram, linspace
import os
from time import time as now

import matplotlib
import matplotlib.cbook
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
import matplotlib.cm as cm
from matplotlib.figure import Figure
import matplotlib.patches as mpatches

from django.conf import settings

from retrieval_libs.dblayer.DbAccess import DbAccess

###########################################################################
# GLOBALS

#logger = logging.getLogger('retrieval')

# implemented as a singleton
dbaccess = DbAccess()

FLAG_TRACEGFX = settings.FLAG_TRACEGFX
DEFAULT_DPI = settings.DPI_TRACEGFX

DEFAULT_TNLEVEL = 5#7
ALPHA_COLOR = 1.0#0.6
ALPHA_GREY = 1.0#0.75
ALPHA_AX = 1.0#0.75

COLORS = None

###########################################################################
# COLORS

COLORS = ['r', 'g', 'b', 'c', 'm', 'y', '#afffaa', '#afeeee', '#ffffff']

###########################################################################
# CLASS

class TraceGfx(object):
    """
    Provides the trace representation for display.
    """

    #######################################################################
    # FUNCTION

    def _generate_image2D_trix(self, rankDict, dispDict, filename):

        fig = Figure(figsize=(4, 4), dpi=DEFAULT_DPI)
        ax = fig.add_subplot(111)

        keys = rankDict.keys()
        keys.sort()
        for key in keys:

            #c = COLORS[key]
            xyz = rankDict[key]

            if key == -1:
                for x, y, z, p in zip(xyz['x'], xyz['y'], xyz['z'], xyz['p']):
                    z = pow(2, -z)
                    #print 'z', z, z/2.0, '%.10f' % (x-z/2.0), '%.10f' % (y-z/2.0)
                    c = cm.gist_yarg(p, 1)
                    art = mpatches.Rectangle((x-z/2.0, y-z/2.0), z, z, facecolor=c, edgecolor=c, linewidth=0.0, alpha=ALPHA_GREY)
                    ax.add_patch(art)

            else:
                continue
                for x, y, z, p in zip(xyz['x'], xyz['y'], xyz['z'], xyz['p']): 
                    z = pow(2, -z)
                    c = cm.gist_yarg(p, 1)
                    art = mpatches.Rectangle((x-z/2.0, y-z/2.0), z, z, facecolor=c, edgecolor=c, linewidth=0.0, alpha=ALPHA_GREY)
                    art.set_zorder(0)
                    ax.add_patch(art)

        keys = rankDict.keys()
        keys.sort()
        for key in keys:

            c = COLORS[key]
            xyz = rankDict[key]
            #x = xyz['x']
            #y = xyz['y']

            #sx, sy = 4./80., 4./80.
            #S = [ 80. * 4. * pow(2, (-s)) for s in xyz['z']]
            #sx, sy = transforms.ScaledTranslation(sx, sy, fig.dpi_scale_trans)

            #golden_mean = (pow(5, 0.5) - 1.0) / 2.0
            #golden_mean = pow(2, 0.5) / 4.0
            #S = [ 80. * 4. * pow(2, (-s)) for s in xyz['z']]
            #print xyz['z'], S

            if key == -1:
                continue
                for x, y, z in zip(xyz['x'], xyz['y'], xyz['z']):            
                    z = pow(2, -z)
                    #art = mpatches.Rectangle((x-z/2.0, y-z/2.0), z, z, facecolor=c, edgecolor='k', alpha=0.75)
                    art = mpatches.Rectangle((x-z/2.0, y-z/2.0), z, z, facecolor=c, edgecolor=c, linewidth=0.0, alpha=ALPHA_COLOR)
                    ax.add_patch(art)

            else:
                for x, y, z in zip(xyz['x'], xyz['y'], xyz['z']):       
                    z = pow(2, -z)
                    #art = mpatches.Rectangle((x-z/2.0, y-z/2.0), z, z, facecolor=c, edgecolor='k', alpha=0.75)
                    art = mpatches.Rectangle((x-z/2.0, y-z/2.0), z, z, facecolor=c, edgecolor=c, linewidth=0.0, alpha=ALPHA_COLOR)
                    #art.set_zorder(0)
                    ax.add_patch(art)

        keys = dispDict.keys()
        keys.sort()
        #logger.debug('dispDict.keys: %s, %s' % (keys, dispDict))
        for key in keys:

            xyz = dispDict[key]
            x = xyz.get('x', None)[0]
            y = xyz.get('y', None)[0]
            z = xyz.get('z', None)[0]
            p = xyz.get('p', None)[0]
            c = cm.gist_yarg(p, 1)            
            c2 = cm.gist_yarg(p/4.0, 1)

            keystr = '%s' % key

            z = pow(2, -z)
            #c = cm.gist_yarg(p, 1)
            art = mpatches.Rectangle((x-z/2.0, y-z/2.0), z, z, facecolor="k", edgecolor=c, linewidth=0.0, alpha=1.0)
            #art.set_zorder(0)
            ax.add_patch(art)

            ax.annotate(keystr, size=10, color="k", xy=(x, y),  xycoords='data',
                    xytext=(3.5, 2.5), textcoords='offset points',
                    bbox=dict(boxstyle="round4", facecolor=c2, edgecolor=c, alpha=1.0))                             
#            ax.annotate(keystr, size=10, xy=(u[0], v[0]),  xycoords='data',
#                    xytext=(-2.5, -3.5), textcoords='offset points',
#                    bbox=dict(boxstyle="round", facecolor=c, edgecolor="k"))
                                
        # transparency
        ax.set_alpha(ALPHA_AX)

        ax.set_xlabel('')
        xbins = array([])
        ax.set_xticks(xbins)
        ax.set_xticklabels([], ha='center', fontdict=6)
        ax.set_xlim(0.0, 1.0)

        ax.set_ylabel('')
        ybins = array([])
        ax.set_yticks(xbins)
        ax.set_yticklabels([], ha='center', fontdict=6)
        ax.set_ylim(0.0, 1.0)

        #ax.set_title(tinfo)
        ax.grid(False)

        canvas = FigureCanvas(fig)
        canvas.print_figure(filename, dpi=DEFAULT_DPI)

    def _generate_image2D_prob(self, rankDict, dispDict, filename):

        fig = Figure(figsize=(4, 4), dpi=DEFAULT_DPI)
        ax = fig.add_subplot(111)

        keys = rankDict.keys()
        keys.sort()
        for key in keys:

            c = COLORS[key]
            xyz = rankDict[key]

            if key == -1:
                for x, y, z, p in zip(xyz['x'], xyz['y'], xyz['z'], xyz['p']):
                    z = pow(2, -z)
                    c = cm.gist_yarg(p, 1)
                    #c = cm.gist_heat(1-p, 1)
                    art = mpatches.Rectangle((x-z/2., y-z/2.), z, z, facecolor=c, edgecolor=c, linewidth=0.0, alpha=ALPHA_GREY)
                    ax.add_patch(art)

            else:
                for x, y, z, p in zip(xyz['x'], xyz['y'], xyz['z'], xyz['p']):
                    z = pow(2, -z)
                    c = cm.gist_yarg(p, 1)
                    art = mpatches.Rectangle((x-z/2., y-z/2.), z, z, facecolor=c, edgecolor=c, linewidth=0.0, alpha=ALPHA_GREY)
                    art.set_zorder(0)
                    ax.add_patch(art)

        keys = dispDict.keys()
        keys.sort()
        for key in keys:

            xyz = dispDict[key]
            x = xyz.get('x', None)[0]
            y = xyz.get('y', None)[0]
            z = xyz.get('z', None)[0]
            p = xyz.get('p', None)[0]
            c = cm.gist_yarg(p, 1)            
            c2 = cm.gist_yarg(p/4.0, 1)

            keystr = '%s' % key

            z = pow(2, -z)
            art = mpatches.Rectangle((x-z/2.0, y-z/2.0), z, z, facecolor="k", edgecolor=c, linewidth=0.0, alpha=1.0)
            ax.add_patch(art)

#            ax.annotate(keystr, size=10, color="k", xy=(x, y),  xycoords='data',
#                    xytext=(3.5, 2.5), textcoords='offset points',
#                    bbox=dict(boxstyle="round4", facecolor=c2, edgecolor=c, alpha=1.0))  
                    
        # transparency
        ax.set_alpha(ALPHA_AX)

        ax.set_xlabel('')
        xbins = array([])
        ax.set_xticks(xbins)
        ax.set_xticklabels([], ha='center', fontdict=6)
        ax.set_xlim(0.0, 1.0)

        ax.set_ylabel('')
        ybins = array([])
        ax.set_yticks(xbins)
        ax.set_yticklabels([], ha='center', fontdict=6)
        ax.set_ylim(0.0, 1.0)

        #ax.set_title(tinfo)
        ax.grid(False)

        canvas = FigureCanvas(fig)
        canvas.print_figure(filename, dpi=DEFAULT_DPI)

    #######################################################################
    #

    def _prepare_data(self, rankdata, dispdata):

        #logger.debug('START')
        #start_time = now()

        ###
        dispDict = {}

        if type(dispdata) is dict:


            dispdbid = dispdata['imdbid']
            dispclix = dispdata['tnclix']
            dispprob = dispdata['improb']
            if 'tnlevel' in dispdata:
                displn = dispdata['tnlevel']
            else:
                displn = empty(dispdbid.shape, dtype=float32, order='c')
                displn[:] = DEFAULT_TNLEVEL
                #displn = self.dbtree.level_of_seed_dbid(dispdbid)
                #displn = [item[1] for item in displn]

            dispfeat = self.dbimg.featdata(dispdbid, -1, False)

            for i, clix in enumerate(dispclix):

                xyz = dispDict.get(clix, {})
                x = xyz.get('x', [])
                y = xyz.get('y', [])
                z = xyz.get('z', [])
                p = xyz.get('p', [])

                x.append(dispfeat[0][i,0])
                y.append(dispfeat[0][i,1])
                p.append(dispprob[i])
                z.append(displn[i])

                xyz['x'] = x
                xyz['y'] = y
                xyz['z'] = z
                xyz['p'] = p
                dispDict[clix] = xyz

        #else:
        #    dispDict = {}

        ###
        rankDict = {}

        if type(rankdata) is dict:

            rankdbid = rankdata['imdbid']
            rankclix = rankdata['tnclix']
            rankprob = rankdata['improb']
            if 'tnlevel' in dispdata:
                rankln = rankdata['tnlevel']
            else:
                #rankln = empty(rankdbid.shape, dtype=float32, order='c')
                #rankln[:] = DEFAULT_TNLEVEL
                rankln = self.dbtree.level_of_seed_dbid(rankdbid)
                rankln = [item[1] for item in rankln]
                
            rankfeat = self.dbimg.featdata(rankdbid, -1, False)

            rankclixln = len(rankclix)
            for i, clix in enumerate(rankclix):

                if rankclixln > 20000 and rankln[i] != DEFAULT_TNLEVEL:
                    continue
                    
                xyz = rankDict.get(clix, {})
                x = xyz.get('x', [])
                y = xyz.get('y', [])
                z = xyz.get('z', [])
                p = xyz.get('p', [])

                x.append(rankfeat[0][i,0])
                y.append(rankfeat[0][i,1])
                p.append(rankprob[i])
                z.append(rankln[i])

                xyz['x'] = x
                xyz['y'] = y
                xyz['z'] = z
                xyz['p'] = p
                rankDict[clix] = xyz

        #elif type(rankdata) is ndarray:
        #    rankDict = {}
        #else:
        #    rankDict = {}

        #stop_time = now()
        #logger.debug('DONE in %f seconds' % (stop_time - start_time))

        return rankDict, dispDict

    #######################################################################
    # main functionality

    def init(self):
        """
        Initialize the object.
        """

        self.dbimg = dbaccess.dbimg()
        self.dbtree = dbaccess.dbtree()
        self.dbfig = dbaccess.dbfig()

    def traceplot2(self, rankdata, dispdata, interactive=False):
        """
        Prepare the trace plot.
        """

        return None

    def traceplot(self, rankdata, dispdata, interactive=False):
        """
        Prepare the trace plot.
        """

        # since it is expensive, one may want to skip it
        if not FLAG_TRACEGFX:
            return None

        fileurls = []

        TABLE = dbaccess.dbtable()

        if TABLE.startswith('ImageSynthetic'):

            rankDict, dispDict = self._prepare_data(rankdata, dispdata)

            if rankDict is not None and dispDict is not None:

                if FLAG_TRACEGFX.get('TRIX_2D', False):
                    filename, fileurl = self.dbfig.filename('traceplot', 'trix_2D', '.png')
                    self._generate_image2D_trix(rankDict, dispDict, filename)
                    # check if now the file exists
                    if os.access(filename, os.F_OK):
                        fileurls.append(fileurl)

                if FLAG_TRACEGFX.get('PROB_2D', False):
                    filename, fileurl = self.dbfig.filename('traceplot', 'prob_2D', '.png')
                    self._generate_image2D_prob(rankDict, dispDict, filename)

                    if os.access(filename, os.F_OK):
                        fileurls.append(fileurl)

        if not fileurls:
            fileurls = None

        return fileurls

    def traceplot_r(self, rankdata, dispdata, interactive=False):
        """
        Prepare the trace plot.
        """

        # since it is expensive, one may want to skip it
        if not FLAG_TRACEGFX:
            return None

        TABLE = dbaccess.dbtable()

        if TABLE.startswith('ImageSynthetic'):

            rankDict, dispDict = self._prepare_data(rankdata, dispdata)

            if rankDict is not None and dispDict is not None:

                if FLAG_TRACEGFX.get('TRIX_2D', False):
                    filename, fileurl = self.dbfig.filename('traceplot', 'trix_2D', '.png')
                    self._generate_image2D_trix(rankDict, dispDict, filename)

                if FLAG_TRACEGFX.get('PROB_2D', False):
                    filename, fileurl = self.dbfig.filename('traceplot', 'prob_2D', '.png')
                    self._generate_image2D_prob(rankDict, dispDict, filename)

        else:
            pass