#!/usr/bin/env python

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Manages the workflow within one iteration of the retrieval process.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, zeros, ones
import sys
from time import time as now

from retrieval_libs.search.im_rnd.QueryImg import QueryImg as QueryImg_r
from retrieval_libs.search.im_orig.QueryImg import QueryImg as QueryImg_0
from retrieval_libs.search.im_heat.QueryImg import QueryImg

###########################################################################
# GLOBALS

logger = logging.getLogger('retrieval')

###########################################################################
# CLASS

class SearchMain(object):
    """
    Manages the workflow within one iteration of the retrieval process.
    """

    #######################################################################
    #  local functionality

    def _perform_img_random(self, rankdata, rankdata_0, dispfdbk):

        logger.debug('Image random - START')
        start_time = now()

        dispdata = dispfdbk[0]
        if dispdata is not None:
            sys.stderr.write('\nERROR DISPDATA\n')
            sys.exit(1)

        fdbkdata = dispfdbk[1]
        if fdbkdata is not None:
            sys.stderr.write('\nERROR FDBKDATA\n')
            sys.exit(1)

        self.queryimg.perform_query_random(rankdata, rankdata_0, dispfdbk)

        self.rankdata = self.queryimg.rankdata
        self.rankdata_0 = self.queryimg.rankdata_0
        self.dispdata = self.queryimg.seeddata
        self.tracedata = self.queryimg.tracedata

        stop_time = now()
        logger.debug('Image random - DONE in %f seconds' % (stop_time - start_time))

        return self.rankdata

    def _perform_img_search(self, rankdata, rankdata_0, dispfdbk):

        logger.debug('Image search - START')
        start_time = now()

        dispdata = dispfdbk[0]
        if dispdata is None:
            sys.stderr.write('\nERROR DISPDATA\n')
            sys.exit(1)

        fdbkdata = dispfdbk[1]
        if fdbkdata is not None and not isinstance(fdbkdata, int32):
            sys.stderr.write('\nUNKNOWN FDBKDATA %s, %s\n' % (type(fdbkdata), str(fdbkdata)))
            sys.exit(1)

        self.queryimg.perform_query_fdbk(rankdata, rankdata_0, dispfdbk)

        self.rankdata = self.queryimg.rankdata
        self.rankdata_0 = self.queryimg.rankdata_0
        self.dispdata = self.queryimg.seeddata
        self.tracedata = self.queryimg.tracedata

        stop_time = now()
        logger.debug('Image search - DONE in %f seconds' % (stop_time - start_time))

        return self.rankdata

    #######################################################################
    # main functionality

    def init(self, TRACE, SYS_THARGS, SYS_WEIGHTS, SYS_CONSISTENCY, SYS_VERSION):
        """
        Initialize the object.
        """

        if SYS_VERSION == -1:
            self.queryimg = QueryImg_r()
            self.queryimg.init(SYS_THARGS, SYS_WEIGHTS, SYS_CONSISTENCY)
        elif SYS_VERSION == 0:
            self.queryimg = QueryImg_0()
            self.queryimg.init(SYS_THARGS, SYS_WEIGHTS, SYS_CONSISTENCY)
        elif SYS_VERSION == 2:
            self.queryimg = QueryImg()
            self.queryimg.init(TRACE, SYS_THARGS, SYS_WEIGHTS, SYS_CONSISTENCY)
        else:
            sys.stderr.write('\nUNKNOWN SYS_VERSION %s\n\n' % SYS_VERSION)
            sys.exit(1)

    def set_data(self, dispfdbkH, thargsH, weightsH, consistencyH):
        """
        FWD
        """

        self.queryimg.set_data(dispfdbkH, thargsH, weightsH, consistencyH)

    def trim_data(self):
        """
        FWD
        """

        self.queryimg.trim_data()

    #######################################################################
    # update functionality

    def update_thargs(self, rankdata, input_data, dispfdbkH, thargsH, weightsH):
        """
        Update the thargs.

        HINT: this is kept separated from the retrieval routine, for easy
        changes in the flow of operations (ie. different treads possible)
        """

        action = input_data.get('action', None)

        if action in ('new_rnd', 'fdbk_im'):
            dispfdbk = input_data.get('dispfdbk', None)
            thargs = self.queryimg.update_thargs(rankdata, dispfdbk, dispfdbkH, thargsH, weightsH)
        else:
            thargs= None

        return thargs

    def update_weights(self, rankdata, input_data, dispfdbkH, thargsH, weightsH):
        """
        Update the weights.

        HINT: this is kept separated from the retrieval routine, for easy
        changes in the flow of operations (ie. different treads possible)
        """

        action = input_data.get('action', None)

        if action in ('new_rnd', 'fdbk_im'):
            dispfdbk = input_data.get('dispfdbk', None)
            weights = self.queryimg.update_weights(dispfdbk, dispfdbkH, thargsH, weightsH)
        else:
            weights = None

        return weights

    def update_consistency(self, rankdata, input_data, dispfdbkH, thargsH, weightsH, consistencyH):
        """
        Update the consistency.

        HINT: this is kept separated from the retrieval routine, for easy
        changes in the flow of operations (ie. different treads possible)
        """

        action = input_data.get('action', None)

        if action in ('new_rnd', 'fdbk_im'):
            dispfdbk = input_data.get('dispfdbk', None)
            consistency = self.queryimg.update_consistency(rankdata, dispfdbk, dispfdbkH, thargsH, weightsH, consistencyH)
        else:
            consistency= 1.0

        return consistency

    #######################################################################

    def search(self, rankdata, rankdata_0, input_data):
        """
        Perform the retrieval routine (ie. the pipeline of algorithms).
        """

        action = input_data.get('action', None)
        data = input_data.get('data', None)

        if action not in ('new_rnd', 'fdbk_im'):
            sys.stderr.write('\nUNKNOWN ACTION %s\n' % action)
            sys.exit(1)

        # refresh the CLASS variables
        self.rankdata = None
        self.rankdata_0 = None
        self.dispdata = None
        self.tracedata = None

        if action == 'new_rnd':
            if data is not None:
                sys.stderr.write('\nERROR %s\n' % action)
                sys.exit(1)

            dispfdbk = input_data.get('dispfdbk', None)
            self._perform_img_random(rankdata, rankdata_0, dispfdbk)
            self.rankdata = self.queryimg.rankdata
            self.tracedata = self.queryimg.tracedata

        # image relevance feedback
        elif action == 'fdbk_im':
            dispfdbk = input_data.get('dispfdbk', None)
            self._perform_img_search(rankdata, rankdata_0, dispfdbk)
            self.rankdata = self.queryimg.rankdata
            self.tracedata = self.queryimg.tracedata

        else:
            sys.stderr.write('\nUNKNOWN ACTION %s\n' % action)
            sys.exit(1)

        return self.rankdata
