#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Manages the HEAT retrieval algorithm.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, float32, ndarray, array, ones, empty
from random import Random
import sys
from time import time as now

from retrieval_libs.dblayer.DbAccess import DbAccess
from retrieval_libs.display.TraceGfx import TraceGfx

from libc.math cimport abs as c_abs

from retrieval_libs.common.index cimport get_index0
from retrieval_libs.common.extend cimport extend_float32_ndim1

from retrieval_libs.common.candidate cimport skipped_images

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

# implemented as a singleton
cdef object dbaccess = DbAccess()

random = Random()
random.jumpahead(8)

# the number of image/seeds selected for display
cdef cnp.int32_t SIZE = 8

cdef bint TEST = False

cdef cnp.float32_t ACCURACY = pow(10, -6)

# maximum number of collapse/expand loops to stabilize the trace
NB_EXPCOL_MAX = 10

###########################################################################
# FUNCTIONS

def _test_rankdata(rankdata, rankdata_0):
    """
    Test if the corresponding probabilities in rankdata and randata_0
    are equal.
    """

    flag = True
    counter = 0

    imdbid = rankdata['imdbid']
    for trix, dbid in enumerate(imdbid):

        # identify the corresponding trix_0
        trix_0 = dbid - 1 # for efficiency, this is the case for mm_0
        if len (rankdata_0['imdbid']) <= trix_0 or \
           rankdata_0['imdbid'][trix_0] != dbid:
            # if the guess was not correct
            trix_0 = get_index0(dbid, rankdata_0['imdbid'], False)
        if trix_0 == -1:
            continue

        # compare the corresponding probabilities
        improb = rankdata['improb'][trix]
        improb_0 = rankdata_0 ['improb'][trix_0]
        if c_abs(improb - improb_0) > ACCURACY:
            logger.error('RANKDATA %s/%s/%s IMDBID %s: IMPROB %s != %s' %
                         (len(rankdata['imdbid']), len(rankdata_0['imdbid']),
                          counter, dbid, improb, improb_0))
            flag = False
            break

        counter += 1

    if flag == True:
        logger.info('RANKDATA TEST OK, OVERLAP %s/%s/%s SAMPLE %s' %
                    (len(rankdata['imdbid']), len(rankdata_0['imdbid']),
                     counter, random.sample(rankdata['improb'], 4)))
    else:
        sys.stderr.write('\nRANKDATA ERROR\n')
        sys.exit(1)

###########################################################################
# CLASS

cdef class QueryImg(object):
    """
    Manages the HEAT retrieval algorithm.
    """

    #######################################################################
    # main functionality

    def init(self, TRACE, SYS_THARGS, SYS_WEIGHTS, SYS_CONSISTENCY):
        """
        Initialize the object.
        """

        cdef DbTableImg dbimg

        self.thargs = SYS_THARGS[0]
        self.weights = SYS_WEIGHTS[0]

        self.rankdata = None
        self.rankdata_0 = None
        self.seeddata = None
        self.tracedata = None

        #self.thargsH = None
        #self.weightsH = None
        #self.dispfdbkH = None

        self.dbimg = dbaccess.dbimg()
        self.dbtree = dbaccess.dbtree()

        self.traceupdater = TraceUpdater()
        self.traceupdater.init(self.dbimg, self.dbtree, TRACE)
        self.consistencyupdater = ConsistencyUpdater()
        self.consistencyupdater.init(self.dbimg, SYS_CONSISTENCY)
        self.probupdater = ProbabilityUpdater()
        self.probupdater.init(self.dbimg)
        self.dispselector = DisplaySetSelector()
        self.dispselector.init(self.dbimg, SIZE)

        if TEST == True:
            self.traceupdater_0 = TraceUpdater_0()
            self.traceupdater_0.init()
            #self.traceupdater_0.init(self.dbimg, TRACE)
            self.probupdater_0 = ProbabilityUpdater()
            self.probupdater_0.init(self.dbimg)
            self.dispselector_0 = DisplaySetSelector_0()
            self.dispselector_0.init(self.dbimg, SIZE)

        self.COUNT_EXPAND = 0
        self.COUNT_COLLAPSE = 0

        self.traceimg = TraceGfx()
        self.traceimg.init()

    def set_buffers(self):
        """
        FWD
        """

        self.dbimg.set_buffers()

    def trim_buffers(self):
        """
        FWD
        """

        self.dbimg.trim_buffers()

    def set_data(self, dispfdbkH, thargsH, weightsH, consistencyH):

        self.thargsH = thargsH
        self.weightsH = weightsH
        self.dispfdbkH = dispfdbkH
        self.consistencyH = consistencyH

        self.thargs = thargsH[-1]
        self.weights = weightsH[-1]
        #self.dispfdbk = dispfdbkH[-1]
        #self.consistency = consistencyH[-1]

    def trim_data(self):

        self.thargsH = None
        self.weightsH = None
        self.dispfdbkH = None
        self.consistencyH = None

    #######################################################################
    # update functionality

    def update_thargs(self, rankdata, dispfdbk, dispfdbkH, thargsH, weightsH):
        """
        Update the thargs.
        """

        return self.thargs

    def update_weights(self, dispfdbk, dispfdbkH, thargsH, weightsH):
        """
        Update the weights.
        """

        return self.weights

    def update_consistency(self, rankdata, dispfdbk, dispfdbkH, thargsH, weightsH, consistencyH):
        """
        Update the consistency.
        """

        return self.consistencyupdater.update(dispfdbk, dispfdbkH, thargsH, weightsH, consistencyH)

    #######################################################################
    # main functionality

    def perform_query_random(self, rankdata, rankdata_0, dispfdbk):
        """
        Generate the seeddata (i.e. the next dispdata).

        - NO relevance feedback >> NO update for the image probabilities;
        - Voronoi partitioning starting from a random image/seed;
        """

        # initialize skipdisp
        skipdisp = skipped_images(self.dispfdbkH)
        if len(skipdisp) > 0:
            skipped_nodes = self.dbimg.parentnode_dbid(skipdisp)
            self.traceupdater.set_skipped_nodes(skipped_nodes)

        self.rankdata = self.traceupdater.rankdata(rankdata)
        self.probupdater.update(self.rankdata, dispfdbk, self.thargs, self.weights)
        self.dispselector.set_consistency(self.consistencyH)

        # save/store/checkpoint
        self._add_tracedata(self.rankdata, dispfdbk[0], ['initial', 'RAND'])

        #rankdata_ref = self.rankdata
        #self.rankdata = self._collapse_rankdata(self.rankdata, None)
        #self.rankdata = self._expand_rankdata(self.rankdata, None, None, rankdata_ref)

        self.seeddata = self.dispselector.update(self.rankdata, self.thargs, self.weights, self.dispfdbkH)

        self._add_tracedata(self.rankdata, self.seeddata, ['final', 'RAND'])

        self.rankdata = self.traceupdater.clean_rankdata(self.rankdata)

        if TEST == True:
            self.rankdata_0 = self.traceupdater_0.rankdata(rankdata_0)
            self.probupdater_0.update(self.rankdata_0, dispfdbk, self.thargs, self.weights)

            #self.dispselector_0.given_seed_candidates = self.seeddata
            #unused_seeddata_0 = self.dispselector_0.update(self.rankdata_0, self.thargs, self.weights, self.dispfdbkH)
            #print 'self.seeddata', self.seeddata['imdbid'], unused_seeddata['imdbid']
            #self.traceupdater.conditional_entropy(self.rankdata, self.rankdata_0)
            self.rankdata_0 = self.traceupdater_0.clean_rankdata(self.rankdata_0)

            _test_rankdata(self.rankdata, self.rankdata_0)

    def perform_query_fdbk(self, rankdata, rankdata_0, dispfdbk):
        """
        Generate the seeddata (i.e. the next dispdata).

        - update the image probabilities based on the given relevance feedback
        - perform the Voronoi partitioning algorithm starting from
          the image/seed with the highest probability
        """

        # initialize skipdisp
        skipdisp = skipped_images(self.dispfdbkH)
        if len(skipdisp) > 0:
            skipped_nodes = self.dbimg.parentnode_dbid(skipdisp)
            self.traceupdater.set_skipped_nodes(skipped_nodes)

        self.rankdata = self.traceupdater.rankdata(rankdata)
        self.probupdater.update(self.rankdata, dispfdbk, self.thargs, self.weights)
        self.dispselector.set_consistency(self.consistencyH)

        # save/store/checkpoint
        self._add_tracedata(self.rankdata, dispfdbk[0], ['initial', 'FDBK'])

        self.probupdater.init_update_full(self.dispfdbkH, self.thargsH, self.weightsH)

        # collapse/expand until the trace is stable or NB_EXPCOL_MAX is reached
        dbid = self.rankdata['imdbid']
        seeddata = dispfdbk[0]

        #for i, item in enumerate(self.dbimg.FFBUFFER[2]):
        #    if len(item) > 500:
        #        print 'FFBUFFER', self.dbimg.FFBUFFER[1][i], len(item), item[0:8]

        current_set = sorted(set(self.rankdata['pndbid']))
        previous_set = set([])
        counter = 0
        while current_set != previous_set:
            rankdata_ref = self.rankdata
            #logger.debug('    COLLAPSE START %s' % counter)
            self.rankdata = self._collapse_rankdata(self.rankdata, None)
            #logger.debug('    COLLAPSE END / EXPAND START %s' % counter)
            self.rankdata = self._expand_rankdata(self.rankdata, None, seeddata, rankdata_ref)
            #logger.debug('    EXPAND END %s' % counter)
            counter += 1

            # HINT: check it here, because computing current_set is expensive
            if counter >= NB_EXPCOL_MAX:
                break

            previous_set = current_set
            current_set = sorted(set(self.rankdata['pndbid']))

        newdbid = self.rankdata['imdbid']

        #for i, item in enumerate(self.dbimg.FFBUFFER[2]):
        #    if len(item) > 500:
        #        print 'FFBUFFER', self.dbimg.FFBUFFER[1][i], len(item), item[0:8]

        # clean
        self.probupdater.init_update_full(None, None, None)

        self.traceupdater.expcol_databuffer(None, newdbid, dbid)
        logger.info('    EXPCOL WAS STABLE AFTER %s ITERATIONS' % counter)

        self.seeddata = self.dispselector.update(self.rankdata, self.thargs, self.weights, self.dispfdbkH)

        self._add_tracedata(self.rankdata, self.seeddata, ['final', 'FDBK'])

        self.rankdata = self.traceupdater.clean_rankdata(self.rankdata)

        if TEST == True:
            self.rankdata_0 = self.traceupdater_0.rankdata(rankdata_0)
            self.probupdater_0.update(self.rankdata_0, dispfdbk, self.thargs, self.weights)

            #self.dispselector_0.given_seed_candidates = self.seeddata
            #unused_seeddata = self.dispselector_0.update(self.rankdata_0, self.thargs, self.weights, self.dispfdbkH)
            #print 'self.seeddata', self.seeddata['imdbid'], unused_seeddata['imdbid']
            #self.traceupdater.conditional_entropy(self.rankdata, self.rankdata_0)
            self.rankdata_0 = self.traceupdater_0.clean_rankdata(self.rankdata_0)

            _test_rankdata(self.rankdata, self.rankdata_0)

    #######################################################################
    # expand functionality

    cdef dict _expand_rankdata(self, dict rankdata, tuple nodes, dict seeddata, dict rankdata_ref):
        """
        Perform the expanding operation.

        Replace all the nodes in the rankdata with their children nodes.
        """

        #cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] seeddbid
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] trix_to_copy_from
        cdef dict deltadata

        logger.debug('START')
        start_time = now()

        # save/store/checkpoint
        self._add_tracedata(rankdata, seeddata, ['expand', 'IN'])

        trix_to_copy_from, deltadata = self.traceupdater.expand_specs(rankdata, nodes)
        if deltadata is None:
            # save/store/checkpoint
            self._add_tracedata(rankdata, seeddata, ['expand', 'OUT'])
            stop_time = now()
            logger.debug('    DONE [UNCHANGED] in %f seconds'
                         % (stop_time - start_time))
            return rankdata

#        self.probupdater.update_full(deltadata)

        #dbid = rankdata['imdbid']
        rankdata = self.traceupdater.expcol_rankdata(rankdata, trix_to_copy_from, deltadata)
        #newdbid = rankdata['imdbid']
        #seeddbid = None #seeddata['imdbid']
        #self.traceupdater.expcol_databuffer(seeddbid, newdbid, dbid)

        dbid = rankdata_ref['imdbid']
        newdbid = rankdata['imdbid']
        deltatrix, deltadbid = self.traceupdater._reformat_specs2(dbid, newdbid)
        deltadata2 = {'imdbid': deltadbid}

        if len(deltadbid) > 0:
            self.probupdater.update_full(deltadata2)
        else:
            deltadata2 = {'improb': array([], dtype=float32, order='c')}

        rankprob = extend_float32_ndim1(rankdata_ref['improb'], deltatrix, deltadata2['improb'])
        rankdata['improb'] = rankprob

        self.COUNT_EXPAND += 1

        # save/store/checkpoint
        self._add_tracedata(rankdata, seeddata, ['expand', 'OUT'])

        stop_time = now()
        logger.info('    EXPAND DONE in %f seconds' % (stop_time - start_time))

        return rankdata

    #######################################################################
    # collapse functionality

    cdef dict _collapse_rankdata(self, dict rankdata, dict seeddata):
        """
        Perform the collapsing operation.

        First, rankdata is arranged in a dictionary format. Second, the
        recursive collapsing routine is called. Third, rankdata is rebuilt.
        """

        cdef dict deltadata

        logger.debug('START')
        start_time = now()

        # save/store/checkpoint
        self._add_tracedata(rankdata, seeddata, ['collapse', 'IN'])

        # reformat rankdata
        rankmeta = self.traceupdater.init_collapse(rankdata)

        # recursive collapsing operation
        while True:
            #deltadata = self.traceupdater.collapse_specs(rankmeta, self.rankdata_0)
            deltadata = self.traceupdater.collapse_specs(rankmeta, None)
            if deltadata is None:
                break

            self.probupdater.update_full(deltadata)

            # rankmeta is modified in place
            self.traceupdater.do_collapse(rankmeta, deltadata)

            self.COUNT_COLLAPSE += 1

        # reformat rankdata
        trix_to_copy_from, deltadata = self.traceupdater.exit_collapse(rankmeta)

        #dbid = rankdata['imdbid']
        rankdata = self.traceupdater.expcol_rankdata(rankdata, trix_to_copy_from, deltadata)
        #newdbid = rankdata['imdbid']
        #self.traceupdater.expcol_databuffer(None, newdbid, dbid)

        # save/store/checkpoint
        self._add_tracedata(rankdata, seeddata, ['collapse', 'OUT'])

        stop_time = now()
        logger.info('    COLLAPSE DONE in %f seconds' % (stop_time - start_time))

        return rankdata

    #######################################################################
    # trace functionality

    cdef _add_tracedata(self, dict rankdata, dict seeddata, list action):
        """
        Prepare and append the trace data.
        """

        # if necessary, initialize tracedata
        if self.tracedata is None:
            self.tracedata = []

        if action[0] == 'initial':
            actioninfo = [' '.join(action), self.COUNT_EXPAND, len(rankdata['imdbid'])]
        elif action[0] == 'seed':
            actioninfo = [' '.join(action), len(seeddata['imdbid']), len(rankdata['imdbid'])]
        elif action[0] == 'expand':
            actioninfo = [' '.join(action), self.COUNT_EXPAND, len(rankdata['imdbid'])]
        elif action[0] == 'collapse':
            actioninfo = [' '.join(action), self.COUNT_COLLAPSE, len(rankdata['imdbid'])]
        elif action[0] == 'final':
            COUNT_TOTAL = self.COUNT_EXPAND + self.COUNT_COLLAPSE
            actioninfo = [' '.join(action), COUNT_TOTAL, len(rankdata['imdbid'])]
        else:
            sys.stderr.write('\nUNKNOWN ACTION %s\n' % action[0])
            sys.exit(1)

        fileurl = None
        if logger.getEffectiveLevel() < logging.DEBUG or \
           action[0] == 'final':
           #action[0] == 'initial' or \
            fileurl = self.traceimg.traceplot(rankdata, seeddata)

        actioninfo.append(fileurl)
        traceinfo = tuple(actioninfo)

        self.tracedata.append(traceinfo)
