#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Manages the rankdata.

.. note:: Rankdata has the size of the trace within the
          hierarchical tree, and it is dynamically updated by
          collapse/expand operations.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, float32, ndarray, array, zeros, ones, empty, fromstring
from math import log
import operator
from random import Random
import sys
from time import time as now

from retrieval_libs.common.candidate cimport candidate_nodes, candidate_nodes2
from retrieval_libs.common.extend cimport extend_int32_ndim1, extend_int64_ndim1, extend_float32_ndim1
from retrieval_libs.common.index cimport get_index0

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

random = Random()
random.jumpahead(8)

cdef cnp.float32_t PACCURACY = 0.000001
cdef cnp.float32_t EPSILON = 0.000001

###########################################################################
# CLASS

cdef class TraceUpdater(object):
    """
    Manages the rankdata.

    .. note:: Rankdata has the size of the trace within the
              hierarchical tree, and it is dynamically updated by
              collapse/expand operations.
    """

    ###########################################################################
    # local functionality

    cdef dict _tracedata(self, cnp.ndarray tndbid, bint unlock):

        cdef cnp.ndarray imdbid
        cdef cnp.int32_t dataln
        cdef tuple shape
        cdef dict data

        if unlock == True and tndbid is None:
            imdbid = self.dbimg.all_ids()

            dataln = len(imdbid)
            shape = (dataln,)

            data = {}
            data['imdbid'] = imdbid
            data['tnsize'] = ones(shape, dtype=int32, order='c')
            data['tndbid'] = zeros(shape, dtype=int32, order='c')
            data['pndbid'] = ones(shape, dtype=int32, order='c')
            data['chsize'] = zeros(shape, dtype=int32, order='c')
            data['tnlevel'] = empty(shape, dtype=int32, order='c')
            data['tnlevel'][:] = -1

        else:
            data = self.dbtree.tracedata(tndbid)

        return data

    cdef dict _deltadata(self, cnp.ndarray tndbid):

        data = self.dbtree.expand_tracedata(tndbid)

        return data

    #######################################################################
    # local functionality

    cdef dict _expcol_rankdata(self, dict rankdata, list trix, dict deltadata):
        """
        Returns a new rankdata, composed from rankdata and deltadata
        according to the indexes in trix.
        """

        logger.debug('START')
        start_time = now()

        newrankdata = {}

        # HINT: the keys that are NOT in deltadata are ignored
        for key, deltaitem in deltadata.iteritems():
            item = rankdata[key]

            if isinstance(item, ndarray) and item.dtype == int32:
                newrankdata[key] = extend_int32_ndim1(item, trix, deltaitem)
            elif isinstance(item, ndarray) and item.dtype == int:
                newrankdata[key] = extend_int64_ndim1(item, trix, deltaitem)
            elif isinstance(item, ndarray) and item.dtype == float32:
                newrankdata[key] = extend_float32_ndim1(item, trix, deltaitem)
            else:
                sys.stderr.write('\nUNKNOWN %s\n' % type(item))
                sys.exit(1)

        # verify the rankdata
        if logger.getEffectiveLevel() < logging.DEBUG:
            dataln = len(newrankdata['imdbid'])
            for key, item in newrankdata.iteritems():
                if len(item) != dataln:
                    sys.stderr.write('\nRANKDATA MISMATCH %s, %d, %d\n' %
                                     (key, len(item), dataln))
                    sys.exit(1)

        stop_time = now()
        logger.debug('DONE in %f seconds' % (stop_time - start_time))

        return newrankdata

    #######################################################################
    # local functionality

    cdef dict _update_rankdata(self, dict rankdata):
        """
        Initialize the rankdata.

        FORM: {'imdbid': array(),
               'improb': array(),
               'tndbid': array(),
               'pndbid': array(),
               'tnsize': array(),
               'chsize': array(),
               'tnclix': array(),}
        """

#        cdef list TRACE
        cdef cnp.int32_t dataln
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] tndbid

        logger.debug('START')
        start_time = now()

        if rankdata is None:
            # initialize the trace, then initialize rankdata
            TRACE = self._init_trace()
            rankdata = self._deltadata(TRACE)

        else:
            # verify, then re-initialize rankdata
            if not isinstance(rankdata, dict):
                sys.stderr.write('\nUNKNOWN TYPE %s\n' % type(rankdata))
                sys.exit(1)

            if 'tnclix' in rankdata:
                sys.stderr.write('\nRANKCLIX OVERWRITTEN\n')
                sys.exit(1)

        # verify the rankdata
        if logger.getEffectiveLevel() < logging.DEBUG:
            tndbid = rankdata['tndbid']
            dataln = tndbid.shape[0]
            for key, item in rankdata.iteritems():
                if len(item) != dataln:
                    sys.stderr.write('\nRANKDATA MISMATCH %s, %d, %d\n' % (key, len(item), dataln))
                    sys.exit(1)

        stop_time = now()
        logger.debug('DONE in %f seconds' % (stop_time - start_time))

        return rankdata

    #######################################################################
    # local functionality

    cdef dict _init_trace_recursive(self, dict trace, bint recursive):
        # a trace is a list of tree node dbid
        # RECURSIVE

        trixtoexpand = candidate_nodes(trace, 4)
        if trixtoexpand is None:
            return trace

        tndbid = trace['tndbid']
        tnlevel = trace['tnlevel']
        tnaccess = trace['tnaccess']

        tndbidtoexpand = [tndbid[i] for i in trixtoexpand]
        childdata = self.dbtree.children_data(tndbidtoexpand)

        # first, mark the nodes that should not be accesses recursively
        for i, item in enumerate(childdata):
            ictbl = item[2]

            # HINT: fromstring misbehaves for empty string
            #if ictbl == 'tree' and len(icdbid) > 0:
            if ictbl == 'tree':
                pass

            elif ictbl == 'img':
                itrix = trixtoexpand[i]
                tnaccess[itrix] = -1

            else:
                sys.stderr.write('\nUNKNOWN TABLE %s\n' % ictbl)
                sys.exit(1)

        # second, expand the other nodes
        for i, item in enumerate(childdata):
            idbid, ilevel, ictbl, ichsize, icdbid = item

            # HINT: fromstring misbehaves for empty string
            #if ictbl == 'tree' and len(icdbid) > 0:
            if ictbl == 'tree':
                icdbid = fromstring(icdbid, dtype=int32, sep=' ')

                #itrix = trixtoexpand[i] # here, it does NOT stay valid anymore because of DEL
                itrix = tndbid.index(idbid)
                del tndbid[itrix]
                del tnlevel[itrix]
                del tnaccess[itrix]

                tndbid.extend(icdbid)
                tnlevel.extend([ilevel + 1] * len(icdbid))
                tnaccess.extend([0] * len(icdbid))

            elif ictbl == 'img':
                pass

            else:
                sys.stderr.write('\nUNKNOWN TABLE %s\n' % ictbl)
                sys.exit(1)

        if recursive == True and len(tndbid) == 1:
            trace = None
        elif recursive == True and len(tndbid) < self.TRACE_LN:
            trace = self._init_trace_recursive(trace, recursive)
        else:
            pass

        return trace

    cdef cnp.ndarray _init_trace(self):
        # a trace is a list of tree node dbid

        logger.debug('START')
        start_time = now()

        # expand RECURSIVELY the ROOT node
        trace0 = {'tndbid': [1,], 'tnlevel': [0,], 'tnaccess': [0,]}
        trace = self._init_trace_recursive(trace0, True)

        trace = trace['tndbid']
        trace = array(trace, dtype=int32, order='c')

        stop_time = now()
        logger.debug('DONE in %f seconds' % (stop_time - start_time))

        return trace

    #######################################################################
    #  main functionality

    cdef set_skipped_nodes(self, cnp.ndarray skipped):
        """
        """

        self.skipped_nodes = skipped

    cdef init(self, DbTableImg dbimg, DbTableTree dbtree, tuple TRACE):
        """
        Initialize the object.
        """

        self.dbimg = dbimg
        self.dbtree = dbtree

        # initial size/expansion of the trace
        #nb_leaves = self.dbtree.nb_leaves()
        #self.TRACE_LN = min (TRACE[0], nb_leaves)
        self.TRACE_LN = TRACE[0]

        self.skipped_nodes = None

    cdef dict rankdata(self, dict rankdata):
        """
        Initialize the rankdata.
        """

        rankdata = self._update_rankdata(rankdata)

        return rankdata

    cdef dict clean_rankdata(self, dict rankdata):
        """
        Clean the rankdata.
        """

        #if 'closestclix' in rankdata:
        del rankdata['closestclix']
        #if 'closestcldd' in rankdata:
        del rankdata['closestcldd']
        if 'sortedarg' in rankdata:
            del rankdata['sortedarg']
        if 'immass' in rankdata:
            del rankdata['immass']
        #if 'tnclix' in rankdata:
#        del rankdata['tnclix']

        return rankdata

    #######################################################################
    # expand/collapse functionality

    cdef _reformat_specs(self, cnp.int32_t dataln, cnp.ndarray trix_to_copy_from, cnp.int32_t deltaln):

        trix_to_copy_ln = len(trix_to_copy_from)

        trix_to_copy_to = range(0, trix_to_copy_ln, 1)
        trix_to_copy_to = array(trix_to_copy_to, dtype=int32, order='c')

        trix_to_add = range(trix_to_copy_ln, (trix_to_copy_ln + deltaln), 1)
        trix_to_add = array(trix_to_add, dtype=int32, order='c')

        return [trix_to_copy_to, trix_to_copy_from, trix_to_add]

    cdef _reformat_specs2(self,
            cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] old_imdbid,
            cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] new_imdbid):

        cdef cnp.ndarray[cnp.int_t, ndim=1, mode='c'] old_argsort, new_argsort

        cdef cnp.int32_t old_ln, new_ln, old_i, new_i
        cdef cnp.int32_t old_index, new_index
        cdef cnp.int32_t old_current, new_current

        #logger.debug('START')
        start_time = now()

        old_ln = len(old_imdbid)
        new_ln = len(new_imdbid)

        old_argsort = old_imdbid.argsort()
        new_argsort = new_imdbid.argsort()

        trix_to_copy_from = []
        trix_to_copy_to = []
        trix_to_add = []

        old_i = 0
        new_i = 0
        while (old_i < old_ln) and (new_i < new_ln):
            old_index = old_argsort[old_i]
            new_index = new_argsort[new_i]

            old_current = old_imdbid[old_index]
            new_current = new_imdbid[new_index]

            if old_current > new_current:
                # new index is missing from the olddata
                trix_to_add.append(new_index)
                new_i += 1
            elif old_current < new_current:
                # old index is missing from the newdata
                old_i += 1
            else: #if old_current == new_current:
                # old index is present in the newdata
                trix_to_copy_from.append(old_index)
                trix_to_copy_to.append(new_index)
                old_i += 1
                new_i += 1

        while new_i < new_ln:
            new_index = new_argsort[new_i]
            # new index is missing from the olddata
            trix_to_add.append(new_index)
            new_i += 1

        # recover the order
        trix_to_add = sorted(trix_to_add, reverse=False)
        deltadbid = [new_imdbid[i] for i in trix_to_add]
        deltadbid = array(deltadbid, dtype=int32, order='c')

        trix_to_add = array(trix_to_add, dtype=int32, order='c')
        trix_to_copy_to = array(trix_to_copy_to, dtype=int32, order='c')
        trix_to_copy_from = array(trix_to_copy_from, dtype=int32, order='c')

        stop_time = now()
        logger.debug('OLD %s NEW %s TRIX_TO_COPY %s' %
                     (old_ln, new_ln, len(trix_to_copy_to)))
        logger.debug('DONE in %f seconds' % (stop_time - start_time))

        return [trix_to_copy_to, trix_to_copy_from, trix_to_add], deltadbid

    #######################################################################

    cdef tuple expand_specs(self, dict rankdata, nodes):
        """
        Expand all non leaf nodes.
        """

        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] tndbid = rankdata['tndbid']
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] chsize = rankdata['chsize']

        cdef cnp.int32_t i, dataln

        deltatndbid = []
        trix_to_copy_from = []

        dataln = tndbid.shape[0]
        for i from 0 <= i < dataln:
            if chsize[i] > 0:
                # the ids of the non leaf nodes
                deltatndbid.append(tndbid[i])
            else:
                # the indexes of the leaf nodes
                trix_to_copy_from.append(i)

        trix_to_copy_from = array(trix_to_copy_from, dtype=int32, order='c')

        if len(deltatndbid) > 0:
            # get the deltadata
            deltatndbid = array(deltatndbid, dtype=int32, order='c')
            deltadata = self._deltadata(deltatndbid)
        else:
            deltadata = None

        return trix_to_copy_from, deltadata

    #######################################################################

    cdef dict collapse_specs(self, dict rankmeta, dict rankdata_true):
        """
        Choose the nodes to be collapsed based on the collapsing criterion.

        If rankdata_true is None, the statistics are approximated from the node's children.
        Otherwise, the statistics are computed from the node's omega sets.
        """

        cdef cnp.int32_t ii
        cdef cnp.int_t i

        cdef cnp.float32_t pnmass, pnmean, pnvar, pnvalue
        cdef cnp.int32_t pnsize, tmp0, dataln, itemln
        cdef cnp.int32_t ipndbid
        cdef list delta_pndbid, delta_improb, delta_tnsize, delta_value, delta_pnmass
        cdef dict rankdict, ipndict
        cdef list candidates

        ###
        rankdict = rankmeta['rankdict']
        dataln = rankmeta['traceln']

        # STOP CRITERION
        if dataln < self.TRACE_LN:
            #print 'EXIT: TRACE_LN'
            return None

        ### !! keep only FULL parents
        delta_pndbid = []
        delta_value = []
        delta_pnmass = []

        keys = sorted(rankdict.keys())
        #for ipndbid, ipndict in rankdict.iteritems():
        for ipndbid in keys:

            if ipndbid in self.skipped_nodes:
                #print 'HURA'
                continue

            ipndict = rankdict[ipndbid]

            # do NOT collapse nodes under a certain depth
            #tnlevel = ipndict.get('tnlevel', -1)
            #if tnlevel >= 0 and tnlevel <= 1:
            #    continue

            # already computed, just take them into account
            pnvalue = ipndict.get('value', -1.0)
            if pnvalue >= 0.0:
                ###
                pnmass = ipndict['mass']
                ###
                delta_pndbid.append(ipndbid)
                delta_value.append(pnvalue)
                delta_pnmass.append(pnmass)
                continue

            improb = ipndict['improb']
            chsize = ipndict['chsize']
            itemln = len(improb)
            if itemln == chsize:

                if rankdata_true is None:
                    # probabilities of the node's children
                    #improb = ipndict['improb']
                    improbln = len(improb)
                else:
                    # probabilities of the node's omega sets
                    improb = []
                    omega = self.dbtree.omega([ipndbid])
                    for dbid in omega:
                        ### shift dbid to match trix
                        improb.append(rankdata_true['improb'][dbid-1])
                        #except:
                        #    print 'dbid', ipndbid, len(omega), dbid
                        #    print 'rankdata_true', rankdata_true['imdbid']
                        #    print 'rankdata_true', rankdata_true['improb']
                    improbln = len(improb)

                ### determine pnvalue (ie. according to which the nodes are ranked)
                criterion = 1
                if criterion == 0:
                    # mean-variance or only variance ?
                    pnmass = 0.0
                    pnsize = 0
                    for ii from 0 <= ii < improbln:
                        tmp0 = 1
                        pnmass += improb[ii] * tmp0
                        pnsize += tmp0

                    pnmean = pnmass / pnsize

                    # HINT: keep ordering on the mean, even if var is 0.0
                    pnvar = 0.0
                    for ii from 0 <= ii < improbln:
                        tmp0 = 1
                        pnvar += pow((improb[ii] - pnmean), 2) * tmp0

                    pnvar /= pnsize

                    pnvalue = pnmean * pnvar + pnmass * EPSILON
                    #pnvalue = pnvar

                elif criterion == 1:
                    # maximum probability
                    pnmass = 0.0 #sum(improb)
                    pnvalue = max(improb)

                elif criterion == 2:
                    # sampled maximum probability
                    pnmass = 0.0 #sum(improb)
                    if improbln > chsize:
                        pnvalue = max(random.sample(improb, chsize))
                    else:
                        pnvalue = max(improb)

                elif criterion == 3:
                    # 0.95 population maximum probability
                    pnmass = 0.0 #sum(improb)
                    if improbln > chsize:
                        improb_sorted = sorted(improb)
                        index_special = int(0.95 * improbln)
                        pnvalue = improb_sorted[index_special]
                    else:
                        pnvalue = max(improb)

                # progress info
                if criterion in (2, 3) and pnvalue != max(improb):
                    logger.info('    PNVALUE: TRUE_MAX %f, APPROX_MAX %f' %
                                (max(improb), pnvalue))

                ###
                ipndict['value'] = pnvalue
                ipndict['mass'] = pnmass

                ###
                delta_pndbid.append(ipndbid)
                delta_value.append(pnvalue)
                delta_pnmass.append(pnmass)

        deltaln = len(delta_pndbid)
        if deltaln == 0:
            #sys.stderr.write('\nTRACE ERROR\n')
            #sys.exit(1)
            #print 'EXIT: DELTALN'
            return None

        ###
        delta_data = array(delta_value, dtype=float32, order='c')
        candidates = candidate_nodes2(delta_data)

        tmp = [delta_pndbid[i] for i in candidates]
        tmp = array(tmp, dtype=int32, order='c')
        #print 'tmp', tmp
        ###
        deltadata = self._tracedata(tmp, False)

        # later, one can compare the[children] estimated mass with the[parent] one
        if logger.getEffectiveLevel() < logging.DEBUG:
            #print 'candidates', len(candidates), len(delta_pndbid), len(delta_pnmass), len(delta_value), candidates
            for ii in candidates:
                 if delta_pndbid[ii] in (2, 43692):
                    logger.info('    MASS [pndbid %d]: CHILDRENSUM %f, MEANVAR %f' %
                                (delta_pndbid[ii], delta_pnmass[ii], delta_value[ii]))
            for ii in xrange(0, len(delta_pndbid), 1):
                if delta_pndbid[ii] in (2, 43692):
                    logger.info('    MASS [pndbid %d]: CHILDRENSUM %f, MEANVAR %f XX' %
                                (delta_pndbid[ii], delta_pnmass[ii], delta_value[ii]))

        return deltadata

    #######################################################################

    cdef dict init_collapse(self, dict rankdata):
        """
        Prepare for the collapsing operation.
        Rankdata is arranged in a dictionary format.
        """

        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] tndbid = rankdata['tndbid']
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] pndbid = rankdata['pndbid']
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] improb = rankdata['improb']
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] tnsize = rankdata['tnsize']
        #cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] tnlevel = rankdata['tnlevel']

        cdef cnp.int32_t i, dataln, ipndbid, ichsize, itnlevel
        cdef list tmplist, pndbid_set
        cdef dict rankmeta, rankdict, ipndict

        # rankdata >> rankdict >> rankmeta
        rankdict = {}
        dataln = pndbid.shape[0]
        for i from 0 <= i < dataln:

            ipndbid = pndbid[i]

            ipndict = rankdict.get(ipndbid, {})

            tmplist = ipndict.get('tntrix', [])
            tmplist.append(i)
            ipndict['tntrix'] = tmplist

            tmplist = ipndict.get('tndbid', [])
            tmplist.append(tndbid[i])
            ipndict['tndbid'] = tmplist

            tmplist = ipndict.get('improb', [])
            tmplist.append(improb[i])
            ipndict['improb'] = tmplist

            tmplist = ipndict.get('tnsize', [])
            tmplist.append(tnsize[i])
            ipndict['tnsize'] = tmplist

            rankdict[ipndbid] = ipndict

        pndbid_set = rankdict.keys()
        chdata = self.dbtree.chsize_level(pndbid_set)
        for ipndbid, ichsize, itnlevel in chdata:
            ipndict = rankdict[ipndbid]
            ipndict['chsize'] = ichsize
            ipndict['tnlevel'] = itnlevel

        # rankmeta
        rankmeta = {}
        rankmeta['rankdict'] = rankdict
        rankmeta['traceln'] = dataln

        return rankmeta

    cdef tuple exit_collapse(self, dict rankmeta):
        """
        Prepare for the collapsing operation.
        Rankdata is arranged in a dictionary format.
        """

        cdef cnp.int32_t i, itrix, ipndbid
        cdef list newimprob
        cdef dict rankdict, ipndict
        cdef list itntrix, itndbid, itnprob

        ###
        rankdict = rankmeta['rankdict']

        #for key, value in rankdict.iteritems():
        #    print value['chsize'], value['tnsize']

        ### rankmeta >> rankdict >> rankdata
        trix_to_copy_from = []
        newtndbid = []
        newimprob = []
        for ipndbid, ipndict in rankdict.iteritems():

            itntrix = ipndict.get('tntrix', None)
            itndbid = ipndict.get('tndbid', None)
            itnprob = ipndict.get('improb', None)

            for i, itrix in enumerate(itntrix):
                if itrix < 0:
                    newtndbid.append(itndbid[i])
                    newimprob.append(itnprob[i])
                else:
                    trix_to_copy_from.append(itrix)

        trix_to_copy_from = array(trix_to_copy_from, dtype=int32, order='c')

        if len(newtndbid) > 0:
            newtndbid = array(newtndbid, dtype=int32, order='c')
            deltadata = self._tracedata(newtndbid, False)
            deltadata['improb'] = array(newimprob, dtype=float32, order='c')
        else:
            deltadata = None

        return trix_to_copy_from, deltadata

    cdef do_collapse(self, dict rankmeta, dict deltadata):
        """
        Collapse the rankdata, recursively.
        """

        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] pndbid
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] tndbid
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] improb
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] tnsize

        cdef cnp.int32_t i, deltaln, dataln, itndbid, ipndbid
        cdef dict rankdict, ipndict
        cdef list tmplist

        ###
        pndbid = deltadata['pndbid']
        tndbid = deltadata['tndbid']
        improb = deltadata['improb']
        tnsize = deltadata['tnsize']

        ###
        rankdict = rankmeta['rankdict']
        dataln = rankmeta['traceln']

        deltaln = pndbid.shape[0]
        for i from 0 <= i < deltaln:

            ipndbid = pndbid[i]
            itndbid = tndbid[i]

            # HINT: the corresponding trix become obsolete
            dataln -= rankdict[itndbid]['chsize']
            del rankdict[itndbid]

            ipndict = rankdict.get(ipndbid, {})

            # set an invalid trix
            tmplist = ipndict.get('tntrix', [])
            tmplist.append(-1)
            ipndict['tntrix'] = tmplist

            tmplist = ipndict.get('tndbid', [])
            tmplist.append(itndbid)
            ipndict['tndbid'] = tmplist

            tmplist = ipndict.get('improb', [])
            tmplist.append(improb[i])
            ipndict['improb'] = tmplist

            tmplist = ipndict.get('tnsize', [])
            tmplist.append(tnsize[i])
            ipndict['tnsize'] = tmplist

            if ipndict.get('chsize', None) is None:
                children_size = self.dbtree.children_size((ipndbid,))
                ipndict['chsize'] = children_size[0][1]

            ###
            rankdict[ipndbid] = ipndict

            ###
            dataln += 1

        ###
        rankmeta['rankdict'] = rankdict
        rankmeta['traceln'] = dataln

    #######################################################################
    # expand/collapse functionality

    cdef dict expcol_rankdata(self, dict rankdata, cnp.ndarray trix_to_copy_from, dict deltadata):
        """
        Update efficiently the rankdata for the new trace.
        """

        deltaln = len(deltadata['imdbid'])
        dataln = len(rankdata['imdbid'])
        deltatrix = self._reformat_specs(dataln, trix_to_copy_from, deltaln)

        new_rankdata = self._expcol_rankdata(rankdata, deltatrix, deltadata)

        return new_rankdata

    cdef expcol_databuffer(self, cnp.ndarray seeddbid, cnp.ndarray newdbid, cnp.ndarray dbid):
        """
        Update efficiently the databuffer for the new trace.
        """

        deltatrix, deltadbid = self._reformat_specs2(dbid, newdbid)

        self.dbimg.expcol_featdata(newdbid, dbid, deltatrix, deltadbid)
        if seeddbid is not None:
            self.dbimg.expcol_dd(seeddbid, newdbid, dbid, deltatrix, deltadbid)