#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2010                   - IDIAP Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Estimates the consistency based on the relevance feedback history.

.. note:: The consistency scores are used to compute the zoom-in factor.
"""

###########################################################################
# IMPORTS

import logging
import sys
from time import time as now

from scipy.special import erf
from libc.math cimport sqrt as c_sqrt

from retrieval_libs.common.index cimport get_index0

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

###########################################################################
# CLASS

cdef class ConsistencyUpdater(object):
    """
    Estimates the consistency based on the relevance feedback history.
    
    .. note:: The consistency scores are used to compute the zoom-in factor.
    """

    #######################################################################
    #  local functionality

    cdef cnp.float32_t _update(self, tuple dispfdbk, list dispfdbkH, list thargsH, list weightsH, list consistencyH):
        """
        Update the consistency score.

        dispfdbk  - the current relevance feedback

        dispfdbkH - the relevance feedback history (i.e. the list of displayed
                    images and the corresponding relevant images);
        thargsH   - the thresholds history (i.e. the list of thresholds
                    that were applied at each iteration);
        weightsH  - the weights history (i.e. the list of weights
                    that were applied at each iteration);
        """

#        cdef tuple thargs
#        cdef cnp.ndarray[cnp.float32_t, ndim=2, mode='c'] weights

        # the posteriori probabilities
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] hprob

        # the P_{+} and P_{-} probabilities
        #cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] pp, pm

        cdef cnp.int32_t kindex, hindex

        version = 1
        if version == 0:
            # get the current dispdbid and the corresponding dispstar
            hdbid = dispfdbk[0]['imdbid']
            hstar = dispfdbk[1]
            # get the current probabilities of dispdbid
            hprob = dispfdbk[0]['improb']

            # get the previous dispdbid and the corresponding star
#            kdbid = dispfdbkH[-1][0]['imdbid']
#            kstar = dispfdbkH[-1][1]

            # get the previous thargs and weights
#            thargs = thargsH[-1]
#            weights = weightsH[-1]

#            kindex = get_index0(kstar, kdbid, True)
            hindex = get_index0(hstar, hdbid, True)
#            dd = self.dbimg.distdata(kdbid, hdbid, -1, True, True)

            #
            #ddp, ddm = ddpddm(dd, thargs, weights)

            #consistency = hprob[hindex] / hprob.max(0) #* hprob.std(0)
            #consistency = hprob[hindex] / hprob.mean(0) #* hprob.std(0)
            #consistency = ddp[kindex, hindex] / ddp[kindex, :].sum(0)

            if hprob.max(0) > hprob.min(0):
                cmin = 0.5
                cmax = 2.0
                A = (cmax - cmin) / (hprob.max(0) - hprob.min(0))
                B = (cmin * hprob.max(0) - cmax * hprob.min(0)) / (hprob.max(0) - hprob.min(0))
                consistency = A * hprob[hindex] + B
            else:
                consistency = 1.0

        elif version == 1:
            # get the current dispdbid and the corresponding dispstar
            hdbid = dispfdbk[0]['imdbid']
            hstar = dispfdbk[1]
            # get the current probabilities of dispdbid
            hprob = dispfdbk[0]['improb']

            hindex = get_index0(hstar, hdbid, True)

            if hprob.max(0) > hprob.min(0):
                mu = hprob.mean()
                sigma = hprob.std()
                x = hprob[hindex]
                consistency = 0.5 * (1.0 + erf((x - mu) / (sigma * c_sqrt(2.0))))
#                print 'ZZZZZZZZZZZ', hindex, x, hprob
#                print 'CCCCCCCCCCC', mu, sigma, (x - mu) / (sigma * c_sqrt(2.0)), consistency
                consistency = 0.5 + 1.5 * consistency
            else:
                consistency = 1.0

        elif version == 2:
            # get the current dispdbid and the corresponding dispstar
            hdbid = dispfdbk[0]['imdbid']
            hstar = dispfdbk[1]
            # get the current probabilities of dispdbid
            hprob = dispfdbk[0]['improb']

            hindex = get_index0(hstar, hdbid, True)

            if hprob[hindex] == hprob.max(0):
                consistency = 2.0
            else:
                consistency = 1.0

        else:
            sys.stderr.write('\nUNKNOWN VERSION %s\n' % version)
            sys.exit(1)

        return consistency

    #######################################################################
    #  main functionality

    cdef init(self, DbTableImg dbimg, tuple CONSISTENCY):
        """
        Initialize the object.
        """

        self.dbimg = dbimg
        self.consistency_init = CONSISTENCY[0]
        self.consistency_type = CONSISTENCY[1]

    cdef cnp.float32_t update(self, tuple dispfdbk, list dispfdbkH, list thargsH, list weightsH, list consistencyH):
        """
        Update the consistency score.

        dispfdbk  - the current relevance feedback

        dispfdbkH - the relevance feedback history (i.e. the list of displayed
                    images and the corresponding relevant images);
        thargsH   - the thresholds history (i.e. the list of thresholds
                    that were applied at each iteration);
        weightsH  - the weights history (i.e. the list of weights
                    that were applied at each iteration);
        consistencyH - the consistency history;
        """

        cdef cnp.float32_t consistency

        #logger.debug('START')
        start_time = now()

        # HINT: there can be iterations without image feedback
        if self.consistency_type == 'FIXED' or \
           dispfdbk is None:
           #len(dispfdbkH) < 1:
            consistency = self.consistency_init

        #elif dispfdbkH[-1][1] is None:
        #    consistency = self.consistency_init

        elif dispfdbk[1] is None:
            consistency = self.consistency_init

        elif self.consistency_type == 'ADAPTIVE':
            consistency = self._update(dispfdbk, dispfdbkH, thargsH, weightsH, consistencyH)

        else:
            sys.stderr.write('\nUNKNOWN CASE\n')
            sys.exit(1)

        stop_time = now()
        logger.info('    consistency: %s' % consistency)
        logger.debug('DONE in %f seconds' % (stop_time - start_time))

        return consistency
