#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Selects the image set for display.

.. note:: The images are chosen by performing an heuristic algorithm.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, float32, array, zeros, ones, empty, append
import sys
from time import time as now

from retrieval_libs.common.candidate cimport candidate_image, \
                                             skipped_images
from retrieval_libs.common.index cimport get_index0
from retrieval_libs.common.thdist cimport ddw

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

###########################################################################
# CLASS

cdef class DisplaySetSelector(object):
    """
    Selects the image set for display.

    .. note:: The images are chosen by performing an heuristic algorithm.
    """

    #######################################################################
    # local functionality

    cdef _prob_mass(self,
            cnp.int32_t nb_seeds,
            dict rankdata,
            cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] skipdisp):
        """
        Update the total probability mass of the rankdata.
        """

        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] rankdbid = rankdata['imdbid']
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] rankprob = rankdata['improb']
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] rankmass

        cdef cnp.int32_t i, ni = rankprob.shape[0]
        #cdef cnp.int32_t idbid, ii, ini = skipdisp.shape[0]
        cdef cnp.float32_t immass, total_mass
        #cdef bint skipflag

        rankmass = empty((ni,), dtype=float32, order='c')

        # total probability mass
        total_mass = 0.0
        for i from 0 <= i < ni:

            # skip the images already displayed
            #skipflag = False
            #idbid = rankdbid[i]
            #for ii from 0 <= ii < ini:
            #    if skipdisp[ii] == idbid:
            #        skipflag = True
            #        break
            #if skipflag == True:
            #    continue

            immass = rankprob[i]
            rankmass[i] = immass
            total_mass += immass

        rankdata['immass'] = rankmass

        self.total_mass = total_mass

        #logger.debug('Total mass %f' % total_mass)

    #######################################################################

    cdef _group(self, list clusdata, dict rankdata):
        """
        Group the images into clusters around the seeds.

        First, the images are order by their distances to their closest
        clusters; Then, images are added to their closest cluster.
        """

        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] closestcldd = rankdata['closestcldd']

        # order images by their distances, in increasing order
        # HINT: keep the sorted arguments, for within the expand/collapse
        # routines, during which 'closestcldd' is unchanged.
        rankdata['sortedarg'] = closestcldd.argsort()

        self._group_iteration(clusdata, rankdata)

    cdef _group_iteration(self, list clusdata, dict rankdata):
        """
        Images are added to their closest cluster, as ordered, one by one,
        until the probability mass of each cluster reach the optimal value, or
        until all the 'booked-as-closest' images are added.
        """

        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] seeddbid

        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] clussize = clusdata[1]
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] clusmass = clusdata[2]
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] clusdone = clusdata[3]
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] clusrad = clusdata[4]

        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] rankmass = rankdata['immass']
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] rankclix = rankdata['tnclix']

        cdef cnp.ndarray[cnp.int_t, ndim=1, mode='c'] closestclix = rankdata['closestclix']
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] closestcldd = rankdata['closestcldd']
        cdef cnp.ndarray[cnp.int_t, ndim=1, mode='c'] sortedarg = rankdata['sortedarg']

        cdef cnp.int32_t i, ni = rankmass.shape[0]
        cdef cnp.float32_t clmass
        cdef cnp.int_t trix, clix
        cdef dict seeddata

        ###
        seeddata = clusdata[0]
        seeddbid = seeddata['imdbid']

        # cluster the images, as ordered, one by one, around the closest seed
        # HINT: the ~clouds of images are NOT built/saved effectively,
        #       only their sizes and their masses are computed/saved
        for i from 0 <= i < ni:

            trix = sortedarg[i]

            # skip the images already clustered
            if rankclix[trix] >= 0:
                continue

            clix = closestclix[trix]
            # do NOT check if clmass > self.optim_clmass
            # HINT: when DONE, clmass is going to be a fraction smaller than
            # the optimum clmass; otherwise, if it would be a fraction larger,
            # then the last cluster may be too obstructed.
            if clusdone[clix] > 0:
                continue

            clmass = clusmass[clix]

            clmass += rankmass[trix]
            if clmass > self.optim_clmass:
                # mark the cluster as finished/full
                # HINT: do NOT add it to the cluster
                clusdone[clix] = 1
                continue

            # update the cluster size and mass
            clussize[clix] += 1
            clusmass[clix] = clmass
            clusrad[clix] = closestcldd[trix]
            # update the cluster membership
            rankclix[trix] = clix

        # progress info
        if logger.getEffectiveLevel() < logging.DEBUG:
            seeddataln = len(clusdata[0]['imdbid'])
            for clix in xrange(0, seeddataln, 1):
                logger.debug(
                        '    Cluster %s: seed %d, size %d, '
                        'mass %f, optimum mass %f, rad %f'
                        % (clix, clusdata[0]['imdbid'][clix], clusdata[1][clix],
                           clusdata[2][clix], self.optim_clmass, clusdata[4][clix]))

    cdef _sort(self,
            cnp.int_t sdtrix,
            cnp.int_t sdclix,
            cnp.ndarray ddC,
            list clusdata,
            dict rankdata):
        """
        Update the 'closest-data' altered by the last added seed.

        For each image, the closest seed and the corresponding minimum distance
        are booked. If the image is closer to the new/last added seed,
        the 'closest-data' is updated, and if clustered already, its existing
        clustering is undone.

        The seed is added to its own cluster (i.e. the cluster is initialized).
        """

        cdef cnp.ndarray[cnp.float32_t, ndim=3, mode='c'] dd = ddC

        cdef cnp.ndarray[cnp.int_t, ndim=1, mode='c'] closestclix = rankdata['closestclix']
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] closestcldd = rankdata['closestcldd']

        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] rankmass = rankdata['immass']
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] rankclix = rankdata['tnclix']

        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] seedclix = clusdata[0]['tnclix']
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] clussize = clusdata[1]
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] clusmass = clusdata[2]
        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] clusdone = clusdata[3]
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] clusrad = clusdata[4]

        cdef cnp.int32_t i, ni, isize
        cdef cnp.float32_t imass

        ni = rankmass.shape[0]

        if sdclix == 0:
            # for the first seed, copy the distances
            # HINT: initially, the images belong to the first seed
            for i from 0 <= i < ni:
                # HINT: closestclix was just initialized
                #closestclix[i] = 0
                closestcldd[i] = dd[0,0,i]

        else:
            # for successive seeds, update the distances
            # and the (index of the) corresponding seed
            for i from 0 <= i < ni:
                # HINT: choose the new seed, if it is closer
                if closestcldd[i] > dd[0,0,i]:
                    # step 1: if necessary, undo the clusdata
                    if rankclix[i] >= 0:
                        # the image is currently part of a valid cluster
                        #logger.debug('DECLUSTER the image')

                        # get the current cluster
                        clix = closestclix[i]
                        # if rankclix is valid, it should match clix
                        #if rankclix[i] != clix:
                        #    logger.error('    RANKCLIX MISMATCH %d != %d' %
                        #                 (rankclix[i], clix))

                        clussize[clix] -= 1
                        # update the cluster probability mass
                        clusmass[clix] -= rankmass[i]
                        clusrad[clix] = 0.0
                        clusdone[clix] = 0
                        rankclix[i] = -1

                    # step 2: update the closest data
                    closestclix[i] = sdclix
                    closestcldd[i] = dd[0,0,i]

        # update the clusdata: add the seed/image to its own cluster
        # HINT: guarantee that the seed/image is part of its own cluster
        # ... since more images CAN have 'zero' distance to the seed
        if sdtrix >= 0:
            i = sdtrix

            # the size and the probability mass of the current seed/image
            isize = 1
            imass = rankmass[i]

            # step 1: if necessary, undo the clusdata
            if rankclix[i] >= 0:
                # the seed/image is currently part of a valid cluster
                #logger.debug('DECLUSTER the seed/image')

                # get the current cluster
                clix = closestclix[i]
                # if rankclix is valid, it should match clix
                #if rankclix[i] != clix:
                #    logger.error('    RANKCLIX MISMATCH %d != %d' %
                #                 (rankclix[i], clix))

                clussize[clix] -= isize
                # update the cluster probability mass
                clusmass[clix] -= imass
                #clusrad[clix] = 0.0
                clusdone[clix] = 0
                #rankclix[i] = -1

            # step 2: update the closest data
            closestclix[i] = sdclix
            closestcldd[i] = dd[0,0,i]

            # step 3: update the clusdata
            clussize[sdclix] += isize
            # update the cluster probability mass
            clusmass[sdclix] += imass
            clusrad[sdclix] = closestcldd[i]
            clusdone[sdclix] = 0
            # update the cluster membership
            seedclix[sdclix] = sdclix
            rankclix[i] = sdclix

    #######################################################################
    # local functionality

    cdef cnp.int32_t _update(self,
            list clusdata,
            dict rankdata,
            tuple thargs,
            cnp.ndarray weights,
            cnp.ndarray skipdisp):

        cdef cnp.ndarray[cnp.int32_t, ndim=1, mode='c'] rankdbid

        cdef cnp.int32_t i, ni

        cdef cnp.int32_t sdtrix, sdclix

        self._prob_mass(self.SIZE, rankdata, skipdisp)

        # LOOP for selecting the seed/images, one by one
        for sdclix from 0 <= sdclix < self.SIZE:

            # this is more efficient than retrieving from the tables
            # HINT: items of rankdata are CHANGING
            rankdbid = rankdata['imdbid']

            # select a new seed/image
            if self.given_seed_candidates is None:
                sdtrix = candidate_image(sdclix, rankdata, skipdisp)
                if sdtrix < 0:
                    logger.error('    INVALID CANDIDATE %d' % sdtrix)
                    return -1
                sddbid = rankdbid[sdtrix]
            else:
                sddbid = self.given_seed_candidates['imdbid'][sdclix]
                sdtrix = get_index0(sddbid, rankdbid, True)
                if sdtrix < 0:
                    logger.error('    INVALID CANDIDATE %d' % sdtrix)
                    return -1

            # save/keep/register the seed/image MAIN INFO
            seeddata = clusdata[0]
            for key, item in seeddata.iteritems():
                seeddata[key] = append(item, rankdata[key][sdtrix:(sdtrix+1)], axis=0)

            # plot the trace
            if self.traceimg is not None:
                tmp = seeddata['tnclix'][-1]
                seeddata['tnclix'][-1] = len(seeddata['tnclix']) - 1
                fileurl = self.traceimg.traceplot_r(rankdata, seeddata)
                seeddata['tnclix'][-1] = tmp

            # terminate the LOOP if there are enough seed/images
            # HINT: if only the seeds are needed, the last clustering can be skipped
            if sdclix >= (self.SIZE - 1) and self.FLAG_CLUSTER == False:
                break

            # get the distances of the new seed
            sddbid = array([sddbid,], dtype=int32, order='c')
            distdata = self.dbimg.distdata(sddbid, rankdbid, -1, True, True)
            dd = ddw(distdata, thargs, weights)

            # update the 'closest-data' and de-cluster
            self._sort(sdtrix, sdclix, dd, clusdata, rankdata)

            # uniform optim_clmass
            self.optim_clmass = self.total_mass / self.SIZE  / self.zoom_in

            # re-cluster
            self._group(clusdata, rankdata)

            # plot the trace
            if self.traceimg is not None:
                fileurl = self.traceimg.traceplot_r(rankdata, seeddata)

        if self.FLAG_CLUSTER == False:
            # approximate the last cluster from all unclustered images
            # HINT: used only for display
            rankprob = rankdata['improb']
            clusdata[1][sdclix] = rankprob.shape[0] - clusdata[1].sum()
            clusdata[2][sdclix] = self.total_mass - clusdata[2].sum()

        return 0

    #######################################################################
    #  main functionality

    cdef set_consistency(self, list consistencyH):
        """
        """

        self.zoom_in = 1.0

        for consistency in consistencyH:
            self.zoom_in *= consistency

        if self.zoom_in < 1.0:
            self.zoom_in = 1.0
        elif self.zoom_in > (self.SIZE / 2.0):
            self.zoom_in = self.SIZE / 2.0

        logger.info('    ZOOM IN %s' % self.zoom_in)

    def set_traceimg(self, traceimg):
        """
        """

        self.traceimg = traceimg

    #######################################################################
    #  main functionality

    cdef init(self, DbTableImg dbimg, cnp.int32_t size):
        """
        Initialize the object.
        """

        self.dbimg = dbimg

        self.FLAG_CLUSTER = True #False
        #if logger.getEffectiveLevel() <= logging.DEBUG:
        #    self.FLAG_CLUSTER = True

        self.SIZE = size
        self.given_seed_candidates = None

        self.traceimg = None

    cdef dict update(self,
            dict rankdata,
            tuple thargs,
            cnp.ndarray weights,
            list dispfdbkH):
        """
        Return the set of images for display, by performing an heuristic
        sampling algorithm on the ranked images.

        HINT: seeddata and clusdata are modified in place,
        but their referenced arrays are NOT modified in place!
        """

        logger.debug('START')
        start_time = now()

        # initialize seeddata
        seeddata = {}
        seeddata['imdbid'] = array([], dtype=int32, order='c')
        seeddata['improb'] = array([], dtype=float32, order='c')
        seeddata['tnclix'] = array([], dtype=int32, order='c')

        # initialize clusdata
        clusdata = []
        shape = (self.SIZE,)
        clussize = zeros(shape, dtype=int32, order='c')
        clusmass = zeros(shape, dtype=float32, order='c')
        clusdone = zeros(shape, dtype=int32, order='c')
        clusrad = zeros(shape, dtype=float32, order='c')
        clusdata.append(seeddata)
        clusdata.append(clussize)
        clusdata.append(clusmass)
        clusdata.append(clusdone)
        clusdata.append(clusrad)

        # extra rankdata
        rankdataln = len(rankdata['improb'])
        shape = (rankdataln,)
        # for each image, keep (the index of) its closest seed
        rankdata['closestclix'] = zeros(shape, dtype=int, order='c')
        # for each image, keep the distance to its closest seed
        rankdata['closestcldd'] = empty(shape, dtype=float32, order='c')
        #rankdata['sortedarg'] = array([], dtype=int, order='c')
        # for each image, keep (the index of) the cluster to which it belongs
        rankdata['tnclix'] = empty(shape, dtype=int32, order='c')
        rankdata['tnclix'][:] = -1

        # initialize skipdisp
        skipdisp = skipped_images(dispfdbkH)

        err = self._update(clusdata, rankdata, thargs, weights, skipdisp)
        if err == -1:
            sys.stderr.write('\nALL IMAGES HAVE BEEN CLUSTERED ALREADY\n\n')
            sys.exit(1)

        seeddataln = len(seeddata['imdbid'])

        # progress info
        if logger.getEffectiveLevel() <= logging.DEBUG:
            for clix in xrange(0, seeddataln, 1):
                logger.info(
                        '    Cluster %s: seed %d, size %d, '
                        'clmass %f, rad %f'
                        % (clix, seeddata['imdbid'][clix], clusdata[1][clix],
                           clusdata[2][clix], clusdata[4][clix]))

        if seeddataln > 8:
            for key, item in seeddata.iteritems():
                # keep only the first 8
                seeddata[key] = item[:8]
                # keep only the last 8
                #seeddata[key] = item[-8:]

        stop_time = now()
        logger.debug('DONE in %f seconds' % (stop_time - start_time))

        return seeddata
