#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Updates the probabilities of image relevance.
"""

###########################################################################
# IMPORTS

import logging
from numpy import float32, empty
import sys
from time import time as now

from retrieval_libs.common.index cimport get_index0
from retrieval_libs.common.probability cimport update_prob, update_prob_full, dwdate_prob
from retrieval_libs.common.thdist cimport pppm

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

# initial probability of image relevance
cdef cnp.float32_t INITPROB = 0.5

###########################################################################
# CLASS

cdef class ProbabilityUpdater(object):
    """
    Updates the probabilities of image relevance.
    """

    #######################################################################
    # main functionality

    cdef init(self, DbTableImg dbimg):
        """
        Initialize the object.
        """

        self.dbimg = dbimg

    cdef update(self, dict rankdata, tuple dispfdbk, tuple thargs, cnp.ndarray weights):
        """
        Update (one iteration) the probabilities of image relevance.

        rankdata  - the data for which the probabilities have to be updated;
        dispfdbk  - the current relevance feedback (i.e. the displayed images
                    and the image indicated as relevant);
        thargs    - the thargs to be applied;
        weights   - the weights to be applied;
        """

        cdef cnp.ndarray rankdbid, fdbkdbid, rankprob, pp, pm
        cdef tuple dd
        cdef cnp.int32_t fdbkstar

        logger.debug('START')
        start_time = now()

        rankdbid = rankdata['imdbid']

        # if necessary, initialize the probabilities
        if 'improb' not in rankdata:
            shape = (rankdbid.shape[0],)
            rankprob = empty(shape, dtype=float32, order='c')
            rankprob[:] = INITPROB
            rankdata['improb'] = rankprob

        # update the probabilities
        if dispfdbk[0] is not None:
            rankprob = rankdata['improb']

            fdbkdbid = dispfdbk[0]['imdbid']
            fdbkstar = dispfdbk[1]

            fdbkindex = get_index0(fdbkstar, fdbkdbid, True)

            dd = self.dbimg.distdata(fdbkdbid, rankdbid, -1, True, True)
            pp, pm = pppm(dd, fdbkindex, thargs, weights)

            update_prob(rankprob, pp, pm)

        stop_time = now()
        logger.debug('DONE in %f seconds' % (stop_time - start_time))

    #######################################################################
    # update full functionality

    cdef init_update_full(self, list dispfdbkH, list thargsH, list weightsH):
        """
        Update (from scratch) the probabilities of image relevance.

        rankdata  - the data for which the probabilities have to be updated;
        dispfdbkH - all relevance feedback iterations;
        thargsH   - the corresponding thargs (ie. that were applied);
        weightsH  - the corresponding weights (ie. that were applied);
        """

        cdef cnp.int32_t fdbkstar

        self.fdbkfeatH = []
        self.fdbkindexH = []
        self.thargsH = []
        self.weightsH = []

        if dispfdbkH is None:
            return

        for i in xrange(1, len(dispfdbkH), 1):

            dispfdbk = dispfdbkH[i]
            if dispfdbk[0] is None:
                continue

            fdbkdbid = dispfdbk[0]['imdbid']
            fdbkstar = dispfdbk[1]

            fdbkfeat = self.dbimg.featdata(fdbkdbid, -1, True)
            self.fdbkfeatH.append(fdbkfeat)

            fdbkindex = get_index0(fdbkstar, fdbkdbid, True)
            self.fdbkindexH.append(fdbkindex)

            self.thargsH.append(thargsH[i])
            self.weightsH.append(weightsH[i])

    cdef init_update_full2(self, list fdbkfeatH, list fdbkindexH, list thargsH, list weightsH):
        """
        Update (from scratch) the probabilities of image relevance.

        rankdata  - the data for which the probabilities have to be updated;
        dispfdbkH - all relevance feedback iterations;
        thargsH   - the corresponding thargs (ie. that were applied);
        weightsH  - the corresponding weights (ie. that were applied);
        """

        self.fdbkfeatH = fdbkfeatH
        self.fdbkindexH = fdbkindexH
        self.thargsH = thargsH
        self.weightsH = weightsH

    cdef update_full(self, dict rankdata):
        """
        Update (from scratch) the probabilities of image relevance.
        """

        cdef cnp.ndarray rankdbid
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] rankprob
        cdef cnp.ndarray[cnp.float32_t, ndim=1, mode='c'] ppM, pmM, pp, pm
        cdef tuple fdbkfeat, rankfeat, dd
        cdef cnp.int32_t fdbkindex, i, ni

        cdef tuple thargs
        cdef cnp.ndarray weights

        #logger.debug('    UPDATE-FULL START')
        #start_time = now()

        rankdbid = rankdata['imdbid']

        # CLEAN START, from the initial/default probabilities
        #if 'improb' in rankdata:
        #    sys.stderr.write('\nRANKPROB OVERWRITTEN\n')
        #    sys.exit(1)
        shape = (rankdbid.shape[0],)
        rankprob = empty(shape, dtype=float32, order='c')
        rankprob[:] = INITPROB

        if len(self.fdbkfeatH) == 0:
            rankdata['improb'] = rankprob
            #stop_time = now()
            #logger.debug('DONE (NOUPDATE) in %f seconds' % (stop_time - start_time))
            return

        rankfeat = self.dbimg.featdata(rankdbid, -1, True)

        ni = len(self.fdbkfeatH)

        for i from 0 <= i < ni:
            fdbkfeat = self.fdbkfeatH[i]
            fdbkindex = self.fdbkindexH[i]

            thargs = self.thargsH[i]
            weights = self.weightsH[i]

            dd = self.dbimg.distdata_feat(fdbkfeat, rankfeat)
            pp, pm = pppm(dd, fdbkindex, thargs, weights)

            if i > 0:
                ppM *= pp
                pmM *= pm
            else:
                ppM = pp
                pmM = pm

        update_prob_full(rankprob, ppM, pmM)

        rankdata['improb'] = rankprob

        #stop_time = now()
        #logger.debug('    UPDATE-FULL DONE in %f seconds' % (stop_time - start_time))
