#!/usr/bin/env python

#cython: boundscheck=False
#cython: wraparound=False
#cython: embedsignature=True
#cython: profile=True

###########################################################################
#                                                                         #
# Copyright (c) 2011                   - Idiap Research Institute         #
#                                        http://www.idiap.ch/             #
#                                                                         #
# See file COPYING for the license associated to this software            #
#                                                                         #
###########################################################################

"""
Manages the original retrieval algorithm.
"""

###########################################################################
# IMPORTS

import logging
from numpy import int32, array
import sys
from time import time as now

from retrieval_libs.dblayer.DbAccess import DbAccess
from retrieval_libs.display.TraceGfx import TraceGfx

###########################################################################
# GLOBALS

cdef object logger = logging.getLogger('retrieval')

# implemented as a singleton
cdef object dbaccess = DbAccess()

# the number of image/seeds selected for display
cdef cnp.int32_t SIZE = 8

###########################################################################
# CLASS

cdef class QueryImg(object):
    """
    Manages the original retrieval algorithm.
    """

    #######################################################################
    #  main functionality

    def init(self, SYS_THARGS, SYS_WEIGHTS, SYS_CONSISTENCY):
        """
        Initialize the object.
        """

        self.thargs = SYS_THARGS[0]
        self.weights = SYS_WEIGHTS[0]

        self.rankdata = None
        self.rankdata_0 = None
        self.seeddata = None
        self.tracedata = None

        #self.dispfdbkH = None

        self.dbimg = dbaccess.dbimg()

        self.traceupdater = TraceUpdater()
        self.traceupdater.init()
        self.consistencyupdater = ConsistencyUpdater()
        self.consistencyupdater.init(self.dbimg, SYS_CONSISTENCY)
        self.probupdater = ProbabilityUpdater()
        self.probupdater.init(self.dbimg)
        self.dispselector = DisplaySetSelector()
        self.dispselector.init(self.dbimg, SIZE)

        self.traceimg = TraceGfx()
        self.traceimg.init()
        #self.dispselector.set_traceimg(self.traceimg)

    def set_buffers(self):

        self.dbimg.set_buffers()

    def trim_buffers(self):

        self.dbimg.trim_buffers()

    def set_data(self, dispfdbkH, thargsH, weightsH, consistencyH):

        self.dispfdbkH = dispfdbkH
        self.consistencyH = consistencyH

    def trim_data(self):

        self.dispfdbkH = None
        self.consistencyH = None

    #######################################################################
    # update functionality

    def update_thargs(self, rankdata, dispfdbk, dispfdbkH, thargsH, weightsH):
        """
        Update the thargs.
        """

        return self.thargs

    def update_weights(self, dispfdbk, dispfdbkH, thargsH, weightsH):
        """
        Update the weights.
        """

        return self.weights

    def update_consistency(self, rankdata, dispfdbk, dispfdbkH, thargsH, weightsH, consistencyH):
        """
        Update the consistency.
        """

        return self.consistencyupdater.update(dispfdbk, dispfdbkH, thargsH, weightsH, consistencyH)

    #######################################################################
    # main functionality

    def perform_query_random(self, rankdata, rankdata_0, dispfdbk):
        """
        Generate the seeddata (i.e. the next dispdata).

        - NO relevance feedback >> NO update for the image probabilities;
        - Voronoi partitioning starting from a random image/seed;
        """

        self.rankdata = self.traceupdater.rankdata(rankdata)
        self.probupdater.update(self.rankdata, dispfdbk, self.thargs, self.weights)
        self.dispselector.set_consistency(self.consistencyH)

        self.seeddata = self.dispselector.update(self.rankdata, self.thargs, self.weights, self.dispfdbkH)

        # save/store/checkpoint
        self._add_tracedata(self.rankdata, self.seeddata, ['initial', 'RAND'])

        self.rankdata = self.traceupdater.clean_rankdata(self.rankdata)

    def perform_query_fdbk(self, rankdata, rankdata_0, dispfdbk):
        """
        Generate the seeddata (ie. the next dispdata)
        - relevance feedback >> update the image probabilities;
        - Voronoi partitioning starting from the image/seed with the highest probability;
        """

        self.rankdata = self.traceupdater.rankdata(rankdata)
        self.probupdater.update(self.rankdata, dispfdbk, self.thargs, self.weights)
        self.dispselector.set_consistency(self.consistencyH)

        self.seeddata = self.dispselector.update(self.rankdata, self.thargs, self.weights, self.dispfdbkH)

        # save/store/checkpoint
        self._add_tracedata(self.rankdata, self.seeddata, ['initial', 'FDBK'])

        self.rankdata = self.traceupdater.clean_rankdata(self.rankdata)

    #######################################################################
    # trace functionality

    cdef _add_tracedata(self, dict rankdata, dict seeddata, list action):
        """
        Prepare and append the trace data.
        """

        # if necessary, initialize tracedata
        if self.tracedata is None:
            self.tracedata = []

        COUNT_EXPAND = 0
        COUNT_COLLAPSE = 0

        if action[0] == 'initial':
            actioninfo = [' '.join(action), COUNT_EXPAND, len(rankdata['imdbid'])]
        elif action[0] == 'seed':
            actioninfo = [' '.join(action), len(seeddata['imdbid']), len(rankdata['imdbid'])]
        elif action[0] == 'expand':
            actioninfo = [' '.join(action), COUNT_EXPAND, len(rankdata['imdbid'])]
        elif action[0] == 'collapse':
            actioninfo = [' '.join(action), COUNT_COLLAPSE, len(rankdata['imdbid'])]
        elif action[0] == 'final':
            COUNT_TOTAL = COUNT_EXPAND + COUNT_COLLAPSE
            actioninfo = [' '.join(action), COUNT_TOTAL, len(rankdata['imdbid'])]
        else:
            sys.stderr.write('\nUNKNOWN ACTION %s\n' % action[0])
            sys.exit(1)

        #fileurl = None
        #if logger.getEffectiveLevel() < logging.DEBUG or \
        #   action[0] == 'final':
           #action[0] == 'initial' or \
        fileurl = self.traceimg.traceplot(rankdata, seeddata)

        actioninfo.append(fileurl)
        traceinfo = tuple(actioninfo)

        self.tracedata.append(traceinfo)
